Package: isobar
Title: Analysis and quantitation of isobarically tagged MSMS proteomics
        data
Description: isobar provides methods for preprocessing, normalization, and
    report generation for the analysis of quantitative mass spectrometry
    proteomics data labeled with isobaric tags, such as iTRAQ and TMT.
    Features modules for integrating and validating PTM-centric datasets
    (isobar-PTM). More information on http://www.ms-isobar.org.
Version: 1.57.0
Author: Florian P Breitwieser <florian.bw@gmail.com> and
   Jacques Colinge <jacques.colinge@inserm.fr>, with contributions from 
   Alexey Stukalov <stukalov@biochem.mpg.de>,
   Xavier Robin <xavier.robin@unige.ch> and Florent Gluck  <florent.gluck@unige.ch>
Maintainer: Florian P Breitwieser <florian.bw@gmail.com>
biocViews: ImmunoOncology, Proteomics, MassSpectrometry,
        Bioinformatics, MultipleComparisons, QualityControl
Depends: R (>= 2.10.0), Biobase, stats, methods
Imports: distr, plyr, biomaRt, ggplot2
Suggests: MSnbase, OrgMassSpecR, XML, RJSONIO, Hmisc, gplots,
        RColorBrewer, gridExtra, limma, boot, DBI, MASS
LazyLoad: yes
License: LGPL-2
URL: https://github.com/fbreitwieser/isobar
BugReports: https://github.com/fbreitwieser/isobar/issues
Collate: utils.R ProteinGroup-class.R IBSpectra-class.R isobar-import.R
        IBSpectra-plots.R NoiseModel-class.R Tlsd-class.R
        ratio-methods.R distr-methods.R sharedpep-methods.R
        report-utils-xls.R report-utils-tex.R report-utils.R
        metareport-utils.R ptm-methods.R MSnSet-methods.R zzz.R
git_url: https://git.bioconductor.org/packages/isobar
git_branch: devel
git_last_commit: 0225060
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-10-31
NeedsCompilation: no
Packaged: 2025-10-31 22:27:40 UTC; biocbuild

import("methods")

importFrom("grDevices", "boxplot.stats", "dev.off", "pdf", "png")
importFrom("graphics", "abline", "axTicks", "axis", "hist", "legend",
           "lines", "pairs", "par", "points", "rect", "strwidth",
           "text")
importFrom("stats", "IQR", "confint", "dcauchy", "dnorm", "dt",
           "hclust", "integrate", "lm", "mad", "median", "nlminb",
           "p.adjust", "p.adjust.methods", "pchisq", "pnorm", "pt",
           "qnorm", "qt", "quantile", "reshape", "rt", "setNames",
           "summary.lm", "t.test", "var")
importFrom("utils", "Sweave", "URLdecode", "URLencode",
           "capture.output", "combn", "compareVersion",
           "download.file", "head", "packageDescription", "read.delim",
           "read.table", "setTxtProgressBar", "tail", "txtProgressBar",
           "write.table", "zip")

importFrom("biomaRt", useMart, getBM)
importFrom("distr", Norm, df, location, "df<-", "location<-", "scale",
           "scale<-")
import("Biobase")
import("plyr")
import("ggplot2")
importFrom("graphics", "plot")
importFrom("stats", "sd")
importFrom("distr", "Cauchy", "Norm", "UnivarMixingDistribution",
           "SphericalSymmetry")

exportClasses(IBSpectra, ProteinGroup)
exportMethods("as.data.frame")

export(IBSpectraTypes)
exportPattern("^[a-zA-Z%]")

isobar - R package for iTRAQ and TMT proteomics data
================================================================================================

isobar provides methods for preprocessing, normalization, and  
report generation for the analysis of proteomics data generated by Mass 
spectrometers using isobaric peptide tags, such as iTRAQ and TMT. 
PDF and Excel reports can automatically be generated based on a configuration file.
Include modules for validating and integrating data focussing on 
post-translational protein modifications (isobar-PTM).

## Installation

The official Bioconductor version can be found at http://bioconductor.org/packages/release/bioc/html/isobar.html

To install isobar from Bioconductor, call

    ## try http:// if https:// URLs are not supported
    if (!requireNamespace("BiocManager", quietly=TRUE))
        install.packages("BiocManager")
    BiocManager::install("isobar")

To install the development version of isobar, the easiest way is using the devtools package:

    ## install.package("devtools") # if you do not have the devtools package
    devtools::install_github("fbreitwieser/isobar")

The following packages are suggested for some uses in isobar, and maybe should be installed as well: MSnbase, OrgMassSpecR, biomaRt, Hmisc, gplots and limma.

## References

- Florian P. Breitwieser, André Müller, Loïc Dayon, Thomas Köcher, Alexandre Hainard,  Peter Pichler, Ursula Schmidt-Erfurth, Giulio Superti-Furga, Jean-Charles Sanchez, Karl Mechtler, Keiryn L. Bennett, and Jacques Colinge (June 2011). “General statistical modeling of data from protein relative expression isobaric tags.” In: J Proteome Res
10.6, pp. 2758–2766 
- Florian P. Breitwieser and Jacques Colinge (Sept. 2013). “Isobar(PTM): a software tool for the quantitative analysis of post-translationally modified proteins.” eng.  In: J Proteomics 90, pp. 77–84

For publications citing isobar please see Google scholar for [isobar](https://scholar.google.at/scholar?cites=3978696839477939063&as_sdt=20000005&sciodt=0,21&hl=de) and [isobar PTM](https://scholar.google.at/scholar?cites=988743004087549055&as_sdt=20000005&sciodt=0,21&hl=de).

Many thanks to Alexey Stukalov, Xavier Robin and Florent Gluck who contributed to the development of the package.

## Class: TlsParameter
setClass("TlsParameter",
          representation = representation(df = "numeric",
                                          location = "numeric",
                                          scale = "numeric"),
          prototype = prototype(df = 1, location = 0, scale = 1, name =
                      gettext("Parameter of a Student/T location scale distribution")
                      ),
          contains = "Parameter"
          )

setClass("Tlsd",
          prototype = prototype(
                      r = function(n) { rt(n, df)*scale + location },
                      d = function(x, log = FALSE){ 
                            1/scale * dt((x - location)/scale, df, log = log) },
                      p = function(q, lower.tail = TRUE, log.p = FALSE ){
                            pt((x - location)/scale, df,lower.tail=lower.tail,log.p=log.p)},
                      q = function(q, lower.tail = TRUE, log.p = FALSE ){
                            qt(p, df,lower.tail=lower.tail,log.p=log.p)*scale + location},
                      param = new("TlsParameter"),
                     .logExact = TRUE,
                     .lowerExact = TRUE,
                     Symmetry = new("SphericalSymmetry",
                                     type = "univariate symmetric distribution",
                                     SymmCenter = 0)
                      ),
          contains = "AbscontDistribution"
          )

## TODO: log=TRUE does not work correctly
## Class: TlsDistribution
setMethod("initialize", "Tlsd",
          function(.Object, df = 1, location = 0, scale = 1) {
            .Object@img <- new("Reals")
            .Object@param <- new("TlsParameter", df = df, location = location,
                                  scale = scale)
            .Object@r <- function(n){}
            .Object@d <- function(x, log = FALSE){}
            .Object@p <- function(q, lower.tail = TRUE, log.p = FALSE){}
            .Object@q <- function(p, lower.tail = TRUE, log.p = FALSE){}
            
            rtls = function(n,df,location,scale) rt(n, df)*scale + location 
            dtls = function(x,df,location,scale,log = FALSE)
                     1/scale * dt((x - location)/scale, df, log = log) 
            ptls = function(q,df,location,scale,lower.tail = TRUE, log.p = FALSE )
                     pt((q - location)/scale, df,lower.tail=lower.tail,log.p=log.p) 
                     
            qtls = function(p,df,location,scale, lower.tail = TRUE, log.p = FALSE )
                            qt(p, df,lower.tail=lower.tail,log.p=log.p)*scale + location
                            
            body(.Object@r) <- substitute(
                           { rtls(n, df=dfSub,location=locationSub,scale=scaleSub) },
                             list(dfSub = df,
                                  locationSub = location,
                                  scaleSub = scale)
                                          )
            body(.Object@d) <- substitute(
                           { dtls(x, df=dfSub, location = locationSub,
                                     scale = scaleSub, log = log) },
                             list(dfSub = df,locationSub = location, scaleSub = scale)
                                          )
            body(.Object@p) <- substitute(
                           { ptls(q, df=dfSub, location = locationSub,
                                     scale = scaleSub, lower.tail = lower.tail,
                                     log.p = log.p) },
                             list(dfSub = df,locationSub = location, scaleSub = scale)
                                          )
            body(.Object@q) <- substitute(
                           { qtls(p, df=dfSub, location = locationSub,
                                     scale = scaleSub, lower.tail = lower.tail,
                                     log.p = log.p) },
                             list(dfSub = df,locationSub = location, scaleSub = scale)
                                          )
            .Object@.withSim   <- FALSE
            .Object@.withArith <- FALSE
            .Object@Symmetry <- SphericalSymmetry(location)
            .Object
          })


## Access Methods
setMethod("df", "TlsParameter", function(x, ...) x@df)
setMethod("location", "TlsParameter", function(object) object@location)
setMethod("scale", "TlsParameter",
           function(x, center = TRUE, scale = TRUE) x@scale)
## Replace Methods
setReplaceMethod("df", "TlsParameter",
                  function(object, value){ object@df <- value; object})
setReplaceMethod("location", "TlsParameter",
                  function(object, value){ object@location <- value; object})
setReplaceMethod("scale", "TlsParameter",
                  function(object, value){ object@scale <- value; object})

setValidity("TlsParameter", function(object){
  if(length(df(object)) != 1)
    stop("df has to be a numeric of length 1")
  if(df(object) <= 0)
    stop("df has to be positive")
  # TODO: more tests
  else return(TRUE)
})

################################
##
## Class: Student T Location Scale distribution
##
################################

Tlsd <- function(df = 1, location = 0, scale = 1) new("Tlsd", df = df, location = location, scale = scale)

## wrapped access methods
setMethod("df", "Tlsd", function(x, ...) df(param(x)))
setMethod("location", "Tlsd", function(object) location(param(object)))
setMethod("scale", "Tlsd",
           function(x, center = TRUE, scale = TRUE) scale(param(x)))
           ### odd arg-list due to existing function in base package 


## wrapped replace methods
#setMethod("df<-", "Tlsd",
#           function(object, value) new("Tlsd", df = value, ncp = ncp(object)))

##
## Isobar properties.R file
##     for automatic report generation
## 
## It is standard R code and parsed using sys.source

#####################################################################
## General properties

## Report type: Either 'protein' or 'peptide'
# report.level="peptide"
report.level="protein"
#attr(report.level,"allowed.values") <- c("protein","peptide")

## Isobaric tagging type. Use one of the following:
# type='iTRAQ4plexSpectra'
# type='iTRAQ8plexSpectra'
# type='TMT2plexSpectra'
# type='TMT6plexSpectra'
type=NULL
#attr(type,"allowed.values") <- IBSpectraTypes()

isotope.impurities=NULL
correct.isotope.impurities=TRUE

## Name of project, by default the name of working directory
## Will be title and author of the analysis reports.
name=basename(getwd())
author=paste0("isobar R package v",packageDescription("isobar")$Version)

## specifes the IBSpectra file or object
##   - can be a data.frame (e.g. ibspectra=as.data.frame(ibspiked_set1) )
##   - if it is a character string, it is assumed to be a file
##     - if it ends on .rda, then it is assumed to be a R data object
##     - if it does not exists, then it is may generated based on 
##        the peaklist and identifications properties
ibspectra=paste(name,"ibspectra.csv",sep=".")

## When replicates or 'samples belonging together' are analyzed, a
## ProteinGroup object based on all data should be constructed
## beforehand. This then acts as a template and a subset is used.
protein.group.template=NULL

## Via database or internet connection, informations on proteins (such
## as gene names and length) can be gathered. protein.info.f defines
## the function which takes a ProteinGroup object as argument
protein.info.f=getProteinInfoFromTheInternet

## Where should cached files be saved? Will be created if it does not
## exist
# cachedir="."
cachedir="cache"
## Regenerate cache files? By default, chache files are used.
regen=FALSE

## An ibspectra object can be generated from peaklists and
## identifications.

## peaklist files for quantitation, by default all mgf file in
## directory
peaklist=list.files(pattern="*\\.mgf$")
## id files, by default all id.csv files in directory
identifications=list.files(pattern="*\\.id.csv$")
## mapping files, for data quantified and identified with different but
## correspoding spectra. For example corresponding HCD-CID files.

## masses and intensities which are outside of the 'true' tag mass 
## +/- fragment.precision/2 are discarded
fragment.precision=0.01
## filter mass outliers
fragment.outlier.prob=0.001

## Additional arguments of readIBSpectra can be set here
##  decode.titles should be set to TRUE for Mascot search results
##  as Mascot encodes the spectrum title (e.g. space -> %20)
readIBSpectra.args = list(
    mapping.file=NULL,
    decode.titles=FALSE
)

#####################################################################
## Quantification properties

normalize=TRUE
# if defined, normalize.factors will be used for normalization
normalize.factors=NULL 
normalize.channels=NULL
normalize.use.protein=NULL
normalize.exclude.protein=NULL
normalize.function=median
normalize.na.rm=FALSE

peptide.specificity=REPORTERSPECIFIC

use.na=FALSE

## the parameter noise.model can be either a NoiseModel object or a file name
data(noise.model.hcd)
noise.model=noise.model.hcd
## If it is a file name, a noise model is estimated as non one-to-one
## and saved into the file. otherwise, the noise model is loaded from
## the file
# noise.model="noise.model.rda"

## Define channels for creation of a noise model, ideally a set of 
##  channels which are technical replicates. 
noise.model.channels=NULL

## If noise.model.is.technicalreplicates is FALSE, the intensities
##   are normalized for protein means, creating artifical technical
##   replicates. For this procedure, only proteins with more than
##   noise.model.minspectra are considered.
noise.model.is.technicalreplicates=FALSE
noise.model.minspectra=50

## class labels. Must by of type character and of same length as
## number of channels I. e. 4 for iTRAQ 4plex, 6 for TMT 6plex Example
## for iTRAQ 4plex:
## Class definitions of the isobaric tag channels. 
## A character vector with the same length as channels 
##   (e.g. 4 for iTRAQ 4plex, 6 for TMT 6plex)
## Example for iTRAQ 4plex:
# class.labels=as.character(c(1,0,0,0))
# class.labels=c("Treatment","Treatment","Control","Control")
## Also names are possible - these serves as description in the report
##  and less space is used in the rows
# class.labels=c("Treatment"="T","Treatment"="T","Control"="C","Control"="C")
class.labels=NULL

## The following parameters define which ratios are calculated.

## summarize ratios with equal class labels, set to TRUE when replicates are used
summarize=FALSE

## combn.method defines which ratios are calculated - versus a channel or a class,
##   all the ratios within or across classes, or all possible combinatioins.
## When summarize=TRUE is set, use "interclass", "versus.class", or "intraclass"
# combn.method="global"
# combn.method="versus.class"
# combn.method="intraclass"
# combn.method="interclass"
combn.method="versus.channel"
vs.class=NULL

cmbn=NULL

## Arguments given to 'proteinRatios' function. See ?proteinRatios
ratios.opts = list(
    sign.level.sample=0.05,
    sign.level.rat=0.05,
    groupspecific.if.same.ac=TRUE)

quant.w.grouppeptides=c()

min.detect=NULL

preselected=c()

### Biological Variability Ratio Distribution options
## ratiodistr can be set to a file or a 'Distribution object. ' If
##  NULL, or the specified file is not existent, the biological
##  variability of ratios is estimated on the sample at hand and
##  written to cachedir/ratiodistr.rda or the specified file.
ratiodistr=NULL

## Ideally, when the biological variability is estimated for the
## sample at hand, a biological replicate is present (/ie/ same class
## defined in class labels).  Classes can also be assigned just for
## estimation of the ratio distribution, /eg/ to choose biologically
## very similar samples as pseudo replicates.
ratiodistr.class.labels=NULL

## Function for fitting. Available: fitCauchy, fitTlsd
ratiodistr.fitting.f=fitCauchy

## Use symetrical ratios - i.e. for every ratio r add a ratio -r
##   prior to fitting of a distribution
ratiodistr.symmetry=TRUE

## If defined, use z-score instead of ratio distribution
# zscore.threshold=2.5
zscore.threshold=NULL

####################################################################
## PTM properties

## PhosphoSitePlus dataset which can be used to annotate known
## modification sites. Download site:
## http://www.phosphosite.org/staticDownloads.do
phosphosite.dataset <- NULL

## Modification to track. Use 'PHOS' for phosphorylation.
# ptm <- c('ACET','METH','UBI','SUMO', 'PHOS')
ptm <- NULL

## file name of rda or data.frame with known modification sites
## gathered with ptm.info.f.  defaults to 'cachedir/ptm.info.rda'
ptm.info <- NULL

## Function to get PTM modification sites from public datasets
# ptm.info.f <- getPtmInfoFromNextprot
# ptm.info.f <- function(...)
#    getPtmInfoFromPhosphoSitePlus(...,modification="PHOS")
# ptm.info.f <- function(...)
#    getPtmInfoFromPhosphoSitePlus(...,modification=ptm)
ptm.info.f <- getPtmInfoFromNextprot

## A protein quantification data.frame (generated with
##  'proteinRatios').  The ratio and variance are used to correct the
##  observed modified peptide ratios Needs to have the experimental
##  setup as the modified peptide experiment
correct.peptide.ratios.with <- NULL
## Protein groups to use with correct.peptide.ratios()
correct.peptide.ratios.with_protein.group <- NULL

## The correlation between peptide and protein ratios defines the
## covariance

##  Var(ratio m) = Var(ratio mp) + Var(ratio p)
##                                    + 2 * Cov(ratio mp, ratio p),
##  Cov(ratio mp, ratio p) = 2 * cor * Sd(ratio mp) * Sd(ratio p),
##  with m = modifcation, mp = modified peptide, p = protein
peptide.protein.correlation <- 0

## quantification table whose columns are attached to the XLS
## quantification table
compare.to.quant <- NULL

#####################################################################
## Report properties 

write.qc.report=TRUE
write.report=TRUE
write.xls.report=TRUE

## Use name for report, ie NAME.quant.xlsx instead of
## isobar-analysis.xlsx
use.name.for.report=TRUE

## PDF Analysis report sections: Significant proteins and protein
## details
show.significant.proteins=FALSE
show.protein.details=TRUE

### QC REPORT OPTIONS ###
#qc.maplot.pairs=FALSE # plot one MA plot per tag (versus all others)
qc.maplot.pairs=TRUE # plot MA plot of each tag versus each tag

### XLS REPORT OPTIONS ###
## Spreadsheet format: Either 'xlsx' or 'xls'
# spreadsheet.format="xlsx"
spreadsheet.format="xlsx"

## XLS report format 'wide' or 'long '.

## 'wide' format outputs ratios in separate columns of the same record
## (i.e. one line per protein)
## 'long' format outputs ratios in separate records (i.e. one line per
## ratio)
# xls.report.format="wide"
xls.report.format="long"

## XLS report columns in quantification tab
##  possible values: ratio, is.significant, CI95.lower, CI95.upper, 
##                   ratio.minus.sd, ratio.plus.sd,
##                   p.value.ratio, p.value.sample, n.na1, n.na2, 
##                   log10.ratio, log10.variance,
##                   log2.ratio, log2.variance
##  only for summarize=TRUE: n.pos, n.neg
xls.report.columns <- c("ratio","is.significant","ratio.minus.sd",
                        "ratio.plus.sd","p.value.ratio","p.value.sample",
                        "log10.ratio","log10.variance")

## Perl command to be used for Excel report generation
# perl.cmd = "C:/Strawberry/perl/bin/perl.exe"
# perl.cmd = "C:/Perl/bin/perl.exe"
# perl.cmd = "perl5"
perl.cmd = "perl"

#####################################################################
## Etc

sum.intensities=FALSE

datbase="Uniprot"

scratch=list()

##
#  compile LaTeX reports into PDF files
compile=TRUE

# zip final report files into archive
zip=FALSE

# warning level (see 'warn' in ?options)
warning.level=1


###########################################################
## Novel options

shrink.mean=TRUE
use.t.stat=TRUE

#!/bin/Rscript

{
print_help_and_exit <- function() {
  cat("create_reports.R       Command line tool to create reports using the\n",
      "                       isobar R/Bioconductor package.\n",
      "\n",
      "Usage:\n",
      "  Rscripts create_reports.R -- 'type=\"iTRAQ4plexSpectra\"'\n",
      "\n",
      "  create_reports.R expects all files to be present in the working directory.\n",
      "  Options can be specified on command line or in a properties.conf file\n",
      "  in the working directory.",
      "\n",
      "Options:\n",
      "  Options defined on the command line overwrite the local properties file\n",
      "  overwrite the global properties file. For reference of possible options\n",
      "  take a look at the file ",
      system.file("report","properties.R",package="isobar"),".\n");
  quit("no");
}

args <- commandArgs(TRUE)

if ("--help" %in% args || "-h" %in% args) {
  print_help_and_exit();
} 

get.arg <- function(name) {
  if (name %in% args) {
    args <<- setdiff(args,name)
    return(TRUE)
  } else {
    return(FALSE)
  }
}

meta.report <- get.arg("--meta")

if (meta.report) 
  properties.file <- "meta-properties.R"
else 
  properties.file <- "properties.R"

if (length(args) > 0 && file.exists(args[length(args)])) {
  properties.file <- args[length(args)]
  args[length(args)] <- NULL
}

xls.report <- get.arg("--xls")
xlsx.report <- get.arg("--xlsx")
qc.report <- get.arg("--qc")
pdf.report <- get.arg("--pdf")

## TODO: parse further arguments

message("started at ",date())
message("Loading package isobar v",packageDescription("isobar")$Version," ...")
suppressPackageStartupMessages(library(isobar))

create.reports.args <- list(properties.file=properties.file,
                            args=args)

if (xls.report || xlsx.report || qc.report || pdf.report) {
  create.reports.args$write.xls.report <- (xls.report || xlsx.report)
  if (create.reports.args$write.xls.report) {
    if (xls.report) create.reports.args$spreadsheet.format <- 'xls'
    if (xlsx.report) create.reports.args$spreadsheet.format <- 'xlsx'
  }

  create.reports.args$write.qc.report <- qc.report
  create.reports.args$write.report <- pdf.report
  do.compile <- TRUE
}


tryCatch({
  if (meta.report) create.reports.f <- create.meta.reports
  else create.reports.f <- create.reports

  do.call(create.reports.f,create.reports.args)},
         error=function(e) {
           save.image(file="isobar.fail.rda")
           stop("create.reports exited with an error - saving session to isobar.fail.rda.\n\n  Message: ",
                 as.character(e))
         }
         )
message("finished at ",date())
}

# ==========================================================================
# isobar package initialization
# ==========================================================================

#opts_isobar <- list(verbose=FALSE)

.onAttach <- function(libname, pkgname) {
  packageStartupMessage(paste("Welcome to isobar (v ",
                              packageDescription("isobar")$Version,")\n",
                              "   'openVignette(\"isobar\")' and '?isobar' provide help on usage.\n", sep=""))
  addVigs2WinMenu("isobar") 
}

### R code from vignette source 'isobar-ptm.Rnw'

###################################################
### code chunk number 1: init
###################################################
  require(ggplot2)
  dir.create(file.path("graphics"), showWarnings = FALSE)


###################################################
### code chunk number 2: load-isobar
###################################################
  library(isobar) ## load the isobar package


###################################################
### code chunk number 3: isobar-ptm.Rnw:81-87 (eval = FALSE)
###################################################
## # Generate PhosphoRS XML input file based on MGF and identification file
## #  massTolerance: fragment ion mass tolerance (in Da)
## #  activationType: CID, HCD, or ETD
## writePhosphoRSInput("phosphors.in.xml",
##                       "identifications.id.csv","peaklist.mgf",
##                        massTolerance=0.5,activationType="CID")


###################################################
### code chunk number 4: isobar-ptm.Rnw:94-98 (eval = FALSE)
###################################################
## # Read PhosphoRS XML output file
## #   simplify: if TRUE, a data.frame is returned, else a list
## #   besthit.only: if TRUE, only the best localization per spectrum is returned
## readPhosphoRSOutput("phosphors.out.xml",simplify=TRUE,besthit.only=TRUE)


###################################################
### code chunk number 5: isobar-ptm.Rnw:104-107 (eval = FALSE)
###################################################
## getPhosphoRSProbabilities("identifications.id.csv","peaklist.mgf",
##                             massTolerance=0.5,activationType="CID",
##                             phosphors.cmd="java -jar phosphoRS.jar")


###################################################
### code chunk number 6: isobar-ptm.Rnw:125-140 (eval = FALSE)
###################################################
## # filterSpectraDeltaScore calls calc.delta.score 
## #   if no column named delta.score is present in the data frame
## # identifications below a min.delta.score are REMOVED
## ib <- readIBSpectra("identifications.id.csv","peaklist.mgf",
##                       annotate.spectra.f=function(...) 
##                         filterSpectraDeltaScore(...,min.delta.score=10))
## 
## 
## # filterSpectraPhosphoRS calls PhosphoRS to caluclate PhosphoRS probabilities
## # identifications below a min.prob (PhosphoRS peptide isoform probability) 
## # are marked to be NOT QUANTIFIED (use.for.quant=FALSE), but not removed
## ib <- readIBSpectra("identifications.id.csv","peaklist.mgf",
##                       annotate.spectra.f=
##                         function(...) filterSpectraPhosphoRS(...,min.prob=0.9,
##                            phosphors.cmd="java -jar PhosphoRS.jar"))


###################################################
### code chunk number 7: isobar-ptm.Rnw:161-165
###################################################
data(ib_phospho)
data(noise.model.hcd)
head(proteinGroup(ib_phospho)@peptideInfo)
10^estimateRatio(ib_phospho,noise.model.hcd,peptide="SPLSPTETFSWPDVR")


###################################################
### code chunk number 8: isobar-ptm.Rnw:170-176
###################################################
pep.n.modif <- unique(apply(fData(ib_phospho)[,c("peptide","modif")],2,cbind))
print(head(pep.n.modif))
estimateRatio(ib_phospho,noise.model.hcd,channel1="114",channel2="115",
                peptide=head(pep.n.modif),combine=FALSE)[,c("lratio","variance",
                                                            "n.spectra","p.value.rat")]



###################################################
### code chunk number 9: ratiodistr
###################################################
suppressPackageStartupMessages(library(distr))
suppressPackageStartupMessages(library(ggplot2))
peptide.ratios <- peptideRatios(ib_phospho,noise.model=noise.model.hcd,
                                  cmbn=matrix(c("114","116"),ncol=1))

lim <- max(abs(peptide.ratios$lratio),na.rm=TRUE)
peptide.distr.cauchy <- fitCauchy(peptide.ratios$lratio)

pseq <- seq(from=-lim,to=lim,length.out=1000)
ggplot() + 
  geom_histogram(aes(x=lratio,y=..density..),data=peptide.ratios,binwidth=0.05,
                 color="darkgreen",fill="white") +
  geom_line(aes(x=x,y=y),color="black",
            data=data.frame(x=pseq,y=d(peptide.distr.cauchy)(pseq)))


###################################################
### code chunk number 10: isobar-ptm.Rnw:233-243
###################################################
peptides <- pep.n.modif[1:5,]

orig.ratio <- estimateRatio(ib_phospho,noise.model.hcd,channel1="114",channel2="115",
                              peptide=peptides,combine=FALSE)[,c("lratio","variance")]

peptides.c <- cbind(peptides,correct.ratio=c(0,-1,1,2,-2))
corr.ratio <- estimateRatio(ib_phospho,noise.model.hcd,channel1="114",channel2="115",
                              peptide=peptides.c,combine=FALSE)[,c("lratio","variance")]

data.frame(peptides.c,orig.ratio,corr.ratio)


###################################################
### code chunk number 11: isobar-ptm.Rnw:264-266 (eval = FALSE)
###################################################
## ptm.info <- getPtmInfoFromPhosphoSitePlus(proteinGroup(ib_phospho),modif="PHOS")
## ptm.info <- getPtmInfoFromNextprot(proteinGroup(ib_phospho))


###################################################
### code chunk number 12: isobar-ptm.Rnw:268-269
###################################################
head(ptm.info)

#!/usr/bin/Rscript
suppressPackageStartupMessages(library(gplots))
suppressPackageStartupMessages(library(boot))
suppressPackageStartupMessages(library(RColorBrewer))
suppressPackageStartupMessages(library(isobar))

source("meta-properties.R")
load("protein.group.rda")
proteinInfo(protein.group) <- getProteinInfoFromBiomart(protein.group)

if (!exists("cols"))
  cols <- c()
cols <- unique(c("ac","r1","r2","lratio","variance",cols))

message("Calculating dNSAF ...")
if (file.exists("dnsaf.rda")) {
  load("dnsaf.rda")
} else {
  dnsaf <- calculate.dNSAF(protein.group)
  save(dnsaf,file="dnsaf.rda")
}
names(dnsaf) <- reporterProteins(protein.group)

message("Merging tables ...")
merged.table <- get.merged.table(samples,cols=cols)
merged.table <- subset(merged.table,r1==merged.table$r1[1])

tbl.wide <- reshape(merged.table,idvar="ac",timevar=c("r2"),direction="wide",drop="r1")
rownames(tbl.wide) <- tbl.wide$ac
all.names <- do.call(rbind,lapply(tbl.wide[,"ac"],get.names,protein.group=protein.group))
tbl.wide$dNSAF <- dnsaf[as.character(tbl.wide$ac)]

summarized.table <- write.summarized.table(tbl.wide,all.names,cols=unique(merged.table$r2))

ratio.matrix <- as.matrix(tbl.wide[,grep("lratio",colnames(tbl.wide))])
variance.matrix <- as.matrix(tbl.wide[,grep("var",colnames(tbl.wide))])
sel <- !apply(is.na(ratio.matrix),1,any)
ratio.matrix <- ratio.matrix[sel,]
variance.matrix <- variance.matrix[sel,]

m.median <- apply(ratio.matrix,2,median)
normalized.ratio.matrix <- ratio.matrix-matrix(m.median,nrow=nrow(ratio.matrix),ncol=ncol(ratio.matrix),byrow=T)

plot.heatmaps(ratio.matrix)
plot.pairs()

# returns a list of all combinations of a group reporter and a group member
# where there are peptides which are only shared between reporter and member
.proteins.w.shared.peptides <- function(protein.group) {
  proteins.w.shared.pep = lapply(reporterProteins(protein.group), function(reporter.protein.g) {    
    gmp <- groupMemberPeptides(protein.group,reporter.protein.g,TRUE)           
    reporter.sp.sel <- gmp$peptide.info$specificity == REPORTERSPECIFIC      
    quant.sel <- gmp$peptide.info$n.shared.groups == 1 &                        
                 gmp$peptide.info$n.shared.proteins == 2                        
    gd.proteins <- apply(gmp$group.member.peptides[quant.sel,,drop=FALSE],2,any)         
                                                                                
    return(gd.proteins)                                                         
                                                                                
  })
  names(proteins.w.shared.pep) <- reporterProteins(protein.group)
  do.call(sum,proteins.w.shared.pep)

  proteins.w.shared.pep.dfl <- lapply(seq_len(length(proteins.w.shared.pep)), function(i) {
    gd.proteins <-proteins.w.shared.pep[[i]]
    if (any(gd.proteins))
      return(data.frame(reporter.protein.g=names(proteins.w.shared.pep)[i],
                        protein.g=names(gd.proteins)[gd.proteins]))
  })

  # remove entries w/o such cases
  proteins.w.shared.pep.dfl <- proteins.w.shared.pep.dfl[!sapply(proteins.w.shared.pep.dfl,is.null)]

  proteins.w.shared.pep.df <- do.call(rbind,proteins.w.shared.pep.dfl)
  proteins.w.shared.pep.df$reporter.protein.g <- as.character(proteins.w.shared.pep.df$reporter.protein.g)
  proteins.w.shared.pep.df$protein.g <- as.character(proteins.w.shared.pep.df$protein.g)

  return(proteins.w.shared.pep.df)
}

shared.ratios <- function(ibspectra,noise.model,channel1=NULL,channel2=NULL,protein=reporterProteins(proteinGroup(ibspectra)),...) {
  protein.group <- proteinGroup(ibspectra)
  l <- lapply(protein, function(reporter.protein.g) {    
    gmp <- groupMemberPeptides(protein.group,reporter.protein.g,TRUE)           
    reporter.sp.sel <- gmp$peptide.info$specificity == REPORTERSPECIFIC      
    quant.sel <- gmp$peptide.info$n.shared.groups == 1 &                        
                 gmp$peptide.info$n.shared.proteins == 2                        
    gd.proteins <- apply(gmp$group.member.peptides[quant.sel,,drop=FALSE],2,any)
    gd.proteins <- names(gd.proteins)[gd.proteins]
    gd.proteins <- gd.proteins[gd.proteins != reporter.protein.g]

    if (length(gd.proteins) > 0) {
      ratio.1 <- estimateRatio(ibspectra,noise.model=noise.model,               
                               protein=reporter.protein.g,channel1=channel1,      
                               channel2=channel2,...)
      if (is.na(ratio.1['lratio']))
        return(NULL)
      
      quant.df <- data.frame(reporter.protein=reporter.protein.g,
                             protein2=gd.proteins,
                             ratio1=ratio.1['lratio'],ratio1.var=ratio.1['variance'],
                             n.spectra.1=ratio.1['n.spectra'],
                             ratio2=NA,ratio2.var=NA,n.spectra.2=NA,                             
                             stringsAsFactors=FALSE)
      reporter.pep <- gmp$peptide.info$peptide[reporter.sp.sel]

      for (protein.i in seq_along(gd.proteins)) {
        protein.gd <- quant.df[protein.i,"protein2"]
        protein.pep <-  rownames(gmp$group.member.peptides)[gmp$group.member.peptides[,protein.gd]]
        quant.pep <- gmp$peptide.info[ quant.sel &
                                        gmp$peptide.info$peptide %in% protein.pep,
                                      "peptide"]
        ratio.2 <- estimateRatio(ibspectra,noise.model=noise.model,protein=NULL,
                                 peptide=quant.pep, channel1=channel1,channel2=channel2,combine=T,...)
        quant.df[protein.i,"ratio2"] <- ratio.2['lratio']
        quant.df[protein.i,"ratio2.var"] <- ratio.2['variance']
        quant.df[protein.i,"n.spectra.2"] <- ratio.2['n.spectra']
      }
      return(quant.df)
    } else {
      return(NULL)
    }                                                                                
  })
  l <- l[!sapply(l,is.null)]
  do.call(rbind,l)
}

shared.ratios.sign <- function(ress,z.shared,min.spectra=1,plot=TRUE) {
    ratio1.issmaller <- ress$ratio1+z.shared*sqrt(ress$ratio1.var) < ress$ratio2 - z.shared*(sqrt(ress$ratio2.var)) 
    ratio1.isbigger  <- ress$ratio1-z.shared*sqrt(ress$ratio1.var) > ress$ratio2 + z.shared*(sqrt(ress$ratio2.var))
 
    rr <- ress[(ratio1.issmaller | ratio1.isbigger) &
               !is.na(ratio1.issmaller) & !is.na(ratio1.isbigger) & 
               ress$n.spectra.1 >= min.spectra & ress$n.spectra.2 >= min.spectra,]

    rr<-rr[order(rr$ratio1,decreasing=T),]
    
    rr$proteins <- as.character(paste(rr$reporter.protein,"\nvs",rr$protein2))
    
    rr$n.spectra.txt <-"> 10"
    rr$n.spectra.txt[rr$n.spectra<=10]<-"6 - 10"
    rr$n.spectra.txt[rr$n.spectra<=5]<-"2 - 5"
    rr$n.spectra.txt[rr$n.spectra==1]<-"1"

    rr$n.spectra.txt.1 <-"> 10"
    rr$n.spectra.txt.1[rr$n.spectra.1<=10]<-"6 - 10"
    rr$n.spectra.txt.1[rr$n.spectra.1<=5]<-"2 - 5"
    rr$n.spectra.txt.1[rr$n.spectra.1==1]<-"1"
    xx=reshape(rr,list(c("ratio1","ratio2"),c("ratio1.var","ratio2.var"),
                       c("n.spectra.txt","n.spectra.txt.1")),
               direction="long",v.names=c("ratio","var","n.spectra"),timevar="g")
    xx$proteins=factor(as.character(xx$proteins),levels=rr$proteins,labels=rr$proteins)
#    xx$g <- as.character(xx$g)
#    xx$g[xx$g==1] <-"reporter"
#    xx$g[xx$g==2] <-"member"
    xx$g <- factor(xx$g,levels=c(1,2),labels=c("reporter","member"))
    if (min.spectra == 2)
      xx$n.spectra <- factor(as.character(xx$n.spectra),levels=c("2 - 5","6 - 10","> 10"))
    else
      xx$n.spectra <- factor(as.character(xx$n.spectra),levels=c("1","2 - 5","6 - 10","> 10"))

  if (plot) {
    breaks <- c(0.1,0.25,0.5,1,2,3,4,5,10,20,30,40,50)
    breaks <- breaks[log10(breaks) %inrange% range(xx$ratio)]

    print(ggplot(xx,aes_string(x="ratio",y="proteins")) +
          geom_vline(xintercept=0,alpha=0.5) + 
          geom_point(aes_string(colour="g",shape="g"),size=4) + 
          geom_errorbarh(aes(xmax=ratio+sqrt(var),xmin=ratio-sqrt(var),colour=factor(g),height=0.2)) + 
          scale_x_continuous("Ratio",breaks=log10(breaks),labels=breaks) +
          scale_colour_manual("group",values = c("blue","darkgreen")) + 
          scale_shape("group") + 
          scale_size(guide="none"))
  }
  xx
}

setAs("MSnSet","IBSpectra",function(from) {
  ions <- exprs(from)
  q <- qual(from)
  lowerMz <- matrix(q$lowerMz,byrow=TRUE,ncol=ncol(ions))
  upperMz <- matrix(q$upperMz,byrow=TRUE,ncol=ncol(ions))
  maxInt <- matrix(q$maxInt,byrow=TRUE,ncol=ncol(ions))
  
  data <- fData(from)[,c("ProteinAccession","PeptideSequence","charge","precursor.mz","retention.time","spectrum")]
  colnames(data) <- c("accession","peptide","charge","exp.mass","retention.time","spectrum")

  my.class <- switch(colnames(ions)[1],
                     iTRAQ4.114 = "iTRAQ4plexSpectra",
                     iTRAQ8.113 = "iTRAQ8plexSpectra",
                     TMT2.126 = "TMT2plexSpectra",
                     TMT6.126 = "TMT6plexSpectra",
                     TMT10.126 = "TMT10plexSpectra",
                     "unknown")
  if (identical(my.class,"unknown"))
    stop("I do not know how to map MSnSet w/ columns [",paste(colnames(ions),collapse=","),"] to an IBSpectra object.")
         
  o <- new(my.class)
  rownames(ions) <- data$spectrum
  colnames(ions) <- o@reporterTagNames
  mass <- 0.5*(lowerMz+upperMz)
  rownames(mass) <- data$spectrum
  colnames(mass) <- o@reporterTagNames
  assayDataElements <- list(lowerMz=lowerMz,upperMz=upperMz,maxInt=maxInt)
  for (elem in names(assayDataElements)) {
    rownames(assayDataElements[[elem]]) <- data$spectrum
    colnames(assayDataElements[[elem]]) <- o@reporterTagNames
  }

  ib <- new(my.class,data,ions,mass,
            assayDataElements=assayDataElements)
  if (validObject(ib))
    return(ib)
})

setAs("IBSpectra","MSnSet",function(from) {
  requireNamespace("MSnbase")
  ## based on quantify.MSnExp from MSnbase
  
  elems <- assayDataElementNames(from)
  exprs <- reporterIntensities(from)

  get.elem <- function(x,y) {
    if (is.element(x,elems)) {
      return(assayDataElement(from,x))
    } else {
      if (is.matrix(y)) return(y)
      else return(assayDataElement(from,y))
    }
  }

  if (is(from,"iTRAQ4plexSpectra")) {
    reporters <- MSnbase::iTRAQ4
  } else if (is(from,"iTRAQ8plexSpectra")) {
    reporters <- MSnbase::iTRAQ8
  } else if (is(from,"TMT6plexSpectra")) {
    reporters <- MSnbase::TMT6
  } else if (is(from,"TMT10plexSpectra")) {
    stop("conversion for TMT10plexSpectra not supported yet")
    #reporters <- TMT10
  } else {
    stop("Cannot convert object")
  }
  
  lowerMz <- get.elem("lowerMz","mass")
  upperMz <- get.elem("upperMz","mass")
  maxInt  <- get.elem("maxInt",matrix(NA,nrow=nrow(lowerMz),ncol=ncol(lowerMz)))
  nMaxInt  <- get.elem("nMaxInt",matrix(1,nrow=nrow(lowerMz),ncol=ncol(lowerMz)))
  baseLength  <- get.elem("baseLength",matrix(NA,nrow=nrow(lowerMz),ncol=ncol(lowerMz)))

  .qual <- data.frame(maxInt=as.numeric(t(maxInt)),
                  nMaxInt=as.numeric(t(nMaxInt)),
                  baseLength=as.numeric(t(baseLength)),
                  lowerMz=as.numeric(t(lowerMz)),
                  upperMz=as.numeric(t(upperMz)),
                  reporter=mz(o),
                  precursor=rep(fData(from)$exp.mass,each=length(mz(o))))

  .exprs <- reporterIntensities(from)
  colnames(.exprs) <- reporters@reporterTagNames
  mapping <- c(spectrum="spectrum",
               ProteinAccession="accessions",
               ProteinDescription="NA",
               PeptideSequence="peptide",
               index="NA",
               file="NA",
               retention.time="retention.time",
               precursior.mz="exp.mass",
               peaks.count="NA",
               tic="NA",
               ms.level="NA",
               charge="charge",
               collision.energy="NA")
  df <- ibSpectra.as.concise.data.frame(from)
  df[,"NA"] <- NA
  
  fd <- df[,mapping]
  rownames(fd) <- fd[,"spectrum"]
  colnames(fd) <- names(mapping)
  .featureData <- new("AnnotatedDataFrame",data=fd[rownames(.exprs),])
  .phenoData <- new("AnnotatedDataFrame",
                    data=data.frame(mz=reporters@mz,
                      reporters=reporters@name,
                      row.names=reporters@reporterTagNames))
  
  msnset <- new("MSnSet",
                qual=.qual,
                exprs=.exprs,
                phenoData=.phenoData,
                featureData=.featureData,
                processingData=new("MSnProcess",processing=paste("created from IBSpectra object:",date())),
                annotation="No annotation")
  
  if (validObject(msnset))
    return(msnset)
})

getPhosphoRSProbabilities <- function(
  id.file,mgf.file,massTolerance,activationType,simplify=FALSE,
  mapping.file=NULL,mapping=c(peaklist="even",id="odd"),pepmodif.sep="##.##",besthit.only=TRUE,
  phosphors.cmd=paste("java -jar",system.file("phosphors","phosphoRS.jar",package="isobar")),
  file.basename=tempfile("phosphors.")) {

  infile <- paste0(file.basename,".in.xml")
  outfile <- paste0(file.basename,".out.xml")

  writePhosphoRSInput(infile,
                      id.file,mgf.file,massTolerance,activationType,
                      mapping.file,mapping,pepmodif.sep)
  
  system(paste(phosphors.cmd,shQuote(infile),shQuote(outfile)))
  readPhosphoRSOutput(outfile,simplify=simplify,pepmodif.sep=pepmodif.sep,besthit.only=besthit.only)
}

.iTRAQ.mass = c(monoisotopic = 144.102063, average = 144.1544)
.iTRAQ8.mass = c(monoisotopic = 304.2, average = 304.308)
.CysCAM.mass = c(monoisotopic = 57.021464, average =  57.0513)
.OxidationM.mass = c(monoisotopic = 15.994915, average =  15.9994)
.TMT6.mass = c(monoisotopic = 229.162932, average = 229.2634)
  
writePhosphoRSInput <- 
  function(phosphoRS.infile,id.file,mgf.file,massTolerance,activationType,
           mapping.file=NULL,mapping=c(peaklist="even",id="odd"),pepmodif.sep="##.##",
           modif.masses=
           rbind(c("PHOS",       "1","1:Phospho:Phospho:79.966331:PhosphoLoss:97.976896:STY"),
                 c("Oxidation_M","2","2:Oxidation:Oxidation:15.994919:null:0:M"),
                 c("Cys_CAM",    "3","3:Carbamidomethylation:Carbamidomethylation:57.021464:null:0:C"),
                 c("iTRAQ4plex", "4","4:iTRAQ4:iTRAQ4:144.1544:null:0:KX"),
                 c("iTRAQ8plex", "5","5:iTRAQ8:iTRAQ8:304.308:null:0:KX"),
                 c("TMT6plex",   "7","7:TMT6:TMT6:229.162932:null:0:KX"),
                 c("TMTsixplex",   "6","6:TMT6:TMT6:229.162932:null:0:KX"))) {

  if (is.data.frame(id.file)) 
    ids <- id.file
  else
    ids <- isobar:::.read.idfile(id.file,id.format="ibspectra.csv",log=NULL)

  ids <- unique(ids[,c("peptide","modif","spectrum")])
  # data[,SC['PEPTIDE']] <- gsub("I","L",data[,SC['PEPTIDE']])

  
  if (!is.null(mapping.file)) {
    mapping.quant2id <- do.call(rbind,lapply(mapping.file,function(f) {
      read.table(f,header=TRUE,sep=',',stringsAsFactors=FALSE)
    }))
    cn <-  colnames(mapping.quant2id)
    if (!all(mapping %in% cn))
      stop("mapping not correct")
    
    colnames(mapping.quant2id)[cn == mapping['id']] <- 'id'
    colnames(mapping.quant2id)[cn == mapping['peaklist']] <- 'peaklist'  
  }
  
  con.out <- file(phosphoRS.infile,'w')
  cat.f <- function(...) cat(...,"\n",file=con.out,sep="",append=TRUE)
  cat.f("<phosphoRSInput>")
  cat.f("  <MassTolerance Value='",0.02,"' />")
  cat.f("  <Phosphorylation Symbol='1' />")
  cat.f("  <Spectra>")

  input <- c()
  for (f in mgf.file) {
    con <- file(f,'r')
    input <- c(input,readLines(con))
    close(con)
  }

  begin_ions <- which(input=="BEGIN IONS")+1
  end_ions <- which(input=="END IONS")-1
  titles <- gsub("TITLE=","",grep("TITLE",input,value=TRUE),fixed=TRUE)
  if (!all(ids$spectrum %in% titles))
    stop("Not all id spectrum titles are in MGF titles!\n",
         .sum.bool.c(ids$spectrum %in% titles))
  
  if (length(begin_ions) != length(end_ions))
    stop("mgf file is errorneous, non-matching number",
         " of BEGIN IONS and END IONS tags");

  modif <- "PHOS"
  ids$modifrs <- .convertModifToPhosphoRS(ids$modif,modif.masses)

  pepid <- 0
  for (title in unique(ids[grep(modif,ids$modif),"spectrum"])) {

    spectrum_i <- which(titles==title)
    spectrum <- input[begin_ions[spectrum_i]:end_ions[spectrum_i]]
    ## read header
    header <- .strsplit_vector(spectrum[grep("^[A-Z]",spectrum)],"=")
    peaks <- gsub(" ?","",spectrum[grep("^[0-9]",spectrum)],fixed=TRUE)

    if (length(peaks) > 0) {
      cat.f("    <Spectrum ID='",URLencode(header["TITLE"],reserved=TRUE),"'",
            " PrecursorCharge='",sub("+","",header["CHARGE"],fixed=TRUE),"'",
            " ActivationTypes='",activationType,"'>")
    
      cat.f("    <Peaks>",paste(gsub("\\s+",":",peaks),collapse=","),"</Peaks>")

      for (id_i in which(ids$spectrum==title)) {
        pepid <- pepid + 1
        cat.f("      <IdentifiedPhosphorPeptides>")
        cat.f("        <Peptide ID='",ids[id_i,"peptide"],pepmodif.sep,ids[id_i,"modif"],"'",
              " Sequence='",ids[id_i,"peptide"],"'",
              " ModificationInfo='",ids[id_i,"modifrs"],"' />")
        cat.f("      </IdentifiedPhosphorPeptides>")
      }
      cat.f("    </Spectrum>")
    }
  }

  cat.f("  </Spectra>")
  cat.f("  <ModificationInfos>")
  for (i in seq_len(nrow(modif.masses))) {
    cat.f("    <ModificationInfo Symbol='",modif.masses[i,2],"' Value='",modif.masses[i,3],"' />")
  }
  cat.f("  </ModificationInfos>")
  cat.f("</phosphoRSInput>")
  close(con.out)
}

calc.delta.score <- function(my.data) {
  pep.n.prot <- unique(my.data[,c("accession","peptide","start.pos")])
  my.data$accession <- NULL
  my.data$start.pos <- NULL
  my.data <- unique(my.data)
  if (!any(by(my.data$score,my.data$spectrum, length)>1)) {
    stop("Cannot calculate delta score: Only one hit per spectrum available")
  }

  my.data$delta.score <- my.data$score
  my.data$n.pep <- 1
  my.data$n.loc <- 1

  res <- ddply(my.data,"spectrum",function(x) {
    if (nrow(x) == 1) return(x);
    res <- x[which.max(x$score),,drop=FALSE]
    res$n.pep <- length(unique(x$peptide))
    x <- x[-which.max(x$score),] # remove best hit from x
    res$delta.score <- res[,'score'] - x[which.max(x$score),'score'] # calc delta score w/ max
    res$delta.score.pep <- res$delta.score
    x <- x[x$peptide == res[,'peptide',],] # only keep same peptide hits in x
    if (nrow(x) == 0) return(res);
    res$delta.score.pep <- res[,'score'] - x[which.max(x$score),'score'] # calc delta score w/ max (same pep)
    res$n.loc <- nrow(x) + 1
    return(res);
  })

  my.data <- merge(pep.n.prot,res,by="peptide",all.y=TRUE)

  return(my.data[order(my.data[,"accession"],my.data[,"peptide"]),])
}

calc.pep.delta.score <- function(y,spectrum.col='spectrum',score.col='score',peptide.col='peptide') {
  y$delta.score <- y$score
  y$delta.score.pep <- y$score
  y$delta.score.notpep <- y$score

  y$n.pep <- 1
  y$n.loc <- 1
    
  ddply(y,"spectrum",function(x) {
    if (nrow(x) == 1) return(x);
    res <- x[which.max(x$score),,drop=FALSE]
    res$n.pep <- length(unique(x[,peptide.col]))
    x <- x[-which.max(x$score),] # remove best hit from x
    res$delta.score <- res[,'score'] - x[which.max(x$score),'score'] # calc delta score w/ max
    res$delta.score.pep <- res$delta.score
    y <- x[x$peptide == res[,'peptide'],] # only keep same peptide hits in x
    if (nrow(y) > 0) {
      res$delta.score.pep <- res[,'score'] - y[which.max(y$score),'score'] # calc delta score w/ max (same pep)
      res$n.loc <- nrow(x) + 1
    }
    y <- x[x$peptide != res[,'peptide'],] # only keep different peptide hits
    if (nrow(y) > 0) {
      res$delta.score.notpep <- res[,'score'] - y[which.max(y$score),'score'] # calc delta score w/ max (different pep)
    }
    return(res);
  })
}


filterSpectraDeltaScore <- function(my.data, min.delta.score=10, do.remove=FALSE) {
  if (!"delta.score" %in% colnames(my.data))
    my.data <- calc.delta.score(my.data)
  
  if (!is.null(min.delta.score)) {
    sel.mindeltascore <- my.data[,"delta.score"] >= min.delta.score
    my.data[,"use.for.quant"] <- my.data[,"use.for.quant"] & sel.mindeltascore
    if (isTRUE(do.remove))
      my.data <- my.data[,sel.mindeltascore]
  }
  return(my.data)
}


.getModifOnPosition <- function(modifstring,pos=NULL) {
  splitmodif <- strsplit(paste0(modifstring," "),":")
  sapply(seq_along(splitmodif),function(i) {
    x <- splitmodif[[i]]
    x[length(x)] <- sub(" $","",x[length(x)])
    if (!is.null(pos))
      if (length(pos)==1)
        x[pos+1]
      else
        x[pos[i]+1]
    else
      x[seq(from=2,to=length(x)-1)]
  })
}

.convertModifToPhosphoRS <- function(modifstring,modifs) {
  sapply(strsplit(paste0(modifstring," "),":"),function(x) {
    x[length(x)] <- sub(" $","",x[length(x)])
    xx <- x
    xx[x==""] <- 0
    xx[x!=""] <- NA
    for (i in seq_len(nrow(modifs))) 
      xx[grep(paste0("^",modifs[i,1]),x)] <- modifs[i,2]

    if(any(is.na(xx))) stop("Could not convert [",paste0(x[is.na(xx)],collapse=" and "),"] modifstring ",modifstring)

    y <- c(xx[1],".",xx[2:(length(xx)-1)],".",xx[length(xx)]);
    paste(y,collapse="") })
}

writeHscoreData <- function(outfile,ids,massfile="defs.txt") {
  # command line call: [python Hscorer.py --myDir .  --quantmeth itraq --massfile defs.txt]
  modif.masses <- read.delim(text=grep("\t",readLines(massfile),value=TRUE),
                       stringsAsFactors=FALSE)
  modif.masses <- as.matrix(modif.masses[,c(3,1)])
  modif.masses <- modif.masses[nchar(modif.masses[,1]) > 1,]
  modifstring <- gsub(".","",.convertModifToPhosphoRS(ids[,'modif'],modif.masses),fixed=TRUE)

  write.table(cbind(ids[,'spectrum'],ids[,'peptide'],modifstring),file=outfile,
              col.names=FALSE,row.names=FALSE,sep="\t",quote=FALSE)
}

## TODO:
#.convertModifToPosAndName <- function(modifstring,modif=c(p="PHOS",o="Oxidation_M",c="Cys_CAM",
#                                          me="METH_KR",me="METH_K",me="METH_R",
#                                          me2="BIMETH_KR",me2="BIMETH_K",me2="BIMETH_R",
#                                          me3="TRIMETH_K",
#                                          ac="ACET_K"),collapse="&",simplify=TRUE) {
#  sapply(strsplit(paste0(modifstring," "),":"),function(x) {
#    x[length(x)] <- sub(" $","",x[length(x)])
#    res <- sapply(seq_along(x),function(x.i) {
#      which(modif==indiv.modif)
#    })
#    if (!is.null(collapse))
#      paste(which(x%in%modif)-1,collapse=collapse)
#    else
#      which(x%in%modif)-1
#  },simplify=simplify)
#}
#
.convertModifToPos <- function(modifstring,modif="PHOS",collapse="&",simplify=TRUE,and.name=FALSE) {
  split.modif <- strsplit(paste0(modifstring," "),":")
  if (and.name) {
    name.modif <- .names.as.vector(modif)
  }
  sapply(split.modif,function(x) {
    x[length(x)] <- sub(" $","",x[length(x)])
    modification.pos <- which(x%in%modif)-1 
    if (!is.null(collapse))
      paste(modification.pos,collapse=collapse)
    else {
      if (and.name) 
        data.frame(modif.pos=modification.pos,modif=name.modif[x[modification.pos+1]],stringsAsFactors=FALSE)
      else
        modification.pos
    }
  },simplify=simplify)
}



.convertPhosphoRSPepProb <- function(peptide,pepprob,round.to.frac=NULL) {
  mapply(function(pep,pprob) {
           pprob <- as.numeric(pprob)
           prob <- rep(-1,length(pep))
           pep.pos <- pprob[seq(from=1,to=length(pprob),by=2)]
           pep.prob <- pprob[seq(from=2,to=length(pprob),by=2)]
           if (!is.null(round.to.frac)) 
             pep.prob <- round(pep.prob*round.to.frac)/round.to.frac
           prob[pep.pos] <- pep.prob

           prob_mask <- !is.na(prob) & prob>=0
           pep[prob_mask] <- paste0(pep[prob_mask],"(",prob[prob_mask],")")
           paste0(pep,collapse="")
         },
         strsplit(peptide,""),
         strsplit(pepprob,"[;:]"))
}

.convertPeptideModif <- function(peptide,modifstring,
                                 modifs=c(p="PHOS",o="Oxidation_M",c="Cys_CAM",
                                          me="METH_KR",me="METH_K",me="METH_R",
                                          dime="BIMETH_KR",dime="BIMETH_K",dime="BIMETH_R",
                                          trime="TRIMETH_K",
                                          ac="ACET_K")) {
  names(letters) <- LETTERS
  if (length(peptide)==0 || all(nchar(peptide)==0))
	  stop("peptide length=0")
  mapply(function(pep,m) {
           m <- m[-c(1,length(m))]
           if (is.null(names(modifs))) {
             for (mm in modifs) 
               pep[m==mm] <- letters[pep[m==mm]]
           } else {
             for (i in seq_along(modifs)) 
               pep[m==modifs[i]] <- paste0(pep[m==modifs[i]],"(",names(modifs)[i],")")
           }
           paste(pep,collapse="")
         },
         strsplit(peptide,""),
         strsplit(paste0(modifstring," "),":")
  )
}

readPhosphoRSOutput <- function(phosphoRS.outfile,simplify=FALSE,pepmodif.sep="##.##",
                                besthit.only=TRUE) {
  requireNamespace("XML")
  doc <- XML::xmlTreeParse(phosphoRS.outfile,useInternalNodes=TRUE)  
  spectra <- XML::xmlRoot(doc)[["Spectra"]]
  res <- XML::xmlApply(spectra,function(spectrum) {
    spectrum.id <- URLdecode(XML::xmlAttrs(spectrum)["ID"])
    #message(spectrum.id)
    res.s <- XML::xmlApply(spectrum[["Peptides"]],function(peptide) {
      pep.id <- strsplit(XML::xmlAttrs(peptide)["ID"],pepmodif.sep,fixed=TRUE)[[1]]    
      #message(pep.id[1])
      site.probs <- t(XML::xmlSApply(peptide[["SitePrediction"]],XML::xmlAttrs))
      isoforms <- t(XML::xmlSApply(peptide[["Isoforms"]],function(isoform) {
        seqpos <- XML::xmlSApply(isoform[["PhosphoSites"]],XML::xmlGetAttr,"SeqPos")

        # get right modif string
        modifstring <- strsplit(paste0(pep.id[2]," "),":")[[1]]
        modifstring <- gsub(" $","",modifstring)
        modifstring[modifstring=='PHOS'] <- ''
        modifstring[as.numeric(seqpos)+1] <- 'PHOS'
        modifstring <- paste(modifstring,collapse=":")
 
        if (length(seqpos > 1)) seqpos <- paste(seqpos,collapse="&")
        c(modif=modifstring,
          pepscore=XML::xmlAttrs(isoform)[['PepScore']],
          pepprob=XML::xmlAttrs(isoform)[['PepProb']],
          seqpos=seqpos)
      }))   

      #rownames(isoforms) <- NULL
      storage.mode(site.probs) <- "numeric"
      site.probs[,2] <- round(site.probs[,2],2)
      if (isTRUE(simplify))
        data.frame(peptide=pep.id[1],isoforms,
                   site.probs=paste(apply(site.probs,1,paste,collapse=":"),collapse=";"),
                   stringsAsFactors=FALSE,row.names=NULL)
      else
        list(peptide=pep.id,
             site.probs=site.probs,
             isoforms=isoforms)
    })
    if (simplify)
      data.frame(spectrum=spectrum.id,do.call(rbind,res.s),
                 stringsAsFactors=FALSE,row.names=NULL)
    else
      res.s
  })
  if (simplify) {
    res <- do.call(rbind,res)
    res$pepscore <- as.numeric(res$pepscore)
    res$pepprob <- as.numeric(res$pepprob)
    rownames(res) <- NULL
  } else {
    names(res) <- sapply(XML::xmlChildren(spectra),XML::xmlGetAttr,"ID")
  }
  if(besthit.only & simplify) {
    res <- ddply(res,'spectrum',function(d) d[which.max(d$pepprob),])
    rownames(res) <- res$spectrum
  }
  res
}

filterSpectraPhosphoRS <- function(id.file,mgf.file,...,min.prob=NULL, do.remove=FALSE) {
  if (is(id.file,"character"))
    id.file <- .read.idfile(id.file)
  probs <- getPhosphoRSProbabilities(id.file,mgf.file,...,simplify=TRUE)
  ## probs excludes non-PHOS peptides - we do filter them for now? (about 8-10%)
  id.file$peptide <- NULL
  id.file$modif <- NULL

  colnames.both <- intersect(colnames(id.file),colnames(probs))
  if (length(colnames.both) > 1) {
    stop("id.file includes colnames which are added by the function filterSpectraPhosphoRS: \n",
         "\t",paste(sort(colnames.both[colnames.both != 'spectrum']),collapse=", "),
         "\nRemove them prior to calling filterSpectraPhosphoRS.")
  }

  id.file <- merge(id.file,probs,by="spectrum")
  if (!is.null(min.prob)) {
    if (!'use.for.quant' %in% colnames(id.file)) id.file$use.for.quant <- TRUE
    sel.minprob <- id.file[,"pepprob"] >= min.prob
    id.file[,"use.for.quant"] <- id.file[,"use.for.quant"] & sel.minprob
    if (isTRUE(do.remove))
      id.file <- id.file[,sel.minprob]
  }
  return(id.file)
}

modif.sites <- function(protein.group,protein.g=reporterProteins(protein.group),modif) {
  pi <- protein.group@peptideInfo
  ip <- indistinguishableProteins(protein.group)
  sapply(protein.g,
         function(my.protein.g) {
           isoform.acs <- names(ip)[ip==my.protein.g]
           res <- lapply(isoform.acs,function(ac) {
                         sel <- pi[,"protein"] == ac
                         pep.pos <- .convertModifToPos(pi[sel,"modif"],modif,simplify=FALSE,collapse=NULL) 
                         modif.pos <- unlist(mapply(function(start.pos,pep.posi) start.pos + pep.posi -1,
                                                    pi[sel,"start.pos"],pep.pos))

                         seen.sites <- rep(FALSE,max(modif.pos))
                         seen.sites[modif.pos] <- TRUE
                         which(seen.sites)
                 })
           names(res) <- isoform.acs
           res
         })
}

modif.site.count <- function(protein.group,protein.g=reporterProteins(protein.group),modif,take=max) {
  modif.sites <- modif.sites(protein.group,protein.g,modif)
  sapply(modif.sites,function(modifs.protein.g) {
         n.modifs.protein.acs <- sapply(modifs.protein.g,length)
         take(n.modifs.protein.acs)
  })
}

observedKnownSites <- function(protein.group,protein.g,ptm.info,modif,modification.name=NULL) {
  modif.sites <- modif.sites(protein.group,protein.g,modif)

  lapply(modif.sites,function(modifs.protein.g) {
         lapply(names(modifs.protein.g),function(isoform.ac) {
                observed.sites <- modifs.protein.g[[isoform.ac]]

                sel.known.sites <- ptm.info$isoform_ac==ifelse(grepl("-[0-9]$",isoform.ac),
                                                   isoform.ac,paste(isoform.ac,"-1",sep=""))
                if (!is.null(modification.name))
                  sel.known.sites <- sel.known.sites & grepl(modification.name,ptm.info$modification.name)

                c(n.observed.sites=length(observed.sites),
                  n.known.sites=sum(sel.known.sites),
                  n.known.sites.observed=sum(ptm.info[sel.known.sites,"first_position"] %in% observed.sites))
         })
  })
}


.proteinPtmInfo <- function(isoform.ac,protein.group,ptm.info,modif,modification.name=NULL,simplify=TRUE) {
  requireNamespace("OrgMassSpecR")
  if (length(proteinInfo(protein.group)) == 0)
    stop("no protein info attached to protein.group: see ?getProteinInfoFromUniprot on how to get it.")

  protein.length <- as.numeric(proteinInfo(protein.group,protein.ac=isoform.ac,select="length") )
  if(all(is.na(protein.length))) 
    stop("no protein info for ",isoform.ac,"; need protein length and sequence")

 
  obs.peptides <- observable.peptides(proteinInfo(protein.group,protein.ac=isoform.ac,select="sequence"),nmc=2)
  possible.sites <- t(sapply(seq_len(protein.length),function(p) c(possible.nmc1=any(p>=obs.peptides$start & p<=obs.peptides$stop & obs.peptides$mc <=1),
                                                                   possible.nmc2=any(p>=obs.peptides$start & p<=obs.peptides$stop & obs.peptides$mc <=2))))
  my.ptm.info <- ptm.info[ptm.info$isoform_ac==ifelse(grepl("-[0-9]$",isoform.ac),
                                                   isoform.ac,paste(isoform.ac,"-1",sep="")),]
  if (!is.null(modification.name)) 
    my.ptm.info <- my.ptm.info[my.ptm.info$modification.name%in%modification.name,]
 
  # TO CHECK: first_position might be bigger than protein.length
  known.sites <- rep(FALSE,protein.length)
  if (nrow(my.ptm.info) > 0)
    known.sites[my.ptm.info$first_position] <- TRUE

  
  pi <- protein.group@peptideInfo
  sel.has.modif <- sapply(strsplit(pi[,"modif"],":"),function(x) any(x %in% modif))
  pi <- pi[pi[,"protein"]==isoform.ac & sel.has.modif,]

  pep.pos <- .convertModifToPos(pi[,"modif"],modif,simplify=FALSE,collapse=NULL)

  if (nrow(pi) > 0)
    modif.pos <- unlist(mapply(function(start.pos,pep.posi) start.pos + pep.posi -1,
                               pi[,"start.pos"],pep.pos))
  else
    modif.pos <- NULL

  seen.sites <- rep(FALSE,protein.length)
  seen.sites[modif.pos] <- TRUE

  if (simplify) {
  return(
         data.frame(observed.site.pos=paste(which(seen.sites),collapse=","),
           observed.sites=sum(seen.sites),
           known.sites=sum(known.sites),
           oberserved.known.sites=sum(known.sites&seen.sites),
           observable.known.sites.1mc=sum(known.sites&possible.sites[,"possible.nmc1"]),
           observable.known.sites.2mc=sum(known.sites&possible.sites[,"possible.nmc2"]),
           stringsAsFactors=FALSE)
         )
  } else {
  return(list(peptideInfo=pi,modif.pos=modif.pos,
              observable.peptides=obs.peptides,
              known.sites=my.ptm.info))

  }
}


getPeptideModifContext <- function(protein.group,modif,n.aa.up=7,n.aa.down=7) {
  if (length(proteinInfo(protein.group)) == 0)
    stop("no protein info attached to protein.group: see ?getProteinInfoFromUniprot on how to get it.")

  peptide.info <- unique(peptideInfo(protein.group)[,c('modif','protein','peptide','real.peptide')])
  protein.sequences <- paste0(paste0(rep("_",n.aa.down),collapse=""),
                              #gsub("I","L",proteinInfo(protein.group)[,'sequence']),
                              proteinInfo(protein.group)[,'sequence'],
                              paste0(rep("_",n.aa.up),collapse="")) # enlarge sequence in case peptide starts in the beginning / end
  names(protein.sequences) <- proteinInfo(protein.group)[,'accession']
  
  pep.modif.context <- mapply(function(pepmodifs,protein.ac,pep) {
    my.seq <- protein.sequences[protein.ac]
    if (is.na(my.seq)) {
      warning("No sequence for ",protein.ac)
      return(NA)
    }
    
    peptide.startpos <- gregexpr(pep,my.seq)[[1]]
    if (peptide.startpos[1] == -1) {
      warning("Peptide [",pep,"] could not be matched to ",protein.ac)
      return(NA)    
    }

    pepmodifs[length(pepmodifs)] <- sub(" $","",pepmodifs[length(pepmodifs)])
    modification.positions <- which(pepmodifs%in%modif)-1
    if (length(modification.positions) == 0) 
      return(NA)
    
    paste(sapply(peptide.startpos,function(pep.pos) {
      modification.positions.in.protein <- pep.pos + modification.positions - 1
      res <- paste(sapply(modification.positions.in.protein,function(pos) 
                          substr(my.seq,pos-n.aa.down,pos+n.aa.up)),collapse=",")
    
      if (nchar(res) < n.aa.up+n.aa.down+1) 
        stop("extracted pattern does not have the length it should have")
      res
    }),collapse=",")
  },strsplit(paste0(peptide.info[,'modif']," "),":"),
    peptide.info[,'protein'], 
    peptide.info[,'real.peptide'])

  # get pep-modif context
  context.df <- unique(data.frame(peptide=peptide.info[,'peptide'],modif=peptide.info[,'modif'],context=pep.modif.context,stringsAsFactors=FALSE))
  ddply(context.df,
        c("peptide","modif"),
        function(x) 
          data.frame(peptide=x[1,'peptide'],modif=x[1,'modif'],
                     context=paste(x[,'context'],collapse=";"),stringsAsFactors=FALSE))

}

if(!isGeneric("as.data.frame")) setGeneric("as.data.frame", useAsDefault=as.data.frame)

# for R version pre 2.14
#if (!exists("paste0")) # will show a warning on newer R versions
  paste0 <- function(...,sep="") paste(...,sep=sep)

"%inrange%" <- function(a,b) {
  if (!is.numeric(a) || !is.numeric(b)) stop("Arguments must be numeric")
  if (length(b) != 2) stop("Second argument must have 2 elements")
  return(a >= b[1] & a <=b[2])
}

.cn <- function(x,y) {
  x[ , y ]
}

.gg_theme <- function(...) {
  theme(...)
}

.gg_element_text <- function(...) {
  element_text(...)
}

.abbrev <- function(strings,n=4,collapse=NULL) {
  ll <- length(strings)
  if (ll > n) {
    strings <- strings[seq_len(n-1)]
    strings[n] <- sprintf("... (%s in total)",ll)
  }
  if (!is.null(collapse))
    paste(strings,collapse=collapse)
  else
    strings
}


# are 'most' (default: 90%) of the values in x TRUE?
.most <- function(x,fraction=0.9) {
  if (is.null(dim(x)) || !is.logical(x)) stop(".most function works on logical vectors")
  sum(x)/length(x) > fraction
}

.check.isfunction <- function(f) {
  if (!is.function(f))
    stop(paste(deparse(substitute(f)),"must be a function!"))
  TRUE
}

.unique.or.collapse <- function(x,collapse=";") {
  if (is.null(x)) NA
  else
    ifelse(all(x==x[1]),x[1],paste0(x,collapse=collapse))
}

.paste_unique <- function(x,...,na.rm=TRUE) {
  x <- unique(x)
  if (na.rm)
    x <- x[!is.na(x)]
  paste(x,...)
}

.grep_columns <- function(df,pattern,...,logical=TRUE) {
  if (logical)
    grepl(pattern,colnames(df),...)
  else
    grep(pattern,colnames(df),...)
}

#TODO: unify factor.as.character and factor.to.chr
.factor.as.character <- function(df) {
  for (col_i in seq_len(ncol(df))) {
    if (is.factor(df[,col_i]))
      df[,col_i] <- as.character(df[,col_i])
  }
  df
}

.names.as.vector <- function(x) {
  if (!is.null(names(x)))
    vec <- names(x)
  else 
    vec <- x
  setNames(vec,x)
}

# from Gavin Simpson [http://stackoverflow.com/questions/9788026/change-the-order-of-columns]
.moveToFirstCol <- function(df, colname) {
  cnams <- colnames(df)
  want <- which(colname == cnams)
  df[, c(cnams[want], cnams[-want])]
}

.factor.to.chr <- function(df) {
  for (col in colnames(df))
    if (is.factor(df[,col])) df[,col] <- as.character(df[,col])
  df
}

.all.duplicates <- function(x,n=2) {
  t <- table(x)
  names(t)[t>=n]
}

# get a number range. E.g. 1,2,3,5,6 -> 1-3,5,6
.string.number.ranges <- function(numbers) {
  n <- number.ranges(numbers)
  if (is.na(n)) return("")
  else return(sprintf("[%s]",n))
}

number.ranges <- function(numbers) {
  if (all(is.na(numbers))) { return(NA) }
  numb=c()
  numb_string=c()

  for (i in sort(unique(as.numeric(numbers)))) {
    if (length(numb) == 0) {
      numb = c(i)
    } else {
      if (i == numb[length(numb)]+1) {
        numb = c(numb,i)
      } else {
        if (length(numb) <= 2) {
          numb_string = c(numb_string,numb)
        } else {
          numb_string = c(numb_string,paste(min(numb),max(numb),sep="-"))
        }
        numb = c(i)
      }
    }
  }
  if (length(numb) <= 2) {
        numb_string = c(numb_string,numb)
  } else {
       numb_string = c(numb_string,paste(min(numb),max(numb),sep="-"))
  }
  return(paste(numb_string,collapse=","))
}

.return.equal.or.na <- function(df) {
  apply(df, 1, function(x) {

    if (all(is.na(x))) return(NA)

    y <- x[!is.na(x)]
    if (!all(y == y[1])) return(NA)
    return(y[1])
  })
}

.all.duplicate.rows <- function(df,column,n=2) {
  t <- table(df[,column])
  res <- df[df[,column] %in% names(t)[t>=n],]
  res[order(res[,column]),]
}

.vector.as.data.frame <- function(vect,colnames=NULL,stringsAsFactors=FALSE) {
	as.data.frame(.as.matrix(vect,colnames),stringsAsFactors=stringsAsFactors)
}

.as.matrix <- function(vect,colnames=NULL) {
  mat <- matrix(c(names(vect),vect),ncol=2,byrow=FALSE)
  colnames(mat) <- colnames
  mat
}

.as.vect <- function(my.matrix,col.data=2,col.names=1) {
  setNames(my.matrix[,col.data],my.matrix[,col.names])
}

.stopifnot <- function(cond,...) if (!cond) stop(...)

.stopiflengthnotequal <- function(x,y,...) {
  if (length(x) != length(y)) {
    stop(..., " length(x) = ",length(x),", length(y) = ",length(y),"")
  }
}


.stopifna  <- function(data,...) if (any(is.na(data))) stop(...)


.expand.w.vector <- function(v,m,by.m,v.name='v',...) {
  df <- do.call(rbind,lapply(names(v),function(x) {
    pos <- which(m[,by.m] %in% v[x])
    data.frame(x,m[pos,],row.names=NULL,stringsAsFactors=FALSE,...)
  }))
  colnames(df)[1] <- v.name
  df
}

.trim <- function(x,side="both") {
  if (side=="left" | side=="both")
    x <- sub("^\\s*","",x)
  if (side=="right" | side=="both")
    x <- sub("\\s*$","",x)
  x
}

.bnds <- function(x,bnd=NULL,min.x=-bnd,max.x=bnd) {
  max(
      min(x,max.x),
      min.x)
}


# split string into named vector
.strsplit_vector <- function(x,pattern) {
  pos <- sapply(x,regexpr,pattern="=",fixed=TRUE)
  res <- substring(x,pos+1)
  names(res) <- substring(x,rep(1,length(x)),pos-1)
  res
}

.sum.bool.na  <- function(x) 
  c('TRUE'=sum(x,na.rm=TRUE),'FALSE'=sum(!x,na.rm=TRUE),'TRUE %'=round(sum(x,na.rm=TRUE)/length(x)*100,1),length=length(x),n.na=sum(is.na(x)))

# provide summary of boolean vector values
.sum.bool  <- function(x) 
  c('TRUE'=sum(x),'FALSE'=sum(!x),'TRUE %'=round(sum(x)/length(x)*100,1),length=length(x))

# provide character summary of boolean vectors
.sum.bool.c  <- function(x) 
  paste('TRUE: ',sum(x),'; FALSE: ',sum(!x),'; TRUE %: ',round(sum(x)/length(x)*100,1),"; length: ",length(x),sep="")

.combine.fisher <- function(p.values,signs) {
  if (length(signs) != length(p.values))
    stop("lratios and pvalues must have equal length!")
  
  sel.notna <- !is.na(p.values)
  p.values <- p.values[sel.notna]
  signs <- signs[sel.notna]
  k <- length(p.values)

  if (length(p.values) == 1) return(p.values)

  ## require that the direction is the same for all p-values
  if (!all(signs == signs[1])) return(1)

  return(pchisq(-2*sum(log(p.values)),2*k,lower.tail=FALSE))
}

.combine.fisher.tblwide <- function(my.df) {
  lr.cols <- grepl("^lratio.",colnames(my.df))
  p.cols <- grepl("^p.value.rat.",colnames(my.df)) & !grepl("adj",colnames(my.df))

  if (sum(lr.cols) != sum(p.cols))
    stop("unequal number of '^lratio.' and '^p.value.rat' columns")

  combined.p <- rep(1,nrow(my.df))
  signs.equal <- apply(sign(my.df[,lr.cols]),1,function(x) { y=x[!is.na(x)]; all(y==y[1])})
  ks <- apply(!is.na(my.df[,p.cols]),1,sum)
  logsums <- rowSums(log(my.df[,p.cols]),na.rm=TRUE)
  sel <-  ks > 1 & signs.equal
  combined.p[sel] <- pchisq(-2*logsums[sel],2*ks[sel],lower.tail=FALSE)

  combined.p[ks==1] <- as.numeric(apply(my.df[ks==1,p.cols],1,function(x) { y=x[!is.na(x)]; ifelse(length(y)==0,NA,y) } ))
  return(combined.p)
}

.call.cmd <- function(cmd,stdout.to=NULL) {
    if (is.null(stdout.to)) {
      message("  calling system command [",cmd,"]",appendLF=FALSE)
      if (system(cmd) != 0) stop("\nError executing [",cmd,"]")
      message(" finished.")
    } else {
      message("  calling system command [",cmd," > ",stdout.to,"]",appendLF=FALSE)
      if (system(paste(cmd,">",stdout.to)) != 0) 
        stop("\nError executing [",cmd,"]: \n\n ...\n",
             paste(tail(readLines(stdout.to),n=10),collapse="\n"))
      message(" finished.")
    }
}


.get.cmbn <- function(combn,tags,cl) {
  if (!all(unlist(combn) %in% cl))
    stop("incorrect combn specification")

  res <- c()
  for (cc in combn)
    for (tag1 in tags[cl==cc[1]&!is.na(cl)]) 
      for (tag2 in tags[cl==cc[2]&!is.na(cl)])
        res <- cbind(res,c(r1=tag1,r2=tag2,
                     class1=cc[1],
                     class2=cc[2]))
  res
}


.sanitize.sh <- function(str) {
  gsub("[^a-zA-Z\\.0-9_\\-]","", str)
}

.weighted.cor <- function( x, y, w = rep(1,length(x)),use='complete.obs') {
  ## (c) Heather Turner, Vincent Zoonekynd at http://stackoverflow.com/questions/9460664/weighted-pearsons-correlation
  stopifnot(length(x) == dim(y)[2] )
  if (use=='complete.obs') {
    sel <- !is.na(x) & !is.na(y) & !is.na(w)
    x <- x[sel]
    y <- y[sel]
    w <- w[sel]
  }

  w <- w / sum(w)
  # Center x and y, using the weighted means
  x <- x - sum(x * w)
  ty <- t(y - colSums(t(y) * w))
  # Compute the variance
  vx <- sum(w * x * x)
  vy <- colSums(w * ty * ty)
  # Compute the covariance
  vxy <- colSums(ty * x * w)
  # Compute the correlation
  vxy / sqrt(vx * vy)
}


.concensus.il.peptide <- function(peptide) { 
    pep <- do.call(rbind,strsplit(peptide,""))
    paste0(apply(pep,2,function(x) {
      xu <- unique(x)
      if (length(xu)==1) xu
      else "L"
    }),collapse="")
  }

.fix.il.peptide <- function(from,sub.il=TRUE) {
   if (.PEPTIDE.COLS['REALPEPTIDE'] %in% colnames(from))
     return(from)

   from[,.PEPTIDE.COLS['REALPEPTIDE']] <- from[,.SPECTRUM.COLS['PEPTIDE']]
   l.peptide <- gsub("I","L",from[,.SPECTRUM.COLS['PEPTIDE']])
   if (sub.il) {
     from$peptide <- l.peptide
   } else {
     from$peptide <- as.vector( tapply(from[,.SPECTRUM.COLS['PEPTIDE']],l.peptide,function(x) {
           if (all(x == x[1])) x[1]
           else 
              .concensus.il.peptide(unique(x))
     }))[l.peptide]
   }
   return(from)
}

#!/usr/bin/Rscript

create.protein.group <- function(id.files,file="protein.group.rda") {
	message("Identification files used for ProteinGroup creation: \n\t",paste(id.files,collapse="\n\t"),"\n")
	suppressPackageStartupMessages(library(isobar))

	# create an IBSpectra object with all identifications
	protein.group <- readProteinGroup(id.files)
	proteinInfo(protein.group) <- getProteinInfoFromUniprot(protein.group)
	print(protein.group)
	# save ProteinGroup object
	save(protein.group,file=file)
	message("\nSaved ProteinGroup object to ",file,".\n")
}

message("## CREATE PROTEINGROUP SCRIPT ##\n")
if (length(commandArgs(TRUE)) == 0) 
  stop("Provide at least one id file as argument!\n")

args <- commandArgs(TRUE)
create.protein.group(args[2:length(args)],args[1])

### =========================================================================
### NoiseModel objects.
### -------------------------------------------------------------------------
###
### Class definition.

# A virtual class representing the interface to noise models.
# Use ExponentialNoiseModel or InverseNoiseModel
setClass("NoiseModel",
    representation(
        na.region="numeric",
        low.intensity="numeric",
        f="function",
        parameter="numeric",
        "VIRTUAL"),
    contains = "VersionedBiobase",
    prototype = prototype(
        new("VersionedBiobase",versions=c(NoiseModel="1.0.0"))
    )
)

#b*e^(cx)
setClass("ExponentialNoANoiseModel",
    contains = "NoiseModel",
    prototype = prototype(
        new("VersionedBiobase",versions=c(classVersion("NoiseModel"),ExponentialNoANoiseModel="1.0.0")),
        f = function(data,parameter) { parameter[1] * exp (- data * parameter[2]) },
        parameter = c(10,1),
        na.region = 0,
        low.intensity = 0
    )
)

#a + b*e^(cx)
setClass("ExponentialNoiseModel",
    contains = "NoiseModel",
    prototype = prototype(
        new("VersionedBiobase",versions=c(classVersion("NoiseModel"),ExponentialNoiseModel="1.0.0")),
        f = function(data,parameter) { parameter[1] + parameter[2] * exp (- data * parameter[3]) },
        parameter = c(0,10,1),
        na.region = 0,
        low.intensity = 0
    )
)

#a + bx^c
setClass("InverseNoiseModel",
    contains = "NoiseModel",
    prototype = prototype(
        new("VersionedBiobase",versions=c(classVersion("NoiseModel"),InverseNoiseModel="1.0.0")),
        f = function(data,parameter) { parameter[1] + parameter[2]*data^-parameter[3] },
        parameter = c(0,1,1),
        na.region = 0,
        low.intensity = 0
    )
)

# bx^c
setClass("InverseNoANoiseModel",
    contains = "NoiseModel",
    prototype = prototype(
        new("VersionedBiobase",versions=c(classVersion("NoiseModel"),InverseNoiseModel="1.0.0")),
        f = function(data,parameter) {  parameter[1]*data^-parameter[2] },
        parameter = c(1,1),
        na.region = 0,
        low.intensity = 0
    )
)

setClass("GeneralNoiseModel",
    contains = "NoiseModel",
    prototype = prototype(
        new("VersionedBiobase",versions=c(classVersion("NoiseModel"),
                                          GeneralNoiseModel="1.0.0")),
        f = function(data,parameter) {   },
        parameter = c(),
        na.region = 0,
        low.intensity = 0
    )
)


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Validity.
###

.valid.NoiseModel.slots <- function(object) {
  NULL
}

.valid.NoiseModel <- function(object) {
	.valid.NoiseModel.slots(object)
}

setValidity("NoiseModel",.valid.NoiseModel)


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Constructor.
###

setMethod("initialize","NoiseModel",
    function(.Object,ibspectra=NULL,reporterTagNames=NULL,one.to.one=TRUE,
             min.spectra=10,max.n.proteins=50,plot=F,pool=FALSE,
             nm.col=c("#E41A1C","#377EB8","#4DAF4A","#984EA3","#FF7F00","#FFFF33","#A65628","#F781BF","#999999"),nm.cmp=NULL,...) {
      if (!is.null(ibspectra))
      {
        if (is.null(reporterTagNames)) reporterTagNames <- reporterTagNames(ibspectra)
        if (!is.character(reporterTagNames))
          stop("reporterTagNames has to be of class character.")
        
        if (is.matrix(reporterTagNames)) {
          # a combination matrix is supplied
          if (nrow(reporterTagNames) != 2)
            stop("reporterTagNames has to be either a vector or a combination matrix (with two rows).")
          reporterTagNames.combn <- reporterTagNames
        } else {
          reporterTagNames.combn <- combn(reporterTagNames,2)
        }

        ri <- reporterIntensities(ibspectra,na.rm=FALSE)
        
        if (plot & !pool)
          par(mfrow=rep(ceiling(sqrt(ncol(reporterTagNames.combn))),2))

        # matrix to store fitted paramters of reporter combinations
        # will be averaged in the end
        all.parameters <-
          matrix(nrow=ncol(reporterTagNames.combn),ncol=length(parameter(.Object)))
        
        pool.ch1 <- c()
        pool.ch2 <- c()

        if (one.to.one) {
          
          # for each reporter combination, fit a noise function on the ratios
          for (ch_i in seq_len(ncol(reporterTagNames.combn))) {
            ch1 <- ri[,reporterTagNames.combn[1,ch_i]]
            ch2 <- ri[,reporterTagNames.combn[2,ch_i]]
            pool.ch1 <- c(pool.ch1,ch1)
            pool.ch2 <- c(pool.ch2,ch2)

            all.parameters[ch_i,] <-
              .fitNoiseFunction(ch1,ch2,noiseFunction(.Object),parameter(.Object),...)

             if (plot) {
                x.data <- log10(sqrt(ch1*ch2))
                plot(x.data,ch1/ch2,log="y",
                       main=sprintf("%s vs %s",reporterTagNames.combn[1,ch_i],
                       reporterTagNames.combn[2,ch_i]),ylim=c(0.2,5))
                .lines.nf(noiseFunction(.Object),parameter=all.parameters[ch_i,],xlim=range(x.data,na.rm=TRUE))
              }

          }
        } else {

          # scale the ratios of proteins down to 1, and learn noise model on
          # the differences in their spectra
          
          pg <- proteinGroup(ibspectra)
          pnp <- peptideNProtein(pg)
          pepnprot <- as.data.frame(pnp[pnp[,'peptide'] %in% 
              peptides(pg,protein=reporterProteins(pg),
                       specificity=REPORTERSPECIFIC),],
              stringsAsFactors=FALSE)
          
          pepnspec <-
            as.data.frame(.as.matrix(spectrumToPeptide(pg),
                                     c("spectrum","peptide")),stringsAsFactors=FALSE)
          pgdf <- merge(pepnprot,pepnspec)

          # only take non-NA spectra
          pgdf <- pgdf[pgdf$spectrum %in%
                       rownames(ri)[apply(ri,1,function(d) !any(is.na(d)))],]
          pgdf$protein.g <- as.character(pgdf$protein.g)
          
          t <- table(pgdf$protein.g)
          sel.proteins <-
            names(t)[t>=min.spectra][order(t[t>=min.spectra],decreasing=TRUE)]
          sel.proteins <- sel.proteins[seq_len(min(length(sel.proteins),max.n.proteins))]
          sel = pgdf$protein.g %in% sel.proteins
          cat(sprintf("%s proteins with more than %s spectra, taking top %s.\n",
                  length(unique(pgdf$protein.g)),
                  min.spectra,max.n.proteins))
          pgdf <- cbind(pgdf[sel,],ri[pgdf[sel,"spectrum"],])
          

          for (ch_i in seq_len(ncol(reporterTagNames.combn))) {
            ch1 <- c()
            ch2 <- c()
            pool.ch1 <- c(pool.ch1,ch1)
            pool.ch2 <- c(pool.ch2,ch2)
            # compute protein ratios
            protein.ratios <- estimateRatio(ibspectra=ibspectra, 
                channel1=reporterTagNames.combn[1,ch_i],
                channel2=reporterTagNames.combn[2,ch_i],
                protein=unique(pgdf$protein.g),combine=F)
            
            # bring protein ratio to ratio 1
            for (protein in rownames(protein.ratios)) {
              if (abs(protein.ratios[protein,"lratio"]) < 1) {
              ch1 <- c(ch1,pgdf[pgdf$protein.g==protein,reporterTagNames.combn[1,ch_i]])
              ch2 <- c(ch2,pgdf[pgdf$protein.g==protein,reporterTagNames.combn[2,ch_i]] * 
                  10^protein.ratios[protein,"lratio"])
              }
            }
            
            if (!pool) {
              all.parameters[ch_i,] <- .fitNoiseFunction(ch1,ch2,
                  noiseFunction(.Object),parameter(.Object),...)
            
             if (plot) {
                x.data <- log10(sqrt(ch1*ch2))
                plot(x.data,ch1/ch2,log="y",
                       main=sprintf("%s vs %s",reporterTagNames.combn[1,ch_i],
                       reporterTagNames.combn[2,ch_i]),ylim=c(0.2,5))
                .lines.nf(noiseFunction(.Object),parameter=all.parameters[ch_i,],xlim=range(x.data))
              }
              
              print(all.parameters[ch_i,])
            }
            pool.ch1 <- c(pool.ch1,ch1)
            pool.ch2 <- c(pool.ch2,ch2)
          }
        }
        if (!pool) {
          parameter(.Object) <- .averageParameters(.Object,all.parameters) 
        } else {
          parameter(.Object) <- .fitNoiseFunction(pool.ch1,pool.ch2,noiseFunction(.Object),parameter(.Object),...)
        }
        if (plot) {
          x.data <- log10(sqrt(pool.ch1*pool.ch2))
          plot(x.data,pool.ch1/pool.ch2,log="y",
               main="All channels",ylim=c(0.2,5))
          .lines.nf(noiseFunction(.Object),parameter=parameter(.Object),xlim=range(x.data,na.rm=TRUE),col=nm.col[1])
          if (!is.null(nm.cmp)) {
            if (is(nm.cmp,"NoiseModel")) nm.cmp <- c(nm.cmp)
            for (nm_i in seq_along(nm.cmp)) {
              .lines.nm(nm.cmp[[nm_i]],xlim=range(x.data,na.rm=TRUE),col=nm.col[nm_i+1])
            }
          }

        }

        sel.allna <- apply(is.na(ri),1,all)
        sel.anyna <- apply(is.na(ri),1,any)
        ri.anyna <- as.numeric(ri[!sel.allna & sel.anyna,])
        na.val <- log10(as.numeric(ri.anyna))
        na.val <- na.val[!is.na(na.val)]

        print(parameter(.Object))
        lowIntensity(.Object) <- quantile(unlist(ri[,reporterTagNames]),prob=0.05,na.rm=T)
        naRegion(.Object) <- na.val
      }
      callNextMethod(.Object)
    }
)

.lines.nm <- function(nm,xlim,col="red",parameter=NULL,lwd=2,...) {
  xx <- seq(from=xlim[1],to=xlim[2],length.out=100)
  lines(xx,10^(1.96*sqrt(variance(nm,xx))),col=col,lwd=lwd)
  lines(xx,10^(-1.96*sqrt(variance(nm,xx))),col=col,lwd=lwd)
}


.lines.nf <- function(f,xlim,col="red",parameter=NULL,lwd=2,...) {
  xx <- seq(from=xlim[1],to=xlim[2],length.out=100)
  lines(xx,10^(1.96*sqrt(0.5)*f(xx,parameter=parameter)),col=col,lwd=lwd,...)
  lines(xx,10^(-1.96*sqrt(0.5)*f(xx,parameter=parameter)),col=col,lwd=lwd,...)
}

setGeneric("NoiseModel",function(ibspectra,...) standardGeneric("NoiseModel"))

setMethod("NoiseModel",signature(ibspectra="IBSpectra"),
  function(ibspectra,type="exponential",...) {
    if (type=="exponential") new("ExponentialNoiseModel",ibspectra,...)
    else if (type=="exponential.noa") new("ExponentialNoANoiseModel",ibspectra,...)
    else if (type=="inverse") new("InverseNoiseModel",ibspectra,...)
  }
)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### .fitNoiseFunction and .averageParameters.
### (called by initialize)

setGeneric(".fitNoiseFunction",function(data1,data2,f,initial.parameters,...)
  standardGeneric(".fitNoiseFunction"))
setGeneric(".averageParameters",function(x,parameters)
  standardGeneric(".averageParameters"))

setMethod(".fitNoiseFunction", 
    signature=c(data1="numeric",data2="numeric",f="function",
      initial.parameters="numeric"),
    function(data1,data2,f,initial.parameters,n.bins=30,min.n.bin=50,
        na.rm=is.null(set.na.to),set.na.to=NULL,...) {

      fit.f <- function(param,x,y) 
           -sum(dnorm(x,mean=0,sd=f(data=y,parameter=param),log=T))
      
      # remove points which are NA in both channels
      sel <- is.na(data1) & is.na(data2)
      data1 <- data1[!sel]; data2 <- data2[!sel]
      
      if (na.rm) {
        sel <- is.na(data1) | is.na(data2)
        data1 <- data1[!sel]; data2 <- data2[!sel]
      } else {
        if (is.null(set.na.to)) 
          set.na.to <- quantile(10^abs(log10(data1/data2)),probs=0.99,na.rm=T)
      }
      
      ch1 <- log10(data1)
      ch2 <- log10(data2)
      
      lratio <- ch1 - ch2
      avg <- 0.5*(ch1 + ch2) 
      
      if (!is.null(set.na.to)) {
        sel <- is.na(data1);
        lratio[sel] <- log10(set.na.to)
        avg[sel] <- ch2[sel]
        sel <- is.na(data2); 
        lratio[sel] <- -log10(set.na.to)
        avg[sel] <- ch1[sel]
      }
      
      result <- nlminb(initial.parameters,fit.f,
          lower=rep(1e-10,length(initial.parameters)),x=lratio,y=avg,...)
      
      result$par
    }
)


setMethod(".averageParameters",
    signature(x="ExponentialNoiseModel",parameters="matrix"),
    function(x,parameters) {
      parameter <- numeric()
      parameter[1] <- mean(parameters[,1])
      parameter[2] <- mean(parameters[,2])
      parameter[3] <-
        -log(mean(parameters[,2]*exp(-parameters[,3]))) + log(parameter[2])
      parameter
    }
)

setMethod(".averageParameters",
    signature(x="ExponentialNoANoiseModel",parameters="matrix"),
    function(x,parameters) {
      parameter <- numeric()
      parameter[1] <- mean(parameters[,1])
      parameter[2] <-
        -log(mean(parameters[,1]*exp(-parameters[,2]))) + log(parameter[1])
      parameter
    }
)


setMethod(".averageParameters",
    signature(x="InverseNoiseModel",parameters="matrix"),
    function(x,parameters) { # Coincidentally the same average formula as for the exponential model!
      parameter <- numeric()
      parameter[1] <- mean(parameters[,1])
      parameter[2] <- mean(parameters[,2])
      parameter[3] <-
        -log(mean(parameters[,2]*exp(-parameters[,3]))) + log(parameter[2])
      parameter
    }
)

setMethod(".averageParameters",
    signature(x="InverseNoANoiseModel",parameters="matrix"),
    function(x,parameters) { # Coincidentally the same average formula as for the exponential model!
      parameter <- numeric()
      parameter[1] <- mean(parameters[,1])
      parameter[2] <- -log(mean(parameters[,1]*exp(-parameters[,2]))) + log(parameter[1])
      parameter
    }
)


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Accessor methods.
###

setGeneric("variance",function(x,channel1,channel2) standardGeneric("variance"))
setGeneric("stddev",function(x,...) standardGeneric("stddev"))
setGeneric("noiseFunction",function(x) standardGeneric("noiseFunction"))
setGeneric("parameter",function(x) standardGeneric("parameter"))
setGeneric("parameter<-",function(x,value) standardGeneric("parameter<-"))
setGeneric("lowIntensity",function(x) standardGeneric("lowIntensity"))
setGeneric("lowIntensity<-",function(x,value) standardGeneric("lowIntensity<-"))
setGeneric("naRegion",function(x) standardGeneric("naRegion"))
setGeneric("naRegion<-",function(x,value) standardGeneric("naRegion<-"))

## could support switching to fitting of the square-root standard deviation 
##  instead of the standard deviations themselves (as done in limma):
#setMethod("noiseFunction","NoiseModel",function(x) function(...) x@f(...)**2)
## as far as it was tested, it had little impact on the fit of the noise model

setMethod("noiseFunction","NoiseModel",function(x) x@f)

setMethod("variance",signature(x="NoiseModel",channel1="numeric",channel2="missing"),
    function(x,channel1) 0.5*(noiseFunction(x)(data=channel1,parameter=parameter(x))^2)
)
setMethod("variance",signature(x="NoiseModel",channel1="numeric",channel2="numeric"),
    function(x,channel1,channel2) {
      var1 <- 0.5*(noiseFunction(x)(data=channel1,parameter=parameter(x))^2)
      var2 <- 0.5*(noiseFunction(x)(data=channel2,parameter=parameter(x))^2)
      var1 + var2
    }
)
setMethod("stddev",signature(x="NoiseModel"),
    function(x,...) sqrt(variance(x,...))
)


setMethod("parameter","NoiseModel",function(x) x@parameter)
setReplaceMethod("parameter","NoiseModel",function(x,value) {
      x@parameter <- value
      x }
)

setMethod("lowIntensity","NoiseModel",function(x) x@low.intensity)
setReplaceMethod("lowIntensity","NoiseModel",function(x,value) {
      x@low.intensity <- value
      x }
)

setMethod("naRegion","NoiseModel",function(x) x@na.region)
setReplaceMethod("naRegion","NoiseModel",function(x,value) {
      x@na.region <- value
      x }
)
### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### show method.
###

setMethod("show",signature(object="NoiseModel"),
    function(object){
      cat(class(object),"\n")
      cat("  definition: ")
      print(noiseFunction(object))
      cat("  parameter:     [",paste(sprintf("%.2f",object@parameter),collapse="; "),"]\n")
      q <- quantile(object@na.region)
      cat("  na region:     [",paste(sprintf("%s",names(q)),sprintf("%.2f",q),collapse="; "),"]\n")
      cat("  low.intensity: ",sprintf("%.2f",object@low.intensity),"\n")
    }
)


plot.NoiseModel <- function(x,y=NULL,...) {
  args = list(...)
  if (is.null(args$xlim)) args$xlim = c(2,10)
  ss <- seq(from=args$xlim[0],to=args$xlim[1],length.out=100)
  nm.sd <- 1.96*sqrt(variance(x,ss))
  plot(ss,10^nm.sd,ylim=c(10^-max(nm.sd),10^max(nm.sd)),col="red",lwd=2,type="l",log="y",xlab="log10 intensity",ylab="ratio",...)
  lines(ss,10^(-nm.sd),col="red",lwd=2)
}

#################################################
## Helper Functions for LaTeX Report Generation

load.tex.properties <- function(env) {
  get.property <- function(name) { get(name,env$properties.env,inherits=FALSE) }
  if (!exists("properties.env",envir=env)) {
    warning("No properties.env - loading properties")
    assign("properties.env",load.properties(),envir=env)
  }
  needed.objects <- c('ibspectra','noise.model','quant.tbl','ratiodistr')
  if (!all(needed.objects %in% objects(envir=env))) {
    warning("Not all necessary objects present - calling initialize.env")
    initialize.env(env,env$properties.env) 
  }
  assign("get.property",get.property,envir=env)
}

write.tex.commands <- function() {
  get.property <- get("get.property")
  cat("\\newcommand{\\analysisname}{",sanitize(get.property('name'),dash=FALSE),"}\n")
  cat("\\newcommand{\\analysisauthor}{",sanitize(get.property('author'),dash=FALSE),"}\n")

  cat("\\newcommand{\\isobarthanks}{\\thanks{This report was 
    generated using the \\texttt{isobar} R package version ",
    packageDescription("isobar")$Version,
    " [built using ",packageDescription("isobar")$Built,"]",
    ". If you use it in published work, please cite ",
    "'Breitwieser FP \\textit{et~al.}: General statistical ",
    " modeling of data from protein relative expression isobaric tags, ",
    " \\textit{Journal of Proteome Research} 2011' and ",
    "'Breitwieser FP and Colinge J: isobarPTM: A software tool for the quantitative analysis of post-translationally modified proteins,",
    "\\textit{Journal of Proteomics} 2013","'}}\n",sep="")
}

print_longtablehdr <- function(level,is.single.comparision,is.quant.tbl,file="") {

  draw.channels <- !is.single.comparision
  draw.signcol <- is.quant.tbl

  if (is.single.comparision) {
    coldef <- "@{}rXfrrrr@{}r@{}r@{}"
    coldef.s <- "@{}rXfrrrr@{}r@{}"
    ncol.p <- 9
  } else {
    coldef <- "@{}rXfrrrrrrr@{}r@{}"
    coldef.s <- "@{}rXfrrrrrrr@{}"
    ncol.p <- 11
  }

  if (!is.quant.tbl) {
    coldef <- coldef.s 
    ncol.p <- ncol.p -1
  }


  mycat <- function(...,append=TRUE,sep="") 
    cat(...,file=file,append=append,sep=sep)

  mycat("\\begin{longtable}{",coldef,"}\n",
        "  \\# ",paste(" & \\textbf{",level,"}")," & \\rh{group}",
        " & \\rh{peptides}"," & \\rh{spectra}",append=FALSE)
  
  # no ch1/ch2 columns when only one comparision available
  if (draw.channels) 
    mycat("\n  & \\rh{ch1} & \\rh{ch2}")
  
  mycat("\n  & \\rh{quant} & \\textbf{ratio}")
  if (draw.signcol) 
    mycat(" & ")
  
  mycat("\n  & {\\hfill \\drawaxis{3pt}{south}}",
        "\n\\endhead \n",rep(" &",  ncol.p-1),
       " {\\hfill \\drawaxis{-3pt}{north}}",
       "\n \\endfoot \n")
}

print_longtablehdr_peptide <- function(coldef,draw.channels,ncol.p,draw.signcol) {
  #cat("\n\n\\renewcommand{\\arraystretch}{0.75}\n")
  cat("\\begin{longtable}{",coldef,"}",'\n',sep="")
  cat("\t \\# ",paste(" & \\textbf{peptide}"))  
  # no ch1/ch2 columns when only one comparision available
  if (draw.channels) {
    cat('\n')
    cat(" & \\rh{ch1}",
        " & \\rh{ch2}",sep="\n\t")
  }
  cat("\n\t"," & \\rh{quant}")
  cat("\n\t"," & \\textbf{ratio}")
  if (draw.signcol) {
    cat("\n\t"," & ")
    #cat("\n\t"," & \\textbf{*}")
  }
  cat("\n\t"," & {\\hfill \\drawaxis{3pt}{south}}",
      "\n \\endhead \n",rep(" &",  ncol.p-1),
      " {\\hfill \\drawaxis{-3pt}{north}}",
      "\n \\endfoot \n")
}


draw.boxplot <- function(lratio,sd,bnd) { 
  if (is.na(lratio) || is.na(sd) || !is.finite(lratio) || !is.finite(sd)) {
    return("")
  }

  ratio.smaller.bnd <- lratio - sd < -bnd
  ratio.bigger.bnd  <- lratio + sd > bnd

  return(sprintf("\\boxplot{%.2f}{%.2f}{%s!%s}{%s}{%s}{%.0f}\n",
                  lratio,
                  sd,
                  ifelse(lratio > 0,"green","red"),
                  min(floor(abs(lratio)/bnd*100),100),
                  ifelse(ratio.smaller.bnd, "\\divcol", "black!1"),
                  ifelse(ratio.bigger.bnd,  "\\divcol", "black!1"),
                  sd))
}

transform_pepmodif <- function(pep.n.modif) {
  message(paste(pep.n.modif,collapse=" - "))
  pep <- strsplit(pep.n.modif[1],"")[[1]]
  modif <- strsplit(pep.n.modif[2],":")[[1]]
  if (length(pep)+1 != length(modif))
    stop("lengthwise i dont like them: ",pep.n.modif[1],
         " and ",pep.n.modif[2])
#  if (modif[length(modif)] != "")
#    stop("modif contains something in the end:",
#         pep.n.modif[2])
  
  modif <- modif[seq(from=2,to=length(modif))]
  pos <- as.numeric(sapply(modif,function(m) which(modifs[,1] == m)))

  pep[!is.na(pos)] <- paste("\\pdftooltip{\\textcolor{",modifs[pos[!is.na(pos)],3],
                           "}{\\underline{",pep[!is.na(pos)],"}}}{",
                            modifs[pos[!is.na(pos)],1],"}",sep="")
  return(paste(pep,collapse=""))
}

modifs <-
  matrix(c(
           "iTRAQ4plex_Nterm","i","black",
           "iTRAQ4plex_K","i","black",
           "Oxidation_M","o","black",
           "Cys_CAM","c","black",
           "PHOS","c","purple"),
         byrow=TRUE,ncol=3)

.bnds <- function(x,bnd=NULL,min.x=-bnd,max.x=bnd) {
 max(
      min(x,max.x),
      min.x)
}

# tex helper functions
draw.proteingroup.row <- function(name,protein.group,reporter.protein.g,file=file) {

  mycat <- function(...,append=TRUE,sep="") 
    cat(...,file=file,append=append,sep=sep)

  gmp <- groupMemberPeptides(protein.group,reporter.protein.g,TRUE)
  pgt <- proteinGroupTable(protein.group)

  n.multirow <- length(unique(protein.ac(protein.group,pgt[pgt$reporter.protein==reporter.protein.g,"protein.g"])))
  mycat(name,"&",sep=" ")
  n.row <- 1
  show.pos <- TRUE
  for (protein.i in seq_len(ncol(gmp$group.member.peptides))) {
    x = colnames(gmp$group.member.peptides)[protein.i]

    my.protein.info <- my.protein.info(protein.group,x)
    my.acs <- unique(my.protein.info$accession)
    for (ac.i in seq_along(my.acs)) {
      if (n.row > 1) mycat(" & ") 
      sel <- my.protein.info$accession == my.acs[ac.i]
      var.string <- number.ranges(my.protein.info$splicevariant[sel])
      #cat(protein.i,"&")
      if (show.pos && ncol(gmp$group.member.peptides)>1) {
        if (ac.i == 1)
          mycat("{\\small (",protein.i,")}",sep="")
        else
          mycat ("$\\mathord{\\cdot}$")
      }

      mycat(" & ")
      mycat(sprintf("\\uniprotlink{%s}%s",sanitize(my.acs[ac.i],dash=FALSE),ifelse(is.na(var.string),"",paste0("-",var.string)[1])),
          sanitize(unique(my.protein.info$gene_name[sel]))[1],
          sanitize(unique(my.protein.info$protein_name[sel]))[1],
          sep=", ")

      if (n.row==1) {
        # PEPTIDE GROUPING
        mycat(" & ")
        if (n.multirow > 1) mycat(" \\multirow{",n.multirow ,"}{*}{%\n",sep="")
        mycat(tikz.proteingroup(protein.group,reporter.protein.g,show.pos,show.header=FALSE))
        if (n.multirow > 1) mycat(" }")
        mycat(" \\\\ \n")
      } else {
        mycat(" & \\\\ \n")
      }
      n.row <- n.row + 1
          
    }
    #human.protein.name <- human.protein.names(my.protein.info)
  } 
}


draw.protein.group <- function(protein.group,reporter.protein.g) {
  gmp <- groupMemberPeptides(protein.group,reporter.protein.g,TRUE)
  pgt <- proteinGroupTable(protein.group)
  show.pos <- ncol(gmp$group.member.peptides) > 1
# cat("\\paragraph{Proteins}\n")
  if (show.pos) {
    cat("\\begin{tabular}{rl@{}llp{5cm}l}\n")
    cat(paste("pos","accession","","gene name","protein name",sep=" & "))
  } else {
    cat("\\begin{tabular}{l@{}llp{5.5cm}l}\n")
    cat(paste("accession","","gene name","protein name",sep=" & "))
  }
  cat(" & \\multirow{",length(unique(protein.ac(protein.group,
                                                pgt[pgt$reporter.protein==reporter.protein.g,"protein.g"])))+1,
      "}{*}{%\n",sep="")
  tikz.proteingroup(protein.group,reporter.protein.g,show.pos)
  cat("} \\\\ \n")
  for (protein.i in seq_len(ncol(gmp$group.member.peptides))) {
    x = colnames(gmp$group.member.peptides)[protein.i]

    my.protein.info <- my.protein.info(protein.group,x)
    for (ac in unique(my.protein.info$accession)) {
      sel <- my.protein.info$accession == ac
      var.string <- number.ranges(my.protein.info$splicevariant[sel])
      if (show.pos) cat(protein.i,"&")
      cat(paste(sprintf("\\uniprotlink{%s}",sanitize(ac,dash=FALSE)),
          ifelse(is.na(var.string),"",var.string),
          sanitize(unique(my.protein.info$gene_name[sel])),
          sanitize(unique(my.protein.info$protein_name[sel])),"",
          sep=" & "),"\\\\ \n")
          
    }
    #human.protein.name <- human.protein.names(my.protein.info)
  } 
  cat("\\end{tabular}\n")

}

tikz.proteingroup <- function(protein.group,reporter.protein.g,show.pos,show.header=TRUE,
                              colors=c("blue","red","green","cyan","magenta","yellow")) {

  gmp <- groupMemberPeptides(protein.group,reporter.protein.g,TRUE)
  reporter.sp.sel <- gmp$peptide.info$specificity == "reporter-specific"
  quant.sel <- gmp$peptide.info$n.shared.groups == 1 & 
               gmp$peptide.info$n.shared.proteins == 2
  group.sp.sel <- gmp$peptide.info$specificity == "group-specific" & !quant.sel
  unspecific.sel <- gmp$peptide.info$specificity == "unspecific"
  
  peptide.styles <- rep("us",nrow(gmp$peptide.info))
  peptide.styles[reporter.sp.sel] <- "rs"
  peptide.styles[group.sp.sel] <- "gs"

  res <- c()

  gd.proteins <- apply(gmp$group.member.peptides[quant.sel,,drop=FALSE],2,any)
  col.i <- 1
  for (quant.prot.g in names(gd.proteins)[gd.proteins]) {
    if (quant.prot.g != reporter.protein.g) {
      quant.p.sel <- quant.sel & gmp$group.member.peptides[,quant.prot.g]
      peptide.styles[quant.p.sel] <- sprintf("gs %s",colors[col.i])
      if (col.i == length(colors)) {
        col.i <- 1
      } else {
        col.i <- col.i + 1
      }
    }
  }

  n.peptides <- length(reporter.sp.sel)
  max.n.peptides <- 10
  tikz.y <- 0.5
  tikz.x <- 0.4
  tikz.minnodesize <- 0.2

  if (n.peptides > max.n.peptides) {
    tikz.x <- round(tikz.x*max.n.peptides/n.peptides,2)
    tikz.minnodesize <- round(tikz.minnodesize*max.n.peptides/n.peptides,2)
  }
  
  res <- paste(res,sprintf("\\begin{tikzpicture}[x=%scm,y=%scm,every node/.style={minimum size=%scm}]\n",tikz.x,tikz.y,tikz.minnodesize))
  if (show.header)
    res <- paste(res,"  \\node at (1,0)[anchor=west] {peptides};\n")

  n.groupmember <- ncol(gmp$group.member.peptides)
  protein.peptides.df <- data.frame(
    i=seq_len(n.groupmember),
    rs=0,
    gs=0,
    us=0)

  for (protein.i in seq_len(n.groupmember)) {
    x = colnames(gmp$group.member.peptides)[protein.i]

    my.protein.info <- my.protein.info(protein.group,x)
    human.protein.name <- human.protein.names(my.protein.info)
    
    name.i <- 1
    table.name.nolink <- human.protein.name$ac_nolink[1]
    nch.tablename <- nchar(table.name.nolink)
    while (name.i < length(human.protein.name$ac_link)) {
      nch.tablename <- nch.tablename + nchar(human.protein.name$ac_nolink[name.i])
      if (nch.tablename < 20) {
        table.name.nolink <- paste(table.name.nolink,
                             human.protein.name$ac_nolink[name.i],sep=",")
      } else {
        table.name.nolink <- paste(table.name.nolink,"\\dots",sep=",")
        name.i <- Inf
      }
      name.i <- name.i + 1
    }

    peptide.idx.sel <- gmp$group.member.peptides[,protein.i]
    protein.peptides.df[protein.i,] <- c(protein.i,
                sum(peptide.idx.sel&reporter.sp.sel),
                sum(peptide.idx.sel&(group.sp.sel|quant.sel)),
                sum(peptide.idx.sel&unspecific.sel))
   
     res<- paste(res,.print.proteinrow(
        -protein.i,paste(which(peptide.idx.sel),peptide.styles[peptide.idx.sel],sep="/")
     ))
   
     if (n.peptides < 15) {
       res <- paste(res,connect.nodes(-protein.i,which((reporter.sp.sel | group.sp.sel | unspecific.sel | quant.sel)&gmp$group.member.peptides[,protein.i])))
     }
  }

  res <- paste(res,sprintf("  \\node [left=.25cm of prot-1 pep1] {%s/%s/%s};\n",
              sum(reporter.sp.sel),
              sum(group.sp.sel),
              sum(unspecific.sel)))


  if (show.header) {
    res <- paste(res,"\n","  \\matrix[ampersand replacement=\\&,matrix anchor=us node.east,anchor=base] at (0,0) {\n")
    res <- paste(res,"\n","    \\node{}; \\& \\node {rs};  \\& \\node {gs}; \\& \\node (us node){us};  \\\\\n")
    for (protein.i in seq_len(n.groupmember)) {
       if (show.pos) res <- paste(res,"\n",sprintf("    \\node[draw,gray]{%s};",protein.i))
       res <- paste(res,"\n",sprintf("    \\& \\node{%s};\\& \\node{%s};\\& \\node{%s}; \\\\\n",
                   protein.peptides.df[protein.i,"rs"],
                   protein.peptides.df[protein.i,"gs"],
                   protein.peptides.df[protein.i,"us"]))
    }
    res <- paste(res,"\n","  };\n")
  }
  res <- paste(res,"\\end{tikzpicture}\n")

  return(res)
}

.tex.combinenames <- function(protein_name,is.single.comparision,cmbn) {
  gene.names <- c()
  for (prot.gene.name in protein_name) {
    if (is.na(prot.gene.name)) next;
    prot.gene.name <- ifelse(nchar(prot.gene.name)>80,
                             paste(sanitize(substr(prot.gene.name,0,76)),"\\dots"),
                             sanitize(prot.gene.name))
    gene.names <- c(gene.names,prot.gene.name)
  }
  if (!is.single.comparision) {
    if (length(gene.names) > ncol(cmbn) - 1) {
      gene.names <- gene.names[seq_len(ncol(cmbn)-1)]
      gene.names[length(gene.names)] <- paste(gene.names[length(gene.names)],", \\dots",sep="")
    }
  }
  return(gene.names)
}

.print.proteinrow <- function(protein.idx,peptides.sel) {
  if (length(peptides.sel)>0) {
      sprintf("  \\proteinrow{%s}{%s}{}\n",
                  protein.idx,paste(peptides.sel,collapse=","))
  }
}

connect.nodes <- function(protein.idx,pep.pos) {
   if (length(pep.pos) == 1) return();
#   message(protein.idx," (",length(pep.pos),"):   ",paste(pep.pos,collapse=","),"\n")
   last_pos = pep.pos[1];
   res <- sprintf("  \\draw (prot%s pep%s)",protein.idx,last_pos)
   for (i in 2:length(pep.pos)) {
      if (pep.pos[i] == (last_pos+1)) {
         res <- paste(res,"--")
      }
      last_pos <- pep.pos[i]
      res <- paste(res,sprintf("(prot%s pep%s)",protein.idx,last_pos))
   }
   paste(res,";\n")
}


# function adapted from iquantitator
sanitize <- function(str,dash=TRUE) {
  result <- str
  result <- gsub("\\\\", "SANITIZE.BACKSLASH", result)
  result <- gsub("$", "\\$", result, fixed = TRUE)
  if (dash)
    result <- gsub("-", "\\nobreakdash-", result, fixed = TRUE)
  result <- gsub(">", "$>$", result, fixed = TRUE)
  result <- gsub("<", "$<$", result, fixed = TRUE)
  result <- gsub("|", "$|$", result, fixed = TRUE)
  result <- gsub("{", "\\{", result, fixed = TRUE)
  result <- gsub("}", "\\}", result, fixed = TRUE)
  result <- gsub("%", "\\%", result, fixed = TRUE)
  result <- gsub("&", "\\&", result, fixed = TRUE)
  result <- gsub("_", "\\_", result, fixed = TRUE)
  result <- gsub("#", "\\#", result, fixed = TRUE)
  result <- gsub("^", "\\verb|^|", result, fixed = TRUE)
  result <- gsub("~", "\\~{}", result, fixed = TRUE)
  result <- gsub("SANITIZE.BACKSLASH", "$\\backslash$", result, fixed = TRUE)
  return(result)
}


print_sign_proteins_tbl <- function(file,cmbn,protein.group,quant.tbl,my.protein.infos,bnd) {

  is.single.comparision <- ncol(cmbn) ==1
  mycat <- function(...,append=TRUE,sep="") 
    cat(...,file=file,append=append,sep=sep)

  print_longtablehdr("protein",is.single.comparision,FALSE,file=file)
  draw.hline <- FALSE
  
  for (cmb_i in seq_len(ncol(cmbn))) {
    sel <- quant.tbl[["r1"]]==cmbn[1,cmb_i] &
           quant.tbl[["r2"]]==cmbn[2,cmb_i] &
           quant.tbl[["is.significant"]] == 1
    sel[is.na(sel)] <- FALSE
    if (any(sel)) {
      if (draw.hline) {
        mycat (" \\hline\n")
      } else {
        draw.hline = TRUE
      }
      prot_i <- 1
  
      proteins <- quant.tbl[sel,"ac"][order(quant.tbl[sel,"lratio"])]
      for (protein in proteins) {
        prot.info <- my.protein.infos[[protein]]
        protein.row <- quant.tbl[quant.tbl[["ac"]]==protein & sel,,drop=FALSE]
        reporter.peptides <- peptides(protein.group,protein=protein)
        n.spectra <- length(names(spectrumToPeptide(protein.group))[spectrumToPeptide(protein.group)%in%reporter.peptides])

        protein.names <- prot.info[["collapsed.gene_name"]][["protein_name"]]
        gene.names <- .tex.combinenames(protein.names,is.single.comparision,cmbn)

        mycat(" ",sprintf("\\hyperref[protein.%s]{%s}",
                          protein.row[,"group"],prot_i),
              " & ",paste0(prot.info[["table.name"]],": ",
                           paste(gene.names,collapse=", ")),
              " & ",print_groupsize(prot.info[["n.reporter"]],prot.info[["n.groupmember"]]),
              " & ",length(reporter.peptides),
              " & ",n.spectra)
        if (!is.single.comparision) {
            mycat(" & ",cmbn[1,cmb_i]," & ",cmbn[2,cmb_i])
        }
        mycat(" & ",protein.row[1,"n.spectra"])
        mycat(" & ",sprintf("\\textbf{%.2f}",10^protein.row[1,"lratio"]),sep=" ")
        mycat(" & ",draw.boxplot(protein.row[1,'lratio'],
                               protein.row[1,'sd'],bnd))

        mycat(" \\\\  \n")
        prot_i <- prot_i + 1
      }
    }
  }
  mycat("\\end{longtable}\n")

}


print_groupsize <- function (r,t) {
  if (r>1 || t>1)
    paste0(r,"/",t)
}

get_n_proteins <- function(quant.tbl,sign=FALSE,is.na=FALSE) {
  n.quant <- ddply(quant.tbl,
                   c("class1","class2"),
                   function(x) {
                     sel <- !is.na(x[["lratio"]])
                     if (is.na) sel <- !sel
                     if (sign) sel <- sel & x[["is.significant"]]
                     sum(sel)
                   })
  n.quant <- n.quant[n.quant[["V1"]]>0,]

  paste(sprintf("%s/%s: %s",
                    sanitize(n.quant[["class2"]]),
                    sanitize(n.quant[["class1"]]),
                    n.quant[["V1"]]),collapse="; ")
}

print_protein_quant_tbl <- function(file="",
                                    cmbn,protein.group,
                                    quant.tbl,
                                    my.protein.infos,
                                    bnd) {

  message("Writing protein quantifications table ... ",append.lf=FALSE)

  is.single.comparision <- ncol(cmbn) ==1
  mycat <- function(...,append=TRUE,sep="") 
    cat(...,file=file,append=append,sep=sep)

  mr <- function(text,...) if (is.single.comparision) { text } else { sprintf("\\mr{%s}",text) }
  mrp <- function(text,...) if (is.single.comparision) { text } else {sprintf("\\mrp{%s}",text)}

  print_longtablehdr("protein",is.single.comparision,TRUE,file=file)

  proteins.n.group <- unique(quant.tbl[,c('ac','group')])
  proteins <- proteins.n.group[order(proteins.n.group[["group"]]),"ac"]
  
  for (protein in proteins) {
    protein.rows <- quant.tbl[quant.tbl[["ac"]]==protein,,drop=FALSE]
    protein.rows <- protein.rows[order(protein.rows[["r1"]],protein.rows[["r2"]]),,drop=FALSE]
    protein.groupnumber <- protein.rows[1,"group"]
    prot.info <- my.protein.infos[[protein]]
  
    reporter.peptides <- peptides(protein.group,protein=protein)
    spectra <- names(spectrumToPeptide(protein.group))[spectrumToPeptide(protein.group)%in%reporter.peptides]
  
    if (all(is.na(protein.rows[,'lratio']))) {
      next
    }
  
    protein.names <- prot.info[["collapsed.gene_name"]][["protein_name"]]
    gene.names <- .tex.combinenames(protein.names,is.single.comparision,cmbn)
  
    mycat(mr(sprintf(" \\hyperref[protein.%s]{\\textbf{%s}}",
                   protein.groupnumber,protein.groupnumber)),
        " & ",mrp(paste0(prot.info[["table.name"]],ifelse(length(gene.names)>0,": ",""),
                 paste(gene.names,collapse=", "))),
      #  " & ",mr(paste(prot.info[["n.reporter"]],prot.info[["n.groupmember"]],sep="/")),
        " & ",mr(print_groupsize(prot.info[["n.reporter"]],prot.info[["n.groupmember"]])),
        " & ",mr(length(reporter.peptides)),
        " & ",mr(length(spectra)),sep=" ")
  
    for (i in seq_len(ncol(cmbn))) {
      if (i > 1) { mycat(paste(rep("&",4),collapse=" ")) }
      if (!is.single.comparision) {
          mycat(" & ",protein.rows[i,'r1'])
          mycat(" & ",protein.rows[i,'r2'])
      }
      mycat(" & ",ifelse(is.na(protein.rows[i,'n.spectra']) | 
                    protein.rows[i,'n.spectra']==0,"",protein.rows[i,'n.spectra']))
      if (is.na(protein.rows[i,"lratio"])) {
        mycat (" & & & ")
      } else {
        mycat(" & ",sprintf("\\textbf{%.2f}",10^protein.rows[i,"lratio"]))
        mycat(" & ",ifelse(protein.rows[i,"is.significant"] == 1,"*",""))
        mycat(" & ",draw.boxplot(protein.rows[i,'lratio'],
                               protein.rows[i,'sd'],bnd))
      }
      mycat(" \\\\")
      if (i < ncol(cmbn)) mycat("*\n")
    }
    if (protein != proteins[length(proteins)] && ncol(cmbn) > 1) {
        mycat(" \\midrule[0.02em] \n\n");
    }
  }
  mycat("\\end{longtable}\n")
  message("done")
}

print_classlabels_tbl <- function(cl,reporterTagNames) {
  if (!is.null(cl)) {
    cl[is.na(cl)] <- ""
    if (!is.null(cl)) {
      if (is.null(names(cl))) {
        cat("\\begin{tabular}{rp{4cm}}\n")
        cat("Channel & Class \\\\ \\hline \\bigstrut[t]\n")
        cl <- paste("\\emph{",cl,"}",sep="")
      } else {
        cat("\\begin{tabular}{rrM}\n")
        cat("\\multicolumn{3}{c}{Class Labels} \\bigstrut \\\\ \n")
        cat("Channel & Class & Description \\\\ \\hline \\bigstrut[t]\n")
        cl <- paste(paste("\\emph{",cl,"} & "),names(cl),sep="")
      }
      for (i in seq_along(reporterTagNames)) {
        cat(reporterTagNames[i]," & ",
            cl[i],
            "\\\\ \n")
      }
      cat("\\end{tabular}\n")
    }
  }
  
}

print_protein_notquant_tbl <- function(file="",
                                    cmbn,protein.group,
                                    quant.tbl,
                                    my.protein.infos) {

  is.single.comparision <- ncol(cmbn) ==1
  mycat <- function(...,append=TRUE,sep="") 
    cat(...,file=file,append=append,sep=sep)


  tt <- table(quant.tbl[is.na(quant.tbl[["lratio"]]),"ac"])
  proteins.notquantified <- names(tt)[tt==ncol(cmbn)]
  mycat("\\begin{longtable}{rXrrr}",
  "  \\#",
  "  & \\textbf{protein}",
  "  & \\rh{group}",
  "  & \\rh{peptides}",
  "  & \\rh{spectra}",
  "\\endhead",sep="\n",append=FALSE)
  
  for (protein in proteins.notquantified) {
    protein.rows <- quant.tbl[quant.tbl[["ac"]]==protein,,drop=FALSE]
    protein.rows <- protein.rows[order(protein.rows[["r1"]],protein.rows[["r2"]]),,drop=FALSE]
    protein.groupnumber <- protein.rows[1,"group"]
    prot.info <- my.protein.infos[[protein]]
  
    reporter.peptides <- peptides(protein.group,protein=protein)
    spectra <- names(spectrumToPeptide(protein.group))[spectrumToPeptide(protein.group)%in%reporter.peptides]
  
    protein.names <- prot.info[["collapsed.gene_name"]][["protein_name"]]
    gene.names <- .tex.combinenames(protein.names,is.single.comparision,cmbn)
  
    mycat(sprintf(" \\hyperref[protein.%s]{%s}",
                   protein.groupnumber,protein.groupnumber),
        " & ",paste0(prot.info[["table.name"]],": ",
                 paste(gene.names,collapse=", ")),
        " & ",print_groupsize(prot.info[["n.reporter"]],prot.info[["n.groupmember"]]),
        " & ",length(reporter.peptides),
        " & ",length(spectra),"\\\\ \n",sep=" ")
  }
  mycat("\\end{longtable}\n")
}

print_protein_grp_tbl <- function(file="",proteins, protein.group) {
  mycat <- function(...,append=TRUE,sep="") 
    cat(...,file=file,append=append,sep=sep)

  mycat("\\begin{longtable}{rcXp{6cm}}\n",append=FALSE)
  prot_i <- 1
  for (reporter.protein.g in proteins) {
    draw.proteingroup.row(sprintf("\\label{protein.%s} %s",prot_i,prot_i),protein.group,reporter.protein.g,file=file)
    mycat("\\bigstrut\n")
    if (prot_i%%100 == 0) message(".",appendLF=FALSE);
    prot_i <- prot_i + 1
  }
  mycat("\\end{longtable}\n")
  
}

print_protein_grp_info <- function() {
cat("
\\section{Protein Group Details} \\label{sec:proteingroups}

Protein groups are created by grouping according to mass spectrometry evidence:

\\begin{tabularx}{\\textwidth}{rX}
1. & Proteins which are detected with the exact same set of peptides are clustered - they are indistinguishable based on available MS data. Very often these are splice variants. In the following text, proteins are clustered proteins (cluster size one if they have specific peptides).  \\\\ 
2. & Proteins, which have unique peptides (\\emph{i.\\,e.}~peptides not present in other detected proteins) are promoted to \\emph{reporters}. \\\\ 
3. & Remaining proteins are grouped to those reporters which have a superset of the peptides they have. These proteins are termed \\emph{group members} \\\\ 
\\end{tabularx}

Peptides are characterized based on their specificity to one or more proteins, and one or more protein groups:
\\begin{center}
\\begin{tabular}{rll}
\\emph{reporter specific} rs &  \\tikz \\node[rs,minimum size=0.2cm] {};  & Number of peptides specific to reporter. \\\\
\\emph{group specific} gs    &  \\tikz \\node[gs,minimum size=0.2cm] {};  & Number of peptides specific to the group (including rs). \\\\
\\emph{unspecific} us        &  \\tikz \\node[us,minimum size=0.2cm] {};  & Number of peptides shared with any protein. \\\\
\\end{tabular}
\\end{center}
")
}

report.level="protein"
compile=TRUE

## report name
name  <- basename(getwd())

## Samples to be combined:
##  Name subdirectories which contain ibspectras and quant tables
samples <- sub("./","",list.dirs(recursive=FALSE),fixed=TRUE)

## regenerate cache files
regen <- FALSE

## The samples to be combined must have been analyzed using a 
## common protein group template, such that the protein group 
## identifiers are the same
protein.group <- "protein.group.rda"

significance.method='p.value'

## Via database or internet connection, informations on proteins 
## (such as gene names and length) can be gathered. protein.info.f
## defines the function which takes a ProteinGroup object as argument
protein.info.f <- getProteinInfoFromUniprot

## Where should cached files be saved? Will be created if it does not exist
# cachedir="cache"
cachedir="."

## Modification to track. Use 'PHOS' for phosphorylation.
# ptm <- c('ACET','METH','UBI','SUMO', 'PHOS')
ptm <- NULL

## file name of rda or data.frame with known modification sites
## gathered with ptm.info.f.  defaults to 'cachedir/ptm.info.rda'
ptm.info <- NULL

## Function to get PTM modification sites from public datasets
# ptm.info.f <- getPtmInfoFromNextprot
# ptm.info.f <- function(...)
#    getPtmInfoFromPhosphoSitePlus(...,modification="PHOS")
# ptm.info.f <- function(...)
#    getPtmInfoFromPhosphoSitePlus(...,modification=ptm)
ptm.info.f <- getPtmInfoFromNextprot

## XLS report format 'wide' or 'long '.

## 'wide' format outputs ratios in separate columns of the same record
## (i.e. one line per protein)
## 'long' format outputs ratios in separate records (i.e. one line per
## ratio)
xls.report.format="wide"
#xls.report.format="long"

zscore.threshold <- 2.5


spreadsheet.format="xlsx"

create.meta.reports <- function(properties.file="meta-properties.R",
                                global.properties.file=system.file("report","meta-properties.R",package="isobar"),
                                args=NULL,include.not.localized=FALSE,...) {
  requireNamespace("ggplot2")
  if (!exists("properties.env")) {
    properties.env <- load.properties(properties.file,global.properties.file,
                                      args=args,...)
  }

  protein.group <- .get.or.load("protein.group",properties.env,"protein group object","ProteinGroup")

  pg.df <- .create.or.load("pg.df",properties.env,function() {
    if (properties.env[["report.level"]]=="peptide") {
      ptm.info  <- .create.or.load.ptm.info(properties.env,protein.group)
      pg.df <- isobar:::.proteinGroupAsConciseDataFrame(protein.group,
                                                        modif.pos=properties.env$ptm,
                                                        ptm.info=properties.env$ptm.info,
                                                        report.only.modified=TRUE)
    } else {
      acs <- reporterProteins(protein.group)
      pg.df <- data.frame(ac=acs,
                          AC=.protein.acc(acs,protein.group),
                          ID=proteinInfo(protein.group,acs,do.warn=FALSE),
                          Description=proteinInfo(protein.group,acs,select="protein_name",do.warn=FALSE),
                          Gene=proteinInfo(protein.group,acs,select="gene_name",do.warn=FALSE),
                          stringsAsFactors=FALSE)

    }
  })

  if (length(proteinInfo(protein.group)) ==0 && !is.null(properties.env$protein.info.f)) 
    proteinInfo(protein.group) <- 
      .create.or.load("protein.info",envir=properties.env,
                      f=properties.env$protein.info.f,
                      x=protein.group,
                      error=warning,default.value=proteinInfo(protein.group))
  
  if (is.null(proteinInfo(protein.group)) || length(proteinInfo(protein.group))==0)
    stop("No protein information available.")

  #if (properties.env$calculate.dnsaf)
  #  dnsaf <- .create.or.load("dnsaf",envir=properties.env,f=calculate.dNSAF,protein.group=protein.group)

  message("Merging samples ",paste(properties.env$samples,collapse=", ")," ...")
  ac.vars <- switch(properties.env[["report.level"]],
                    protein = "ac",
                    peptide = c("peptide","modif","pep.siteprobs"),
                    stop("properties.env$report.level ",properties.env[["report.level"]]," unknown"))

  merge.cols <- c("class1","class2")

  merged.table <- .create.or.load.merged.table(properties.env=properties.env)
  merged.table.notlocalized <- .create.or.load.merged.table("merged.table.notlocalized","quant.tbl.notlocalized.rda",properties.env=properties.env)
  ac.vars <- intersect(ac.vars,colnames(merged.table))
  
  tbl.pg <- .create.or.load("tbl.pg",envir=properties.env,f=function()
                            .get.tbl.pg(merged.table,pg.df,merge.cols,ac.vars))

  if (exists("merged.table.notlocalized")) {
    tbl.pg.nl <- .create.or.load("tbl.pg.nl",envir=properties.env,f=function()
                                  .get.tbl.pg(merged.table.notlocalized,
                                             pg.df,merge.cols,ac.vars))
  }
#  .create.figures(tbl.pg)

  #tbl.wide <- .create.or.load("tbl.wide",properties.env,function() reshape(merged.table,idvar=intersect(ac.vars,colnames(merged.table)),timevar="comp",direction="wide",drop=merge.cols))

  #.plot.data(properties.env,tbl.wide,ac.vars)

  .write.xls.report(properties.env,tbl.pg,tbl.pg.nl,include.not.localized)
}

.plot.data <- function(properties.env,tbl.wide,ac.vars) {
  message("plotting data")
  ratio.matrix <- as.matrix(tbl.wide[,grep("lratio",colnames(tbl.wide))])
  variance.matrix <- as.matrix(tbl.wide[,grep("var",colnames(tbl.wide))])
  rownames(ratio.matrix)  <- do.call(paste,tbl.wide[,ac.vars,drop=FALSE])
  rownames(variance.matrix)  <- do.call(paste,tbl.wide[,ac.vars,drop=FALSE])
  sel <- !apply(is.na(ratio.matrix),1,any)
  ratio.matrix <- ratio.matrix[sel,]
  variance.matrix <- variance.matrix[sel,]

  m.median <- apply(ratio.matrix,2,median)
  normalized.ratio.matrix <- ratio.matrix-matrix(m.median,nrow=nrow(ratio.matrix),ncol=ncol(ratio.matrix),byrow=T)

  .plot.heatmaps(ratio.matrix,normalized.ratio.matrix,variance.matrix,properties.env$name)
  .plot.pairs(ratio.matrix,variance.matrix,properties.env$name)
}


.write.xls.report <- function(properties.env,tbl.pg,tbl.pg.nl,include.not.localized=FALSE) {
  message("writing Excel report")
  xlsxname <- paste0(properties.env$name,"-combined_report.xlsx")
  tab2spreadsheet.cmd <- switch(properties.env$spreadsheet.format,
                                xlsx=system.file("pl","tab2xlsx.pl",package="isobar"),
                                xls=system.file("pl","tab2xls.pl",package="isobar"),
                                stop("spreadsheet.format property must be either 'xlsx' or 'xls'."))

  csvname <- paste0(properties.env$name,"-combined_report.csv")
  write.table(tbl.pg[order(tbl.pg[,'ID'],tbl.pg[,'start.pos']),],file=csvname,
              row.names=FALSE,sep="\t",na="")
  perl.cl <- paste(tab2spreadsheet.cmd,shQuote(xlsxname),shQuote(csvname))
  
  if (include.not.localized) {
    csvname.nl <- paste0(properties.env$name,".not.localized-combined_report.csv")
    write.table(tbl.pg.nl[order(tbl.pg.nl[,'ID'],tbl.pg.nl[,'start.pos']),],file=csvname.nl,row.names=FALSE,sep="\t",na="")
    perl.cl <- paste(perl.cl,shQuote(csvname.nl))
  }
  ## generate Excel report (using Spreadsheet::WriteExcel)
  message(perl.cl)
  system(perl.cl)
}


.get.tbl.pg <- function(merged.table,pg.df,merge.cols,ac.vars) {
  tbl.wide <- reshape(merged.table,idvar=intersect(ac.vars,colnames(merged.table)),timevar="comp",direction="wide",drop=merge.cols)
  tbl.pg <- merge(pg.df,tbl.wide)#,by=intersect(ac.vars,colnames(tbl.wide)))

  sel <- apply(!is.na(tbl.pg[,grep("^lratio",colnames(tbl.pg))]),1,any)
  tbl.pg <- tbl.pg[sel,]

  tbl.pg <- tbl.pg[order(rowSums(tbl.pg[,grep("is.significant",colnames(tbl.pg))]),
                         rowSums(tbl.pg[,grep("lratio",colnames(tbl.pg))]),
                         decreasing=TRUE),]

  if ("peptide" %in% colnames(tbl.pg)) {
    tbl.pg$peptide <- isobar:::.convertPeptideModif(tbl.pg$peptide,tbl.pg$modif)
    tbl.pg$modif <- NULL
  }
  tbl.pg$ac <- NULL

  print(t(sapply(.grep_columns(tbl.wide,"lratio.P",value=TRUE,logical=FALSE),
                 function(i) isobar:::.sum.bool(!is.na(tbl.wide[,i])))))


  tbl.pg
}

.create.or.load.merged.table <- function(name="merged.table",fname="quant.tbl.rda",properties.env) {
  ac.vars <- switch(properties.env[["report.level"]],
                    protein = "ac",
                    peptide = c("peptide","modif","pep.siteprobs"),
                    stop("properties.env$report.level ",properties.env[["report.level"]]," unknown"))

  merge.cols <- c("class1","class2")

  res <- .create.or.load(name,properties.env,function() {
    merged.table <- .get.merged.table(properties.env$samples,
                                     cols=c(ac.vars,merge.cols,"lratio","variance","p.value.rat"),
                                     merge.by=c(ac.vars,merge.cols),format=properties.env[["xls.report.format"]],
                                     fname=fname)
  
    zscore.any2.5 <- apply(abs(merged.table[,grepl("zscore.indiv",colnames(merged.table))]) >= properties.env[["zscore.threshold"]],1,any,na.rm=TRUE)
    zscore.all1 <- apply(abs(merged.table[,grepl("zscore.indiv",colnames(merged.table))]) >= 1,1,all,na.rm=TRUE)

    merged.table$p.value.rat <- .combine.fisher.tblwide(merged.table)
    merged.table$is.significant.notadj <- (merged.table$p.value.rat < 0.05) & zscore.any2.5 & zscore.all1

    #merged.table$p.value.rat.fdr.adj <- p.adjust(merged.table$p.value.rat)
    merged.table <- ddply(merged.table,c("class1","class2"),function(x) 
                                 cbind(x,p.value.rat.fdr.adj=p.adjust(x$p.value.rat,"fdr")))
    merged.table$is.significant <- (merged.table$p.value.rat.fdr.adj < 0.05) & zscore.any2.5 & zscore.all1  
  
    merged.table$comp <- paste(merged.table[,merge.cols[2]],sep="/",merged.table[,merge.cols[1]])
    merged.table
  })
  #merged.table$p.value.rat <- .combine.fisher.tblwide(merged.table)
  return(res)
}

.calc.zscore <- function(qq) {
  s.median <- median(qq,na.rm=TRUE)
  s.mad <- mad(qq,na.rm=TRUE)
  (qq-s.median)/s.mad
}

.get.merged.table <- function(samples,quant.tbls=paste0(samples,"/cache/",fname),
                              cols=c("ac","r1","r2","lratio","variance"),
                              merge.by=c("ac","r1","r2"),format="wide",fname="quant.tbl.rda") {

  quant.tables <- lapply(seq_along(samples),
                         function(idx) {
                           message("loading ",quant.tbls[idx])
                           env <- new.env()
                           load(quant.tbls[idx],envir=env)
                           quant.tbl <- get(ls(envir=env),envir=env)
                           q <- quant.tbl[,intersect(cols,colnames(quant.tbl))]
                           q <- ddply(q,c("class1","class2"),function(x) 
                                 cbind(x,zscore.indiv=round(.calc.zscore(x[,"lratio"]),4)))
                           q[,"lratio"] <- round(q[,"lratio"],4)
                           q[,"variance"] <- round(q[,"variance"],4)
                           if (format=="wide") {
                             sel <- !colnames(q) %in% merge.by
                             colnames(q)[sel] <- paste(colnames(q)[sel],gsub("[./]","",samples[idx]),sep=".")
                           } else {
                             q$sample <- samples[idx]
                           }
                           message(paste(colnames(q),collapse=":"))
                           q
                         })

  if (format == "wide") {
    merged.table <- quant.tables[[1]]
    for (idx in  2:length(samples)) {
      merge.colnames <- intersect(merge.by,colnames(quant.tables[[idx]]))

      merged.table <- merge(merged.table,
                            quant.tables[[idx]],by=merge.colnames,all=TRUE)
    }
    merged.table$lratio <- rowMeans(merged.table[,.grep_columns(merged.table,"lratio")],na.rm=TRUE)
    merged.table$variance <- rowMeans(merged.table[,.grep_columns(merged.table,"variance")],na.rm=TRUE)
  } else {
    merged.table <- do.call(rbind,quant.tables)
  }
  return(merged.table)
}

.get.names <- function(p,protein.group) 
  apply(my.protein.info(protein.group,p)[,c("name","gene_name","protein_name")],2,
        function(s) paste(unique(sort(gsub("'","",s))),collapse=", "))

.calc.col <- function(tbl.wide,tag,add.name=TRUE) {
  idx.ratio <- grep("lratio",colnames(tbl.wide),fixed=TRUE)
  idx.var <- grep("var",colnames(tbl.wide),fixed=TRUE)
  x <- t(apply(tbl.wide,1,function(r) {
    pos <- grep(tag,colnames(tbl.wide),fixed=TRUE)
    idx.r <- intersect(pos,idx.ratio)
    idx.v <- intersect(pos,idx.var)
  
    r<-as.numeric(r)
    c(weighted.mean=round(10^weightedMean(r[idx.r],1/r[idx.v]),2),
      var=round(weightedVariance(r[idx.r],1/r[idx.v]),4),
      n.up=sum(r[idx.r]>0,na.rm=T),
      n.down=sum(r[idx.r]<0,na.rm=T))
  }))
  if (add.name)
    colnames(x) <- paste(colnames(x),tag,sep=".")
  return(x)
}

.write.t <- function(...,sep="\t",row.names=FALSE,quote=FALSE)
  write.table(...,sep=sep,row.names=row.names,quote=quote)

.write.summarized.table <- function(tbl.wide,allnames,cols) {
  summarized.table <- cbind(ac=rownames(tbl.wide),allnames)
  for (col in cols)
    summarized.table <- cbind(summarized.table,.calc.col(tbl.wide,col,add.name=length(cols)>1))
  summarized.table <- cbind(summarized.table,tbl.wide)
  .write.t(summarized.table,file=paste(name,"_summarized_table.csv",sep=""))
  return(summarized.table)
}

.plot.heatmaps.gd <- function(ratio.matrix,variance.matrix,name) {
  requireNamespace("gplots")
  breaks <- seq(from=-max(abs(ratio.matrix)),to=max(abs(ratio.matrix)),length.out=51)
  pdf(sprintf("heatmap_%s.pdf",name),width=15,height=30,title=name)
  min.max <- quantile(ratio.matrix,probs=c(0.0005,0.9995))
  sel <- apply(ratio.matrix<min.max[1] | ratio.matrix > min.max[2],1,any)
  if (sum(sel) > 2) {
    ratio.matrix2 <- ratio.matrix[sel,,drop=FALSE]
    gplots::heatmap.2(ratio.matrix2,Colv=NA,col=gplots::greenred(50),dendrogram="row",
              margins=c(5,25),main=paste(name,"- above fold change of",round(10^min.max[2],1),
                                "or below",round(10^min.max[1],1)),
              key=FALSE, keysize=1.0, symkey=FALSE, density.info='none',
              trace='none', colsep=1:10,
              #labRow=allnames[sel,"gene_name"],
              sepcolor='white', sepwidth=0.025,
              scale="none",cexRow=2,cexCol=2,
              labCol = colnames(ratio.matrix),                 
              hclustfun=function(c){hclust(c, method='mcquitty')},
              lmat=rbind( c(0, 3), c(2,1), c(0,4) ), lhei=c(0.25, 4, 0.25 ),breaks=breaks)
  }
  dev.off()

  requireNamespace("boot")

  cols <- colnames(ratio.matrix)
  correlation.matrix <- sapply(cols ,function(i) sapply(cols,function(j) boot::corr(ratio.matrix[,c(i,j)],
                                                            w=apply(1/variance.matrix,1,median))))

  pdf(sprintf("heatmap_correlation_%s.pdf",name),width=15,height=30,title=name)
  gplots::heatmap.2(correlation.matrix,scale="none",Rowv=NA,Colv=NA,col=gplots::greenred(50),
            density.info='none',trace="none",cellnote=round(correlation.matrix,1),main="not re")
  dev.off()

}


.plot.heatmaps <- function(ratio.matrix,normalized.ratio.matrix,variance.matrix,name) {
  requireNamespace("gplots")
  breaks <- seq(from=-max(abs(ratio.matrix)),to=max(abs(ratio.matrix)),length.out=51)
  pdf(sprintf("heatmap_%s.pdf",name),width=15,height=30,title=name)
  gplots::heatmap.2(ratio.matrix,Colv=NA,col=gplots::greenred(50),margins=c(5,25),main=paste(name),
#            labRow=allnames[,"gene_name"],
            lmat=rbind( c(0, 3), c(2,1), c(4,0) ), lhei=c(0.5, 5, 0.5 ),
            scale="none",trace="none",labRow=NA,
            hclustfun=function(c){hclust(c, method='mcquitty')},breaks=breaks)

#  gplots::heatmap.2(normalized.ratio.matrix,Colv=NA,col=greenred(50),margins=c(5,25),main=paste(name,"- renormalized"),
 #           labRow=allnames[,"gene_name"],
#            lmat=rbind( c(0, 3), c(2,1), c(4,0) ), lhei=c(0.5, 5, 0.5 ),scale="none",trace="none",cexCol=0.7,labRow=NA,
#            hclustfun=function(c){hclust(c, method='mcquitty')},breaks=breaks)

  min.max <- quantile(ratio.matrix,probs=c(0.005,0.995))
  sel <- apply(ratio.matrix<min.max[1] | ratio.matrix > min.max[2],1,any)
  if (sum(sel) > 2) {
    ratio.matrix2 <- ratio.matrix[sel,,drop=FALSE]
    gplots::heatmap.2(ratio.matrix2,Colv=NA,col=gplots::greenred(50),dendrogram="row",
              margins=c(5,25),main=paste(name,"- above fold change of",round(10^min.max[2],1),
                                "or below",round(10^min.max[1],1)),
              key=FALSE, keysize=1.0, symkey=FALSE, density.info='none',
              trace='none', colsep=1:10,
#              labRow=allnames[sel,"gene_name"],
#              sepcolor='white', sepwidth=0.025,
              labRow=rownames(ratio.matrix2),
              scale="none",cexRow=1.5,cexCol=1,
              labCol = colnames(ratio.matrix),                 
              hclustfun=function(c){hclust(c, method='mcquitty')},
              lmat=rbind( c(0, 3), c(2,1), c(0,4) ), lhei=c(0.25, 4, 0.25 ),breaks=breaks)
  }
  dev.off()

  requireNamespace("boot")
  cols <- colnames(ratio.matrix)
  correlation.matrix <- sapply(cols ,function(i) sapply(cols,function(j) boot::corr(ratio.matrix[,c(i,j)],
                                                            w=apply(1/variance.matrix,1,median))))

  pdf(sprintf("heatmap_correlation_%s.pdf",name),width=7,height=7,title=name)
  gplots::heatmap.2(correlation.matrix,scale="none",Rowv=NA,Colv=NA,col=gplots::greenred(50),
            density.info='none',trace="none",cellnote=round(correlation.matrix,1),main="not re")
  dev.off()

}

.scatter.plot <- function(my.table,name) {

  png(sprintf("scatter_plot%s.png",name),width=7,height=7)
  ggplot(my.table) + 
     geom_point(aes_string(x="lratio.x",y="lratio.y",color="comp")) + 
     facet_wrap(~comp)

}

.plot.pairs <- function(ratio.matrix,variance.matrix,name) {
  weights <- apply(1/variance.matrix,1,median)
  weights <- weights/sum(weights)
  
  lim <- max(abs(ratio.matrix))
  ##pdf("pairwise_correlation.pdf",title="Pairwise Correlation plot",width=10,height=10)
  png(sprintf("pairwise_correlation_%s.png",name),title="Pairwise Correlation plot",
      width=1000,height=1000,pointsize=14)
  pairs(ratio.matrix,xlim=c(-lim,lim),ylim=c(-lim,lim),
        text.panel=.panel.txt,
        upper.panel=.panel.smooth,weights=weights,
        lower.panel=function(...) .panel.cor(...,cex.cor=1.2),pch=1,cex=2)
  dev.off()

}

.panel.cor <- function(x, y, digits=2, prefix="", cex.cor, weights, ...)  {
  requireNamespace("boot")
  usr <- par("usr"); on.exit(par(usr))
  par(usr = c(0, 1, 0, 1))
  r <- abs(boot::corr(matrix(c(x,y),ncol=2),w=weights))
  ##    r <- abs(cor(x,y,method="spearman"))
  txt <- format(c(r, 0.123456789), digits=digits)[1]
  txt <- paste(prefix, txt, sep="")
  if(missing(cex.cor)) cex.cor <- 0.8/strwidth(txt)
  text(0.5, 0.5, txt, cex = cex.cor * sqrt(r) * 2)
}

.panel.smooth <- function(x, y, digits=2, prefix="", cex=1, weights, ...) {
  abline(c(0,0),1,col="red")
  points(x,y,cex=weights/max(weights)/2,...)
}

.panel.txt <- function(x,y,labels,cex,font,...) {
  requireNamespace("RColorBrewer")
  ##abc<-c("1"="rep 1","2"="rep 2");
  
  s <- strsplit(labels,".",fixed=TRUE)[[1]]
  u <- par('usr')
  names(u) <- c("xleft", "xright", "ybottom", "ytop")
  pal <- RColorBrewer::brewer.pal(8,"Pastel1")
  bgcolor <- pal[as.numeric(s[[2]])]
  ##bgcolor <- pal[as.numeric(s[[3]])-114]
  do.call(rect, c(col = bgcolor, as.list(u)))
  ##text(0.5,0.5, paste(abc[s[[2]]],"\n",s[[3]],"/114",sep=""),cex=cex)
  text(0.5,0.5, paste(s[[2]],sep=""),cex=cex)
  #text(0.5,0.5, paste(s[[2]],"\n vs ",s[[3]],sep=""),cex=cex)
}

### =========================================================================
### ProteinGroup objects
### -------------------------------------------------------------------------
###
### Class definition

setClass("ProteinGroup",
    contains = "VersionedBiobase",
    representation(
                   spectrumToPeptide = "character",
                   spectrumId = "data.frame",
                   peptideSpecificity = "data.frame",
                   peptideNProtein = "matrix",
                   indistinguishableProteins = "character",
                   proteinGroupTable = "data.frame",
                   overlappingProteins = "matrix",
                   isoformToGeneProduct= "data.frame",
                   proteinInfo = "data.frame",
                   peptideInfo = "data.frame"
    ),
    prototype = prototype(
        new("VersionedBiobase",versions=c(ProteinGroup="1.0.0")),
        peptideSpecificity = data.frame(peptide=c(),specificity=c())
    )
)


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Validity.
###

.valid.ProteinGroup.slots <- function(object) {
  peptides <- unique(object@peptideNProtein[,"peptide"])
  proteins <- unique(object@peptideNProtein[,"protein.g"])
  msg <- NULL

# TODO: check this if
#  if (!setequal(peptides,peptideSpecificity[,'peptide']) ||
#      !setequal(proteins,indistinguishableProteins[,'protein.g']) ||
#      !setequal(peptides,spectrumToPeptide[,'peptide']))
#    msg <- "slots are not of equal length"
  return(msg)
}

.valid.ProteinGroup <- function(object) {
  .valid.ProteinGroup.slots(object)
}
 
setValidity("ProteinGroup", .valid.ProteinGroup)

UNSPECIFIC="unspecific"
GROUPSPECIFIC="group-specific"
REPORTERSPECIFIC="reporter-specific"
#PROTEINSPECIFIC="protein-specific"
SPECIFICITIES=c(UNSPECIFIC,GROUPSPECIFIC,REPORTERSPECIFIC)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Constructor and readProteinGroup.
###

setGeneric("ProteinGroup",function(from,template=NULL,proteinInfo=data.frame())
           standardGeneric("ProteinGroup"))


setMethod("ProteinGroup",signature(from="data.frame",template="ProteinGroup",proteinInfo="ANY"),
          function(from,template,proteinInfo=data.frame()) {
      # TODO: exclude groups from beeing reporters when
      #       there's no reporter-specific peptide ?
      message("Creating ProteinGroup from template ... ",appendLF=FALSE)

      from <- .factor.as.character(from)
      if ('accession' %in% colnames(from)) 
        colnames(from)[colnames(from)=='accession'] <- 'protein'
        
      required.cols <- c("spectrum","peptide","modif","protein")
      required.cols.present <- sapply(required.cols,function(x) x %in% colnames(from))
      if (!all(required.cols.present)) {
        stop("Not all required columns ['spectrum','peptide','modif','accession' or 'protein'] present:\n",
             paste(required.cols[!required.cols.present],collapse=" and ")," missing")
      }
      if (!'start.pos' %in% colnames(from)) {
        warning('start.pos not in supplied data.frame, setting to NA')
        from$start.pos <- NA
      }
      
      ## Substitute Isoleucins with Leucins (indistinguishable by Masspec)
      if (!.PEPTIDE.COLS['REALPEPTIDE'] %in% colnames(from)) 
        from <- .fix.il.peptide(from)

      peptideNProtein <- peptideNProtein(template)[peptideNProtein(template)[,"peptide"] %in% from[,'peptide'],]

      protein.group.table <- proteinGroupTable(template)[proteinGroupTable(template)$protein.g %in% peptideNProtein[,"protein.g"],]
      ipt <- indistinguishableProteins(template)
      indistinguishableProteins <- ipt[ipt %in% peptideNProtein[,"protein.g"]]

      peptideSpecificity <-
        peptideSpecificity(template)[peptideSpecificity(template)[["peptide"]] %in% from[,"peptide"],]
      spectrumId <- unique(from[,setdiff(colnames(from),c("protein","start.pos"))])
      spectrumToPeptide <- spectrumToPeptide(template)[
                     names(spectrumToPeptide(template)) %in% from[,"spectrum"]]
      
      isoforms <-template@isoformToGeneProduct[names(indistinguishableProteins),]
      peptideInfo <- template@peptideInfo[template@peptideInfo$peptide %in% from[,'peptide'] & template@peptideInfo$modif %in% from[,"modif"],]
      ## TODO: overlappingProteins are missing

      if (length(proteinInfo) == 0 && length(template@proteinInfo) > 0) {
        if (proteinInfoIsOnSpliceVariants(template@proteinInfo))
          proteinInfo <- template@proteinInfo[template@proteinInfo$accession %in% from$protein,]
        else
          proteinInfo <- template@proteinInfo[template@proteinInfo$accession %in% isoforms$proteinac.wo.splicevariant,]
      }

      attr(proteinInfo,"on.splice.variant") <- attr(template@proteinInfo,"on.splice.variant")

      message("done")

      return(
          new("ProteinGroup",
              peptideNProtein = peptideNProtein,
              peptideSpecificity = peptideSpecificity,
              proteinGroupTable = protein.group.table,
              indistinguishableProteins = indistinguishableProteins,
              spectrumId = spectrumId,
              spectrumToPeptide = spectrumToPeptide,
              isoformToGeneProduct = isoforms,
              proteinInfo = proteinInfo,
              peptideInfo = peptideInfo)
      )
    }
)

readProteinGroup <- function(id.file,...,identifications.format=NULL,sep="\t",
                         decode.titles=TRUE,trim.titles=FALSE) {
  pp <- .read.idfile(id.file,identifications.format=identifications.format,
                     sep=sep,decode.titles=decode.titles,trim.titles=trim.titles)
  return(ProteinGroup(from=pp,...))
}

readProteinGroup2 <- function(id.file,...,identifications.format=NULL,
                              sep="\t",decode.titles=TRUE,trim.titles=FALSE) {
  identifications <- .read.idfile(id.file,sep=sep,
                                  identifications.format=identifications.format,
                                  decode.titles=decode.titles,trim.titles=trim.titles)

  ## Check that obligatory columns are present
  identifications <- .check.columns(identifications)

  SC <- .SPECTRUM.COLS[.SPECTRUM.COLS %in% colnames(identifications)]
  ## Substitute Isoleucins with Leucins (indistinguishable by Masspec)
  if (!.PEPTIDE.COLS['REALPEPTIDE'] %in% colnames(identifications)) 
     identifications <- .fix.il.peptide(identifications)

  ## Separate protein columns (focus on peptide-spectrum matches)
  PC <- setdiff(.PEPTIDE.COLS, SC['PEPTIDE'])
  protein.colnames <- which(colnames(identifications) %in% c(SC['PEPTIDE'],PC))
  pept.n.prot <- unique(identifications[,protein.colnames])
  identifications <- unique(identifications[,-which(colnames(identifications) %in% PC)])

  ## Merge identifications
  if (max(table(identifications[,SC['SPECTRUM']])) > 1) {
    if (SC['DISSOCMETHOD'] %in% colnames(identifications) && 
        length(unique(identifications[,SC['DISSOCMETHOD'] ]))>1) {
      identifications <- ddply(identifications,'dissoc.method',.merge.identifications)
      identifications <- .merge.quant.identifications(identifications)
    } else {
      identifications <- .merge.identifications(identifications)
    }
  }
  identifications <- .remove.duplications(identifications)
 
  # Create ProteinGroup
  proteinGroup <- ProteinGroup(merge(pept.n.prot,identifications,by="peptide"),...)
  
  return(proteinGroup)
}

setMethod("ProteinGroup",signature(from="data.frame",template="NULL",proteinInfo="ANY"),
          function(from,template,proteinInfo) ProteinGroup(from,proteinInfo=proteinInfo))

setMethod("ProteinGroup",signature(from="data.frame",template="missing",proteinInfo="ANY"),
          function(from,proteinInfo) {
      message("Creating ProteinGroup ... ",appendLF=FALSE)

      from <- .factor.as.character(from)
      if ('accession' %in% colnames(from)) 
        colnames(from)[colnames(from)=='accession'] <- 'protein'
        
      required.cols <- c("spectrum","peptide","modif","protein")
      required.cols.present <- sapply(required.cols,function(x) x %in% colnames(from))
      if (!all(required.cols.present)) {
        stop("Not all required columns ['spectrum','peptide','modif','accession' or 'protein'] present:\n",
             paste(required.cols[!required.cols.present],collapse=" and ")," missing")
      }
      if (!'start.pos' %in% colnames(from)) {
        warning('start.pos not in supplied data.frame, setting to NA')
        from$start.pos <- NA
      }
      
      ## Substitute Isoleucins with Leucins (indistinguishable by Masspec)

     if (!.PEPTIDE.COLS['REALPEPTIDE'] %in% colnames(from)) 
       from <- .fix.il.peptide(from)

      subset.s <- function(my.df,j) {
        if (is(my.df,"data.table"))
          my.df[,j,with=FALSE]
        else
          my.df[,j]
      }

      spectrumToPeptide <- .as.vect(unique(subset.s(from,c("spectrum","peptide"))))
      ##spectrumToPeptide <- setNames(from[["peptides"]],from[["spectrum"]])

      spectrumId <- unique(subset.s(from,setdiff(colnames(from),c("protein","start.pos","aa.before","aa.after"))))
      peptideInfo <- unique(subset.s(from,intersect(c("protein","peptide","start.pos","aa.before","aa.after","modif","real.peptide"),colnames(from))))
      peptideInfo <- peptideInfo[order(peptideInfo[["protein"]],
                                       peptideInfo[["start.pos"]],
                                       peptideInfo[["peptide"]]),]
      from <- unique(subset.s(from,c("protein","peptide")))
    
      # use numbers to not exceed the maximum length
      from$peptn <- as.integer(as.factor(from[["peptide"]]))
      from$protn <- as.integer(as.factor(from[["protein"]]))
      
      # with which peptide-combination are proteins detected?
      combn.peptides <- tapply(from[["peptn"]],from[["protein"]],
                               function(s) paste0(sort(unique(s)),collapse="-"))
      
      # group proteins with same peptide combinations together
      # - those are indistiguishable
      combn.proteins <- tapply(names(combn.peptides),combn.peptides,
                               function(x) paste0(x,collapse=","))
      
      # merge data frames
      f.peptides <- combn.peptides[ from[["protein"]] ]
      from[["protein.g"]] <- combn.proteins[ f.peptides ]
                                                
      pep.n.prots <- unique(subset(from,,c("peptide","protein.g")))
      all.protein.g <- table(pep.n.prots[['protein.g']])
      prots.to.prot <- as.matrix(unique(subset.s(from,c("protein.g","protein"))))
      prots.group <- c()
      
      # GROUPING:
      # 1) determine proteins w/ specific peptides [reporterProteins]
      ssp <- table(pep.n.prots[["peptide"]])==1
      sel.ssp <- pep.n.prots[["peptide"]] %in% names(ssp)[ssp]
      # take first those proteins which explain most specific peptides
      tt <- table(pep.n.prots[["protein.g"]][sel.ssp])
      protein.reporterProteins <- names(sort(tt,decreasing=TRUE))
      protein.groupMembers <- setdiff(names(all.protein.g),protein.reporterProteins)
      # proteins to consider for grouping
      pg.length <- length(all.protein.g)

      pgt <- new.env()
      pgt[["protein.group.table"]] <- data.frame(reporter.protein=rep("",pg.length),
                                        protein.g=rep("",pg.length),
                                        'n.reporter-specific'=0,
                                        'n.group-specific'=0,
                                        'n.unspecific'=0,
                                        stringsAsFactors=FALSE,check.names=FALSE)
      pgt[["my.idx"]] <- 1
      pgt[["pep.n.prots"]] <- pep.n.prots
      pgt[["prots.to.consider"]] <- new.env()
      for (my.protein.g in protein.groupMembers) {
        pgt[["prots.to.consider"]][[my.protein.g]] <- 
          subset(pep.n.prots,protein.g == my.protein.g)[["peptide"]]
      }

      # 2) get proteins, which are contained by those detected proteins
      #    (i.e. no specific and no overlapping peptides)
      for (reporter.protein in protein.reporterProteins) {
        .build.protein.group(pgt,reporter.protein)
      }

      # proteins left in pep.n.prots.toconsider are those proteins
      # with overlapping peptides only. 
      # 3. They might have no reporter-specific peptides, but group-specific peptides

      grouped.proteins <- pgt[["protein.group.table"]][["protein.g"]]
      grouped.peptides <- subset(pep.n.prots,protein.g %in% protein.reporterProteins)[["peptide"]]
      avail.peptides.n.prots <- subset(pep.n.prots,!peptide %in% grouped.peptides)
      ungrouped.peptides <- subset(pep.n.prots,!peptide %in% grouped.peptides)[["peptide"]]

      while (length(ls(envir=pgt[["prots.to.consider"]])) > 0) {
        sorted.proteins <- sort(unlist(eapply(pgt[["prots.to.consider"]],length)))
        my.protein.g <- names(sorted.proteins)[length(sorted.proteins)]
       
        remove(list=my.protein.g,envir=pgt[["prots.to.consider"]])
        .build.protein.group(pgt,my.protein.g)
      }

      pgt[["protein.group.table"]]$is.reporter <-
        pgt[["protein.group.table"]]$protein.g == pgt[["protein.group.table"]]$reporter.protein
      
      # define peptide specificity - pep.n.prots.toconsider are ignored here for now
      tmp <- merge(as.data.frame(pgt[["protein.group.table"]],stringsAsFactors=FALSE),
                   pep.n.prots,by.x="protein.g",by.y="protein.g")
      
      peptideSpecificity <- .factor.as.character(ddply(tmp,"peptide",function(d) {
        data.frame(specificity=
                   ifelse(length(unique(d[,"protein.g"]))==1,REPORTERSPECIFIC,
                          ifelse(length(unique(d[,"reporter.protein"]))==1,GROUPSPECIFIC,
                                 UNSPECIFIC)),
                   n.shared.proteins=length(unique(d[,"protein.g"])),
                   n.shared.groups=length(unique(d[,"reporter.protein"])),
                   stringsAsFactors=FALSE)
      }
      ))

      # isoforms (handled for Uniprot only, ATM)
      proteins <- as.character(sort(unique(prots.to.prot[,"protein"])))
      isoforms <- data.frame(database = .get.database(proteins),
                             proteinac.w.splicevariant = proteins,
                             proteinac.wo.splicevariant = proteins,
                             splicevariant = NA,stringsAsFactors=FALSE)
      
      sel.uniprot <- isoforms$database == .UNIPROTDATABASE
      pos.isoforms <- sel.uniprot & grepl("^[^-]*\\-[0-9]*$",proteins)
      isoforms$proteinac.wo.splicevariant[pos.isoforms] <- 
         sub("-[^-]*$","",proteins[pos.isoforms])
      isoforms$splicevariant[pos.isoforms] <- sub(".*-","",proteins[pos.isoforms])
      rownames(isoforms) <- proteins

      message("done")
      
      return(
          new("ProteinGroup",
              peptideNProtein = as.matrix(pep.n.prots),
              peptideSpecificity = peptideSpecificity,
              proteinGroupTable = pgt[["protein.group.table"]],
              indistinguishableProteins =
                .as.vect(prots.to.prot,col.data='protein.g',col.names='protein'),
              spectrumId = spectrumId,
              spectrumToPeptide = spectrumToPeptide,
              isoformToGeneProduct = isoforms,
              peptideInfo = peptideInfo,
              proteinInfo = proteinInfo)
      )
      
    }
)


.uniprot.pattern.ac <- "^[A-Z][0-9][A-Z0-9][A-Z0-9][A-Z0-9][0-9]-?[0-9]*$"
.uniprot.pattern.id <- "^[A-Z0-9]{2,5}_[A-Z9][A-Z0-9]{2,5}$"
.entrez.pattern.id <- "^gi\\|[0-9]{5,15}$"
.numeric.pattern.ac <- "^[0-9]{5,15}$"

.UNIPROTDATABASE <- "UniprotKB"
.UNIPROTDATABASE_ID <- "UniprotKB IDs" # Maybe: add support to get Protein Info for Uniprot IDs, as some people use them (but they are not stable, see http://www.uniprot.org/faq/6)
.ENTREZDATABASE <- "Entrez Protein"

.get.database <- function(acs) {
  res <- rep("unknown",length(acs))
  sel.uniprot <- grepl(.uniprot.pattern.ac,acs)
  sel.entrez <- grepl(.entrez.pattern.id,acs)
  sel.numeric <- grepl(.numeric.pattern.ac,acs)

  if (any(sel.uniprot))
    res[sel.uniprot] <- .UNIPROTDATABASE

  if (any(sel.entrez))
    res[sel.entrez] <- .ENTREZDATABASE 

  if (any(sel.numeric))
    res[sel.numeric] <- "numeric"

  return(res)  
}




.build.protein.group <- function(pgt,reporter.protein) {

    pep.for.reporter <- subset(pgt[["pep.n.prots"]],protein.g==reporter.protein)[["peptide"]]

    # define proteins which have a subset of the reporter protein peptides
    groupmember.pept <- eapply(pgt[["prots.to.consider"]],function(pep.for.prot) {
      if (length(pep.for.prot) < length(pep.for.reporter) & 
              all(pep.for.prot %in% pep.for.reporter))
        pep.for.prot
      else
        NA
    })
   
    groupmember.pept <- groupmember.pept[!is.na(groupmember.pept)]
    groupmembers <- names(groupmember.pept)
    remove(list=groupmembers,envir=pgt[["prots.to.consider"]])

    my.idxes <- seq(from=pgt[["my.idx"]],to=pgt[["my.idx"]]+length(groupmembers))
    pgt[["protein.group.table"]][my.idxes,'reporter.protein'] <- reporter.protein
    pgt[["protein.group.table"]][my.idxes,'protein.g'] <- c(reporter.protein,groupmembers)
    # define number of reporter-/group-/non-specific peptides for group members

    all.peptides <- pep.for.reporter
    sel.unspecific <- 
      pgt[["pep.n.prots"]]$peptide %in% all.peptides &
         !pgt[["pep.n.prots"]]$protein.g %in% c(reporter.protein,groupmembers)
    unspecific.pept <- pgt[["pep.n.prots"]][sel.unspecific,"peptide"]
    groupspecific.pept <- all.peptides[all.peptides %in% as.character(unlist(groupmember.pept)) & !all.peptides %in% unspecific.pept]
    reporterspecific.pept <- setdiff(pep.for.reporter,c(groupspecific.pept,unspecific.pept))

    pgt[["protein.group.table"]][my.idxes,'n.reporter-specific'] <- 
      c(length(reporterspecific.pept),rep(0,length(groupmembers)))
    pgt[["protein.group.table"]][my.idxes,'n.group-specific'] <- 
      c(length(groupspecific.pept),
        sapply(groupmember.pept,function(x) sum(x %in% groupspecific.pept)))
    pgt[["protein.group.table"]][my.idxes,'n.unspecific'] <- 
      c(length(unspecific.pept),
        sapply(groupmember.pept,function(x) sum(x %in% unspecific.pept)))

    pgt[["my.idx"]] = pgt[["my.idx"]]+length(groupmembers)+1
}

getProteinInfoFromBiomart <- function(x,database="Uniprot") {
  protein.acs <- x@isoformToGeneProduct[names(indistinguishableProteins(x)),"proteinac.wo.splicevariant"]

  protein.info <- data.frame(accession=c(),name=c(),protein_name=c(),
                             gene_name=c(),organism=c())
  if (database == "Uniprot") {
    tryCatch({
      mart <- biomaRt::useMart("uniprot_mart",dataset="UNIPROT",
                      host="www.ebi.ac.uk",path="/uniprot/biomart/martservice")
    protein.info <- biomaRt::getBM(attributes=c("accession","name","protein_name","gene_name","organism"),
          filter='accession',values=protein.acs,mart=mart)
    },error=function(e) warning("could not set Biomart"))

  } else {
    stop(sprintf("getProteinInfo for database %s not defined.",database))
  }
  protein.info$gene_name[!is.na(protein.info$gene_name) & protein.info$gene_name == ""] <- NA

  return(protein.info)
}

# for NCBI protein: http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=protein&id=115298678

getProteinInfoFromTheInternet <- function(x) {
  if (is.character(x)) {
    protein.acs <- x
  } else {
    my.df <- x@isoformToGeneProduct
    protein.acs <- unique(my.df[,"proteinac.wo.splicevariant"])
    if ("database" %in% colnames(my.df)) {
      sel.uniprot <- my.df[["database"]] == .UNIPROTDATABASE
      sel.entrez <- my.df[["database"]] == .ENTREZDATABASE
    }
  }

  if (!exists("sel.uniprot")) {
    sel.uniprot <- grepl(.uniprot.pattern.ac,protein.acs)
    sel.entrez <- grepl(.entrez.pattern.id,protein.acs)
  }

  res <- data.frame()
  if (any(sel.uniprot)) 
    res <- getProteinInfoFromUniprot(protein.acs[sel.uniprot])

  if (any(sel.entrez))
    res <- rbind.fill(res,getProteinInfoFromEntrez(protein.acs[sel.entrez]))

  return(res)
}

getProteinInfoFromUniprot <- 
  function(x,splice.by=200, 
           fields = c(accession="id",name="entry%20name",protein_name="protein%20names",
                      gene_name="genes",organism="organism", length="length",
                      sequence="sequence")) {
 
  if (is.character(x))
    protein.acs <- x
  else  
    protein.acs <- x@isoformToGeneProduct[names(indistinguishableProteins(x)),"proteinac.wo.splicevariant"]
  protein.acs <- gsub("%","",protein.acs)
  protein.info <- c()
  i <- 1
  while (i <= length(protein.acs)) {
    cat(".")
    uniprot.url <- paste0("http://www.uniprot.org/uniprot/?query=",
                 paste0("accession:",protein.acs[seq(from=i,to=min(length(protein.acs),i+splice.by-1))],collapse="+OR+"),
                 "&format=tab&compress=no&columns=",
                 paste0(fields,collapse=","))
    if (isTRUE(getOption('isobar.verbose')))
      message("fetching protein info from ",uniprot.url)
    protein.info <- rbind(protein.info,read.delim(url(uniprot.url),stringsAsFactors=FALSE,col.names=names(fields)))
    i <- i + splice.by
  }
  if (!is.null(protein.info) && nrow(protein.info) > 0) {
    protein.info$protein_name <- sapply(strsplit(protein.info$protein_name," (",fixed=TRUE),function(x) x[1])
    protein.info$gene_name <- sapply(strsplit(protein.info$gene_name," "),function(x) x[1])
    protein.info$sequence <- gsub(" ","",protein.info$sequence)
  } else {
    warning("getProteinInfoFromUniprot returned no results for ",protein.acs," accessions")
  }
  return(protein.info)
}

getProteinInfoFromEntrez <- function(x,splice.by=200) {
  requireNamespace("XML")

  eutils.url <- "http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi?db=protein&id="
  if (is.character(x))
    protein.acs <- x
  else  
    protein.acs <- x@isoformToGeneProduct[names(indistinguishableProteins(x)),"proteinac.wo.splicevariant"]

  if (all(grepl("^gi|",protein.acs))) {
    new.protein.acs <- sapply(strsplit(protein.acs,"|",fixed=TRUE),function(x) x[2]) # strip ^gi| 
    protein.conv <- setNames(protein.acs,new.protein.acs)
    protein.acs <- new.protein.acs
  }

  protein.info <- c()
  i <- 1
  while (i < length(protein.acs)) {
    cat(".")

    full.url <- paste0(eutils.url,paste(as.character(protein.acs[seq(from=i,to=min(length(protein.acs),i+splice.by-1))]),collapse=","))
    res <- XML::xmlToList(full.url)
    res.l <- lapply(res, function(x) {
      unlist(setNames(lapply(x[names(x) != 'Id'], function(y) 
        if ('.attrs' %in% names(y) && y$.attrs['Name'] != 'Comment') 
          setNames(y$text,y$.attrs['Name'])),NULL))
    })

    res.l <- res.l[!sapply(res.l,is.null)]
  
    enames.to.isobar <- c(Gi="gi_accession",Caption="name",Title="protein_name")
    res.df <- ldply(res.l, function(x) { 
                    setNames(x[names(enames.to.isobar)],enames.to.isobar)
    })
  
   protein.info <- rbind(protein.info,res.df)
    i <- i + splice.by
  }
  protein.info[,c('protein_name','organism')] <- t(sapply(strsplit(sub("^(.*) \\[(.*)\\]$","\\1\t\\2",protein.info[,'protein_name']),"\t"),function(x) x))
 
  protein.info$gene_name <- NA
  protein.info$.id <- NULL
  protein.info$accession <- protein.conv[protein.info$gi_accession]
  protein.info
}

getProteinInfoFromNextProt <- function(x) {
  fields <- c(accession="id",name="entry%20name",protein_name="protein%20names",
              gene_name="genes",organism="organism",
              length="length",sequence="sequence")

  stop("NOT IMPLEMENTED YET")
   
  protein.acs <- x@isoformToGeneProduct[names(indistinguishableProteins(x)),]
  protein.info  <- ddply(protein.acs,"proteinac.wo.splicevariant",
                         function(y) {
                         url <- sprintf("http://www.nextprot.org/rest/protein/NX_%s?format=json",
                                        unique(y$proteinac.wo.splicevariant))
                         # PARSE JSON, esp isoform sequences
  })
  attr(protein.info,"on.splice.variant") <- TRUE
  return(protein.info)
}

proteinInfoIsOnSpliceVariants <- function(protein.info) {
  return(isTRUE(attr(protein.info,"on.splice.variant")))
}


getProteinInfoFromBioDb <- function(x,...,con=NULL) {
  requireNamespace("DBI")
  if (is.null(con)) {
    con <- DBI::dbConnect(...)
    do.disconnect <- TRUE
  } else {
    do.disconnect <- FALSE
  }

  if (is.character(x))
    protein.acs <- x
  else  
    protein.acs <- names(indistinguishableProteins(x))
  query <- paste("SELECT primaryac AS accession,id AS name,",
                 "  description AS protein_name,",
                 "  (SELECT string_agg(g.genename,',') FROM genenames g WHERE g.entryid=d.entryid AND g.synonym=FALSE AND g.sourcedb=3) AS gene_name,",
                 "  os AS organism, seqlength as length, sequence",
                 "FROM dbentries d ",
                 "WHERE dbid IN (2,3) AND primaryac IN (",paste0("'",protein.acs,"'",collapse=","),")")
  res <- DBI::dbGetQuery(con,query)
  if (do.disconnect)
    DBI::dbDisconnect(con)
  attr(res,"on.splice.variant") <- TRUE
  return(res)
}

getPtmInfoFromPhosphoSitePlus <- function(protein.group,file.name=NULL,modif="PHOS",
                                          psp.url="http://www.phosphosite.org/downloads/",
                                          mapping=c(PHOS="Phosphorylation_site_dataset.gz",
                                                    ACET="Acetylation_site_dataset.gz",
                                                    METH="Methylation_site_dataset.gz",
                                                    SUMO="Sumoylation_site_dataset.gz",
                                                    UBI="Ubiquitination_site_dataset.gz")) {
  if (length(modif) > 1) {
    return(do.call(rbind,lapply(modif,function(m) getPtmInfoFromPhosphoSitePlus(protein.group,file.name,m,psp.url,mapping))))
  }

  if (is.null(file.name)) file.name <- mapping[modif]

  if (!file.exists(file.name) && is.null(modif)) stop("provide PhosphoSitePlus file or modif name")
  if (!file.exists(file.name) && !is.null(modif)) {
    download.file(paste0(psp.url,mapping[modif]),mapping[modif])
    file.name <- mapping[modif]
  }

  sites <- read.delim(file.name,
                      sep="\t",header=TRUE,skip=3,stringsAsFactors=FALSE)
  ac.column <- ifelse("ACC_ID" %in% colnames(sites),"ACC_ID","ACC.")
  species.column <- intersect(c("ORG", "ORGANISM", "SPECIES"), colnames(sites))
  residue.column <- ifelse("MOD_RSD" %in% colnames(sites),"MOD_RSD","RSD")
  domain.column <- intersect(c("DOMAIN", "IN_DOMAIN"), colnames(sites))
  if (!("MOD_TYPE" %in% colnames(sites))) {
    sites$MOD_TYPE = sapply(gsub("^.+-", "", sites$MOD_RSD), function(mod_code) {
                       switch(mod_code,
                         p = "PHOSPHORYLATION",
                         { warning("unknown PTM code: ", mod_code)
                           mod_code })
		     })
    sites$MOD_RSD = gsub("-.+$", "", sites$MOD_RSD)
  }

  sites <- sites[gsub("-.*","",sites[,ac.column]) %in% gsub("-.*","",names(indistinguishableProteins(protein.group))),]

  lit_colnames = c(PUBMED_LTP="n.publ ltp", PUBMED_MS2="n.publ htp",
		   LS_LIT="n.publ ltp", MS_LIT="n.publ htp")
  for (coln in names(lit_colnames)) {
    sites[!is.na(sites[[coln]]), coln] <- paste0(lit_colnames[[coln]],": ",sites[!is.na(sites[[coln]]), coln])
  }

  data.frame(.id=sites[,ac.column],
             isoform_ac=sapply(sites[,ac.column],function(ac) ifelse(grepl("-[0-9]$",ac),ac,paste0(ac,"-1"))),
             species=tolower(sites[,species.column]),
             modification.name=tolower(sites[,'MOD_TYPE']),
             description=apply(sites,1,function(x) {
                               y <- tolower(x['MOD_TYPE'])
                               if (nchar(x[domain.column]) > 0)
                                 y <- paste0(y," (domain ",x[domain.column],")")
                               y
                             }),
             evidence=apply(sites[,intersect(colnames(sites),names(lit_colnames))],1,
                            function(x) { x<-x[!is.na(x)]; paste(x,collapse=";")}),
             position=as.integer(substr(sites[,residue.column],2,nchar(sites[,residue.column]))),
             stringsAsFactors=FALSE)
}

getPtmInfoFromNextprot <- function(protein.group,
                                   nextprot.url="http://www.nextprot.org/rest/entry/NX_XXX/ptm?format=json",
                                   url.wildcard="XXX") {
  protein.acs <- unique(protein.group@isoformToGeneProduct$proteinac.wo.splicevariant)
  requireNamespace("RJSONIO")
  pb <- txtProgressBar(max=length(protein.acs),style=3)

  nextprot.ptmInfo <- 
    lapply(seq_along(protein.acs),function(ac_i) {
           setTxtProgressBar(pb,ac_i)
           tryCatch(RJSONIO::fromJSON(sub(url.wildcard,protein.acs[ac_i],nextprot.url)),
                    error=function(e) {
              if (isTRUE(getOption('isobar.verbose')))
                warning("Could not fetch from ",
                        sub(url.wildcard,protein.acs[ac_i],nextprot.url),":",
                        e$message)
              return()})
  })
  names(nextprot.ptmInfo) <- protein.acs
  sel.length0 <- sapply(nextprot.ptmInfo,length)==0
  if (any(sel.length0))
    warning("Could not fetch neXtProt results for some proteins:",
            " ",.abbrev(protein.acs[sel.length0],4,", "),".")
  nextprot.ptmInfo <- nextprot.ptmInfo[sapply(nextprot.ptmInfo,length)>0]
  ptm.info <- ldply(nextprot.ptmInfo,
                    function(x) 
                      ldply(x,function(y) {
                            y[sapply(y,is.null)] <- NA
                            y$modification.name <- y$modification['name']
                            y$modification.accession <- y$modification['accession']
                            y$modification <- NULL
                            data.frame(y,stringsAsFactors=FALSE)
                      })
                    )
  ptm.info$isoform_ac <- sub("^NX_","",ptm.info$isoform_ac)
  ptm.info$position <- ptm.info$first_position
  ptm.info
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Coercion.
###

as.data.frame.ProteinGroup <- function(x,...) as(x,"data.frame")

setAs("data.frame","ProteinGroup",function(from) ProteinGroup(from))
setMethod("as.data.frame",signature(x="ProteinGroup"),
          function(x, row.names=NULL, optional=FALSE, ...) as(x,"data.frame"))

setAs("ProteinGroup","data.frame",function(from) {
  sp.df <- from@spectrumId
  ip.df <- .vector.as.data.frame(indistinguishableProteins(from),
                                 colnames=c("protein","protein.g"))
  
  ## merge spectrum to peptide to indistinguishable proteins (protein.g)
  spectrum.to.proteing <-
    merge(sp.df, as.data.frame(peptideNProtein(from),stringsAsFactors=FALSE))
 
  ## merge with indist. proteins
  spectrum.to.protein <- merge(ip.df,spectrum.to.proteing)
  spectrum.to.protein.w.peptideinfo <- merge(spectrum.to.protein,from@peptideInfo)

  return(spectrum.to.protein.w.peptideinfo[,c("spectrum","peptide","modif","start.pos","protein")])
})

## Export a /concise/ data.frame which is peptide centric:
##  - for each peptide there is one line only
##  - column proteins concatenates all the different protein/protein group ACs
##    - indistinguishable peptides are separated by ',', protein groups by ';'
##  - column n.groups: number of groups a peptide appears
##  - column n.acs: number of acs a peptide appears in (without splice variant!)
##      counts double if the same ACs are in different groups 
##  - column n.variants: number of acs a peptide appears in (with splice variant!)
proteinGroup.as.concise.data.frame <-
      function(from) {
        ip.df <- .vector.as.data.frame(indistinguishableProteins(from),
                                       colnames=c("protein","protein.g"))
        ip.df <- merge(ip.df,from@isoformToGeneProduct,
                       by.x="protein",by.y="proteinac.w.splicevariant")
        pep.n.prot <- as.data.frame(peptideNProtein(from),stringsAsFactors=FALSE)
        in.df <- unique(ddply(ip.df, "protein.g",
                        function(x) 
                          c(n.acs=length(unique(x[,"proteinac.wo.splicevariant"])),
                            n.variants=length(unique(x[,"protein"])))))
        pep.n.prot <- merge(pep.n.prot,ip.df)
        pep.n.prot <- merge(pep.n.prot,proteinGroupTable(from)[,c("protein.g","reporter.protein")])
        res <- ddply(pep.n.prot,"peptide",
                     function(x) {
                       res <- data.frame(n.acs=length(unique(x[,"proteinac.wo.splicevariant"])),
                                         n.variants=length(unique(x[,"protein"])))
                       x <- unique(x[,c("reporter.protein","protein.g")])
                       protein.gs <- unique(x[,'reporter.protein'])
                       res <- cbind(proteins=paste(tapply(x$protein.g,factor(x$reporter.protein),
                                                          paste,collapse=","),collapse=";"),
                                    n.groups=length(protein.gs),
                                    res,stringsAsFactors=FALSE)
                       if (!is.null(attr(from,"protein.group.ids"))) 
                         res  <- cbind(groups=paste(attr(from,"protein.group.ids")[protein.gs],collapse=","),
                                       res,stringsAsFactors=FALSE)
                       res

                     })
        return(unique(res))
      }


.proteinGroupAsConciseDataFrame1 <- 
  function(from,only.reporters=TRUE,show.proteinInfo=TRUE,
           human.protein.acs=TRUE,show.startpos=TRUE,modif.pos=NULL,
           ptm.info=NULL,link.url="http://www.uniprot.org/uniprot/") {

        pep.n.prot <- merge(as.data.frame(peptideNProtein(from),stringsAsFactors=FALSE),
                            from@peptideInfo)
        p.ac <- .protein.acc(reporterProteins(from),from)
        ip.df <- .vector.as.data.frame(indistinguishableProteins(from),
                                       colnames=c("protein","protein.g"))
        if (only.reporters)
          ip.df <- ip.df[ip.df$protein.g %in% reporterProteins(from),]
        ip.df <- merge(ip.df,from@isoformToGeneProduct,
                       by.x="protein",by.y="proteinac.w.splicevariant")
        ip.df <- merge(ip.df,proteinGroupTable(from)[,c("protein.g","reporter.protein")])
        ipp.df <- merge(ip.df,pep.n.prot)

        in.df <- ddply(ipp.df, c("reporter.protein"),
          function(x) {
            # for each 'protein AC' (no splice variant), summarize the splice variants (eg P123-[1-3,5])
            merged.splicevariants <- ddply(x,"proteinac.wo.splicevariant",.summarize.splice.variants,link.url=link.url)

            # for each modified peptide, get one record summarizing its information
            merged.pepmodifs <- ddply(x,c("peptide","modif"),.summarize.pepmodif,modif.pos=modif.pos,ptm.info=ptm.info,from=from)
            
            data.frame(proteinn=paste(merged.splicevariants$ac,collapse=","),
                       link=merged.splicevariants$link[1],
                       merged.pepmodifs,stringsAsFactors=FALSE)
        })
        merge(ip.df,in.df)
  }



# nexprot url: "http://www.nextprot.org/db/entry/NX_"
.proteinGroupAsConciseDataFrame <- 
  function(from,only.reporters=TRUE,show.proteinInfo=TRUE,
           human.protein.acs=TRUE,show.startpos=TRUE,modif.pos=NULL,
           ptm.info=NULL,link.url="http://www.uniprot.org/uniprot/",
           report.only.modified=FALSE) {
        i = 0

        pep.n.prot <- merge(as.data.frame(peptideNProtein(from),stringsAsFactors=FALSE),
                            from@peptideInfo,by="peptide")
        if (!is.null(modif.pos) && isTRUE(report.only.modified))
          pep.n.prot <- pep.n.prot[grepl(modif.pos,pep.n.prot[,'modif']),]

        p.ac <- .protein.acc(reporterProteins(from),from)

        # ip.df will contain information on
        #   protein.g,protein,proteinac.wo.splicevariant,splicevariant,reporter.protein

        ip.df <- .vector.as.data.frame(indistinguishableProteins(from),
                                       colnames=c("protein","protein.g"))
        if (only.reporters)
          ip.df <- ip.df[ip.df[,'protein.g'] %in% reporterProteins(from),]
        ip.df <- merge(ip.df,from@isoformToGeneProduct,
                       by.x="protein",by.y="proteinac.w.splicevariant")
        ip.df <- merge(ip.df,proteinGroupTable(from)[,c("protein.g","reporter.protein")],by="protein.g")
        ipp.df <- merge(ip.df,pep.n.prot)

        in.df <- ddply(ipp.df, c("reporter.protein"),
          function(x) {
            ## for each 'protein AC' (no splice variant), 
            ##  summarize the splice variants (eg P123-[1-3,5])
            merged.splicevariants <- 
              ddply(x,"proteinac.wo.splicevariant",
                    .summarize.splice.variants,link.url=link.url)
            # for each modified peptide, get one record summarizing its information
            merged.pepmodifs <- ddply(x,c("peptide","modif"),.summarize.pepmodif,
                                      modif.pos=modif.pos,ptm.info=ptm.info,from=from)
            data.frame(proteinn=paste(merged.splicevariants[,'ac'],collapse=","),
                       link=merged.splicevariants[1,'link'],
                       merged.pepmodifs,stringsAsFactors=FALSE)
        },.parallel=isTRUE(getOption('isobar.parallel')))

        cols <- c("proteinn","link","start.pos","real.peptide","modif","modif.pos","modif.comment")
        if (show.proteinInfo) 
          pnd <- proteinNameAndDescription(from,ip.df[,'protein.g'])


        my.res <- 
          ddply(merge(ip.df,in.df,by="reporter.protein"),c("peptide","modif"),
                function(x) {
                  #protein.gs <- unique(x[,'reporter.protein'])
                  protein.gs <- unique(x[,'protein.g'])
                  x <- unique(x[,cols[cols %in% colnames(x)]])

                  res <- data.frame(start.pos=.unique.or.collapse(x[,'start.pos'],";"),
                                    proteins=paste(unique(x[,'proteinn']),collapse=";"),
                                    uniprot=paste0("@link=",x[1,'link'],"@x",collapse=";"),
                                    stringsAsFactors=FALSE)

                  ## n.acs=length(unique(x[,"proteinac.wo.splicevariant"])),
                  ## n.variants=length(unique(x[,"protein"])),
                  ## n.variants=length(protein.gs),

                  if ('real.peptide' %in% colnames(x))
                    res <- 
                      cbind(res,
                            real.peptide=.unique.or.collapse(x[,'real.peptide'],";"),
                            stringsAsFactors=FALSE)

                  if (!is.null(modif.pos)) {
                    null.comments <- x[,'modif.comment'] == ""
                    res <- cbind(res,
                                 modif.pos=ifelse(!any(x[,'modif.pos']!=0),"",
                                                  paste(x[,'modif.pos'],collapse=";")),
                                 comment=ifelse(all(null.comments),"",
                                                paste(x[,'modif.comment'][!null.comments],
                                                collapse="\n")),
                                 stringsAsFactors=FALSE)
                  }

                  if (show.proteinInfo && nrow(pnd) > 0) 
                    res <- cbind(res,
                                 lapply(pnd[ip.df[,'protein.g'] %in% protein.gs,],
                                        .paste_unique,collapse=";"))

                  if (!is.null(attr(from,"from.ids"))) 
                    res  <- cbind(groups=paste(attr(from,"from.ids")[protein.gs],
                                               collapse=","),
                                  res,stringsAsFactors=FALSE)

                  res

                },.parallel=isTRUE(getOption('isobar.parallel')))

        # res[,'peptide'] <- .convertModifToPos(res[,'peptide'],res[,'modif'])
        return(unique(my.res))
}

.summarize.pepmodif <- function(x,modif.pos,ptm.info,from) {
  res <- data.frame(peptide=x$peptide[1],start.pos=paste(x$start.pos,collapse=";"),
                    stringsAsFactors=FALSE)
  if ('real.peptide' %in% colnames(x)) 
    res <- cbind(res,real.peptide=.unique.or.collapse(x[,'real.peptide']),stringsAsFactors=FALSE)
  
  if (!is.null(modif.pos)) 
    res <- cbind(res,.get.modif.pos(x,modif.pos,ptm.info,from))
  res
}

.summarize.splice.variants <- function(x,link.url) {
  res <- c(ac=unique(x$proteinac.wo.splicevariant),
           link=paste0(link.url,unique(x$proteinac.wo.splicevariant)))
  if (!all(is.na(x$splicevariant))) {
    if (length(unique(x$splicevariant))==1) { res['ac'] <- x$protein[1] 
    } else { res['ac'] <- sprintf("%s-[%s]",unique(x$proteinac.wo.splicevariant),
                              number.ranges(as.integer(x$splicevariant)))
    }
  }
  res
}

.get.modif.pos <- function(x,modif.pos,ptm.info,from) {
   pepseq <- strsplit(x$peptide[1],"")[[1]]
   # get modification position foreach protein (in peptide) from modification string
   modification.positions.foreach.protein <- 
     .convertModifToPos(x$modif,modif.pos,collapse=NULL,simplify=FALSE,and.name=!is.null(names(modif.pos)))

   modif.posi <- t(mapply(.get.modif.pos.for.ac,x$protein,x$splicevariant,modification.positions.foreach.protein,x$start.pos,
                          MoreArgs=list(from=from,pepseq=pepseq,ptm.info=ptm.info)))

   my.gene <- sprintf("%s %s",modif.posi[,3],.string.number.ranges(modif.posi[,4]))
   if (is.na(modif.posi[1,1]) || all(modif.posi[,1]==modif.posi[1,1])) {
     my.modif.posi <- modif.posi[1,1]
   } else {
     my.modif.posi <- paste(modif.posi[,1],collapse=";")
   }


   cbind(modif=unique(x$modif),
         modif.pos=my.modif.posi,
         modif.comment=ifelse(any(!is.na(modif.posi[,2])),
                              paste(modif.posi[!is.na(modif.posi[,2]),2],collapse="\n"),""),
         stringsAsFactors=FALSE)
} 

.get.modif.pos.for.ac <- function(ac,sv,pep.pos.n.modif,start.pos,from,pepseq,ptm.info) {
  gene_name <- proteinInfo(from,protein.ac=ac,select="gene_name",do.warn=FALSE,collapse=",")
  if (length(pep.pos.n.modif) == 0) {
    return(c(NA,NA,ifelse(length(gene_name)==0,NA,gene_name),sv))
  }

  if (is.data.frame(pep.pos.n.modif)) {
    pep.pos <- pep.pos.n.modif[,1]
    modif <- pep.pos.n.modif[,2]
  } else {
    pep.pos <- pep.pos.n.modif
    modif <- NULL
  }
  if (all(pep.pos == 0))
    residue <- 'Nterm'
  else
    residue <- pepseq[pep.pos]

  poss <- start.pos + pep.pos - 1
  if (!is.null(ptm.info) && all(c("isoform_ac","position") %in% colnames(ptm.info))) {
    comments <- sapply(poss,function(pp) {
                       if (grepl("-[0-9]$",ac)) 
                         sel <- ptm.info[,"isoform_ac"]==ac
                       else
                         sel <- ptm.info[,"isoform_ac"]==paste(ac,"-1",sep="")
                       sel <- sel & ptm.info[,"position"]==pp
                       sel[is.na(sel)] <- FALSE

                       if (any(sel)) {
                         res <- apply(ptm.info[sel,],1,
                                      function(pi) paste(sprintf("%s pos %g: %s",pi["isoform_ac"],
                                                                 as.integer(pi["position"]),pi["description"])))
                         paste(res,collapse="\n")
                       } else {
                         NA
                       }
                    })
    #known.pos <- sapply(poss,function(pp) any(ptm.info[,"isoform_ac"]==ac & ptm.info[,"position"]==pp))
    known.pos <- !is.na(comments)
  } else {
    comments <- NA
    known.pos <- rep(FALSE,seq_along(start.pos))
  }
  null.comments <- is.na(comments)

  if (!is.null(modif))
    poss <- paste0(poss,modif)
  if (sum(known.pos) > 0)
    poss[known.pos] <- paste0(residue[known.pos],poss[known.pos],"*")
  poss[!known.pos] <- paste0(residue[!known.pos],poss[!known.pos])

  c(paste(poss,collapse="&"),
    ifelse(all(null.comments),NA,paste(comments[!null.comments],collapse="\n")),
    ifelse(length(gene_name)==0,NA,gene_name),
    sv)
}



proteinID <- function(protein.group,protein.g=reporterProteins(protein.group)) {
  proteinInfo(protein.group,protein.g,do.warn=FALSE,collapse=",")
}
proteinDescription <- function(protein.group,protein.g=reporterProteins(protein.group)) {
  proteinInfo(protein.group,protein.g,select="protein_name",do.warn=FALSE,collapse=",")
}
proteinGeneName <- function(protein.group,protein.g=reporterProteins(protein.group)) {
  proteinInfo(protein.group,protein.g,select="gene_name",do.warn=FALSE,collapse=",")
}

proteinNameAndDescription <- function(protein.group,protein.g=reporterProteins(protein.group),collapse=FALSE) {
  if (collapse)
    data.frame(ID=.paste_unique(proteinID(protein.group,protein.g),collapse=","),
               Description=.paste_unique(proteinDescription(protein.group,protein.g),collapse=";"),
               Gene=.paste_unique(proteinGeneName(protein.group,protein.g),collapse=","),
               stringsAsFactors=FALSE)
  else
    data.frame(ID=proteinID(protein.group,protein.g),
               Description=proteinDescription(protein.group,protein.g),
               Gene=proteinGeneName(protein.group,protein.g),
               stringsAsFactors=FALSE)
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Accessor-like methods.
###

setGeneric("peptideNProtein",function(x) standardGeneric("peptideNProtein"))
setGeneric("peptideSpecificity", function(x) standardGeneric("peptideSpecificity") )
setGeneric("peptideInfo", function(x) standardGeneric("peptideInfo"))
setGeneric("peptides",function(x,protein,...) standardGeneric("peptides"))
setGeneric("indistinguishableProteins",
           function(x,protein,protein.g) standardGeneric("indistinguishableProteins"))
setGeneric("reporterProteins",function(x, require.reporter.specific=FALSE) standardGeneric("reporterProteins"))
setGeneric("proteinGroupTable",function(x) standardGeneric("proteinGroupTable"))
setGeneric("spectrumToPeptide", function(x) standardGeneric("spectrumToPeptide"))
setGeneric("proteinInfo",function(x,protein.g,protein.ac,...) standardGeneric("proteinInfo"))
setGeneric("proteinInfo<-",function(x,value) standardGeneric("proteinInfo<-"))

setMethod("peptideSpecificity", "ProteinGroup", function(x) x@peptideSpecificity)
setMethod("peptideInfo", "ProteinGroup", function(x) x@peptideInfo)
setMethod("peptideNProtein",   "ProteinGroup", function(x) x@peptideNProtein)
setMethod("proteinGroupTable",      "ProteinGroup", function(x) x@proteinGroupTable )

setMethod("indistinguishableProteins",signature(x="ProteinGroup",protein="missing",protein.g="missing"),
          function(x) x@indistinguishableProteins)
setMethod("indistinguishableProteins",signature(x="ProteinGroup",protein="missing",protein.g="character"),
          function(x,protein.g) {
            sel <- x@indistinguishableProteins %in% protein.g
            if (!any(sel)) {
              warning(protein.g," is no protein group - see reporterProteins",
                      " for a list of all reporter groups")
              return(NA)
            }
            names(x@indistinguishableProteins)[sel]
          }
)
setMethod("indistinguishableProteins",signature(x="ProteinGroup",protein="character",protein.g="missing"),
          function(x,protein) {
            protein.groups <- x@indistinguishableProteins[protein]
            if (length(protein.groups) == 0) {
              warning(protein," is in not in the list")
              return(NA)
            }
            return(protein.groups)
          }
)

setMethod("spectrumToPeptide",   "ProteinGroup", function(x) x@spectrumToPeptide)

setMethod("peptides",signature(x="ProteinGroup",protein="missing"),
    function(x,...) peptides(x,reporterProteins(x),...)
)

setMethod("peptides",signature(x="ProteinGroup",protein="character"),
    function(x,protein,
             specificity=c(UNSPECIFIC,GROUPSPECIFIC,REPORTERSPECIFIC),
             columns=c('peptide'),set=union,drop=TRUE,
             groupspecific.if.same.ac=FALSE,do.warn=TRUE,
             modif=NULL) {

      if (!all(specificity %in% c(UNSPECIFIC,GROUPSPECIFIC,REPORTERSPECIFIC)))
        stop("specificity should be one or more of ['",REPORTERSPECIFIC,"','",GROUPSPECIFIC,"','",REPORTERSPECIFIC,"']")

      pnp <- peptideNProtein(x)
      ps <- peptideSpecificity(x)
      pi <- merge(peptideInfo(x),peptideSpecificity(x),by="peptide")

      if (groupspecific.if.same.ac) {
        pgt <- proteinGroupTable(x)
        igp <- x@isoformToGeneProduct

        group.proteins <- pgt[pgt[["reporter.protein"]]==protein,"protein.g"]
        group.proteins.all <- indistinguishableProteins(x,protein.g=group.proteins)
        protein.acs <- igp[igp[["proteinac.w.splicevariant"]] %in% group.proteins.all,
                           "proteinac.wo.splicevariant"]
        if (length(unique(protein.acs)) == 1)
          specificity <- c(specificity,GROUPSPECIFIC)
      }
      
      peptides <- lapply(protein, function(p) pnp[pnp[,"protein.g"] == p,"peptide"])
      peptides <- Reduce(set,peptides)

      sel <- pi$peptide %in% peptides                                          # is in protein?
      sel <- sel & pi$peptide %in% ps$peptide[ps$specificity %in% specificity] # has specificity?
      if (!is.null(modif)) sel <- sel & .has.modif(pi$modif,modif)             # has modification?

      peptides <- pi[sel,columns,drop=drop]
      if (length(peptides) == 0 && do.warn)
        warning("No peptide for protein ",protein," with specificity ",paste(specificity,collapse=","))
      return(unique(peptides))
    }
)

.has.modif <- function(modifstring,modif) {
  sapply(strsplit(modifstring,":"),function(m) any(m %in% modif))
}

setMethod("reporterProteins","ProteinGroup",
    function(x,require.reporter.specific=FALSE) {
      if (is.null(require.reporter.specific) || !require.reporter.specific) {
        return(x@proteinGroupTable$protein.g[x@proteinGroupTable$is.reporter])
      } else {
        return(x@proteinGroupTable$protein.g[
                                             x@proteinGroupTable$is.reporter &
                                             x@proteinGroupTable[,'n.reporter-specific'] > 0 ])
      }
    }
)

setMethod("proteinInfo",signature(x="ProteinGroup",protein.g="missing",protein.ac="missing"),function(x) x@proteinInfo)
setMethod("proteinInfo",signature(x="ProteinGroup",protein.g="missing",protein.ac="character"),
    function(x,protein.ac,select="name",collapse=", ",simplify=TRUE,do.warn=TRUE) {
      protein.info <- proteinInfo(x)
      if (length(protein.info) == 0) {
        if (isTRUE(do.warn))
          warning("proteinInfo is empty")
        return()
      }
      if (!all(select %in% colnames(protein.info))) 
        warning("Column(s) ",
                paste(select[!select %in% colnames(protein.info)],collapse=", "),
                " not in proteinInfo (available:",paste(colnames(protein.info),collapse=", "))

      if ((is.null(protein.info) || nrow(protein.info) == 0) && do.warn)
        warning("protein info is NULL! Set for ProteinGroup.")

      res <- sapply(protein.ac,function(p) {
        if (!all(select %in% colnames(protein.info))) 
          return(NA)
        
        if (proteinInfoIsOnSpliceVariants(protein.info))
          protein.acs <- x@isoformToGeneProduct[protein.ac,"proteinac.wo.splicevariant"]
        sel <- protein.info$accession %in% protein.ac
        if (!any(sel)) {
          if (do.warn) warning("No protein info for ",p)
          return(rep(NA,length(select)))
        }
        protein.infos <- protein.info[sel,select]
        if (simplify) {
          if (length(select) > 1)
            apply(protein.infos,2,function(pi) paste(sort(unique(pi)),collapse=", "))
          else 
            paste(sort(unique(protein.infos)),collapse=", ")
        } else {
      if (length(select) == 1)
            names(protein.infos) <- protein.info$accession[sel]
          protein.infos
        }
      },simplify=simplify)
      if (simplify & length(select)>1) t(res)
      else res
    }
)

setMethod("proteinInfo",signature(x="ProteinGroup",protein.g="character",protein.ac="missing"),
    function(x,protein.g,select="name",collapse=", ",simplify=TRUE,do.warn=TRUE) {
      protein.info <- proteinInfo(x)
      if (length(protein.info) == 0) {
        if (isTRUE(do.warn))
          warning("proteinInfo is empty")
        return()
      }

      if (!all(select %in% colnames(protein.info))) 
        warning("Column(s) ",
                paste(select[!select %in% colnames(protein.info)],collapse=", "),
                " not in proteinInfo (available:",paste(colnames(protein.info),collapse=", "))

      if ((is.null(protein.info) || nrow(protein.info) == 0) && do.warn)
        warning("protein info is NULL! Set for ProteinGroup.")

      res <- sapply(protein.g,function(p) {
        if (!all(select %in% colnames(protein.info))) 
          return(NA)
        
        if (proteinInfoIsOnSpliceVariants(protein.info))
          protein.acs <- indistinguishableProteins(x,protein.g=p)
        else
          protein.acs <- x@isoformToGeneProduct[indistinguishableProteins(x,protein.g=p),"proteinac.wo.splicevariant"]
        sel <- protein.info$accession %in% protein.acs
        if (!any(sel)) {
          if (do.warn) warning("No protein info for ",p)
          return(rep(NA,length(select)))
        }
        protein.infos <- protein.info[sel,select]
        if (simplify) {
          if (length(select) > 1)
            apply(protein.infos,2,function(pi) paste(sort(unique(pi)),collapse=", "))
          else 
            paste(sort(unique(protein.infos)),collapse=", ")
        } else {
      if (length(select) == 1)
            names(protein.infos) <- protein.info$accession[sel]
          protein.infos
        }
      },simplify=simplify)
      if (simplify & length(select)>1) t(res)
      else res
    }
)
setReplaceMethod("proteinInfo","ProteinGroup",
                 function(x,value) {
                   x@proteinInfo <- value
                   if ('sequence' %in% colnames(value) && all(is.na(x@peptideInfo[,'start.pos']))) {
                     message("Recaculating peptide start position based on sequence")
                     x@peptideInfo <- calcPeptidePosition(x@peptideInfo,value)
                   }
                   x
                 })

setGeneric("reporter.protein",function(x,protein.g) standardGeneric("reporter.protein"))
setMethod("reporter.protein",signature("ProteinGroup","character"),
          function(x,protein.g) {
            proteinGroupTable(x)[proteinGroupTable(x)[,'protein.g']==protein.g, 'reporter.protein']
          })

setGeneric("protein.g",function(x,pattern,variables=c("AC","name"),...) standardGeneric("protein.g"))
setMethod("protein.g",signature("ProteinGroup","character","ANY"),
          function(x,pattern,variables=c("AC","name"),...) {
  ip <- indistinguishableProteins(x)
  protein.info <- proteinInfo(x)
  result <- c()
  for (p in pattern) {
    protein.acs <- c()
    if ("AC" %in% variables)
      result <- c(result,ip[grep(p,names(ip),...)])
    if ("name" %in% variables) {
      if (length(protein.info) != 0L) {
        protein.acs <- unique(c(
           protein.info$accession[grep(p,protein.info$name,...)],
           protein.info$accession[grep(p,protein.info$gene_name,...)],
           protein.info$accession[grep(p,protein.info$protein_name,...)]
        ))
      }
      
      i.to.gp <- x@isoformToGeneProduct
      sel.isoforms <- i.to.gp[,"proteinac.wo.splicevariant"] %in% protein.acs
      protein.acs.w.isoforms <-
         i.to.gp[sel.isoforms,"proteinac.w.splicevariant"]
      ## TODO: test for proteinInfoIsOnSpliceVariants
      protein.gs <- ip[names(ip) %in% c(protein.acs.w.isoforms,protein.acs)]
      result <- c(result,protein.gs)
   }
 }
 result <- unique(as.character(result))
 if (length(result) == 0)
   warning("Could not find protein group identifier for ",pattern)
 return(result)
})

setGeneric("protein.ac",function(x,protein.g) standardGeneric("protein.ac"))
setMethod("protein.ac",signature("ProteinGroup","character"),
  function(x,protein.g) {
   unique(x@isoformToGeneProduct[indistinguishableProteins(x,protein.g=protein.g),"proteinac.wo.splicevariant"])
}) 

setMethod("protein.ac",signature("ProteinGroup","missing"),
  function(x) {
   unique(x@isoformToGeneProduct[,"proteinac.wo.splicevariant"])
})



### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### show method.
###

setMethod("show","ProteinGroup",function(object) {
      cat("ProteinGroup object\n")
      cat(sprintf("%8d peptides are detected\n",nrow(peptideSpecificity(object))))
      cat(sprintf("%8d protein groups with specific peptides\n",length(reporterProteins(object,require.reporter.specific=TRUE))))
    }
)

get.pep.group <- function(x,protein) {
  pep.specificity <- peptideSpecificity(x)
  protein.group.table <- proteinGroupTable(x)

  message(protein)
  speci <- rep(3,length(pep.specificity$peptide))
  speci[pep.specificity$specificity==GROUPSPECIFIC] <-  2
  speci[pep.specificity$specificity==REPORTERSPECIFIC] <- 1
  names(speci) <- pep.specificity$peptide

  protein.group <- protein.group.table[protein.group.table$reporter.protein==protein,]
  pepnprots <- peptideNProtein(x)
  pp <- as.data.frame(pepnprots[pepnprots[,'protein.g'] %in%
                                protein.group$protein.g,,drop=FALSE],stringsAsFactors=FALSE)

  # get all peptides of protein
  peptides <- unique(pp[,'peptide'])
  t <- table(pp$protein.g)
  proteins <- names(sort(t,decreasing=TRUE))

  # to which proteins are peptides assigned?
  pep.groups <-
    ddply(pp[pp[,'protein.g'] %in% proteins,],"peptide",
          function(s) data.frame(peptide=unique(s$peptide),
                                 protein.g=s['protein.g'],
                                 group=paste(sort(unique(s$protein.g)),collapse="-"),
                                 speci=speci[unique(s$peptide)],
                                 stringsAsFactors=FALSE
                                 )
          )
  pep.groups <- unique(pep.groups[,c("peptide","group","speci")])
  pep.groups[order(pep.groups$speci,pep.groups$peptide),]
}

## create a extended proteinGroup table
## TODO: check function usage
.extend.protein.group <- function(ibspectra,calc.ratio) {
  pg <- proteinGroup(ibspectra)
  protein.group <- proteinGroupTable(pg)
  pnp <- peptideNProtein(pg)
  t <- table(pnp[,"peptide"])
  pnp <- pnp[pnp[,"peptide"] %in% names(t)[t==2],]

  protein.group$lratio <- NA
  protein.group$var    <- NA
  protein.group$is.different    <- FALSE
  protein.group$is.smaller    <- 0
  protein.group$is.bigger    <- 0
  for (i in seq_len(nrow(protein.group))) {
    if (protein.group[i,"is.reporter"]) {
      reporter.protein <- protein.group[i,"protein.g"]
      reporter.ratio <- calc.ratio(ibspectra,protein=reporter.protein)
      protein.group[i,c("lratio","var")] <- reporter.ratio[c("lratio","variance")]
    } else {
      ## reporter.ratio set above - protein.group is sequential
      ## subset protein: get peptide shared with (and only with) reporter
      shared.peptides <- pnp[pnp[,"protein.g"]==protein.group[i,"protein.g"],"peptide"]
      if (length(shared.peptides)>0) {
        shared.ratio <- calc.ratio(ibspectra,peptide=shared.peptides)
        protein.group[i,c("lratio","var")] <- shared.ratio[c("lratio","variance")]
      
        shared.issmaller <- shared.ratio["lratio"] + sqrt(shared.ratio["variance"]) < 
            reporter.ratio["lratio"] - sqrt(reporter.ratio["variance"])
        shared.isbigger <- shared.ratio["lratio"] - sqrt(shared.ratio["variance"]) > 
            reporter.ratio["lratio"] + sqrt(reporter.ratio["variance"])

        protein.group[i,"is.different"] <- shared.issmaller | shared.isbigger
        protein.group[i,"is.smaller"] <- shared.issmaller
        protein.group[i,"is.bigger"] <- shared.isbigger

      }
    }
  }
  return(protein.group)
}


groupMemberPeptides <- function(x,reporter.protein.g,
                                ordered.by.pos=TRUE,only.first.pos=TRUE) {
   peptide.info <- unique(x@peptideInfo[,c("protein","peptide","start.pos")])
   group.table <- proteinGroupTable(x)
   indist.proteins <- x@indistinguishableProteins

   reporter.proteins <- names(indist.proteins)[indist.proteins==reporter.protein.g]
   my.group.table <- group.table[group.table$reporter.protein==reporter.protein.g,]

   # 1. get reporter peptides (w/ position and specificity)
   my.peptide.info <- peptides(x,reporter.protein.g,
                               columns=c("peptide","specificity",
                                         "n.shared.groups","n.shared.proteins"),drop=FALSE,do.warn=FALSE)
 
   if (ordered.by.pos) {     
      reporter.peptide.startpos <- 
         peptide.info[peptide.info$peptide %in% my.peptide.info$peptide &
         peptide.info$protein %in% reporter.proteins[1],c("peptide","start.pos")]

      if (only.first.pos) {
         my.peptide.info$start.pos <- 0
         for (peptide.i in seq_len(nrow(my.peptide.info))) {
            sel <- reporter.peptide.startpos$peptide==my.peptide.info$peptide[peptide.i]
            my.peptide.info$start.pos[peptide.i] <- sort(reporter.peptide.startpos$start.pos[sel])[1]
         }
      } else {
          my.peptide.info <- merge(reporter.peptide.startpos,my.peptide.info)
      }
      my.peptide.info <- my.peptide.info[order(my.peptide.info$start.pos),]
   } 
   rownames(my.peptide.info) <- NULL

   # 2. get group member peptides
   group.member.peptides <- do.call(cbind,lapply(my.group.table$protein.g,
      function(protein.g) my.peptide.info$peptide %in% peptides(x,protein.g,do.warn=FALSE)))
   rownames(group.member.peptides) <- my.peptide.info$peptide
   colnames(group.member.peptides) <- my.group.table$protein.g
   return(list(peptide.info=my.peptide.info,group.member.peptides=group.member.peptides))
}


human.protein.names <- function(my.protein.info) {
  my.df <- my.protein.info[,c("accession","splicevariant","gene_name","protein_name")]
  my.df$gene_name <- sanitize(my.df$gene_name)
  collapsed.splicevariant <- ddply(my.df,"accession",function(x) {
        only_one <- nrow(x) == 1
        if (!all(is.na(x$splicevariant)) & any(!is.na(x$splicevariant)))
          x$splicevariant[is.na(x$splicevariant)] <- 1

        x$splicevariant <- number.ranges(x$splicevariant)
        x <- unique(x)
        if (nrow(x) > 1) 
          x <- as.data.frame(lapply(x,function(y) paste(y[!is.na(y)],collapse=",")),stringsAsFactors=FALSE)
        
        if (is.na(x$splicevariant)) {
          x$ac_link <- sprintf("\\uniprotlink{%s}",sanitize(x$accession,dash=FALSE))
          x$ac_nolink <-  x$accession
        } else {
          x$ac_link <- sprintf("\\uniprotlink{%s}$_{%s}$",sanitize(x$accession,dash=FALSE),x$splicevariant);
          if (only_one) {
            x$ac_nolink <- sprintf("%s-%s",x$accession,x$splicevariant)
          } else {
            x$ac_nolink <- sprintf("%s-[%s]",x$accession,x$splicevariant)
          }
        }
        return(unique(x))
    })
    
    collapsed.gene_name <- ddply(
          unique(collapsed.splicevariant[,c("gene_name","ac_link","ac_nolink","protein_name")]),
          "gene_name",function(x) {
        if (is.na(x$gene_name[1]) || x$gene_name[1] == "") {
          x$name_nolink <- x$ac_nolink
        } else {
          x$ac_link_bold <- sprintf("\\textbf{%s}: %s",unique(x$gene_name),x$ac_link)
          x$ac_link <- sprintf("%s {\\small %s}",unique(x$gene_name),x$ac_link)
          x$ac_nolink <- sprintf("%s: %s",unique(x$gene_name),x$ac_nolink)
          x$name_nolink <- sprintf("%s: %s",x$gene_name,x$protein_name)
        }
        return(unique(x))
    })
    return(collapsed.gene_name)
}


my.protein.info <- function(x,protein.g) {
    protein.acs <- indistinguishableProteins(x,protein.g=protein.g)
    if (all(is.na(protein.acs))) return()
    isoforms <- x@isoformToGeneProduct
    res <- data.frame(protein.ac=protein.acs,
                      accession=isoforms[protein.acs,"proteinac.wo.splicevariant"],
                      splicevariant=as.integer(isoforms[protein.acs,"splicevariant"]), 
                      stringsAsFactors=FALSE)

    if (length(proteinInfo(x)) > 0) {
      if (proteinInfoIsOnSpliceVariants(proteinInfo(x))) {
        res <- merge(res,proteinInfo(x),by.x="protein.ac",by.y="accession",all.x=TRUE,all.y=FALSE)
      } else {
        res <- merge(res,proteinInfo(x),by="accession",all.x=TRUE,all.y=FALSE)
      }
    } else {
      res <- cbind(res,gene_name=NA,protein_name=NA)
    }

    return(.moveToFirstCol(res,"accession"))
}

summary.ProteinGroup <- function(object,only.reporters=TRUE,...) {
  peptide.cnt <- table(peptideNProtein(object)[,"protein.g"])
  peptide.spectra.cnt <- table(spectrumToPeptide(object))
  spectra.cnt <- tapply(peptideNProtein(object)[,"peptide"],peptideNProtein(object)[,"protein.g"],function(pep) sum(peptide.spectra.cnt[pep]))
  
  if (only.reporters) 
    proteins <- reporterProteins(object)
  else
    proteins <- names(spectra.cnt)
  
  proteins <- proteins[order(spectra.cnt[proteins],decreasing=TRUE)]
    
  return(data.frame(peptide.cnt=peptide.cnt[proteins],
                    spectra.cnt=spectra.cnt[proteins]))
  
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Quantification functions (dNSAF).
###  based on peptide or spectral count

peptide.count <- function(protein.group,protein.g=reporterProteins(protein.group),
                          specificity=c("reporter-specific","group-specific","unspecific"),...) {
  sapply(protein.g,
         function(p) length(peptides(protein.group,p,specificity=specificity,...)))
}

spectra.count2 <- function(ibspectra,value=reporterProteins(protein.group),type="protein.g",
                          specificity=c("reporter-specific","group-specific","unspecific"),
                          modif=NULL,combine=FALSE,subset=NULL,require.quant=NULL,...) {
  protein.group <- proteinGroup(ibspectra)
  if (!isTRUE(combine)) {
    spectra.count <- sapply(value, function(p) 
                            spectra.count2(ibspectra,p,type,specificity,modif,combine=TRUE,
                                           subset=subset,require.quant=require.quant,...))
    names(spectra.count) <- value
    return(spectra.count)
  }

  pep <- switch(type,
                protein.g=peptides(protein.group,protein=value,specificity=specificity,...),
                peptide=value)

  ## Calculate unique spectrum counts for all proteins
  if (!is.null(subset)) {
    fd <- subset(fData(ibspectra),eval(subset) & peptide %in% pep)
  } else {
    fd <- subset(fData(ibspectra),peptide %in% pep)
  }

  if (!is.null(require.quant)) {
    spectra <- rownames(reporterIntensities(ibspectra,na.rm=TRUE,na.rm.f=require.quant))
    fd <- fd[fd$spectrum %in% spectra,]
  }

  if (is.null(modif)) {
    peptide.spectra.count <- table(fd$peptide)
    sum(peptide.spectra.count)
  } else {
    has.modif <- sapply(strsplit(fd$modif,":"),function(x) any(x %in% modif))
    sum(has.modif)
  }
}

spectra.count <- function(protein.group,protein.g=reporterProteins(protein.group),
                          specificity=c("reporter-specific","group-specific","unspecific"),
                          modif=NULL,...) {
  ## Calculate unique spectrum counts for all proteins
  if (is.null(modif)) {
    peptide.spectra.count <- table(spectrumToPeptide(protein.group))
    spectra.count <- sapply(protein.g, function(p) 
                            sum(peptide.spectra.count[peptides(protein.group,protein=p,
                                                               specificity=specificity,...)]))
  } else {
    si <- protein.group@spectrumId
    has.modif <- sapply(strsplit(si$modif,":"),function(x) any(x %in% modif))
    spectra.count <- sapply(protein.g, function(p) {
                            pep <- peptides(protein.group,protein=p,specificity=specificity,...)
                            sum(si$peptide %in% pep & has.modif) })
  }

  names(spectra.count) <- protein.g
  return(spectra.count)
}

calcPeptidePosition <- function(peptide.info,protein.info,calc.il.peptide) {
  peptide.info[,'start.pos'] <- NA
  if (missing(calc.il.peptide)) 
    calc.il.peptide <- !('real.peptide' %in% colnames(peptide.info) && !all(is.na(peptide.info[,'real.peptide'])))
  if (isTRUE(calc.il.peptide)) 
    peptide.info[,'real.peptide'] <- NA
  peptide.info <- unique(peptide.info)
  last.elem <- nrow(peptide.info)

  for (p.i in seq_len(nrow(peptide.info))) {
    ## WARNING: Does not yet work on protein groups which have no splice variant specified!
    seqq <- protein.info[protein.info[,'accession']==peptide.info[p.i,'protein'],'sequence'][1]
    pep.i <- peptide.info[p.i,]
    pep <- peptide.info[p.i,'peptide']
    if (isTRUE(getOption("isobar.verbose")))
      message(pep,appendLF=FALSE)
    if (is.na(pep) || nchar(pep) == 0) stop("peptide on position ",p.i," is NA or of zero length")
    if (length(seqq) == 0 || is.na(seqq) || nchar(seqq) == 0) {
      peptide.info[p.i,'real.peptide'] <- paste(pep,"(not matched, no protein sequence)")
      if (isTRUE(getOption("isobar.verbose")))
      message(": no prot sequence")
      next
    } 

    il.seqq <- gsub("I","L",seqq) ## to be confomant with converted peptides
    pep <- gsub("I","L",pep)
    matches <- gregexpr(pep,il.seqq,fixed=TRUE)[[1]]

    pp.i <- p.i
    for (m.i in seq_along(matches)) {
       if (m.i > 1) {
         last.elem <- last.elem + 1
         peptide.info <- rbind(peptide.info,pep.i)
         pp.i <- last.elem
       }
       if (matches[m.i] == -1) {
         peptide.info[pp.i,'start.pos'] <- -1
         peptide.info[pp.i,'real.peptide'] <- paste(pep,"(not matched)")
       } else {
         peptide.info[pp.i,'start.pos'] <- matches[m.i]
         peptide.info[pp.i,'real.peptide'] <- 
                 substr(seqq,matches[m.i],matches[m.i]+attr(matches,"match.length")[m.i]-1)
       }
    }
    if (isTRUE(getOption("isobar.verbose")))
      message(" --> ",peptide.info[pp.i,'real.peptide'])
  }
  if (any(peptide.info[,'start.pos'] == -1,na.rm=TRUE)) {
    sel.bad <- !is.na(peptide.info[,'start.pos']) & peptide.info[,'start.pos'] == -1
    warning("Could not match ",length(unique(peptide.info[sel.bad,'peptide'])),
            " peptides in ",length(unique(peptide.info[sel.bad,'protein']))," proteins")
    peptide.info[sel.bad,'start.pos'] <- NA
  }
  
  peptide.info[order(peptide.info[,'protein'],peptide.info[,'start.pos'],peptide.info[,'peptide']),]
}

# TOFIX: proteinInfo does not contain splice information. With splice sequence, an accurate seqcov could be calculated
sequence.coverage <- function(protein.group,protein.g=reporterProteins(protein.group),
                              specificity=c("reporter-specific","group-specific","unspecific"),
                              simplify=TRUE,...) {

  if (!proteinInfoIsOnSpliceVariants(proteinInfo(protein.group)))
    warning("Protein information is not on splice variants - sequence coverage will be approximate only.")


  if ("length" %in% colnames(proteinInfo(protein.group)) && 
      "start.pos" %in% colnames(protein.group@peptideInfo)) {

    if (all(is.na(protein.group@peptideInfo[,'start.pos']))) {
      protein.group@peptideInfo <- calcPeptidePosition(protein.group@peptideInfo,proteinInfo(protein.group))
    }
    lengths <- proteinInfo(protein.group,protein.g=protein.g,select="length",simplify=FALSE)
    peptides <- peptides(protein.group,protein=protein.g,specificity=specificity,...)
    peptide.info <- unique(protein.group@peptideInfo[,c("protein","peptide","start.pos")])
    peptide.info <- peptide.info[peptide.info[["peptide"]] %in% peptides,]
    protein.ac.wo.splice <- isobar:::.as.vect(protein.group@isoformToGeneProduct)

    res <- sapply(protein.g,function(p) {
             seqcov <- sapply(indistinguishableProteins(protein.group,protein.g=p),function(pp) {
               if (proteinInfoIsOnSpliceVariants(proteinInfo(protein.group)))
                 ppp <- pp
               else
                 ppp <- protein.ac.wo.splice[pp]
               .calc.seqcov(lengths[[p]][as.character(ppp)],
                            peptide.info[peptide.info$protein==pp,])
             })
             if (isTRUE(simplify))
               .simplify.seqcov(seqcov)
             else
               seqcov
    })
    names(res) <- protein.g
    res
  } else {
    stop("Necessary information for sequence coverage not available")
  }
}

.calc.seqcov <- function(length,peptide.info) {
      if (is.na(length) || all(is.na(peptide.info$start.pos)))
        return(NA)
      seqq <- rep(FALSE,length)
      peptide.info <- peptide.info[!is.na(peptide.info$start.pos),]
      peptide.info$peplength <- nchar(peptide.info$peptide)
      peptide.info$end.pos <- peptide.info$start.pos+peptide.info$peplength-1
      if (any(peptide.info[,'end.pos']>length)) {
        warning(".calc.seqcov: peptides present which span over the protein length, removing them")
        peptide.info <- peptide.info[peptide.info[,'end.pos'] <= length,]
      }
      for (i_r in seq_len(nrow(peptide.info)))
        seqq[seq(from=peptide.info[i_r,"start.pos"],to=peptide.info[i_r,"end.pos"])]  <- TRUE
      return(sum(seqq)/length(seqq))
}

.simplify.seqcov <- function(seqcov) {
      if (length(seqcov) > 1)
        mean(seqcov,na.rm=TRUE)
      else
        seqcov
}


calculate.dNSAF <- function(protein.group,use.mw=FALSE,normalize=TRUE,combine.f=mean) {
  if (is.null(proteinInfo(protein.group)) || length(proteinInfo(protein.group)) == 0)
    stop("slot proteinInfo not set - it is needed for sequence length")
  if (!"length" %in% colnames(proteinInfo(protein.group)))
    stop("no column 'length' in proteinInfo slot")

  ip <- indistinguishableProteins(protein.group)
  if(proteinInfoIsOnSpliceVariants(proteinInfo(protein.group))) {
    get.to.acs <- function(my.protein.g) names(ip)[ip==my.protein.g]
  } else {
    get.to.acs <- function(my.protein.g) unique(protein.ac(protein.group,my.protein.g))
  }

  seqlength <- proteinInfo(protein.group)$length
  names(seqlength) <- proteinInfo(protein.group)$accession
  
  spectrum.counts <- table(spectrumToPeptide(protein.group))
  ## Calculate unique spectrum counts for all proteins
  uspc <- sapply(reporterProteins(protein.group), function(p) 
                 sum(spectrum.counts[peptides(protein.group,protein=p,
                                              specificity=c("reporter-specific","group-specific"))]))
  names(uspc) <- reporterProteins(protein.group)
  
  ## Calculate distribution factors for each shared peptide
  pnp <- peptideNProtein(protein.group)
  pnp <- pnp[pnp[,"protein.g"] %in% reporterProteins(protein.group),]
  t <- table(pnp[,"peptide"])
  shared.peptides <- names(t)[t>1]
  distr.factors <- lapply(shared.peptides,function(pep) {
    proteins <- pnp[pnp[,"peptide"]==pep,"protein.g"]
    uspc.p <- uspc[proteins]
    uspc.p/sum(uspc.p)
  })
  names(distr.factors) <- shared.peptides

  ## Calculate dSAF
  dSAF <- sapply(reporterProteins(protein.group), function(p) {
    shared.pep.p <- peptides(protein.group,protein=p,specificity="unspecific",do.warn=FALSE)
    if (length(shared.pep.p) > 0)
      d.sspc <- sum(sapply(shared.pep.p, function(pep) spectrum.counts[pep]*distr.factors[[pep]][p]))
    else
      d.sspc <- 0

    ## for protein length, we take the sum of lengths of all acs
    protein.length <- sum(seqlength[get.to.acs(p)],na.rm=TRUE)
    dSAF <- (uspc[p] + d.sspc) / protein.length
    if (is.nan(dSAF) | !is.finite(dSAF)) dSAF <- NA
    return(dSAF)
  })

  if (use.mw) {
    mw <- .calculate.mw(protein.group,protein.g,combine.f)
    dSAF <- dSAF*mw
  }

  ## normalize to dNSAF
  if (normalize) dNSAF <- dSAF/sum(dSAF,na.rm=TRUE)

  names(dNSAF) <- reporterProteins(protein.group)
  return(dNSAF)
}

n.observable.peptides <- function(...) {
  return(nrow(observable.peptides(...)))
}

# crude calculations:
#   mass of B: (mass_N+mass_D)/2, mass_N=114.04293; mass_D=215.06680
#   mass of Z: (mass_E+mass_Q)/2, mass_E=229.08245; mass_Q=328.13828
#   mass of J: mass of L
observable.peptides <- function(seq,nmc=1,min.length=6,min.mass=600,max.mass=4000,
                                custom=list(code=c("B","Z","J","U"),
                                            mass=c(164.554862,278.61037,213.12392,150.953636)),...) {
  if (is.na(seq) || length(seq)==0 || nchar(seq) == 0)
    return(0)
  seq <- gsub("[ ,]","",seq)
  requireNamespace("OrgMassSpecR")
  pep <- OrgMassSpecR::Digest(seq,missed=nmc,custom=custom,...)
  min.length.ok <- nchar(pep[,"peptide"]) >= min.length
  mass.ok <- pep[,"mz1"] >= min.mass & pep[,"mz3"] <= max.mass
  pep[min.length.ok & mass.ok,]
}

.calculate.mw <- function(protein.group,protein.g,combine.f=mean) {
  requireNamespace("OrgMassSpecR")
  sequences <- proteinInfo(protein.group)$sequence
  ip <- indistinguishableProteins(protein.group)
  names(sequences) <- proteinInfo(protein.group)$accession

  if(proteinInfoIsOnSpliceVariants(proteinInfo(protein.group)))
    get.to.acs <- function(my.protein.g) names(ip)[ip==my.protein.g]
  else
    get.to.acs <- function(my.protein.g) unique(protein.ac(protein.group,my.protein.g))

  sapply(protein.g, function(my.protein.g) 
         do.call(mean,lapply(get.to.acs(my.protein.g),
                                  function(protein.ac) OrgMassSpecR::MolecularWeight(formula = OrgMassSpecR::ConvertPeptide(sequences[protein.ac])))))
}

calculate.emPAI <- function(protein.group,protein.g=reporterProteins(protein.group),
                            normalize=FALSE,observed.pep=c("pep","mod.charge.pep"),
                            use.mw=FALSE,combine.f=mean,...,nmc=0,report.all=FALSE) {
  if (is.null(proteinInfo(protein.group)) || length(proteinInfo(protein.group)) == 0)
    stop("slot proteinInfo not set - it is needed for sequence length")
  if (!"sequence" %in% colnames(proteinInfo(protein.group))) {
    stop("no column 'sequence' in proteinInfo slot")
  }
  ip <- indistinguishableProteins(protein.group)

  sequences <- proteinInfo(protein.group)$sequence
  names(sequences) <- proteinInfo(protein.group)$accession
 
  n.observed.peptides <- switch(observed.pep[1],
                                pep = table(peptideNProtein(protein.group)[,"protein.g"])[protein.g],
                                mod.charge.pep = {
                                  pep.cnt <- table(unique(protein.group@spectrumId[,c("peptide","modif","charge")])[,"peptide"])
                                  tapply(peptideNProtein(protein.group)[,"peptide"],peptideNProtein(protein.group)[,"protein.g"],function(x) sum(pep.cnt[x]))[protein.g]
                                })
  if(proteinInfoIsOnSpliceVariants(proteinInfo(protein.group)))
    get.to.acs <- function(my.protein.g) names(ip)[ip==my.protein.g]
  else
    get.to.acs <- function(my.protein.g) unique(protein.ac(protein.group,my.protein.g))

  n.observable.peptides <- sapply(protein.g, function(my.protein.g) 
                                  do.call(combine.f,lapply(get.to.acs(my.protein.g),
                                                           function(protein.ac) n.observable.peptides(sequences[protein.ac],nmc=nmc,...))))
  if (use.mw) 
    mw <- .calculate.mw(protein.group,protein.g,combine.f)
   else 
    mw <- 1

  empai <- 10^(n.observed.peptides/n.observable.peptides) - 1
  if (report.all) 
    return(data.frame(protein.g,n.observed.peptides,n.observable.peptides,mw,
                      empai,protein_mol=empai/sum(empai,na.rm=TRUE),protein_weight=empai*mw/sum(empai*mw,na.rm=TRUE)))

  if (use.mw) empai <- empai*mw
  if (normalize) empai/sum(empai,na.rm=TRUE)

  return(empai)
}

# pretty format group identifiers
#   Input: c("P1234-1,P1234-2","P6543")
#  Output: c("P1234-[1,2]","P6543")
.protein.acc <- function(my.protein.g,protein.group) {

  if (length(my.protein.g) > 1) 
    return(sapply(my.protein.g,function(p) .protein.acc(p,protein.group)))

  protein.acs <- as.character(indistinguishableProteins(protein.group,protein.g=my.protein.g))

  if (length(protein.acs) == 1) return(protein.acs)
  splice.df <- protein.group@isoformToGeneProduct[protein.acs,]
  if (all(is.na(splice.df[,'splicevariant']))) 
    return(paste0(splice.df[,'proteinac.w.splicevariant'],collapse=", "))

  res <- ddply(splice.df,"proteinac.wo.splicevariant",function(y) {
        if (nrow(y) == 1) return(y[,'proteinac.w.splicevariant'])
        if (all(is.na(y$splicevariant))) return(y[1,'proteinac.w.splicevariant'])

        return(sprintf("%s-[%s]",
                       y[1,'proteinac.wo.splicevariant'],
                       number.ranges(y[,'splicevariant'])))
  })
  return(as.character(paste0(res[,'V1'],collapse=", ")))
}

#!/usr/bin/Rscript

suppressPackageStartupMessages(library(isobar))
suppressPackageStartupMessages(library(plyr))

all.files <- commandArgs(TRUE)
if (length(all.files) < 2)
  stop("usage: msgfpParser.R result.id.csv file1.msgfp.tsv file2.msgfp.tsv ...")

out.file <- all.files[1]
msgfp.files <- all.files[2:length(all.files)]

message("reading MSGF files:")
message("\t",paste(msgfp.files,collapse="\n\t"))

ib.df <- ldply(msgfp.files,isobar:::.read.msgfp.tsv)
ib.df <- ib.df[ib.df[,'msgf.specevalue'] <= 0.05 & ib.df[,'msgf.rawscore'] > 0,]
writeIBSpectra(ib.df,file=out.file,quote=FALSE)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Constructor and readIBSpectra.
###

.check.columns <- function(identifications,data.ions=NULL,data.mass=NULL,allow.missing.columns=FALSE) {
  SC <- .SPECTRUM.COLS[.SPECTRUM.COLS %in% colnames(identifications)]
  PC <- .PROTEIN.COLS[.PROTEIN.COLS %in% colnames(identifications)]
  missing.cols = c()
  for (col in c('SPECTRUM','PEPTIDE','MODIFSTRING'))
    if (!is.element(.SPECTRUM.COLS[col],SC))
      missing.cols <- c(missing.cols,.SPECTRUM.COLS[col])
  for (col in c('PROTEINAC'))
    if (!is.element(.PROTEIN.COLS[col],PC))
      missing.cols <- c(missing.cols,.PROTEIN.COLS[col])

  if (!is.null(data.ions) && !is.null(data.mass)) {
    if (is.null(rownames(data.ions)) || is.null(rownames(data.mass)))
      stop("provide spectrum ids as rownames for data.ions and data.mass")
    if (!identical(rownames(data.ions),rownames(data.mass)))
      stop("data.ions and data.mass do not have identical rownames!")
  }

  # handle missing columns
  if (length(missing.cols) > 0) {
      msg <- paste("not all required columns in identifications, the following are missing: \n\t",
               paste(missing.cols,collapse="\n\t"),
               "\nData:\n")
      #paste(capture.output(print(head(identifications))))
      if (allow.missing.columns) {
         warning(msg)
         print(head(identifications))
         identifications[,missing.cols] <- NA
      } else {
         message(msg)
         print(head(identifications))
         stop()
      }
  }
  message(" data.frame columns OK")
  return(identifications)
}

.get.quant <- function(identifications,colname,reporterTagNames) {
    cols <- sprintf(colname,reporterTagNames)
    if (!all(cols %in% colnames(identifications)))
      stop(" Quantitative information missing: Supply columns '",paste(cols,collapse="', '"),"'.")
    qdata <- unique(identifications[,c(.SPECTRUM.COLS['SPECTRUM'],cols)])
    identifications <<- identifications[,-which(colnames(identifications) %in% cols)]
    rownames(qdata) <- qdata[,.SPECTRUM.COLS['SPECTRUM']]
    qdata <- as.matrix(qdata[,-1])
    colnames(qdata) <- reporterTagNames
    return(qdata)
}

.check.assayDataElements <- function(assayDataElements) {

  for (elem in assayDataElements) {
    if (is.null(rownames(elem)))
      stop("provide rownames for all assayDataElements")
    #if (!identical(rownames(elem),rownames(data.ions)))
    #  stop("all assayDataElements must have the same rownames")
    #if (!identical(dim(data.ions),dim(elem)))
    #  stop("all assayDataElements must have the same dim")
  }
}

.get.quant.elems <- function(assayDataElements,data.ions,data.mass,spectra.ids,fragment.precision,reporterTagMasses) {
  if (is.null(assayDataElements))
    assayDataElements <- list()

  assayDataElements$ions <- data.ions
  assayDataElements$mass <- data.mass
 
  
  if (!identical(spectra.ids,rownames(data.ions))) {
    id.n.quant <- intersect(spectra.ids,rownames(data.mass))
    id.not.quant <- setdiff(spectra.ids,rownames(data.mass))
    quant.not.id <- setdiff(rownames(data.mass),spectra.ids)

    if (length(id.n.quant)==0) stop("No spectra could be matched between identification and quantifications.",
                                    "If the search was preformed with Mascot, set readIBSpectra argument decode.titles=TRUE,\n",
                                    " or in the properties file for report generation:\n",
                                    "   readIBSpectra.args=list(decode.titles=TRUE)")
    
    if (length(quant.not.id) > 0)
      message(" for ",length(quant.not.id)," spectra ",
              "[",round(length(quant.not.id)*100/nrow(data.mass),2),"%],",
              " quantitative information is available,\n",
              "   but no peptide-spectrum match. Spectrum titles: \n\t",
              paste(quant.not.id[1:min(length(quant.not.id),2)],collapse=",\n\t"),", ...")
    if (length(id.not.quant) > 0)
      message(" for ",length(id.not.quant)," spectra ",
              "[",round(length(id.not.quant)*100/length(spectra.ids),2),"%]",
              " with assigned peptides,\n",
              "   no reporter intensities are available. spectrum titles: \n\t",
              paste(id.not.quant[1:min(length(id.not.quant),2)],collapse=",\n\t"),", ...")

    na.matrix <- matrix(NA,nrow=length(spectra.ids),ncol=ncol(data.mass),
                        dimnames=list(spectra.ids,colnames(data.mass)))

    for (elem in names(assayDataElements)) {
      tmp <- na.matrix
      tmp[id.n.quant,] <- assayDataElements[[elem]][id.n.quant,]
      assayDataElements[[elem]] <- tmp
    }
  }

  # filter based on fragment precision
  if (!is.null(fragment.precision)) {
    
    min.masses <- reporterTagMasses - fragment.precision/2
    max.masses <- reporterTagMasses + fragment.precision/2
    for (i in seq_len(nrow(assayDataElements$mass))) {
      bad.mass <- assayDataElements$mass[i,]<min.masses |  assayDataElements$mass[i,] > max.masses
      if (any(bad.mass,na.rm=TRUE)) {
        for (elem in names(assayDataElements)) {
          assayDataElements[[elem]][i,bad.mass] <- NA
        }
      }
    }
  }

  assayDataElements$ions[which(assayDataElements$ions==0)] <- NA
  assayDataElements$mass[which(assayDataElements$mass==0)] <- NA
  if (all(apply(is.na(assayDataElements$mass),2,all))) stop("Unable to extract any reporter m/z values. Try setting 'fragment.precision' higher (current value: ",fragment.precision,"), or to NULL if no filtering is desired.")
  if (all(apply(is.na(assayDataElements$ions),2,all))) stop("Unable to extract any reporter intensities.")

  return(assayDataElements)
}

.get.varMetaData <- function(identifications) {
  nn <- .SPECTRUM.COLS %in% colnames(identifications)

  label.desc <- c(PEPTIDE='peptide sequence',
      MODIFSTRING='modifications of peptide',
      CHARGE='peptide charge state',
      THEOMASS='theoretical peptide mass',
      EXPMASS='experimental peptide mass',
      EXPMOZ='experimental mass-to-charge ratio',
	    PRECURSOR.ERROR='precursor error',
      PARENTINTENS='parent ion intensity',
      RT='retention time',
      DISSOCMETHOD='dissociation METHOD',
      SPECTRUM='spectrum title',
      SPECTRUM.QUANT='title of spectrum used for quantiation',
      PRECURSOR.PURITY="precursor purity",
      RAWFILE="raw file",NMC="nmc",
	    DELTASCORE="delta score",DELTASCORE.PEP="delta score peptide",
      SCANS="scans",
      SCANS.FROM="scans from",SCANS.TO="scans to",
	    MASSDELTA.ABS="massdelta (abs)",MASSDELTA.PPM="massdelta (ppm)",
      SEARCHENGINE='protein search engine',
      SCORE='protein search engine score',
	    SCORE.MSGF='MSGF+ score',

      SCORE.MASCOT="Mascot search score",
      SCORE.PHENYX="Phenyx search score",
      SCORE.PHOSPHORS='PhosphoRS pepscore',
      PROB.PHOSPHORS="PhosphoRS probability",
      SCAFFOLD.PEPPROB="Scaffold: Peptide Probability",
      SEQUEST.XCORR="sequest:xcorr",
      SEQUEST.DELTACN="sequest:deltacn",

      IS.DECOY="is decoy peptide-spectrum match?",
      P.VALUE="p value",
      MSGF.RAWSCORE="MSGF raw score",
      MSGF.DENOVOSCORE="MSGF de novo score",
      MSGF.SPECEVALUE="MSGF specturm EValue",
      MSGF.EVALUE="MSGF EValue",
      MSGF.QVALUE="MSGF QValue",
      MSGF.PEPQVALUE="MSGF PepValue",

      PHOSPHO.SITES="phosphorylation sites",
      USEFORQUANT='use spectrum for quantification',
      SEQPOS='PTM seqpos',
      SITEPROBS='PhosphoRS site.probs',
      PEP.SITEPROBS='PhosphoRS site.probs inside peptide',
      FILE='file',SAMPLE='sample',NOTES='notes'
      )
  if (!all(names(.SPECTRUM.COLS) %in% names(label.desc))) {
    sel.bad <- !names(.SPECTRUM.COLS) %in% names(label.desc)
    warning("Not all SPECTRUM COLS have a label description:\n\t",
            paste(names(.SPECTRUM.COLS)[sel.bad],collapse="\n\t"))
    label.desc[names(.SPECTRUM.COLS)[sel.bad]] <- .SPECTRUM.COLS[sel.bad]
  }


  VARMETADATA=data.frame(labelDescription=label.desc[names(.SPECTRUM.COLS)],
                         row.names=.SPECTRUM.COLS)
	    
  return(VARMETADATA[nn,,drop=FALSE])
}

.remove.duplications <- function(identifications) {
  identifications <- unique(identifications)
  SC <- .SPECTRUM.COLS[.SPECTRUM.COLS %in% colnames(identifications)]
  if (max(table(identifications[,SC['SPECTRUM']])) > 1) {
    t <- table(identifications[,SC['SPECTRUM']])
    warning(sum(t>1)," spectra have diverging identifications after the merging, removing them.")
    print(head(identifications[identifications[,SC['SPECTRUM']] %in% names(t)[t>1],]))

    identifications <- identifications[!identifications[,SC['SPECTRUM']] %in% names(t)[t>1],]
  }

  if ('SEARCHENGINE' %in% names(SC)) {
    message("  Identification details:")
    tt <- sort(table(identifications[,SC['SEARCHENGINE']]))
    stats <- data.frame(perc=sprintf("%.2f %%",tt/sum(tt)*100),n=tt)
    if ('SCORE' %in% names(SC)) {
      scores <- identifications[,SC['SCORE']]
      score.stats <- do.call(rbind,lapply(names(tt),function (se) {
        my.scores <- scores[identifications[,SC['SEARCHENGINE']]==se]
        summary(my.scores,na.rm=TRUE)}))
      stats <- cbind(stats,score.stats)
    } 
    print(stats)
    
  }
  identifications
}

.merge.identifications.full <- function(identifications, ...) {
  SC <- .SPECTRUM.COLS[.SPECTRUM.COLS %in% colnames(identifications)]
  ## Substitute Isoleucins with Leucins (indistinguishable by Masspec)
  if (!.PEPTIDE.COLS['REALPEPTIDE'] %in% colnames(identifications)) 
     identifications <- .fix.il.peptide(identifications)

  ## Separate protein columns (focus on peptide-spectrum matches)
  PC <- unique(c(.SPECTRUM.COLS['PEPTIDE'],.PROTEIN.COLS,.PEPTIDE.COLS))
  protein.colnames <- colnames(identifications)[colnames(identifications) %in% c(PC)]
  pept.n.prot <- unique(identifications[,protein.colnames])
  identifications <- unique(identifications[,-which(colnames(identifications) %in% setdiff(PC,.SPECTRUM.COLS['PEPTIDE']))])

  ## Merge identifications
  if (max(table(identifications[,SC['SPECTRUM']])) > 1) {
    if (SC['DISSOCMETHOD'] %in% colnames(identifications) && 
        length(unique(identifications[,SC['DISSOCMETHOD'] ]))) {
      identifications <- dlply(identifications,'dissoc.method',.merge.identifications,...)
      # rbind separately to assure equal column names
      identifications <- do.call(rbind,identifications)
      identifications <- .merge.quant.identifications(identifications)
    } else {
      identifications <- .merge.identifications(identifications, ...)
    }
  }
  identifications <- .remove.duplications(identifications)
  
  ## Merge back the protein mapping
  merge(pept.n.prot,identifications,by="peptide")
}

setMethod("initialize","IBSpectra",
    function(.Object,identifications=NULL,data.ions=NULL,data.mass=NULL,
             proteinGroupTemplate=NULL,fragment.precision=NULL,
             assayDataElements=list(),allow.missing.columns=FALSE,
             write.excluded.to=NULL,...) { 

  if (is.null(identifications))
    return(callNextMethod(.Object,...))

  reporterTagNames <- reporterTagNames(.Object)
  identifications <- .factor.to.chr(identifications)

  ## Check that obligatory columns are present
  identifications <- .check.columns(identifications, data.ions, data.mass, allow.missing.columns)

  SC <- .SPECTRUM.COLS[.SPECTRUM.COLS %in% colnames(identifications)]

  ## Substitute Isoleucins with Leucins (indistinguishable by Masspec)
  if (!.PEPTIDE.COLS['REALPEPTIDE'] %in% colnames(identifications)) 
     identifications <- .fix.il.peptide(identifications)

  ## Separate protein columns (focus on peptide-spectrum matches)
  PC <- unique(c(.SPECTRUM.COLS['PEPTIDE'],.PROTEIN.COLS,.PEPTIDE.COLS))
  protein.colnames <- colnames(identifications)[colnames(identifications) %in% c(PC)]
  pept.n.prot <- unique(identifications[,protein.colnames])
  identifications <- unique(identifications[,-which(colnames(identifications) %in% setdiff(PC,.SPECTRUM.COLS['PEPTIDE']))])
  ## Merge identifications
  if (max(table(identifications[,SC['SPECTRUM']])) > 1) {
    message("merging identifications")
    if (SC['DISSOCMETHOD'] %in% colnames(identifications) && 
        length(unique(identifications[,SC['DISSOCMETHOD'] ]))) {
      identifications <- dlply(identifications,'dissoc.method',.merge.identifications,...)
      # rbind separately to assure equal column names
      identifications <- do.call(rbind,identifications)
      identifications <- .merge.quant.identifications(identifications)
    } else {
      identifications <- .merge.identifications(identifications, ...)
    }
  }
  identifications <- .remove.duplications(identifications)
 
  # Create ProteinGroup
  proteinGroup <- ProteinGroup(merge(pept.n.prot,identifications,by="peptide"),template=proteinGroupTemplate)
  
  ## Get intensities and masses in assayDataElements
  if (is.null(data.ions))
    data.ions <- .get.quant(identifications,.PEAKS.COLS['IONSFIELD'],reporterTagNames)
  if (is.null(data.mass))
    data.mass <- .get.quant(identifications,.PEAKS.COLS['MASSFIELD'],reporterTagNames)

  if (!all(rownames(data.ions)==rownames(data.mass),na.rm=TRUE))
    stop(sum(rownames(data.ions)==rownames(data.mass))," rownames are not equal between ions and mass")
  if (any(is.na(rownames(data.ions))))
    stop(sum(is.na(rownames(data.ions)))," rownames in data.ions are NA")
  if (any(is.na(rownames(data.mass))))
    stop(sum(is.na(rownames(data.ions)))," rownames in data.ions are NA")

  assayDataElements <- .get.quant.elems(assayDataElements,data.ions,data.mass,
                                        identifications[,SC['SPECTRUM']],fragment.precision,
                                        reporterTagMasses(.Object))

  if (!.SPECTRUM.COLS['USEFORQUANT'] %in% colnames(identifications)) {
    # not perfect - better: set spectra of peptides shared between groups to FALSE
    #                            and spectra with no values
    identifications[,.SPECTRUM.COLS['USEFORQUANT']] <- TRUE
  }  

  fdata <- identifications[,colnames(identifications) %in% .SPECTRUM.COLS]

  rownames(fdata) <- fdata[,'spectrum']
  featureData <- new("AnnotatedDataFrame",data=fdata,
                     varMetadata=.get.varMetaData(fdata))
	   
  assayData=do.call(assayDataNew, assayDataElements, envir=parent.frame())
  
  ## create ibspectra object
  callNextMethod(.Object,
      assayData=assayData,
      featureData=featureData,
      proteinGroup=proteinGroup,
      reporterTagNames=reporterTagNames,...)
})



setGeneric("readIBSpectra", function(type,id.file,peaklist.file,...)
           standardGeneric("readIBSpectra"))

setMethod("readIBSpectra",
          signature(type="character",id.file="character",peaklist.file="missing"),
    function(type,id.file,identifications.format=NULL,
             sep="\t",decode.titles=FALSE,trim.titles=FALSE,...) {
      new(type,
          identifications=.read.idfile(id.file,sep=sep,
                                       identifications.format=identifications.format,
                                       decode.titles=decode.titles,trim.titles=trim.titles),...)
    }
)
setMethod("readIBSpectra",
          signature(type="character",id.file="data.frame",peaklist.file="missing"),
    function(type,id.file,...) new(type,identifications=id.file,...)
)

setMethod("readIBSpectra",
          signature(type="character",id.file="character",peaklist.file="character"),
    function(type,id.file,peaklist.file,sep="\t",
             mapping.file=NULL,mapping=c(quantification.spectrum = "hcd",identification.spectrum = "cid"),
             id.file.domap=NULL,identifications.format=NULL,decode.titles=FALSE,...) {
      
      id.data <- .read.identifications(id.file,sep=sep,
                                       mapping=mapping.file,mapping.names=mapping,
                                       identifications.quant=id.file.domap,
                                       identifications.format=identifications.format,decode.titles=decode.titles)

      readIBSpectra(type,id.data,peaklist.file,...)
})


##' readIBSpectra - read IBSpectra object from files
##'
##' <details>
##' @title 
##' @param type [character] IBSpectra type
##' @param id.file [character] file of format ibspectra.csv or
##' mzid. If format is ibspectra.csv, it may also contain quantitative
##' information, and peaklist.file is not required.
##' @param peaklist.file [character] file in MGF format containing
##' quantitative information (m/z and intensity pairs). If NULL,
##' quantitative information is expected to reside in id.file.
##' @param proteinGroupTemplate [ProteinGroup object] uses certain
##' protein grouping as template. Useful when comparing multiple
##' experiments.
##' @param mapping.file [character] CSV file which maps spectrum
##' titles from peaklist file to those of id file. Usefule when
##' quantitative and identification information reside in different
##' but corresponding spectra. E.g. HCD-CID dissociation
##' @param mapping [named character] column number or name for
##' 'peaklist' and 'id' spectra.
##' @param mapping.file.readopts 
##' @param peaklist.format 
##' @param identifications.format 
##' @return IBSpectra object of type 
##' @author Florian P Breitwieser
setMethod("readIBSpectra",
          signature(type="character",id.file="data.frame",peaklist.file="character"),
    function(type,id.file,peaklist.file,
             annotate.spectra.f=NULL,
             peaklist.format=NULL,scan.lines=0,
             fragment.precision=NULL,fragment.outlier.prob=NULL,...) {
      
      if (is.function(annotate.spectra.f)) {
        id.file <- annotate.spectra.f(id.file,peaklist.file)
      }

      .Object <- new(type)
      quant <- .read.peaklist(peaklist.file,peaklist.format,
                              .Object@reporterTagMasses,.Object@reporterTagNames,
                              scan.lines,fragment.precision,fragment.outlier.prob,
                              id.data=id.file) 
     
      new(type,identifications=id.file,quant[[2]],data.ions=quant[[1]],...)
    }
)

# returns a list with intensities and mass matrices
.read.peaklist <- function(peaklist.file,peaklist.format,
                           reporterMasses,reporterTagNames,
                           scan.lines,fragment.precision,fragment.outlier.prob,
                           id.data=NULL) {
  data.ions=c()
  data.mass=c()
  data.titles=c()
  for (peaklist.f in peaklist.file) {
    peaklist.format.f <- peaklist.format
    if (is.null(peaklist.format.f)) {
      if (grepl(".mgf$",peaklist.f,ignore.case=TRUE))peaklist.format.f <- "mgf"
      else if (grepl(".mcn$",peaklist.f,ignore.case=TRUE))peaklist.format.f <- "mcn"
      else if (grepl(".intensities.csv$",peaklist.f,ignore.case=TRUE))peaklist.format.f <- "csv"
      else
        stop(paste0("cannot parse file ",peaklist.f," - cannot deduce format (mgf or mcn)"))          
    }

    if (tolower(peaklist.format.f) == "mgf") {
      intensities.f <- .read.mgf(peaklist.f,reporterMasses,reporterTagNames,
                                 fragment.precision=fragment.precision,
                                 prob=fragment.outlier.prob,scan.lines=scan.lines)
      if (nrow(intensities.f$ions) == 0) { stop("only NA data in ions/mass") }
      data.titles <- c(data.titles,intensities.f$spectrumtitles)
      data.ions <- rbind(data.ions,intensities.f$ions)
      data.mass <- rbind(data.mass,intensities.f$mass)
    } else if (tolower(peaklist.format.f) == "mcn") {
        if (type != "iTRAQ4plexSpectra")
          stop("mcn format (iTracker) only supports iTRAQ 4plex spectra!")
        mcn.i <- read.table("itraqdta/i-Tracker.mcn",sep=",",skip=2,
                            colClasses=c("character","character",
              "numeric","numeric","numeric","numeric",rep("NULL",36)),
            col.names=c("spectrum","spectrum.t","114","115","116","117",rep("",36)),
            check.names=FALSE)
        data.titles <- c(data.titles,mcn.i$spectrum)
        data.ions <- c(data.ions,as.matrix(mcn.i[,3:6]))
        data.mass <- c(data.mass,
            matrix(rep(114:117,nrow(mcn.i)),nrow=nrow(mcn.i),byrow=TRUE))
    } else if (tolower(peaklist.format.f) == "csv") {
      message("  reading peaklist ",peaklist.f," ...",appendLF=FALSE)
      res <- read.delim(peaklist.f,stringsAsFactors=FALSE,quote="")
      message(" done")
      if (!"spectrum" %in% colnames(res)) stop("'spectrum' column is missing from peaklist CSV")
      if (sum(.grep_columns(res,"ions$")) < length(reporterTagNames)) 
        stop("Not all neccessary intensity columns [",paste0(reporterTagNames,"_ions"),"] in peaklist CSV")
      if (sum(.grep_columns(res,"mass$")) < length(reporterTagNames)) 
        stop("Not all neccessary reporter mass columns [",paste0(reporterTagNames,"_mass"),"] in peaklist CSV")
      
      data.titles <- c(data.titles,res[,"spectrum"])
      data.ions <- rbind(data.ions,as.matrix(res[,.grep_columns(res,"ions$")]))
      data.mass <- rbind(data.mass,as.matrix(res[,.grep_columns(res,"mass$")]))
    }
  }

  if (!is.null(id.data) && .SPECTRUM.COLS['SPECTRUM.QUANT'] %in% colnames(id.data)) {
    data.titles.orig <- data.titles
    data.titles <- .do.map(data.titles,unique(id.data[,.SPECTRUM.COLS[c('SPECTRUM','SPECTRUM.QUANT')]]))
    sel.na <- is.na(data.titles)
    if (any(is.na(data.titles))) {
      message(" for ",sum(sel.na)," of ",length(data.titles)," spectra,",
              " quantitative information is available,\n",
              "   but no peptide-spectrum match. Spectrum titles: \n\t",
              paste(data.titles.orig[sel.na][1:2],collapse=",\n\t"),", ...")
      data.ions <- data.ions[!sel.na,]
      data.mass <- data.mass[!sel.na,]
      data.titles <- data.titles[!sel.na]
    }
  }

  rownames(data.ions)  <- data.titles
  rownames(data.mass)  <- data.titles
  ## TODO: check that all identified spectra are present in intensities

  colnames(data.ions) <- reporterTagNames
  colnames(data.mass) <- reporterTagNames

  if (any(is.na(rownames(data.ions))))
    stop(sum(is.na(rownames(data.ions)))," rownames in data.ions are NA")
  if (any(is.na(rownames(data.mass))))
    stop(sum(is.na(rownames(data.ions)))," rownames in data.ions are NA")
 
  return(list(data.ions,data.mass))
}


.do.map <- function(spectrumtitles,mapping.quant2id) {
  #mapped.spectra.pl <-
  #  mapping.quant2id[,2][ mapping.quant2id[,1] %in% id.spectra]

  #write(mapped.spectra.pl,file='id_spectra.csv')
  #.stopiflengthnotequal(id.spectra,mapped.spectra.pl,
  #                      "not all identified spectra could be matched!\n",
  #                      "\tx ... identified spectra\n",
  #                      "\ty ... mapped spectra\n")
 
  spectra.map <- .as.vect(mapping.quant2id,1,2)
  .stopiflengthnotequal(spectrumtitles,
                        spectra.map[spectrumtitles],
                        "not all spectra could be matched!\n",
                        "\tx ... spectra with quant info\n",
                        "\ty ... identified spectra\n")
  spectra.map[spectrumtitles]
}

.get.dupl.n.warn <- function(df,col,msg="ibspectra",write.to=NULL,f=warning) {
  dupl <- .all.duplicate.rows(df,col)
  if (!is.null(write.to))
    write.table(dupl,file=write.to,row.names=FALSE,sep="\t")
  dupl.msg <- paste(apply(dupl,1,paste,collapse="; "),collapse="\n\t")
  f(sprintf("%s> divergent identifications in %s spectra [%s ids]:\n\t%s",
			  msg,length(unique(dupl[,col])),nrow(dupl),dupl.msg))
  return(unique(dupl[,col]))
}

### READ MzID
read.mzid <- function(filename) {
  requireNamespace("XML")
  doc <- XML::xmlInternalTreeParse(filename)
  ns <- c(x=XML::xmlNamespace(XML::xmlRoot(doc))[[1]])

  root <- ifelse (isTRUE(ns == "http://psidev.info/psi/pi/mzIdentML/1.0"), "/x:mzIdentML", "/x:MzIdentML")
  peptidesequence.name <- ifelse (isTRUE(ns == "http://psidev.info/psi/pi/mzIdentML/1.0"), "peptideSequence", "PeptideSequence")
  dbsequenceref.attrname <- ifelse (isTRUE(ns == "http://psidev.info/psi/pi/mzIdentML/1.0"), "DBSequence_ref", "dBSequence_ref")
  peptideevref.attrname <- ifelse (isTRUE(ns == "http://psidev.info/psi/pi/mzIdentML/1.0"), "PeptideEvidence_ref", "peptideEvidence_ref")
  peptideref.attrname <- ifelse (isTRUE(ns == "http://psidev.info/psi/pi/mzIdentML/1.0"), "Peptide_ref", "peptide_ref")

  searchdatabase.mapping <- data.frame(
    ref=XML::xpathSApply(doc,paste0(root,"/x:DataCollection/x:Inputs/x:SearchDatabase"),
      XML::xmlGetAttr,name="id",namespaces=ns),
    name=XML::xpathSApply(doc,paste0(root,"/x:DataCollection/x:Inputs/x:SearchDatabase"),
      XML::xmlGetAttr,name="name",namespaces=ns),stringsAsFactors=FALSE
  )
   
  modification.mapping.df <- t(do.call(cbind,
                                  XML::xpathApply(doc,paste0(root,"/x:AnalysisProtocolCollection/x:SpectrumIdentificationProtocol/x:ModificationParams/x:SearchModification"),
                                             function(modif) {

    res <- XML::xmlAttrs(modif)
    modparams <- XML::getNodeSet(modif,"x:ModParam",namespaces=ns)
    return (switch(as.character(length(modparams)),
            "0" = c(res,XML::xmlAttrs(modif[['cvParam']])),
            "1" = c(res,XML::xmlAttrs(modparams[[1]]),XML::xmlAttrs(modparams[[1]][['cvParam']])),
            stop("Expecting zero or one ModParam Node in ",
                 "/mzIdentML/AnalysisProtocolCollection/SpectrumIdentificationProtocol/ModificationParams/SearchModification")))
  },namespaces=ns)))

  unknown.modif <- modification.mapping.df[,'name']=='unknown modification' | 
                     modification.mapping.df[,'accession']=='MS:1001460'
  if (any(unknown.modif)) {
    if (! 'value' %in% colnames(modification.mapping.df)) {
      stop('unknown modifications [as defined by nameor accession], and no column value to override.')
    }
    modification.mapping.df[unknown.modif,'name'] <- modification.mapping.df[unknown.modif,'value']
  }

  modif.map <- .as.vect(unique(modification.mapping.df[,c('massDelta','name')]))

  message("peptide and protein mapping")
  peptide.mapping <- do.call(rbind,XML::xpathApply(doc,paste0(root,"/x:SequenceCollection/x:Peptide"),namespaces=ns,
         function(pep) {
           peptide.ref <- XML::xmlGetAttr(pep,name='id')
           peptide.seq <- XML::xmlValue(pep[[peptidesequence.name]])

           loc.n.delta <- XML::xpathSApply(pep,"x:Modification",function(m) c(location=XML::xmlGetAttr(m,name="location"),
                                                                         massDelta=XML::xmlGetAttr(m,name="monoisotopicMassDelta")),
                                      namespaces=ns)

          
           modifstring <- character(nchar(peptide.seq)+2)

           if (length(loc.n.delta) > 0) {
             modif.names <- modif.map[loc.n.delta[2,]]
             if (any(is.na(modif.names)))
               modif.names <- XML::xpathSApply(pep,"x:Modification/x:cvParam[@cvRef='UNIMOD']",XML::xmlGetAttr,name='name',namespaces=ns)

             if (length(modif.names) < ncol(loc.n.delta)) 
               stop("modif names and delta have different size")
 
             modifstring[as.numeric(loc.n.delta[1,])+1] <- modif.names
           }
           
           c(peptide.ref=peptide.ref,peptide=peptide.seq,modif=paste(modifstring,collapse=":"))
         }))

  protein.mapping <- do.call(rbind,XML::xpathApply(doc,paste0(root,"/x:SequenceCollection/x:DBSequence"),function(dbs) {
    c(dbseq.ref = XML::xmlGetAttr(dbs,name='id'),
      accession = XML::xmlGetAttr(dbs,name='accession'),
      length = XML::xmlGetAttr(dbs,name='length'),
      sdb.ref = XML::xmlGetAttr(dbs,name='SearchDatabase_ref'))
      #sequence = XML::xmlValue(dbs[['seq']]))
  },namespaces=ns))
  records.proteinDetections <- do.call(rbind,XML::xpathApply(doc,
    paste0(root,"/x:DataCollection/x:AnalysisData/x:ProteinDetectionList/x:ProteinAmbiguityGroup"),
    namespaces=ns,
    fun=function(pag) {
    ## apply on ProteinAmbiguityGroup
    do.call(rbind,XML::xmlApply(pag,function(pdh) {
      ## apply on ProteinDetectionHypothesis
      cbind(dbseq.ref=XML::xmlGetAttr(pdh,dbsequenceref.attrname),
            peptide.ev.ref=XML::xpathSApply(pdh,"x:PeptideHypothesis",XML::xmlGetAttr,name=peptideevref.attrname,namespaces=ns)
    )}))
  }))

  # map PeptideEvidence attribute names to isobar columns
  pe.attr.names <- c(id="peptide.ev.ref",start="start.pos",end="end.pos",pre="aa.before",post="aa.after",
                     missedCleavages="nmc",isDecoy="is.decoy",DBSequence_Ref="dbseq.ref",dBSequence_Ref="dbseq.ref")
  
  message("spectrum mapping")
  spectrumIdentifications <-
    XML::xpathApply(doc,paste0(root,"/x:DataCollection/x:AnalysisData/x:SpectrumIdentificationList/x:SpectrumIdentificationResult"),
              namespaces=ns,
              fun=function(sir) {
    ## _SpectrumIdentificationResult_ #
    ## All identification from searching one spectrum
    spectrum.id <- XML::xmlGetAttr(sir,"spectrumID")
    spectrum.title.nodes <- XML::getNodeSet(sir,"x:cvParam[@name='spectrum title']",namespaces=ns)
    if (length(spectrum.title.nodes) == 1)
      spectrum.id <- XML::xmlGetAttr(spectrum.title.nodes[[1]],name='value')

    ## get spectrum identifications which pass threshold
      ## _SpectrumIdentificationItem_ #
      ## An identification of a single peptide of a specturm.
      ## Only take the one which passes the threshold.
    sii <- XML::getNodeSet(sir,"x:SpectrumIdentificationItem[@passThreshold='true' and @rank='1']",namespaces=ns)
    if (length(sii) == 0 || length(sii) > 1) return (NULL)
    #"more than one match for [ x:SpectrumIdentificationItem[@passThreshold='true' and @rank='1'] ]")
    # TODO: There will be always two times the same score with peptides with just an I/L difference
    
    sii <- sii[[1]]
 
     scores <- unlist(XML::xpathApply(sii,"x:cvParam",function(cvp) {
                           switch(XML::xmlGetAttr(cvp,name='name'),
                                  "Scaffold: Peptide Probability"=c(scaffold.pepprob=XML::xmlGetAttr(cvp,name="value")),
                                  "mascot:score"=c(score.mascot=XML::xmlGetAttr(cvp,name="value")),
                                  "mascot:expectation value"=c(mascot.evalue=XML::xmlGetAttr(cvp,name="value")),
                                  "sequest:xcorr"=c(sequest.xcorr=XML::xmlGetAttr(cvp,name="value")),
                                  "sequest:deltacn"=c(sequest.deltacn=XML::xmlGetAttr(cvp,name="value")),
                                  NULL
                            )
                 },namespaces=ns))

      if ("PeptideEvidence" %in% names(sii)) {
        pe.attr <- XML::xmlAttrs(sii[["PeptideEvidence"]])
        pe.res <- pe.attr.names[names(pe.attr)[names(pe.attr) %in% names(pe.attr.names)]]
      } else {
        pe.res <- c(peptide.ev.ref=XML::xmlGetAttr(sii[["PeptideEvidenceRef"]],name="peptideEvidence_ref"))
      }

      c(spectrum=spectrum.id,
        peptide.ref   = XML::xmlGetAttr(sii,name=peptideref.attrname),
        theo.mass     = XML::xmlGetAttr(sii,name='calculatedMassToCharge'),
        exp.mass      = XML::xmlGetAttr(sii,name='experimentalMassToCharge'),
        scores,pe.res)
  })

  si.names <- unique(unlist(sapply(spectrumIdentifications,names)))
  records.spectrumIdentifications <- do.call(rbind,lapply(spectrumIdentifications,function(x) x[si.names]))
  records.spectrumIdentifications <- as.data.frame(records.spectrumIdentifications,stringsAsFactors=FALSE)

  peptide.ev.mapping <- XML::xpathApply(doc,paste0(root,"/x:SequenceCollection/x:PeptideEvidence"),namespaces=ns,XML::xmlAttrs)
  if (length(peptide.ev.mapping) > 0) {
    records.spectrumIdentifications <- merge(records.spectrumIdentifications,
                                             as.data.frame(do.call(rbind,peptide.ev.mapping),stringsAsFactors=FALSE))
  }

  spectra.n.peptides <- merge(as.data.frame(records.spectrumIdentifications,stringsAsFactors=FALSE),
                              as.data.frame(peptide.mapping,stringsAsFactors=FALSE),
                              by='peptide.ref')

  XML::free(doc)

  message("merging results")
  protein.mapping1 <- merge(as.data.frame(records.proteinDetections,stringsAsFactors=FALSE),
                            as.data.frame(protein.mapping,stringsAsFactors=FALSE),by="dbseq.ref")
  merge(spectra.n.peptides,
        protein.mapping1,by="peptide.ev.ref")
}

##' .read.mgf: read isobaric reporter tag masses and intensities from MGFs
##'
##' MGF files list m/z and intensities for each spectrum. .read.mgf
##' extracts m/z and intensity pairs of masses corresponding to isobaric
##' tag masses (within a fragment precision).
##' @title 
##' @param filename [character] file name of one or multiple files in MGF format
##' @param type [character] denoting IBSpectra class - used to get
##'             reporter masses and reporter names.
##' @param spectra [character vector] if defined, only export spectra whose TITLE
##'                is listed. Speeds up the function when only identified spectra
##'                are of interest.
##' @param fragment.precision [numeric] take m/z-intensity pairs whose m/z value are
##'                          in a range of +/- fragment.precision/2 of 'true' mass.
##' @param prob [numeric] Filter out m/z-intensitiy values with the prob/2 most
##'             unprecise m/z values on both sides.
##' @param substitute.dta [boolean] internal. replace TITLEs: s/.dta.[0-9]*$/.dta/
##' @return list(ions, mass, spectrumtitles)
##' @author Florian P Breitwieser
.read.mgf <- function(filename,reporterMasses,reporterNames,spectra=NULL,fragment.precision=0.05,
                      prob=NULL,substitute.dta=FALSE,check.id.ok=FALSE,
                      scan.lines=0) {
  if (is.null(fragment.precision)) { fragment.precision=0.05 }
  message("  reading mgf file ",filename,
          " [fragment precision: ",fragment.precision,"]")

  ## get reporter masses and names from type
  nReporter <- length(reporterMasses)
  min.mass <- min(reporterMasses)-fragment.precision/2
  max.mass <- max(reporterMasses)+fragment.precision/2
  
  ## read (and concatenate if multiple) mgf files
  input <- c()
  for (f in filename) {
    con <- file(f,'r')
    if (scan.lines > 0) {
      while(length(f.input <- readLines(con, n=scan.lines)) > 0){
        input <- c(input,grep("^[A-Z]|^1[12][0-9]\\.",f.input,value=T))
      }
    } else {
      input <- c(input,readLines(con))
    }
    close(con)
  }
  if (substitute.dta)
    input <- sub(".dta.*$",".dta",input)
  
  ## index mgf file content: positions of BEGIN and END IONS
  begin_ions <- which(input=="BEGIN IONS")
  end_ions <- which(input=="END IONS")
  if (length(begin_ions) != length(end_ions))
    stop("mgf file is errorneous, non-matching number",
         " of BEGIN IONS and END IONS tags");

  bnd <- data.frame(begin=begin_ions,
                    end=end_ions)
  
  if (!is.null(spectra)) {
    ## filtering to take only spectra defined in spectra
    ## all.titles <- .trim(sapply(strsplit(grep("TITLE",input,value=T),"="),function(x) x[2] )) # not efficient
    all.titles <- .trim(substring(grep("^TITLE=",input,value=TRUE),7))
    if (length(all.titles) != nrow(bnd)) {
      # sanity check that each spectrum has a title
      stop("title not specified for all spectra!");
    }

    spectra.to.take <- all.titles %in% spectra

    if (sum(spectra.to.take) ==0) {
      stop("No identified spectrum is found in MGF file.\n",
           "  TITLEs in MGF [1:",length(all.titles),"]: \n\t",
           paste(all.titles[1:2],collapse=", "),", ...\n",
           "  identified spectra [1:",length(spectra),"]: \n\t",
           paste(spectra[1:2],collapse=", "),", ...\n")
    }
    bnd <- bnd[spectra.to.take,]
  }
  bnd$recordNo = seq_len(nrow(bnd))
  nSpectra <- nrow(bnd)
  message("  ",nSpectra," spectra in MGF file.")

  ## create list with all spectra (header+mass list) as entries
  all.spectra <- apply(bnd,1,function(x) input[x[1]:x[2]])
  rm(input)

  ## if all spectra are of equal length, apply returns a matrix
  ##  convert to list
  if (is.matrix(all.spectra)) 
    all.spectra <- split(all.spectra,col(all.spectra))

  ## extract information from each spectrum
  result <- llply(all.spectra,function(x) {
    header <- .strsplit_vector(x[grep("^[A-Z]",x)],"=")
    numbers <- do.call(rbind,strsplit(x[grep("^1..\\.",x)],"\\s"))
    mzi.mass <- as.numeric(numbers[,1])
    
    rr <- c(title=header["TITLE"])
    
    sel <- mzi.mass > min.mass & mzi.mass < max.mass
    if (any(sel)) {
      mzi.mass <- mzi.mass[sel]
      mzi.ions <- as.numeric(numbers[sel,2])

      rr <- c(rr,do.call(c,lapply(reporterMasses,function(y) {
        m <- abs(y-mzi.mass)
        pos <- which(m == min(m))
        if (length(pos) > 1){
          # (Jacques) added to address cases where 2 peaks (within the required precision) are at the same distance of the reporter theoretical mass
          # Pick most intense
          max.intens <- max(mzi.ions[pos])
          pos <- pos[which(mzi.ions[pos]==max.intens)]
          pos <- pos[1] # in case same intensity as well!
        }
        if (length(pos) > 0 & m[pos] < fragment.precision/2)
          return(c(mzi.mass[pos],mzi.ions[pos]))
        else
          return(c(NA,NA))
      }))) 
    } else {
      rr <- c(rr,rep(NA,nReporter*2))
    }
    return(rr)
  },.parallel=isTRUE(getOption('isobar.parallel')))

  result <- do.call(rbind,result)

  rm(all.spectra)
  if (length(result) == 0 || nrow(result) == 0) {
    stop("error reading MGF file - could not parse masses and intensities.\n",
         "Check the mapping id <-> peaklist.")
  }
 
  mass <- apply(result[,seq(from=2,to=ncol(result),by=2)],2,as.numeric)
  ions <- apply(result[,seq(from=3,to=ncol(result),by=2)],2,as.numeric)
     
  ## only select spectra with itraq masses detected
  sel = apply(mass,1,function(x) any(!is.na(x)))
      
  if (!any(sel)) {
    stop("No values could be extracted from MGF ",filename,"\n",
         "with fragment precision ",fragment.precision,".\n")
  }

  if (!is.null(prob) && prob > 0) {
    ## boundaries: remove extreme outliers of mass spectrum
    ##  to test: another quantile.type might be more appropriate for small datasets
    bnd <- apply(mass[sel,],2,quantile,probs=c(prob*0.5,1-prob*0.5),na.rm=TRUE)
    sel.prob <- !is.na(mass) & (mass < matrix(bnd[1,],byrow=T,nrow=nrow(mass),ncol=ncol(mass)) |
                                mass > matrix(bnd[2,],byrow=T,nrow=nrow(mass),ncol=ncol(mass)))

    ions[sel.prob] <- NA
    mass[sel.prob] <- NA
    message("\tmass boundaries:\n\t",
            paste(colnames(mass),sprintf("%.5f : %.5f",bnd[1,],bnd[2,]),sep="\t",collapse="\n\t"))
  }

  ions <- ions[sel,,drop=FALSE]
  mass <- mass[sel,,drop=FALSE]
 
  spectrumtitles <- .trim(result[sel,1])

  if (ncol(ions) != length(reporterNames) || ncol(mass) != length(reporterNames)) {
    stop("ions or mass matrix have wrong dimension.")
  }

  dimnames(ions) <- list(spectrumtitles,reporterNames)
  dimnames(mass) <- list(spectrumtitles,reporterNames)
  rm(result)
  
  return(list(ions=ions, mass=mass,
              spectrumtitles=spectrumtitles))
}

.read.idfile.df <- function(filename,identifications.format,sep,...) {
    if (is.data.frame(filename))
      return(filename)
    if (!is.character(filename))
      stop("filename should be a data.frame or character")
    if (!file.exists(filename))
      stop("idfile ",filename," defined, but does not exist")
    
    is.format <- function(y,ext)
      identical(identifications.format,y) || 
        any(sapply(ext,function(iext) grepl(paste0(iext,"$"),filename)))
    
    ext.def <- 
    list(mzid      = list(ext=c("mzid"),f=read.mzid),
         rockerbox = list(ext=c("peptides.[ct]sv"),f=.read.rockerbox),
         msgf      = list(ext=c("msgfp.csv","tsv"),f=.read.msgfp.tsv),
         ibspectra = list(ext=c("csv"),
                          f=function(x) read.table(x,header=TRUE,sep=sep,
                                                   stringsAsFactors=FALSE,quote="",...)))

    for (etype in names(ext.def)) {
      if (is.format(etype,ext.def[[etype]][["ext"]])) {
        message("  reading id file ",filename," [type: ",etype,"] ...",appendLF=FALSE)
        res <- ext.def[[etype]][["f"]](filename)
        message(" done")
        return(res)
      }
    }

    stop(paste("cannot parse file ",filename," - cannot deduce format based on extension ",
               "(it is not ibspectra.csv, id.csv, peptides.txt or mzid). ",
               "Please provide id.format to readIBSpectra",sep=""))
}


## TODO: log is not returned
.read.idfile <- function(id.file,identifications.format=NULL,sep="\t",
                         decode.titles=FALSE,trim.titles=FALSE,log=NULL,all.cols=FALSE,...) {
  if (!is.data.frame(id.file)) {
    if (!(is.list(id.file) || is.character(id.file))) 
      stop("id.file argument of .read.idfile should be a data frame, character or list")

    if (all(sapply(id.file,is.character)))
      id.data <- lapply(id.file,.read.idfile.df,sep=sep,
                        identifications.format=identifications.format,...)
    else if (all(sapply(id.file,is.data.frame)))
      id.data <- id.file
    else
      stop()

    rm(id.file)
  
    id.colnames <- lapply(seq_along(id.data),function(s.i) colnames(id.data[[s.i]]))
    colnames.equal <- all(sapply(id.colnames,
                                 function(cn) identical(cn,id.colnames[[1]])))
    if (!colnames.equal) {
      message(" id file colnames are not equal:")
      message(paste(sapply(seq_along(id.data),
                           function(s.i) paste0("    ",s.i,": [",
                                                paste(id.colnames[[s.i]],collapse=","),
                                                "]")),collapse="\n"))
      all.id.colnames <- unique(unlist(id.colnames))
      if (isTRUE(all.cols)) {
        id.data <- do.call(rbind,lapply(id.data,function(i.d) {
          id.d[,all.id.colnames[!all.id.colnames %in% colnames(id.d)]] <- NA
          id.d[,all.id.colnames]
        }))
      } else {
        intersect.colnames <- id.colnames[[1]]
        for (s.i in seq(from=2,to=length(id.colnames))) {
          intersect.colnames <- intersect(intersect.colnames,id.colnames[[s.i]])
        }
        message(" taking intersection: ",paste(intersect.colnames,collapse="; "))
        id.data <- do.call(rbind,lapply(id.data,function(i.d) i.d[,intersect.colnames]))
      }
    } else {
        id.data <- do.call(rbind,id.data)
    }
  }

  #.check.columns(id.data)

  if (!is.character(id.data[,.SPECTRUM.COLS['SPECTRUM']]))
    id.data[,.SPECTRUM.COLS['SPECTRUM']] <- as.character(id.data[,.SPECTRUM.COLS['SPECTRUM']])

  if ('accession' %in% colnames(id.data)) {
    if (all(grepl(sprintf("%s\\|%s",.uniprot.pattern.ac,.uniprot.pattern.id),id.data[,.PROTEIN.COLS['PROTEINAC']]))) {
      split.ac <- strsplit(id.data[,.PROTEIN.COLS['PROTEINAC']],"|",fixed=TRUE)
      id.data[,.PROTEIN.COLS['PROTEINAC']] <- sapply(split.ac,function(x) x[1]) 
      id.data[,.PROTEIN.COLS['NAME']] <- sapply(split.ac,function(x) ifelse(length(x) == 2, x[2], NA))
    }
  }

  if (decode.titles)
    id.data[,.SPECTRUM.COLS['SPECTRUM']] <- sapply(id.data[,.SPECTRUM.COLS['SPECTRUM']],URLdecode)

  if (trim.titles)
    id.data[,.SPECTRUM.COLS['SPECTRUM']] <- .trim(id.data[,.SPECTRUM.COLS['SPECTRUM']])
  return(id.data)
}

.read.rockerbox <- function(filename) {
  data.r <- read.table(filename,header=T,stringsAsFactors=F,sep="\t")
  if (!"scan.title" %in% colnames(data.r))
    stop("no scan.title column in ",filename,"; please use a Rockerbox version >= 2.0.6")

  ## transform modification (TODO)
  data.r$modif <- data.r$modifications
  ## end transform
  
  ## transform 'all.peptide.matches' to ac and start.pos
  split.acs <- strsplit(data.r$all.protein.matches,"; ")
  names(split.acs) <- data.r$all.protein.matches
  ac.n.startpos <- ldply(split.acs,function(x) {
    y <- do.call(rbind,strsplit(x,split="\\[|\\]"))
    start.pos <- sapply(strsplit(y[,2],"-",fixed=TRUE),
                        function(z) if(all(!is.na(as.numeric(z))) && length(z) == 2) as.numeric(z[1])
                        else stop("not numeric"))
    data.frame(accession=y[,1],start.pos=start.pos)  
  })
  data.r <- merge(data.r,ac.n.startpos,by.x="all.protein.matches",by.y=".id")
  data.r$all.protein.matches <- NULL
  ## end transform

  sel <- names(.ROCKERBOX.MAPPING.COLS) %in% names(c(.SPECTRUM.COLS,.PEPTIDE.COLS))
  data.r <- data.r[,.ROCKERBOX.MAPPING.COLS[sel]]
  colnames(data.r) <- c(.SPECTRUM.COLS,.PEPTIDE.COLS)[names(.ROCKERBOX.MAPPING.COLS)[sel]]
  return(data.r)
}

###############################################################################
## MERGE IDENTIFCATIONS


.dissect.search.engines <- function(identifications) {

  SC <- .SPECTRUM.COLS[.SPECTRUM.COLS %in% colnames(identifications)]

  ## scores are merged together
  if (.SPECTRUM.COLS['SCORE'] %in% colnames(identifications)) {
    engines <- strsplit(identifications[,SC['SEARCHENGINE']],"|",fixed=TRUE)
    scores <- strsplit(identifications[,SC['SCORE']],"|",fixed=TRUE)
  } else {
    engine.n.score <- strsplit(identifications[,SC['SEARCHENGINE']],"[ \\|]")
    engines <- lapply(engine.n.score,function(x) x[seq(from=1,to=length(x),by=2)])
    if (length(unique(unlist(engines))>10)) {
      engines <- lapply(engine.n.score,function(x) x[seq(from=1,to=length(x)/2)])
      scores <- lapply(engine.n.score,function(x) as.numeric(x[seq(from=length(x)/2+1,to=length(x))]))
    } else {
      scores <- lapply(engine.n.score,function(x) as.numeric(x[seq(from=2,to=length(x),by=2)]))
    }
  }
  score.columns <- paste0('score.',tolower(unique(unlist(engines))))
  for (engine in unique(unlist(engines))) {
    name <- paste0('score.',tolower(engine))
    e.scores <- mapply(function(e,s) if(any(e==engine)) s[e==engine] else NA,engines,scores)
    identifications[,name] <- as.numeric(e.scores)
  }

  identifications[,SC['SEARCHENGINE']] <- sapply(engines,paste,collapse="&")
  tt <- table(identifications[,SC['SPECTRUM']])
  if (max(tt) > 1) {
    message("  resolving duplicated ids ...")
    id.good <- identifications[identifications$spectrum %in% names(tt)[tt==1],]
    id.bad <- identifications[identifications$spectrum %in% names(tt)[tt>1],]
    id.bad$n.se <- rowSums(!is.na(id.bad[,score.columns]))
    id.bad <- ddply(id.bad,'spectrum',function(x) {
      x[which.max(x$n.se),]
    })
    id.bad$n.se <- NULL
    identifications <- rbind(id.good,id.bad)
  }

  if ('SCORE' %in% names(SC))
    identifications[,SC['SCORE']] <- NULL

  return(identifications)
}

.merge.search.engine.identifications <- function(identifications,...) {

  ## Columns to consolidate after merging
  ##   functions are used to resolve or remove the psm state
  COLS.TO.CONSOLIDATE <- list('peptide'=.consolidate.peptide.ids,
			      'modif'=.consolidate.modification.pos,
			      'charge'=.consolidate.charge,  ## sometimes, charge state 0 is reported when too high
			      'theo.mass'=.mean.na.rm,        ## can differ esp. between Mascot and Phenyx
			      'retention.time'=.mean.na.rm,
			      'parent.intens'=.mean.na.rm)

  COLS.TO.CONSOLIDATE <- COLS.TO.CONSOLIDATE[names(COLS.TO.CONSOLIDATE) %in% colnames(identifications)]

  ## Columns used for merging. Typically 'spectrum' and columns which are the same regardless of search engine
  COLS.FOR.MERGING <- setdiff(.SPECTRUM.COLS,c(.ID.COLS,names(COLS.TO.CONSOLIDATE)))
  COLS.FOR.MERGING <- c(COLS.FOR.MERGING,'is.decoy')
  COLS.FOR.MERGING <- intersect(colnames(identifications),COLS.FOR.MERGING)

  #'delta.score','delta.score.pep','delta.score.notpep','n.pep','n.loc'

  ## Split identifications by search engine
  cn.search.engine <- which(colnames(identifications) == .SPECTRUM.COLS['SEARCHENGINE'])
  ids.split <- lapply(split(identifications,identifications[,cn.search.engine]),
		      function(x) x[,-cn.search.engine])
  if (length(ids.split) < 2) stop("Expecting more than two different search engines.")
  if (length(ids.split) > 10) stop("Split data.frame into more than 10 search.engines - seems unrealistic. ",
				   "search.engine values: ",paste(names(ids.split),collapse=", "))

  clean.names <- gsub("[^[:alnum:]]","",tolower(names(ids.split))) ## lower case alpha-numeric names
  score.cols <- paste0("score.",clean.names)

  for (ii in seq_along(ids.split)) { # fix colnames which are duplicate
    cn <- colnames(ids.split[[ii]])
    cn[!cn %in% COLS.FOR.MERGING] <- paste(cn[!cn %in% COLS.FOR.MERGING],clean.names[ii],sep=".")
    colnames(ids.split[[ii]]) <- cn
  }

  ## Merge identification data frames
  ids.merged <- merge(ids.split[[1]],ids.split[[2]],by=COLS.FOR.MERGING,all=TRUE)
  if (length(ids.split) > 2) {
    for (ii in seq(from=3,to=length(ids.split))) {
      ids.merged <- merge(ids.merged,ids.split[[ii]],by=COLS.FOR.MERGING,all=TRUE)
    }
  }

  if (!.SPECTRUM.COLS['NOTES'] %in% colnames(ids.merged))
    ids.merged[,.SPECTRUM.COLS['NOTES']] <- ''

  ## Consolidate spectra with different identifications with different search engines
  for (colname in names(COLS.TO.CONSOLIDATE)) {
    message("Consolidating ",colname, " [",date(),"]")
    ids.merged <- .resolve.conflicts(ids=ids.merged,resolve.f=COLS.TO.CONSOLIDATE[[colname]],
  	         		     colname=colname,resolve.colnames=paste0(colname,".",clean.names))
  }

  ids.merged <- unique(ids.merged)
  tt <- table(ids.merged[,.SPECTRUM.COLS['SPECTRUM']])
  #spectra.ok <- ids.merged[,'spectrum'] %in% names(tt)[tt==1]
  #resolved.ids <- .resolve.differing.identifications(ids.merged[!spectra.ok,],score.cols)

  #identifications <- rbind(ids.merged[spectra.ok,],resolved.ids)
  if (any(tt>1)) {
    warning("Merging not successful, ",sum(tt>1)," duplicated psms, removing them!")
    ids.merged <- ids.merged[!ids.merged[,.SPECTRUM.COLS['SPECTRUM']] %in% names(tt)[tt>1],]
  }

  ids.merged[,.SPECTRUM.COLS['SEARCHENGINE']] <- 
	  apply(ids.merged[,score.cols],1,function(x) { paste(names(ids.split)[!is.na(x)],collapse="&") })
  return(ids.merged)
}

.na.rm <- function(x) x[!is.na(x)]
.mean.na.rm <- function(x) mean(x,na.rm=TRUE)

# Remove differing peptide ids
.consolidate.peptide.ids <- function(x) NA

# Take a non-zero charge
.consolidate.charge <- function(x) {
	x <- !is.na(x)
	ifelse(any(x > 0),x[x>0][1],0)
}

# Take 'first' modification
.consolidate.modification.pos <- function(x) .na.rm(x)[1]

.searchengine.cols <- function(ids,colname) {

}

.resolve.conflicts <- function(ids, resolve.f, colname, resolve.colnames, keep.cols = FALSE) {

  if (is.character(resolve.colnames))
    resolve.colnames <- which(colnames(ids) %in% resolve.colnames)

  # set result column
  ids[,colname] <- .return.equal.or.na(ids[,resolve.colnames])

  # return resolved conflicting and non-conflicting data
  good.spectra <- !is.na(ids[,colname])
  if (!any(good.spectra))
    stop("No good spectra which don't have to be resolved - something went wrong")
  if (!all(good.spectra)) {
    print(isobar:::.sum.bool(good.spectra))
    ids <- rbind(ids[good.spectra,],
		 .do.resolve.conflicts(ids[!good.spectra,],colname,resolve.colnames, resolve.f))
  }

  if (keep.cols)
    ids
  else
    ids[-resolve.colnames,]
}

.do.resolve.conflicts <- function(ids, colname, resolve.colnames, resolve.f) {
  if (length(ids) == 0)
    return(NULL) 
  if (length(ids) == 1)
    return(ids)

  ids[,colname] <- apply(ids[,resolve.colnames],1,resolve.f)
  if (any(is.na(ids[,colname])))
    message("removing ",sum(is.na(ids[,colname]))) 
  ids <- ids[!is.na(ids[,colname]),]
  if (nrow(ids) == 0)
	  return(NULL)

  ids[,.SPECTRUM.COLS['NOTES']] <- ifelse(ids[,.SPECTRUM.COLS['NOTES']] == '','',
					  paste0(ids[,.SPECTRUM.COLS['NOTES']],"\n"))

  ids[,.SPECTRUM.COLS['NOTES']] <- paste0(ids[,.SPECTRUM.COLS['NOTES']],
					  apply(ids[,resolve.colnames],1,function(x) {
					    x <- x[!is.na(x)]
					    paste(colname,"differing:\n\t",
                                                   paste(names(x),x,sep=": ",collapse="\n\t"))}))
  return(ids)
}

.resolve.modifications <- function(df,colname, modif.cols, standard.modif) {
  message("Resolving modifications")
  if (is.character(modif.cols))
    modif.cols <- which(colnames(df) %in% modif.cols)

  adply(df, 1, function(x) {

    x <- as.data.frame(x)
    modifs <- unlist(x[,modif.cols])
    modifs <- modifs[!is.na(modifs)]
    if (length(modifs) == 0) {
	    print(x)
	    stop ("all modifs are NA!")
    }

    ## coherent modifications
    x$note <- ""
    if (length(modifs) == 1 || all(modifs == modifs[1])) {
      stop("Should be handled before")
      x[,colname] <- modifs[1]
      return(x[,-modif.cols])
    }

    ## resolve incoherent modifs
    note <- paste0("differing modification positions: \n\t",
		   paste(names(modifs),modifs,sep=": ",collapse="\n\t"))

    ## any modification position is more often present?
    tt <- table(modifs)
    if (.max.uniq(tt)) {
      note <- paste0(note,". Taking ", .take.max(tt) )
      stop("TODO: Implement")
    }

    ## take 'first' modification position
    x[,colname] <- modifs[1]
    x[,'note'] <- note
    return(x)
  })  
}

.max.uniq <- function(tt) sum(tt==max(tt)) == 1
.take.max <- function(tt) names(tt)[which.max(tt)]

.resolve.differing.identifications <- function(identifications,score.cols) {
  allequal <- function(x) all(x == x[1])
  by.y <- function(x,ind,fun=sum) sapply(by(x,ind,fun),function(x) return(x) )
  skipna <- function(x) unlist(x[!is.na(x)])
  n.skipped <- 0
  n.max.ids <- 0
  n.modif.pos.dif <- 0

  resolved.identifications <- ddply(identifications,'spectrum', function(x) {
	n.ids <- rowSums(!is.na(x[,score.cols]),na.rm=TRUE)    ## number of identifications for each psm
	if (.max.uniq(n.ids)) {
	  n.max.ids <<- n.max.ids + 1
	  return(x[which.max(n.ids),,drop=FALSE])
	}
        
	## resolve peptide differnces
        if (!allequal(x[,'peptide'])) {                ## different peptides identified
	  pep.ids <- by.y(n.ids,x[,'peptide'],sum)
	  if (.max.uniq(pep.ids)) {                     ##   any peptide has been seen more often than others?
	    x <- x[x[,'peptide'] == .take.max(pep.ids),,drop=FALSE]
	  }  else {
	    n.skipped <<- n.skipped + 1
	    return(NULL)
	  }
	}
	
	## all equal peptides, different modification positions
        if (!allequal(x[,'peptide'])) {
	  message("ERROR while merging: psm peptides should be equal, but they aren't.")
	  print(x)
	  stop()
	}

	## resolve modification differences
        if (!allequal(x[,'modif'])) {                ## different modifs identified
	  modif.ids <- by.y(n.ids,x[,'modif'],sum)

	  modifs <- gsub("Oxidation_M","")

	  if (!.max.uniq(modif.ids))
	    n.modif.pos.dif <<- n.modif.pos.dif + 1

	  #print(x)
	  modif <- paste(x[,.SPECTRUM.COLS['SEARCHENGINE']],x[,'modif'],sep=": ",collapse=' & ')
	  x <- x[x[,'modif'] == .take.max(modif.ids)[1],,drop=FALSE]
	  x[,'modif'] <- modif
	} else if (!allequal(x[,'charge'])) {         ## different charge
          x[1,score.cols] <- as.numeric(sapply(x[,score.cols],skipna))
	  x[1,'charge'] <- max(x[,'charge'])
	  x <- x[1,,drop=FALSE]
	} else { 
          message("Why is the modif equal? ")
	  print(rbind(x,is.equal=apply(x,2,allequal)))
	  stop()
	}
	
	return(x)
  })
  message(" resolving differing ",length(unique(identifications[,'spectrum']))," identifications: " )
  message("   ",n.max.ids," resolved because more evidence was available for one alternative")
  message("   ",n.skipped," removed because of differing peptide ids")
  message("   ",n.modif.pos.dif," kept, as only the modification position was different")
  return(resolved.identifications)
}

.merge.quant.identifications <- function(identifications) {
  SC <- .SPECTRUM.COLS[.SPECTRUM.COLS %in% colnames(identifications)]

  tt <- table(identifications[,SC['SPECTRUM.QUANT']])
  spectra.ok <- identifications[,SC['SPECTRUM.QUANT']] %in% names(tt)[tt==1]

  SC <- .SPECTRUM.COLS[.SPECTRUM.COLS %in% colnames(identifications)]
  score.colname <- .SPECTRUM.COLS[c('SCORE.MASCOT','SCORE.PHENYX','SCORE.MSGF','PROB.PHOSPHORS')]
  score.colname <- score.colname[score.colname %in% colnames(identifications)]

  message("Merging ",sum(!spectra.ok)," identifications from different dissociation methods.")
  ids.quant.merged <- ddply(identifications[!spectra.ok,],.SPECTRUM.COLS['SPECTRUM.QUANT'],function(x) {
      if (nrow(x) == 1) return(x)

      my.args <- as.list(x[,score.colname,drop=FALSE])
      my.args <- lapply(my.args,round,digits=2) ## take two significant digits before taking next score into account as top hitter
      my.args$decreasing=TRUE

      max.hit <- do.call(order,my.args)[1]

      # dismiss peptide-spectrum matches which are different between search engines
      if (!all(x[,SC['PEPTIDE']] == x[1,SC['PEPTIDE']]))
        return(NULL)

      x[max.hit,SC['DISSOCMETHOD']] = paste0("[",x[max.hit,SC['DISSOCMETHOD']],"]")
      if (all(x[,SC['MODIFSTRING']] == x[max.hit,SC['MODIFSTRING']]) && 'DISSOCMETHOD' %in% names(SC)) {
      if ("cid" %in% x[,SC['DISSOCMETHOD']])
      x[max.hit,SC['SPECTRUM']] <- x[x[,SC['DISSOCMETHOD']]=="cid",SC['SPECTRUM']][1]

      x[max.hit,SC['DISSOCMETHOD']] <- paste(x[,SC['DISSOCMETHOD']],collapse="&")
      for (sc in score.colname[score.colname != 'pepprob'])
        x[max.hit,sc] <- gsub("NA","",paste(x[,sc],collapse="&"))
      }
      return(x[max.hit,,drop=FALSE])
  })

  colnames(ids.quant.merged)[colnames(ids.quant.merged)=='SPECTRUM.QUANT'] <- 'spectrum.quant'
  ids.quant.merged <- ids.quant.merged[,colnames(identifications)]
  rbind(identifications[spectra.ok,],ids.quant.merged)
}

.merge.identifications <- function(identifications,...) {
  SC <- .SPECTRUM.COLS[.SPECTRUM.COLS %in% colnames(identifications)]

  ## Merge results of different search engines / on different spectra
  if ('SEARCHENGINE' %in% names(SC) &&                               ## search engine column is present
      length(unique(identifications[,SC['SEARCHENGINE']])) > 1 &&    ## there are multiple search engines defines
      !any(grepl('^score\\.',colnames(identifications))))              ## the individual score.ENGINENAME are not present
  {                     
    if (any(grepl("|",identifications[,SC['SEARCHENGINE']],fixed=TRUE))) {
      identifications <- .dissect.search.engines(identifications)               ## dissect merged id columns
    } else {
      identifications <- .merge.search.engine.identifications(identifications,...)  ## merge identifications
    }
  }

  return(identifications)
}

## end MERGE IDENTIFICATIONS
###############################################################################


.read.identifications <- function(identifications,...,
                                  mapping=NULL,mapping.names=c(quantification.spectrum="hcd",identification.spectrum="cid"),
                                  identifications.quant=NULL) {

  ## load identifications (is either character or data.frame
  if (is.character(identifications) && all(sapply(identifications,file.exists)))
    identifications <- .read.idfile(identifications,...)

  ## load mapping (either character or data.frame)
  if (!is.null(mapping)) {
    if (is.character(mapping) && file.exists(mapping))
      mapping <- do.call(rbind,lapply(mapping,read.table,sep=",",header=TRUE,stringsAsFactors=FALSE,quote=""))
    if (!is.data.frame(mapping)) stop("mapping must be a data.frame or valid file name")
    if (ncol(mapping) > 2) stop("only one column in mapping")

    if (is.null(mapping.names)) { 
      if (ncol(mapping) > 2) stop("more than one column in mapping data - mapping.names are therefore required")
      message("trying to assess right mapping")
      cn <- apply(mapping,2,function(x) sum(x %in% identifications[,.SPECTRUM.COLS['SPECTRUM']]))
      mapping.names <- c(quantification.spectrum=names(cn)[which.min(cn)],
                         identification.spectrum=names(cn)[which.max(cn)])
    }
    if (!all(c('identification.spectrum','quantification.spectrum') %in% names(mapping.names)))
      stop("mapping.names must be given alongside with mapping. names(mapping.names) must be 'identification.spectrum' and 'quantification.spectrum'")
    if (!all(mapping.names %in% colnames(mapping))) stop("mapping.names must be column names of mapping")

    quant2id <- .as.vect(mapping,mapping.names['identification.spectrum'],mapping.names['quantification.spectrum'])
    id2quant <- .as.vect(mapping,mapping.names['quantification.spectrum'],mapping.names['identification.spectrum'])

    identifications[,.SPECTRUM.COLS['SPECTRUM.QUANT']] <- id2quant[identifications[,.SPECTRUM.COLS['SPECTRUM']]]
    identifications[,.SPECTRUM.COLS['DISSOCMETHOD']] <- ifelse(is.null(identifications.quant),"hcd","cid") # temp fix
    #identifications[,.SPECTRUM.COLS['DISSOCMETHOD']] <- names(mapping.names)[mapping.names=='identification.spectrum']

    if (!is.null(identifications.quant)) {
      ## load identifications.quant (either character or data.frame)
      if (is.character(identifications.quant) && all(sapply(identifications.quant,file.exists)))
        identifications.quant <- .read.idfile(identifications.quant,...)
      if (!is.data.frame(identifications.quant)) stop("identifications.quant must be a data.frame or valid file name")
      identifications.quant[,.SPECTRUM.COLS['SPECTRUM.QUANT']] <- identifications.quant[,.SPECTRUM.COLS['SPECTRUM']]
      identifications.quant[,.SPECTRUM.COLS['SPECTRUM']] <- quant2id[ identifications.quant[,.SPECTRUM.COLS['SPECTRUM.QUANT']] ]
      identifications.quant[,.SPECTRUM.COLS['DISSOCMETHOD']] <- "hcd"
      #identifications.quant[,.SPECTRUM.COLS['DISSOCMETHOD']] <- names(mapping.names)[mapping.names=='quantification.spectrum']
      #if (.SPECTRUM.COLS['DATABASE'] %in% colnames(identifications) &&
      #    !.SPECTRUM.COLS['DATABASE'] %in% colnames(identifications.quant)) 
      #  identifications[,.SPECTRUM.COLS['DATABASE']] <- NULL

      intersect.ids <- intersect(colnames(identifications),colnames(identifications.quant))
      cn.i.intersect <- colnames(identifications)[!colnames(identifications) %in% intersect.ids]
      cn.iq.intersect <- colnames(identifications.quant)[!colnames(identifications.quant) %in% intersect.ids]
      if (length(cn.i.intersect) > 0 || length(cn.iq.intersect)>0)
        stop("identifications and identifications.quant dataframes are different:",
             ifelse(length(cn.i.intersect),paste0("\n\t[",paste0(cn.i.intersect,collapse=";"),
                                                  "] only appear in identifications"),""),
             ifelse(length(cn.iq.intersect),paste0("\n\t[",paste0(cn.iq.intersect,collapse=";"),
                                                   "] only appear in identifications.quant"),""))

      identifications <- rbind(identifications[,intersect.ids],identifications.quant[,intersect.ids])
    }
  }
  return(identifications)
}


## read MSGF+ tab-separated identification files
.read.msgfp.tsv <- function(filename,filter.rev.hits=FALSE) {
  if (is.data.frame(filename)) 
    ib.data <- filename
  else
    id.data <- read.delim(filename, sep="\t", stringsAsFactors=FALSE,quote="")
  ib.df <- data.frame(Protein=id.data[,'Protein'],
                      spectrum=id.data[,'Title'],.convert.msgfp.pepmodif(id.data[,'Peptide']),
                      scan.from=id.data[,'ScanNum'],dissoc.method=tolower(id.data[,'FragMethod']),
                      precursor.error=id.data[,'PrecursorError.ppm.'],charge=id.data[,'Charge'],
                      search.engine="MSGF+",score=-log10(id.data[,'SpecEValue']),
                      stringsAsFactors=FALSE)

  sel <- names(.MSGF.MAPPINGCOLS) %in% names(c(.SPECTRUM.COLS,.PEPTIDE.COLS))
  data.r <- id.data[,.MSGF.MAPPINGCOLS[sel]]
  colnames(data.r) <- c(.SPECTRUM.COLS,.PEPTIDE.COLS)[names(.MSGF.MAPPINGCOLS)[sel]]
  
  ib.df <- cbind(ib.df,data.r)

  ib.protnpep <-  .convert.msgfp.protein(unique(ib.df[,c('Protein','peptide')]),filter.rev.hits=filter.rev.hits)
  id.data$Protein <- NULL
  merge(unique(ib.protnpep),ib.df,by='peptide',all=TRUE)
}

.convert.msgfp.pepmodif <- function(peptide,modif.masses=
                                      c(iTRAQ4plex=144.102,Cys_CAM=57.021,Oxidation=15.995,
                                        iTRAQ4_ACET="144.102+42.011",ACET=42.011,
                                        METH=14.016,BIMETH=28.031,TRIMETH=42.047
                                        )) {
  modif.masses.r <- setNames(c(names(modif.masses)),c(paste0("+",modif.masses)))
  modifs <- strsplit(paste0(peptide,"+"),"[A-Z]")
  #sort(table(unlist(modifs)))
  modif.p <- sapply(modifs,function(x) { 
                                         x[length(x)] <- sub(".$","",x[length(x)]);
                                         x[x!=""] <- modif.masses.r[x[x!=""]];
                                         x
                                      })
  if (any(is.na(unlist(modif.p))))
    stop("NA in modifications - did not map")
  #sel <- sapply(modif.p,function(x) any(is.na(x)))
  #modifs[sel]
  
  pep.seq <- gsub("[+0-9\\.]","",peptide)
  if (any(nchar(pep.seq)+1 != sapply(modif.p,length)))
    stop("some modif sequences are not of correct length!")
  #cbind(peptide,pep.seq,nchar(pep.seq),sapply(modif.p,length))
  
  modif.i <- mapply(function(pep,modif) {
    pep <- c("Nterm",pep)
    sel.m <- modif!="" & !grepl("_",modif)
    modif[sel.m] <- paste0(modif[sel.m],"_",pep[sel.m])
    paste(modif,collapse=":")
  } ,strsplit(pep.seq,""),modif.p)

  
  
  return(data.frame(peptide=pep.seq,modif=paste0(modif.i,":"),stringsAsFactors=FALSE))
}


.convert.msgfp.protein <- function(protein.n.peptide,filter.rev.hits=FALSE) {
  # layout: sp|Q60848-1|HELLS_MOUSE(pre=R,post=K);sp|Q60848-2|HELLS_MOUSE(pre=R,post=K)
  # reverse hits: XXX_sp|...
  # TODO: extract pre and post AAs

  split.protein <- strsplit(protein.n.peptide[,1],"[;|]")
  res <- lapply(seq_along(split.protein), function(i) { 
                y <- split.protein[[i]]
                cbind(peptide=protein.n.peptide[i,2],
                      database=y[seq(from=1,to=length(y),by=3)],
                      accession=y[seq(from=2,to=length(y),by=3)])
                        })
  res <- as.data.frame(do.call(rbind,res),stringsAsFactors=FALSE)
  res$is.decoy <- grepl("^XXX_",res[,'database'])
  print(c(is.decoy=.sum.bool(res$is.decoy)))
  
  if (isTRUE(filter.rev.hits))
    res <- res[!res$is.decoy,]

  return(res)
}

### =========================================================================
### IBSpectra objects
### -------------------------------------------------------------------------
###
### Class definition.

setClass("IBSpectra",
         contains = "eSet",
         representation(
             proteinGroup = "ProteinGroup",
             reporterTagNames = "character",
             reporterTagMasses = "numeric",
             isotopeImpurities ="matrix",
             log = "matrix",
             "VIRTUAL"),
         prototype = 
            prototype(log=matrix(c(format(Sys.time(), "%F %T %Z"),"IBSpectra Log"),
                                 ncol=2,
                                 dimnames=list("init",c("Timestamp","Message"))))
)

setClass("iTRAQSpectra",
    contains = "IBSpectra",
    representation("VIRTUAL")
)

setClass("iTRAQ4plexSpectra",
    contains = "iTRAQSpectra",
    prototype = prototype(
        reporterTagNames = as.character(114:117),
        reporterTagMasses = c(114.1112,115.1083,116.1116,117.1150),
        isotopeImpurities = matrix(1/100*c(
                c( 92.90,  5.90,  0.20,  0.00 ),
                c(  2.00, 92.30,  5.60,  0.10 ),
                c(  0.00,  3.00, 92.40,  4.50 ),
                c(  0.00,  0.10,  4.00, 92.30 )),
            nrow=4,byrow=TRUE)
    )
)

setClass("iTRAQ8plexSpectra",
    contains = "iTRAQSpectra",
    prototype = prototype(
      reporterTagNames = as.character(c(113:119,121)),
      reporterTagMasses = c(113.10787,114.11123,115.10826,116.11162,
                            117.11497,118.11201,119.1153 ,121.1220),
      isotopeImpurities = diag(nrow=8)
    )
)

setClass("TMTSpectra",
    contains = "IBSpectra",
    representation("VIRTUAL")
)

setClass("TMT2plexSpectra",
    contains = "TMTSpectra",
    prototype = prototype(
      reporterTagNames = c("126","127"),
      reporterTagMasses = c(126.127725,127.131079),
      isotopeImpurities = diag(nrow=2)
    )
)

setClass("TMT6plexSpectra",
    contains = "TMTSpectra",
    prototype = prototype(
      reporterTagNames = as.character(126:131),
      reporterTagMasses = c(126.127725,127.131079,128.134433,
                         129.137787,130.141141,131.138176),
       isotopeImpurities = diag(nrow=6)
       )
    )

setClass("TMT6plexSpectra2",
    contains = "TMTSpectra",
    prototype = prototype(
      reporterTagNames = as.character(126:131),
      reporterTagMasses = c(126.127725,127.124760,128.134433,
                            129.131468,130.141141,131.138176),
       isotopeImpurities = diag(nrow=6)
       )
    )

setClass("TMT10plexSpectra",
    contains = "TMTSpectra",
    prototype = prototype(
      reporterTagNames = c("126","127N","127C","128N","128C","129N","129C","130N","130C","131N"),
      reporterTagMasses = c(126.1277261,127.124761,127.1310809,
                            128.1281158,128.1344357,129.1314706,
                            129.1377905,130.1348254,130.1411453,
                            131.1381802),
       isotopeImpurities = diag(nrow=10)
       )
    )

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Validity.
###

.valid.IBSpectra.slots <- function(object) {
    n <- length(object@reporterTagNames)
    if (length(object@reporterTagMasses) != n) 
        stop("[IBSpectra: validation] Slot reportMasses [",length(object@reporterTagMasses),"]",
             " has not the same length as reporterTagNames [",n,"]!")

    if (!all(dim(object@isotopeImpurities) == n))
        stop("[IBSpectra: validation] isotopeImpurities has dimensions ",
             paste(dim(object@isotopeImpurities),collapse="x"),"!",
             " Expected is ",n,"x",n,". Set it to diag(",n,") when no known.")

       if (length(classLabels(object)) > 0 && length(classLabels(object)) != n) 
        stop("[IBSpectra: validation] Slot classLabels [",length(classLabels(object)),"]",
             " has not the same length as reporterTagNames [",n,"]!")

    return(TRUE)
}

.valid.IBSpectra <- function(object) {
	.valid.IBSpectra.slots(object)
}

setValidity("IBSpectra",.valid.IBSpectra)


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Definitions.
###

# Definition of column names corresponding to spectrum,
# peptide, and protein level data.

.PEAKS.COLS <- c(MASSFIELD="X%s_mass",IONSFIELD="X%s_ions")

.ID.COLS <- c(SEARCHENGINE="search.engine",SCORE="score",SCORE.MASCOT="score.mascot",
              SCORE.PHENYX="score.phenyx",SCORE.MSGF="score.msgf",IS.DECOY="is.decoy",
              P.VALUE="p.value",MSGF.RAWSCORE="msgf.rawscore",MSGF.DENOVOSCORE="msgf.denvoscore",
              MSGF.SPECEVALUE='msgf.specevalue',MSGF.EVALUE='msgf.evalue',
              MSGF.QVALUE="msgf.qvalue",MSGF.PEPQVALUE="msgf.pepqvalue",
              SCAFFOLD.PEPPROB="scaffold.pepprob",SEQUEST.XCORR="sequest.xcorr",SEQUEST.DELTACN="sequest.deltacn",
              DELTASCORE="delta.score",DELTASCORE.PEP="delta.score.pep")

.PTM.COLS <- c(SCORE.PHOSPHORS='pepscore',PROB.PHOSPHORS='pepprob',SEQPOS='seqpos',
               SITEPROBS='site.probs',PHOSPHO.SITES='phospho.sites',PEP.SITEPROBS='pep.siteprobs')

.SPECTRUM.COLS <- c(PEPTIDE="peptide",MODIFSTRING="modif",CHARGE="charge",
                   THEOMASS="theo.mass",EXPMASS="exp.mass",
                   EXPMOZ="exp.moz",
                   PRECURSOR.ERROR="precursor.error",
                   PARENTINTENS="parent.intens",RT="retention.time",
                   SPECTRUM="spectrum",SPECTRUM.QUANT="spectrum.quant",
                   .ID.COLS,USEFORQUANT="use.for.quant",
                   .PTM.COLS,
                   DISSOCMETHOD="dissoc.method",
                   PRECURSOR.PURITY="precursor.purity",
                   SCANS="scans",SCANS.FROM="scan.from",SCANS.TO="scan.to",
                   RAWFILE="raw.file",NMC="nmc",
                   MASSDELTA.ABS="massdelta.abs",MASSDELTA.PPM="massdelta.ppm",
                   SAMPLE="sample",FILE="file",NOTES="notes")

.PEPTIDE.COLS <- c(PROTEINAC="accession",STARTPOS="start.pos",ENDPOS="end.pos",
                  REALPEPTIDE="real.peptide",ILPEPTIDE='il.peptide',
                  AA.BEFORE="aa.before",AA.AFTER="aa.after")

.PROTEIN.COLS <- c(PROTEINAC="accession",PROTEINAC_CONCISE="accessions",
                   NAME="name",PROTEIN_NAME="protein_name",DATABASE="database",
                   GENE_NAME="gene_name",ORGANISM="organism")

## for mapping
.MSGF.MAPPINGCOLS <- c(MSGF.RAWSCORE="MSGFScore",MSGF.DENOVOSCORE="DeNovoScore",
                       MSGF.SPECEVALUE='SpecEValue',MSGF.EVALUE='EValue',
                       MSGF.QVALUE="QValue",MSGF.PEPQVALUE="PepQValue")

.ROCKERBOX.MAPPING.COLS <- c(PROTEINAC="accession",STARTPOS="start.pos",MODIFSTRING="modif",
                     QN="mascot.query.number",RANK="rank",SCANS="scan.number.s.",RT="retention.time",
                     RAWFILE="raw.file",PEPTIDE="sequence", "phosphosequence",NMC="miscleavages",
                     SCORE="score",DELTASCORE="deltascore",PHOSPHO.DELTASCORE.PEP="phospho_deltascore",
                     EXPMASS="peptide.mr",MASSDELTA.ABS="mass.delta..abs.",MASSDELTA.PPM="mass.delta..ppm.",
                     ROCKERBOX_MOD="modifications",PROTEINACS="all.protein.matches",SPECTRUM="scan.title")


writeIBSpectra <- function(ibspectra,file,sep="\t",row.names=FALSE,...) {
  write.table(as.data.frame(ibspectra),file=file,sep=sep,row.names=row.names,quote=FALSE, ...)
  message("finished writing ibspectra to ",file)
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Coercion.
###

setAs("IBSpectra","data.frame",
    function(from) {

      # prepare ProteinGroup data.frame
      pg.df <- as(proteinGroup(from),"data.frame")
      pg.df.cols <- c("protein","peptide","spectrum","start.pos","real.peptide")
      pg.df <- pg.df[,pg.df.cols[pg.df.cols %in% colnames(pg.df)]]
      colnames(pg.df)[colnames(pg.df) == "protein"] <- .PROTEIN.COLS['PROTEINAC']
      colnames(pg.df)[colnames(pg.df) == "peptide"] <- .SPECTRUM.COLS['PEPTIDE']
      colnames(pg.df)[colnames(pg.df) == "spectrum"] <- .SPECTRUM.COLS['SPECTRUM']

      # prepare fData data.frame
      ri <-reporterIntensities(from)
      colnames(ri) <- paste0("X",reporterTagNames(from),"_ions")
      rm <- reporterMasses(from)
      colnames(rm) <- paste0("X",reporterTagNames(from),"_mass")

      fdata.df <- cbind(fData(from),rm,ri)
      fdata.df[,.SPECTRUM.COLS['PEPTIDE']] <- NULL

      # merge data.frames
      res <- unique(merge(pg.df,fdata.df,by=.SPECTRUM.COLS['SPECTRUM'],all.y=TRUE,sort=FALSE))
      if (.PEPTIDE.COLS['REALPEPTIDE'] %in% colnames(res))
        res$peptide <- res$real.peptide

      # return in order
      res[order(res[,.PROTEIN.COLS['PROTEINAC']],res[,.PEPTIDE.COLS['STARTPOS']],res[,.SPECTRUM.COLS['PEPTIDE']]),
          c(intersect(c(.PROTEIN.COLS,.PEPTIDE.COLS,.SPECTRUM.COLS),colnames(res)),
            colnames(rm),colnames(ri))]
    }
)

ibSpectra.as.concise.data.frame  <- function(from) {

      # prepare ProteinGroup data.frame
      pg.df <- proteinGroup.as.concise.data.frame(proteinGroup(from))
      colnames(pg.df)[colnames(pg.df) == "proteins"] <- .PROTEIN.COLS['PROTEINAC_CONCISE']
      colnames(pg.df)[colnames(pg.df) == "peptide"] <- .SPECTRUM.COLS['PEPTIDE']

      # prepare fData data.frame
      ri <-reporterIntensities(from)
      colnames(ri) <- paste0("X",reporterTagNames(from),"_ions")
      rm <- reporterMasses(from)
      colnames(rm) <- paste0("X",reporterTagNames(from),"_mass")
      fdata.df <- cbind(fData(from),rm,ri)

      # merge data.frames
      res <- unique(merge(pg.df,fdata.df,by=.SPECTRUM.COLS['PEPTIDE'],all=TRUE,sort=FALSE))

      # return in order
      res[order(res[,.PROTEIN.COLS['PROTEINAC_CONCISE']],res[,.SPECTRUM.COLS['PEPTIDE']]),
          c(intersect(c(.PROTEIN.COLS,.PEPTIDE.COLS,"n.groups","n.acs","n.variants",.SPECTRUM.COLS),colnames(res)),
            colnames(rm),colnames(ri))]
    }

.IBSpectraAsConciseDataFrameNew <- function(from,show.phospho.position=FALSE) {
  protein.group <- protein.group(from)

  # prepare ProteinGroup data.frame
  indist.proteins <- indistinguishableProteins(proteinGroup(from))
      pg.df <- proteinGroup.as.concise.data.frame(from)
  colnames(pg.df)[colnames(pg.df) == "proteins"] <- .PROTEIN.COLS['PROTEINAC_CONCISE']
  colnames(pg.df)[colnames(pg.df) == "peptide"] <- .SPECTRUM.COLS['PEPTIDE']

  # prepare fData data.frame
  ri <-reporterIntensities(from)
  colnames(ri) <- paste0("X",reporterTagNames(from),"_ions")
  rm <- reporterMasses(from)
  colnames(rm) <- paste0("X",reporterTagNames(from),"_mass")
  fdata.df <- cbind(fData(from),rm,ri)

  # merge data.frames
  res <- unique(merge(pg.df,fdata.df,by=.SPECTRUM.COLS['PEPTIDE'],all=TRUE,sort=FALSE))

  if (show.phospho.position) {
    res.nice <- data.frame(Sequence=.convertPeptideModif(res$peptide,res$modif),
                           Phospho.Position=.convertModifToPos(res$modif,"PHOS"))
  } else {
    res.nice <- data.frame(Sequence=res$peptide)
  }
  res.nice <- cbind(res.nice,
                    Modification=res$modif,
                    AC=.protein.acc(res[,"accession"],proteinGroup(from)),
                    ID=proteinInfo(protein.group,res[,"accession"],do.warn=FALSE),
                    n=sapply(res[,"accession"],function(p) {length(names(indist.proteins)[indist.proteins == p])}))
  res <- res[,!.grep_columns(res,c("peptide","modif","accession"))]
  # return in order
  res.nice <- cbind(res.nice,res[,c(intersect(.SPECTRUM.COLS,colnames(res)),
                                    colnames(rm),colnames(ri))])
  res.nice[order(res$AC,res$Sequence,res$modif),]
}

setMethod("as.data.frame",signature(x="IBSpectra"), 
		function(x, row.names=NULL, optional=FALSE, ...) as(x,"data.frame"))

as.data.frame.IBSpectra <- function(x,...) as(x,"data.frame")


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Accessor-like methods.
###

IBSpectraTypes <- function() { 
  unlist(sapply(
          names(getClass("IBSpectra")@subclasses),
          function(c) { if (!isVirtualClass(c)) c; } )
  )
}

setGeneric("proteinGroup", function(x) standardGeneric("proteinGroup"))
setGeneric("isotopeImpurities", function(x) standardGeneric("isotopeImpurities"))
setGeneric("proteinGroup<-", function(x,value) standardGeneric("proteinGroup<-"))
setGeneric("isotopeImpurities<-", function(x,value) standardGeneric("isotopeImpurities<-"))


# reporterTagNames in package Biobase are defunct now
setGeneric("reporterTagNames",function(object) standardGeneric("reporterTagNames"))
setMethod("reporterTagNames","IBSpectra",function(object) object@reporterTagNames)
setGeneric("reporterTagMasses",function(object) standardGeneric("reporterTagMasses"))
setMethod("reporterTagMasses","IBSpectra",function(object) object@reporterTagMasses)
setMethod("proteinGroup","IBSpectra", function(x) x@proteinGroup)
setMethod("isotopeImpurities","IBSpectra", function(x) x@isotopeImpurities)

setReplaceMethod("proteinGroup","IBSpectra",function(x,value) {
      x@proteinGroup <- value
      x
    }
)
setReplaceMethod("isotopeImpurities",signature("IBSpectra"),
    function(x,value) {
       x@isotopeImpurities <- value
       validObject(x)
       x
    }
)


# TODO: add replaceMethods for reporterMasses and reporterIntensities
#       with a selection of peptide or protein

setGeneric("reporterData", function(x,...) standardGeneric("reporterData"))
setGeneric("reporterData<-", function(x,...,value) standardGeneric("reporterData<-"))
setGeneric("reporterIntensities", function(x,...)
           standardGeneric("reporterIntensities"))
setGeneric("reporterIntensities<-", function(x,...,value)
           standardGeneric("reporterIntensities<-"))
setGeneric("reporterMasses", function(x,...) standardGeneric("reporterMasses"))
setGeneric("reporterMasses<-", function(x,...,value)
           standardGeneric("reporterMasses<-"))

setMethod("reporterData","IBSpectra",
    function(x,element="ions",na.rm=FALSE,na.rm.f='any',...) {
      sel <- spectrumSel(x,...)
      data <- assayDataElement(x,element)[sel,,drop=FALSE]

      if (na.rm & length(data) > 0)
        return(data[!apply(is.na(data),1,na.rm.f),,drop=FALSE])
      else
        return(data)
    }
    )

setReplaceMethod("reporterData","IBSpectra",
    function(x,element="ions",...,value) {
      sel <- spectrumSel(x,...)
      assayDataElement(x,element)[sel,] <- value
      x
    }
)

setMethod("reporterIntensities","IBSpectra",
		function(x,...) reporterData(x,...,element="ions"))

setReplaceMethod("reporterIntensities",signature("IBSpectra"),
    function(x,...,value) {
	 	reporterData(x,...,element="ions") <- value
      x
    }
)

setMethod("reporterMasses","IBSpectra",
		function(x,...) reporterData(x,...,element="mass"))

setReplaceMethod("reporterMasses",signature("IBSpectra"),
    function(x,...,value) {
	 	reporterData(x,...,element="mass") <- value
      x
    }
)


writeData <- function(x, element="ions", file=paste(element,"csv",sep="."), quote=FALSE,
        sep="\t", ...){
      col.names <- sampleNames(x)
      row.names <- spectrumTitles(x)
      write.table(reporterData(x,element="ions"), file=file, quote=quote, sep=sep,
          col.names=col.names, row.names=row.names, ...)
    }


# TODO: use protein.g instead of protein (when protein group identifier is meant)
setGeneric("spectrumSel", function(x,peptide,protein,...) standardGeneric("spectrumSel"))
setMethod("spectrumSel",signature(x="IBSpectra",peptide="missing",protein="missing"),
    function(x) rep(TRUE,nrow(fData(x))))

setMethod("spectrumSel",signature(x="IBSpectra",peptide="data.frame",protein="missing"),
    function(x,peptide,...) spectrumSel(x,as.matrix(peptide),...) )

setMethod("spectrumSel",signature(x="IBSpectra",peptide="matrix",protein="missing"),
    function(x,peptide,modif=NULL,spectrum.titles=FALSE,
             use.for.quant.only=TRUE,do.warn=TRUE) {
        if (length(peptide) == 0) {
          warning("0L peptide provided")
          return(FALSE)
        }
        
        if (is.null(colnames(peptide)))
          stop("a matrix argument with colnames is needed for peptide")

        if (!all(colnames(peptide) %in% colnames(fData(x)))) {
          warning("not all colnames of peptide [",paste0(colnames(peptide),collapse=";"),"] in fData(x)")
          peptide <- peptide[,colnames(peptide) %in% colnames(fData(x))]
        }

        sel <- rep(TRUE,nrow(fData(x)))
        for (p.i in colnames(peptide)) 
          sel <- sel & fData(x)[[p.i]] %in% peptide[,p.i]

        if (use.for.quant.only && .SPECTRUM.COLS['USEFORQUANT'] %in% colnames(fData(x)))
          sel <- sel & fData(x)[,.SPECTRUM.COLS['USEFORQUANT']] 
        
        for (m in modif)
          sel <- sel & grepl(m,fData(x)[,.SPECTRUM.COLS['MODIFSTRING']])

        if (!any(sel) && do.warn) warning("No spectra for peptide ",paste(peptide,collapse="; modif = "))
        if (spectrum.titles)
          return(rownames(fData(x))[sel])
        else
  	      return(sel)
    }
)

setMethod("spectrumSel",signature(x="IBSpectra",peptide="character",protein="missing"),
    function(x,peptide,modif=NULL,spectrum.titles=FALSE,use.for.quant.only=TRUE,do.warn=TRUE) {
        if (length(peptide) == 0) {
          warning("0L peptide provided")
          return(FALSE)
        }
        sel <- fData(x)[,.SPECTRUM.COLS['PEPTIDE']]  %in% peptide
        for (m in modif)
          sel <- sel & grepl(m,fData(x)[,.SPECTRUM.COLS['MODIFSTRING']])

        if (use.for.quant.only && .SPECTRUM.COLS['USEFORQUANT'] %in% colnames(fData(x)))
          sel <- sel & fData(x)[,.SPECTRUM.COLS['USEFORQUANT']] 
        if (!any(sel) && do.warn) warning("No spectra for peptide ",peptide)
         if (spectrum.titles)
          return(rownames(fData(x))[sel])
        else
  	      return(sel)
    }
)

setMethod("spectrumSel",signature(x="IBSpectra",peptide="missing",protein="character"),
    function(x,protein,specificity=REPORTERSPECIFIC,modif=NULL,spectrum.titles=FALSE,use.for.quant.only=TRUE,
             do.warn=TRUE,...) {
      
      peptides <- peptides(x=proteinGroup(x),protein=protein,specificity=specificity,do.warn=do.warn,...)
      if (length(peptides) == 0)
        return(FALSE)
      sel <- spectrumSel(x,peptide=peptides,spectrum.titles=spectrum.titles,
                         modif=modif,use.for.quant.only=use.for.quant.only,do.warn=do.warn)
      if (do.warn && (isTRUE(spectrum.titles) && length(sel) == 0 || !isTRUE(spectrum.titles) && !any(sel)))
        warning("No spectra for protein ",protein,
                " with specificity ",paste(specificity,collapse=","))
      return(sel)
    }
)

setGeneric("spectrumTitles", function(x,...) standardGeneric("spectrumTitles"))
setMethod("spectrumTitles","IBSpectra",function(x,...) 
    function(x) fData(x)[spectrumSel(x,...),.SPECTRUM.COLS['SPECTRUM']])

#setMethod("reporterTagNames","IBSpectra",function(x) x@reporterTagNames)

setGeneric("classLabels", function(x) standardGeneric("classLabels"))
setGeneric("classLabels<-", function(x,value) standardGeneric("classLabels<-"))

setMethod("classLabels",signature(x="IBSpectra"), 
    function(x) {
      class.labels <- phenoData(x)[["class.labels"]]
      names(class.labels) <- phenoData(x)[["class.label.description"]]
      return(class.labels)
    }
)
setReplaceMethod("classLabels","IBSpectra",
    function(x,value) {
      
      if (length(value) != length(reporterTagNames(x)))
        stop("Class labels [",length(value),"] need to habe the same length as reporterTagNames [",length(reporterTagNames(x)),"]")

      if (is.null(names(value))) {
        phenoData(x)[["class.labels",labelDescription="class labels"]] <- value
      } else {
        phenoData(x)[["class.labels",labelDescription="class labels"]] <- as.character(value)
        phenoData(x)[["class.label.description",labelDescription="class label descriptions"]] <- names(value)
      }

      validObject(x)

      x
    }
)


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### correctIsotopeImpurities, normalize, 
### substractAdditiveNoise and exclude methods.
###

setGeneric("correctIsotopeImpurities",
           function(x) standardGeneric("correctIsotopeImpurities") )
setGeneric("subtractAdditiveNoise",
           function(x,...) standardGeneric("subtractAdditiveNoise") )
setGeneric("exclude",
           function(x,protein,peptide=NULL,...) standardGeneric("exclude"))

setMethod("correctIsotopeImpurities",signature(x="IBSpectra"),
    function(x) {
      
      if (is.logged(x,"isotopeImpurities.corrected")) {
        warning("isotope impurities have been corrected already.");
      }
      
      ri <- reporterIntensities(x)
      AA <- isotopeImpurities(x)
      ri.corrected <- t(apply(ri,1,function(b) {
            ok <- !is.na(b)
            if (sum(ok) > 1){
              A <- AA[ok,ok]
              x <- base::solve(t(A),b[ok])
              b[ok] <-x
            }
            return(b)
          }
      ))
      colnames(ri.corrected) <- reporterTagNames(x)
      ri.corrected[ri.corrected<1] <- NA
      reporterIntensities(x) <- ri.corrected

      x <- do.log(x,"isotopeImpurities.corrected",TRUE)

      return(x)
    }
)

normalize <- function(x,f=median,target="intensity",
                      exclude.protein=NULL, use.protein=NULL, peptide.specificity=NULL,
                      f.doapply=TRUE,log=TRUE,
                      channels=NULL,na.rm=FALSE,per.file=TRUE,
                      normalize.factors=NULL,...){
  
  ## NOTE: median normalizes might normalize too much when a lot of NA
  ##         values are present - mean or colSums better?
  
  if (!is.logged(x,"isotopeImpurities.corrected")) 
    warning("Isotope impurity correction has not been logged",
            " - data might be uncorrected. See ?correctIsotopeImpurities.")

  x <- do.log(x,"is.normalized",TRUE)
  ## save original reporter intensities for noise estimation
  if (is.null(assayDataElement(x,"ions_not_normalized")))
    assayDataElement(x,"ions_not_normalized") <- reporterIntensities(x)


  if (!is.null(normalize.factors)) {
    reporterIntensities(x) <- reporterIntensities(x)/
      rep(normalize.factors,each=nrow(reporterIntensities(x)))
    for (i in seq_along(normalize.factors)) {
      x <- do.log(x,paste("normalization.multiplicative.factor channel",
                          colnames(reporterIntensities(x))[i]),
                  round(normalize.factors[i],4))
    }
    return(x)
  }

  if (per.file) {
    if (!.SPECTRUM.COLS['FILE'] %in% colnames(fData(x))) {
      warning("No 'file' column is specified in IBSpectra, setting normalize per.file to FALSE")
      per.file <- FALSE
    } else if (length(unique(fData(x))[,.SPECTRUM.COLS['FILE']])==1) {
      warning("Only one file - setting normalize per.file to FALSE.")
      per.file <- FALSE
    }
  }

  ##if (is.logged(x,"is.normalized"))
  ##  warning("Normalization is logged already.")
  
  if (!is.null(exclude.protein) & !is.null(use.protein))
    stop("Provide either exclude.protein or use.protein, not both.")

  if (is.null(channels) && length(classLabels(x)) > 0)
    channels <- reporterTagNames(x)[!is.na(classLabels(x))]

  if (!is.null(channels) ) {
    if (is.list(channels)) {
      for (channels.set in channels)
        x <- normalize(x,f=f,target=target,exclude.protein=exclude.protein,
                       use.protein=use.protein,peptide.specificity=peptide.specificity,
                       f.doapply=f.doapply,per.file=per.file,
                       log=log,channels=channels.set,na.rm=na.rm,...)
      return(x)
    } else {
      if (!all(channels %in% reporterTagNames(x)))
        stop("channels must be reporterTagNames.")
                                                         
      message("normalizing channels ",paste(channels,collapse=", "))
      ri <- reporterIntensities(x)[,channels,drop=FALSE]
    }
  } else {
    ri <- reporterIntensities(x)
  } ## else is.null(channels)
  
  if (!is.null(exclude.protein))
    ri <- ri[!spectrumSel(x,protein=exclude.protein,
                          specificity=if(is.null(peptide.specificity)) c(REPORTERSPECIFIC,GROUPSPECIFIC,UNSPECIFIC) else peptide.specificity),]
  
  if (!is.null(use.protein)) 
    ri <- ri[spectrumSel(x,protein=use.protein,
                         specificity=if(is.null(peptide.specificity)) REPORTERSPECIFIC else peptide.specificity),,drop=FALSE]

  if (na.rm) 
    sel.na <- apply(!is.na(ri),1,all)
  else
    sel.na <- rep(TRUE,nrow(ri))
  
  ## TODO: warning when ri is empty

  if (per.file && .SPECTRUM.COLS['FILE'] %in% colnames(fData(x))) {
    fd <- fData(x)
    for (n.file in sort(unique(fd[,.SPECTRUM.COLS['FILE']]))) {
      sel <- sel.na & fd[,.SPECTRUM.COLS['FILE']] == n.file
      message("\tnormalizing ",n.file," [",sum(sel)," spectra]")
      ri.sel <- ri[sel,,drop=FALSE]
      normalize.factors <- .get.normalization.factors(ri.sel,f,target,f.doapply,...)
      reporterIntensities(x)[sel,colnames(ri)] <- 
        reporterIntensities(x)[sel,colnames(ri)]/rep(normalize.factors,each=sum(sel))
      for (i in seq_along(normalize.factors)) {
        x <- do.log(x,paste("normalization.multiplicative.factor file",n.file,"channel",colnames(ri)[i]),
                    round(normalize.factors[i],4))
      } 
    }
  } else {
    normalize.factors <- .get.normalization.factors(ri[sel.na,,drop=FALSE],f,target,f.doapply,...)
    reporterIntensities(x)[,colnames(ri)] <- 
      reporterIntensities(x)[,colnames(ri)]/
      rep(normalize.factors,each=nrow(reporterIntensities(x)))
    for (i in seq_along(normalize.factors)) {
      x <- do.log(x,paste("normalization.multiplicative.factor channel",
                          colnames(ri)[i]),
                  round(normalize.factors[i],4))
    }
  }
  ## FIXME: logging a function for f does not work
  ##       x <- do.log(x,"normalization.method",
  ##              sprintf("%s of %s",as.character(substitute(f)),target))
 
  x
}

.get.normalization.factors <- function(ri,f,target,f.doapply,...) {
  
  if (target=="ratio") {
    rs <- rowSums(ri,na.rm=TRUE)
    if (log)
      ri <- apply(ri,2,function(x) x/rs)
    else
      ri <- apply(ri,2,function(x) x-rs)
  } else if (target=="intensity") {
    ## take ri as specified
  } else {
    stop(paste("target",target,"not known"))
  }
  
  if (!f.doapply)
    res <- f(ri,na.rm=TRUE,...)
  else
    res <- apply(ri,2,f,na.rm=TRUE,...)

  return(res/max(res,na.rm=T))
 
}

setGeneric("normalize")


## Estimates the noise level througth the 1st percentile of the >0
## signals (in the linear space) Method quantile: It take's the prob
## (0.01) quantile to estimate the noise level.  This value is
## subtracted from all intensities, and all remaining intensities have
## to be at least that value.

## If channels are assumed similar in intensity and hence a shared
## noise level is reasonable If not, then one level per channel is
## necessary
setMethod("subtractAdditiveNoise",signature(x="IBSpectra"),
    function(x,method="quantile",shared=TRUE,prob=0.01) {
      if (method=="quantile") {
        if (shared) {
          q <- quantile(reporterIntensities(x),probs=prob,na.rm=T)
          ri.sub <- apply(reporterIntensities(x),2,function(x) {
                res <- x - q
                res[res<=q] <- NA
                res
              }
          )
          
        } else {
          ri.sub <- apply(reporterIntensities(x),2,function(x) {
                q <- quantile(x,probs=prob,na.rm=T)
                res <- x - q
                res[res<=q] <- NA
                res
              }
          )
        }
        reporterIntensities(x) <- ri.sub
        x
      } else {
        stop(paste("Method",method,"not implemented"))
      }
    }
)

setMethod("exclude",
    signature(x="IBSpectra",protein="character",peptide="ANY"),
    function(x,protein,peptide=NULL,specificity=c(UNSPECIFIC,GROUPSPECIFIC,REPORTERSPECIFIC)) {
      
      # select peptides for removal
      sel.peptides <- c(peptides(proteinGroup(x),protein=protein,
                                 specificity=specificity),peptide)
      
      # select spectra to keep
      sel.spectra <- !spectrumSel(x,peptide=sel.peptides)
      #sel.spectra <- !spectrumSel(x,protein=proteins.to.exclude,
      #                            specificity=c(REPORTERSPECIFIC,GROUPSPECIFIC,UNSPECIFIC))
      
      # remove spectra from assayData
        for (aden in assayDataElementNames(x)) {
            xel <- assayDataElement(x,aden)[sel.spectra,]
            x <- assayDataElementReplace(x, aden, xel, FALSE)
      }
      
      # remove from featureData
      featureData(x) <- as(fData(x)[sel.spectra,],"AnnotatedDataFrame")
      
      pg.df <- as.data.frame(proteinGroup(x))
      # remove peptides and proteins from proteinGroup
      proteinGroup(x) <- ProteinGroup(pg.df[!pg.df$peptide %in% sel.peptides,] )

      x
    }
)



subsetIBSpectra <- 
  function(x, protein=NULL, peptide=NULL, direction="exclude",
           specificity=c(REPORTERSPECIFIC,GROUPSPECIFIC,UNSPECIFIC),...) {
  if ((is.null(protein) && is.null(peptide)) || (!is.null(protein) && !is.null(peptide)))
    stop("define either protein or peptide to include or exclude")

  if (!is.null(peptide))
    sel.spectra <- spectrumSel(x,peptide=peptide,...)
  else
    sel.spectra <- spectrumSel(x,protein=protein,specificity=specificity,...)

  sel.spectra <- switch (direction,
                         exclude = !sel.spectra,
                         include = sel.spectra,
                         stop("direction must be either 'exclude' or 'include'."))

  for (aden in assayDataElementNames(x)) {
     xel <- assayDataElement(x, aden)[sel.spectra, ]
     x <- assayDataElementReplace(x, aden, xel, validate = FALSE)
  }

  pg.df <- as.data.frame(proteinGroup(x))
  # remove peptides and proteins from proteinGroup
  proteinGroup(x) <- ProteinGroup(pg.df[pg.df[,"spectrum"] %in% 
                                        fData(x)[sel.spectra,"spectrum"],] )


  # remove from featureData
  featureData(x) <- as(fData(x)[sel.spectra,],"AnnotatedDataFrame")

  x
}






### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### show method.
###

setMethod("show",signature(object="IBSpectra"),
    function(object){
      p <- proteinGroup(object)
      callNextMethod(object)
      cat(nrow(fData(object))," spectra\n")
      show(p)
    }
)


###
### Logging
###

setGeneric("is.logged",function(x,name) standardGeneric("is.logged"))
setMethod("is.logged",signature("IBSpectra","character"), function(x,name) {
    name %in% rownames(x@log)
})

setGeneric("get.log",function(x,name) standardGeneric("get.log"))
setMethod("get.log",signature("IBSpectra","character"), function(x,name) {
  paste("\t",paste(apply(x@log[rownames(x@log) %in% name,,drop=FALSE],
                   1,paste,collapse="\t"),
        collapse="\n\t",sep=""),"\n",sep="")
})

setGeneric("do.log", function(x,name,msg) standardGeneric("do.log"))
setMethod("do.log",signature("IBSpectra","character","ANY"),function(x,name,msg) {
    if (is.logged(x,name)) {
        warning("operation ",name," already logged ",
                sum(rownames(x@log) == name)," times:\n",get.log(x,name))
    }
    if (is.logical(msg)) {
        if (msg) { msg = "TRUE"
        } else   { msg = "FALSE" }
    }
    tmp.matrix <- matrix(c(format(Sys.time(), "%F %T %Z"),msg),
                         ncol=2,dimnames=list(name))
                         
    message("LOG: ",sprintf("%10s: %s",name,msg))
    x@log <- rbind(x@log,tmp.matrix)
    return(x)
})

#########################################
## Functions for XLS Report Generation 

testPerl <- function(perl.cmd) {
  if (system(paste(perl.cmd,"-v")) != 0)
    stop(perl.cmd," does not seem to work. It is required for spreadsheet reports in XLS and XLSX formats. ",
         "Either set 'write.xls.report=FALSE', or 'perl.cmd' to a different executable.") 
}

write.xls.report <- function(properties.env,report.env,file="isobar-analysis.xls") {

    testPerl(properties.env[["perl.cmd"]])

    write.t <- function(...,sep="\t",row.names=FALSE,quote=FALSE)
      write.table(...,sep=sep,row.names=row.names,quote=quote,na="")

    get.val <- function(name) { get(name,report.env,inherits=FALSE) }
    get.property <- function(name) { get(name,properties.env,inherits=FALSE) }
    protein.group <- proteinGroup(get.val('ibspectra'))
    indist.proteins <- indistinguishableProteins(protein.group)
    modificationSites <- NULL
    
    if (identical(properties.env[["report.level"]],"protein")) {
      ## TODO: add groups column to provide a link to the groups in report and quant table
      ## in principle, it works by defining the protein.group.ids attr, but does not here - 
      ## probably due to the environment it does not
      attr(proteinGroup(report.env[["ibspectra"]]),"protein.group.ids") <- .as.vect(unique(get.val('quant.tbl')[,c("ac","group")]))

      protein.id.df <- ibSpectra.as.concise.data.frame(get('ibspectra',report.env))
      ## make columns w/ multiple groups gray
      sel.1group  <- protein.id.df[["n.groups"]] == 1
      #sel.1ac  <- protein.id.df[["n.acs"]] == 1
      #sel.1variant  <- protein.id.df[["n.variants"]] == 1
      #protein.id.df[!sel.1ac & sel.1group,1] <- paste("#color silver#",protein.id.df[!sel.1ac & sel.1group,1],sep="")
      #protein.id.df[!sel.1group | !protein.id.df[["use.for.quant"]],1] <- paste("#color=gray#",protein.id.df[!sel.1group,1],sep="")

    } else {
      protein.id.df <- ibSpectra.as.concise.data.frame(get('ibspectra',report.env))
      protein.group <- proteinGroup(get('ibspectra',report.env))
      if (!is.null(properties.env[["phosphosite.dataset"]])) {
        sites <- do.call(rbind,lapply(properties.env[["phosphosite.dataset"]],
                                      read.delim,sep="\t",header=TRUE,skip=3,stringsAsFactors=FALSE))
        colnames(sites)[colnames(sites)=="ACC."]  <- "accession"
        colnames(sites) <- tolower(colnames(sites))
        modificationSites <- sites[sites$accession %in% proteins,]
      }
      proteins <- c(names(indistinguishableProteins(protein.group)),protein.ac(protein.group))
    }

    ## Analysis Properties:
    nn <- reporterTagNames(get.val('ibspectra'))
    ii <- rbind(c("@centeracross@Analysis Properties",rep("@centeracross@Analysis Properties",length(nn))),
                "",
                c("@centeracross@Isotope Impurity Correction Matrix",
                  rep("@centeracross@Isotope Impurity Correction Matrix",length(nn))),
                cbind(c("",nn),rbind(nn,isotopeImpurities(get.val('ibspectra')))))

    cl <- classLabels(get.val('ibspectra'))
    fill.up <- function(x,w="",n=length(nn)+1)
      if (length(x) <= n) c(x,rep(w,n-length(x)))
      else stop("Can't fill up - length(x) > n")
      
    if (!is.null(cl)) {
      ii <- rbind(ii,
                  "",
                  fill.up(c("@centeracross@Class Labels","@centeracross@Class Labels","")))
      
      for (i in seq_along(nn)) {
        ii <- rbind(ii,fill.up(c(nn[i],cl[i],names(cl)[i])))
      }
    }

    # TODO: add some info on number of quantified proteins
    # n.quant.sign <- ddply(quant.tbl,c("class1","class2"),function(x) sum(!is.na(x[["lratio"]]) & x[["is.significant"]]))
    # n.quant <- ddply(quant.tbl,c("class1","class2"),function(x) sum(!is.na(x[["lratio"]])))
    # n.noquant <- ddply(quant.tbl,c("class1","class2"),function(x) sum(is.na(x[["lratio"]])))

    ## write tables to tab seperated files files:
    protein.quant.f <- paste(get.property('cachedir'),"protein_quant.csv",sep="/")
    quant.notlocalized.f <- paste(get.property('cachedir'),"quant.notlocalized.csv",sep="/")
    protein.id.f <- paste(get.property('cachedir'),"protein_id.csv",sep="/")
    analysis.properties.f <- paste(get.property('cachedir'),"analysis_properties.csv",sep="/")
    log.f <- paste(get.property('cachedir'),"logged_operations.csv",sep="/")

    xls.quant.tbl <- get.val('xls.quant.tbl')
    if (identical(properties.env[["report.level"]],"protein")) {
      xls.quant.tbl <- xls.quant.tbl[order(xls.quant.tbl[,"group"]),]
      ## Create links
    } else {
      ## Create links
      protein.id.df[["peptide.sequence"]] <- protein.id.df[["peptide"]]
      if ("site.probs" %in% colnames(protein.id.df)) {
        protein.id.df[["site.probs"]] <- .convertPhosphoRSPepProb(protein.id.df[["peptide"]],protein.id.df[["site.probs"]])
      }
      protein.id.df[["peptide"]] <- .convertPeptideModif(protein.id.df[,"peptide"],protein.id.df[,"modif"])
      q.links <- sapply(protein.id.df[["peptide"]],function(p) {
                          res=which(xls.quant.tbl[["Sequence"]]==p)[1]
                            if (is.na(res)) ""
                            else paste0("@link=internal:Quantifications!A",res+1,"@q")
                  })
      protein.id.df <- cbind(q=q.links,protein.id.df)
      xls.quant.tbl <- cbind(i=paste0("@link=internal:Identifications!A",
                                      sapply(xls.quant.tbl[["Sequence"]],
                                             function(p) which(protein.id.df[["peptide"]]==p)[1]+1),
                                      "@",xls.quant.tbl[["Spectra"]]),xls.quant.tbl)
      #col_idx <- grep("Spectra", names(xls.quant.tbl))
      #colnames(xls.quant.tbl)[col_idx] <- "i"
      #xls.quant.tbl <- xls.quant.tbl[, c(col_idx, (1:ncol(df))[-col_idx])]

    }
    if (exists("xls.quant.tbl.notlocalized",envir=report.env)) {
      write.t(report.env[["xls.quant.tbl.notlocalized"]],file=quant.notlocalized.f)
    }

    if ('notes' %in% colnames(protein.id.df)) 
      protein.id.df[,'notes'] <- gsub("[\t\n]"," ",protein.id.df[,'notes'])
    write.t(xls.quant.tbl,file=protein.quant.f)
    write.t(protein.id.df,file=protein.id.f)  
    write.t(ii,file=analysis.properties.f,col.names=FALSE)
    write.t(get.val('ibspectra')@log,file=log.f,col.names=NA,row.names=TRUE)

    if (identical(properties.env[["report.level"]],"peptide") && !is.null(modificationSites)) {
      modifsites.f <- paste(get.property('cachedir'),"modification_sites.csv",sep="/")
      write.t(modificationSites,file=modifsites.f)
    }

    ## generate perl command line:
    tab2spreadsheet.cmd <- switch(properties.env[["spreadsheet.format"]],
                                  xlsx=system.file("pl","tab2xlsx.pl",package="isobar"),
                                  xls=system.file("pl","tab2xls.pl",package="isobar"),
                                  stop("spreadsheet.format property must be either 'xlsx' or 'xls'."))


    shq <- function(...) shQuote(paste0(...))
    
    .get.perlcl <- function(name,include.identifications=TRUE) {
      fname <- paste0(ifelse(properties.env[["use.name.for.report"]],
                             sprintf("%s.%s",properties.env[["name"]],name),"isobar-analysis"),
                       ".",properties.env[["spreadsheet.format"]])

      perl.cl <- paste(properties.env[["perl.cmd"]],
                       shq(tab2spreadsheet.cmd),shq(fname),
                       shq(":autofilter,freeze_col=4,name=Quantifications:",protein.quant.f))
      
      if (file.exists(quant.notlocalized.f))
        perl.cl <- paste(perl.cl,shq(":autofilter,freeze_col=4,name=Quantifications (not confidently localized sites):",
                            quant.notlocalized.f))
      
      if (identical(properties.env[["report.level"]],"peptide") && !is.null(modificationSites))
        perl.cl <- paste(perl.cl,shq(":autofilter,freeze_col=3,name=Modification Sites:",modifsites.f))

      if (include.identifications)
        perl.cl <- paste(perl.cl,shq(":autofilter,freeze_col=3,name=Identifications:",protein.id.f))

      perl.cl <- paste(perl.cl,shq(":name=Analysis Properties:",analysis.properties.f),
                                  shq(":name=Log:",log.f))
      perl.cl
    }
    
    ## generate Excel report (using Spreadsheet::WriteExcel)
    .call.cmd(.get.perlcl("quant"))
    .call.cmd(.get.perlcl("quantonly",FALSE))

}



.create.or.load.xls.quant.tbl <- function(env,properties.env,name="xls.quant.tbl",quant.tbl.name="quant.tbl") {
  .create.or.load(name,envir=properties.env,
                  msg.f=paste0("table for Excel export [",name,"]"),f=function() {
    message("XLS report format: ",properties.env[["xls.report.format"]])

    env[[quant.tbl.name]]$is.significant[is.na(env[[quant.tbl.name]]$is.significant)] <- FALSE
                    
    compare.to.quant <- .get.or.load('compare.to.quant',properties.env,class="data.frame",null.ok=TRUE)

    # Create a 'wide' XLS table (one row per protein / peptide)
    if (isTRUE(properties.env[["xls.report.format"]]=="wide")) {
      tbl.input  <- ratiosReshapeWide(env[[quant.tbl.name]],vs.class=properties.env[["vs.class"]],
                                      sep="###",cmbn=properties.env[["combn"]])

      if (!is.null(compare.to.quant))
        compare.to.quant <- lapply(compare.to.quant,ratiosReshapeWide,
                                   vs.class=properties.env[["vs.class"]],sep="###",cmbn=properties.env[["combn"]])
    } else {
      tbl.input <- env[[quant.tbl.name]]
    }

    message("   adding identification columns ",appendLF=FALSE)
    if (identical(properties.env[["report.level"]],"protein"))
      res.tbl <- .create.xls.protein.quant.tbl(tbl.input,proteinGroup(env[["ibspectra"]]))
    else
      res.tbl <- .create.xls.peptide.quant.tbl(tbl.input,env[["ibspectra"]],
                                           properties.env[["ptm"]],env[["ptm.info"]])
 
    message("   adding quantification columns")   
    tbl <- .add.quant.to.xls.tbl(env,properties.env,res.tbl[[1]],res.tbl[[2]],compare.to.quant)
    
    #order.c <- if(isTRUE(properties.env[["xls.report.format"]]=="long"),"Channels",NULL)
    if (identical(properties.env[["report.level"]],"protein"))
      tbl <- tbl[order(tbl[,"group"]),]
    else if (all(c("ID","Sequence") %in% colnames(tbl)))
      tbl <- tbl[order(tbl[,"ID"],tbl[,"Sequence"]),]

    tbl

  })
}

## XLS HELPER FUNCTIONS
.get.cols <- function(df,cc,cc.new=NULL,f=NULL,...) {
  data.cc <- df[,grep(cc,colnames(df)),drop=FALSE]
  if (!is.null(f)) data.cc <- f(data.cc,...)
  if (!is.null(cc.new)) colnames(data.cc) <- gsub(cc,cc.new,colnames(data.cc))
  data.cc
}

.add.quant.to.xls.tbl <- function(env,properties.env,tbl,input.tbl,compare.to.quant=NULL) {
  # Add quantification columns from compare.to.quant
  if (!is.null(compare.to.quant)){
    if (!"ac" %in% colnames(input.tbl)) {
      pnp <- subset(proteinGroup(env[["ibspectra"]])@peptideNProtein,
                    protein.g %in% reporterProteins(proteinGroup(env[["ibspectra"]])))
      pnp <- unlist(tapply(pnp[,"protein.g"],pnp[,"peptide"],paste,collapse=";",simplify=FALSE))
      input.tbl[["ac"]] <- pnp[input.tbl[["peptide"]]]
    }
    if (is.data.frame(compare.to.quant))
      compare.to.quant <- list(proteome=compare.to.quant)

    for (ii in seq_along(compare.to.quant)) 
      input.tbl = merge( input.tbl, compare.to.quant[[ii]], by=c("ac","r1","r2"), all.x=TRUE,
                         suffixes=c("",paste("###",names(compare.to.quant)[ii])) )
  }
  input.tbl <- input.tbl[order(input.tbl[["i"]]),]

  if (properties.env[["sum.intensities"]]) {
    message("summing intensities") 
    protein.intensities <- function(ib,proteins) {
      ri <- reporterIntensities(ib)
      t(sapply(proteins,
               function(p) {
                 sel <- spectrumSel(ib,protein=as.character(p),specificity=REPORTERSPECIFIC)
                 if(sum(sel) < 2)
                   ri[sel,]
                 else
                   colSums(ri[sel,],na.rm=TRUE)
               }
               ))
    }
    
    tbl <- cbind(tbl,protein.intensities(env[["ibspectra"]],tbl[["protein"]]))
  } else {
    if (isTRUE(properties.env[["xls.report.format"]]=="long")) {
     tbl <-cbind(tbl,"Classes"=paste(input.tbl[["class2"]],"/",input.tbl[["class1"]]))
     if (!all(input.tbl[['r2']]==input.tbl[['class2']])) {
       tbl <-cbind(tbl,"Channels"=paste(input.tbl[["r2"]],"/",input.tbl[["r1"]]))
      }
    }

    if ("zscore" %in% properties.env[["xls.report.columns"]]) {
      ## TODO: zscore is calculated across all classes - 
      ##       it is probably more appropriate to calculate it individual for each class
      #input.tbl[,'zscore'] <- .calc.zscore(input.tbl[,'lratio'])
    }
    round.digits <- 4;

.combine.n.append.xls.tbl <- function(cc1,cc2,cc.new,f) {
  A <- .get.cols(input.tbl,cc1)
  B <- .get.cols(input.tbl,cc2)
  data.cc <- sapply(1:ncol(A), function(i) f(A[,i],B[,i]) )
  data.cc <- round(data.cc,round.digits)
  colnames(data.cc) <- gsub(cc1,cc.new,colnames(A))
  data.cc
}
.append.xls.tbl <- function(...) { .get.cols(input.tbl,...) }

.round.n.appendend.xls.tbl <- function(...,digits=round.digits) {
  round(.get.cols(input.tbl,...),digits=digits)
}


  
    for (cc in properties.env[["xls.report.columns"]]) {
      message("    adding column ",cc)
      res <- switch(cc,
            log10.ratio =    .round.n.appendend.xls.tbl("lratio","@conditional_formatting=3_color_scale@log10.ratio"),
            log2.ratio =     .round.n.appendend.xls.tbl("lratio","@conditional_formatting=3_color_scale@log2.ratio",f=function(x) x/log10(2)),
            log10.variance = .append.xls.tbl("variance","log10.var"),
            log2.variance =  .append.xls.tbl("variance","log2.var",f=function(x) (sqrt(x)/log10(2)^2)),
            is.significant = .append.xls.tbl("is.significant"),
            n.na1 =          .append.xls.tbl("n.na1"),
            n.na2 =          .append.xls.tbl("n.na2"),
            p.value =        .append.xls.tbl("p.value"),
            p.value.adjusted = .append.xls.tbl("p.value.adjusted"),
            p.value.ratio =  .append.xls.tbl("p.value.rat"),
            p.value.ratio.adjusted =  .append.xls.tbl("p.value.rat.adjusted"),
            p.value.sample = .append.xls.tbl("p.value.sample"),
            z.score =        .round.n.appendend.xls.tbl("zscore"),
            ratio =          .round.n.appendend.xls.tbl("lratio","ratio",f=function(x) 10^x),
            CI95.lower =     .combine.n.append.xls.tbl("lratio","variance","CI95.lower",f=function(x,y) 10^qnorm(0.025,x,sqrt(y))),
            CI95.upper =     .combine.n.append.xls.tbl("lratio","variance","CI95.upper",f=function(x,y) 10^qnorm(0.975,x,sqrt(y))),
            ratio.minus.sd = .combine.n.append.xls.tbl("lratio","variance","ratio.minus.sd",f=function(x,y) 10^(x-sqrt(y))),
            ratio.plus.sd = .combine.n.append.xls.tbl("lratio","variance","ratio.plus.sd",f=function(x,y) 10^(x+sqrt(y))),
            warning("ignoring unknown column ",cc," in Excel report"))
      
      if (is(res,"data.frame") || is(res,"matrix"))
        tbl <- cbind(tbl,res)
      else
        warning("ignore res",cc)

    }
  }

  if (properties.env[["summarize"]]) {
    .append.xls.tbl("n.pos")
    .append.xls.tbl("n.neg")
  }

  if (length(properties.env[["preselected"]]) > 0) {
    ## tbl <- cbind(tbl,"is.preselected"=tbl[["is.preselected"]])
  }
  tbl[["i"]] <- NULL

  return(tbl)
}

.create.xls.peptide.quant.tbl <- function(input.tbl,ibspectra,ptm,ptm.info) {
  protein.group <- proteinGroup(ibspectra)
  ## PEPTIDE REPORT
  pnp  <- subset(as.data.frame(peptideNProtein(protein.group),stringsAsFactors=FALSE),
                 protein.g %in% reporterProteins(protein.group))

  my.ptm <- "PTM"
  if ("PHOS" %in% ptm) my.ptm="Phosphorylation"
  if ("METH" %in% ptm) my.ptm="Methylation"

  input.tbl[,'ac']  <- NULL
  pnp <- ddply(pnp,'peptide',function(x) c(peptide=x[1,'peptide'],ac=paste(x[,'protein.g'],collapse=";")))

  input.tbl <- merge(pnp,input.tbl,by="peptide")
  input.tbl[["i"]]  <- seq_len(nrow(input.tbl))
  input.tbl[["Spectra"]] <- apply(input.tbl[,grepl('^n.spectra',colnames(input.tbl)),drop=FALSE],1,max)
  pg.df <- .proteinGroupAsConciseDataFrame(protein.group,modif.pos=ptm,ptm.info=ptm.info)

  # show peptide modification context
  if (.proteinInfo.ok(protein.group)) {
    pep.modif.context <- getPeptideModifContext(protein.group,modif=ptm)
    pg.df <- merge(pg.df,pep.modif.context,by=c("peptide","modif"),all=TRUE)
  }
  
  tbl <- merge(pg.df,input.tbl[,intersect(c("peptide","pep.siteprobs","modif","i","Spectra"),colnames(input.tbl))],
               by=c("peptide","modif"),all.y=TRUE)
  tbl[["peptide"]] <- .convertPeptideModif(tbl[,"peptide"],tbl[,"modif"])
  colnames(tbl)[colnames(tbl)=="peptide"] <- "Sequence"
  
  colnames(tbl)[colnames(tbl)=="proteins"] <- "ACs"
  colnames(tbl)[colnames(tbl)=="modif.pos"] <- 
    paste0("@comment=Absolute modification position in protein. Modifications in ",
           "the same protein are separated by '&', in different proteins by ';'. ",
           "Stars denote positions which are annotated as phosphorylated in NextProt.@",
           my.ptm," Position")

  tbl[["start.pos"]] <- NULL
  tbl[["modif"]] <- NULL
  tbl[["pos"]] <- NULL
  tbl[["n.groups"]] <- NULL
  tbl <- tbl[order(tbl[["i"]]),]
  return(list(tbl,input.tbl))
} 

.proteinInfo.ok <- function(protein.group)
  is.data.frame(proteinInfo(protein.group)) && 
  length(proteinInfo(protein.group)) > 0

.create.xls.protein.quant.tbl <- function(input.tbl,protein.group,
                                          specificity=c(GROUPSPECIFIC,REPORTERSPECIFIC)) {

  message(".",appendLF=FALSE)
  input.tbl[["i"]]  <- seq_len(nrow(input.tbl))

  tbl.meta <- data.frame(i=input.tbl[["i"]], group=input.tbl[,"group"],protein.g=input.tbl[,'ac'],
                         stringsAsFactors=FALSE)
 
  # protein information
  protein.gs <- unique(input.tbl[,"ac"])
  tbl <- data.frame(protein.g=protein.gs,
                    AC=.protein.acc(protein.gs,protein.group),
                    stringsAsFactors=FALSE)

  message(".",appendLF=FALSE)
  indist.proteins <- .vector.as.data.frame(indistinguishableProteins(protein.group),colnames=c("protein","protein.g"))
  indist.proteins <- indist.proteins[indist.proteins[,'protein.g'] %in% protein.gs,]
  if (.proteinInfo.ok(protein.group)) {
    protein.info.tbl <- proteinInfo(protein.group,protein.g=protein.gs,select=c("name","protein_name","gene_name"))
    colnames(protein.info.tbl) <- c("ID","Description","Gene")
    tbl <- cbind(tbl,protein.info.tbl)
  }

  message(".",appendLF=FALSE)
  # spectra and peptide counts
  pnp <- as.data.frame(peptideNProtein(protein.group),stringsAsFactors=FALSE)
  ps <- peptideSpecificity(protein.group)
  protein.to.peptides <- merge(pnp[pnp[,'protein.g'] %in% protein.gs,],
                               ps[ps[,'specificity'] %in% specificity,],
                               by="peptide")
  protein.to.spectra <- merge(protein.to.peptides,
                              .vector.as.data.frame(protein.group@spectrumToPeptide,colnames=c("spectrum","peptide")),
                              by="peptide")

  tbl <- cbind(tbl, n=sapply(protein.gs,function(p) {sum(indist.proteins == p)}),
                    "@comment=Number of specific peptides@Peptide Count" = 
                       table(protein.to.peptides[,'protein.g'])[protein.gs],
                    "@comment=Number of specific spectra@Spectral Count" = 
                       table(protein.to.spectra[,'protein.g'])[protein.gs])

  message(".",appendLF=FALSE)
  # sequence coverage
  if (.proteinInfo.ok(protein.group) && "sequence" %in% colnames(proteinInfo(protein.group))) {
    peptide.info <- unique(peptideInfo(protein.group)[,c("protein","peptide","start.pos")])
    peptide.info[,'end.pos'] <- peptide.info[,'start.pos'] + nchar(peptide.info[,'peptide']) - 1
    peptide.info <- peptide.info[peptide.info[,'protein'] %in% indist.proteins[,'protein'],]
    protein.lengths <- nchar(unlist(setNames(proteinInfo(protein.group,protein.g=protein.gs,select=c("sequence"),simplify=FALSE),NULL)))

    if (!proteinInfoIsOnSpliceVariants(proteinInfo(protein.group))) {
      splice.df <- protein.group@isoformToGeneProduct
      splice.df <- splice.df[splice.df[,'proteinac.w.splicevariant'] %in% indist.proteins[,'protein'] &
                             splice.df[,'proteinac.wo.splicevariant'] %in% names(protein.lengths),]

      # only keep first AC
      splice.df.1 <- ddply(splice.df,'proteinac.wo.splicevariant',function(x) x[1,])
      peptide.info <- merge(peptide.info,splice.df.1,by.x='protein',by.y='proteinac.w.splicevariant')
      peptide.info[,'protein'] <- peptide.info[,'proteinac.wo.splicevariant']
    }
  
    seq.covs <- ddply(peptide.info,"protein",function(x) {
      protein.length <- protein.lengths[x[1,'protein']]
      if (is.na(protein.length)) return(c(seq.cov=NA))
      x <- x[x[,'end.pos'] <= protein.length,]
      if (nrow(x) == 0) return(c(seq.cov=NA))
      seqq <- rep(FALSE,protein.length);
      for (i in seq_along(nrow(x))) seqq[x[i,'start.pos']:x[i,'end.pos']] <-  TRUE
      return(c(seq.cov=sum(seqq)/length(seqq)))
    },.parallel=isTRUE(options('isobar.parallel')))
  
    if (!proteinInfoIsOnSpliceVariants(proteinInfo(protein.group))) {
      seq.covs <- merge(seq.covs,splice.df,by.x='protein',by.y='proteinac.wo.splicevariant')
      seq.covs[,'protein'] <- seq.covs[,'proteinac.w.splicevariant']
    }

    proteing.seq.covs <- merge(indist.proteins,seq.covs,by='protein')
    seq.covs <- tapply(proteing.seq.covs[,'seq.cov'],factor(proteing.seq.covs[,'protein.g']),mean)
  
    tbl <- cbind(tbl, "@comment=Coverage of the protein sequence with observed peptides@Sequence Coverage" = 
                        round(seq.covs[protein.gs],digits=4))
  }
  tbl <- merge(tbl.meta,tbl,by='protein.g')
  tbl[["protein.g"]] <- NULL
  tbl <- tbl[order(tbl[,'i']),]
  if (!all(tbl[,'i']==input.tbl[,'i'])) stop ("problem in protein xls report ordering")

  return(list(tbl,input.tbl))
}

calcProbXDiffNormals <- function(X,mu_Y,sd_Y,...,alternative=c("greater","less","two-sided"),progress=FALSE) {
  # calculates a vector of either P(X>=Y) or P(Y>=X)
  # (depending on 'alternative' option),
  # where Y ~ N(mu_Y,sd_Y)
  alternative <- match.arg(alternative)
  if (is.null(X)) {
	return(rep(NA,seq_along(mu_Y)))
  }

  if (!is(X,"Distribution")) stop("X has to be of class Distribution (is ",class(X),")")

  if (length(mu_Y) != length(sd_Y)) 
    stop("mu_Y and sd_Y should have equal length")
 
  if (isTRUE(progress))
    pb <- txtProgressBar(min=1,max=length(mu_Y))

  X_mode <- distr::q(X)(0.5)
  p.values <- sapply(seq_along(mu_Y),function(i) {
    if (isTRUE(progress))
      setTxtProgressBar(pb,i)
    if (is.na(mu_Y[i]) || is.na(sd_Y[i])) {
      return(NA)
    }
    else {
      Y <- Norm(mu_Y[i],sd_Y[i])
      # FIXME P-value not adjusted for two-sided case
      if ( alternative == 'greater'
        || (alternative == 'two-sided' && X_mode < mu_Y[i])
      ){
        calcProbXGreaterThanY(X,Y,...)
      } else {
        calcProbXGreaterThanY(Y,X,...)
      }
    }
  })
  if (isTRUE(progress))
    close(pb)

  p.values
}

calcProbXGreaterThanY <- function(X, Y, rel.tol=.Machine$double.eps^0.25, subdivisions=100L) {
  # numerically calculates P(X>=Y)
  if (!is(X,"Distribution")) stop("X has to be of class Distribution")
  if (!is(Y,"Distribution")) stop("Y has to be of class Distribution")
#  return ( distr::p(Y-X)(0) ) # the easiest way, but it does not seem to give accurate results for Norm+Tlsd

  # FIXME use the distribution info to deduce integration boundaries
  return ( integrate( function(t) distr::d(X)(t)*distr::p(Y)(t),
                      lower=-Inf, upper=Inf, subdivisions = subdivisions,
                      rel.tol = rel.tol )$value )
}

.calcProbXGreaterThanY.orig <- function(X,Y,min.q=10^-6, subdivisions=100L) {
  # numerically calculates P(X>=Y)
  if (!is(X,"Distribution")) stop("X has to be of class Distribution")
  if (!is(Y,"Distribution")) stop("Y has to be of class Distribution")

  steps.seq <- seq(from=min.q,to=1-min.q,length.out=subdivisions)
  steps <- sort(unique(c(q(X)(steps.seq),q(Y)(steps.seq))))
  nsubdivisions <- length(steps)
           
  dens.X <- d(X)(steps)
  dens.Y <- d(Y)(steps)

  dens.sum <- sum(dens.X*dens.Y)*.5 # P( X = Y )
  dens.total <- sum(dens.X)*sum(dens.Y)
  if (isTRUE(all.equal(dens.total,0))) return(0)

  for (i in 1:nsubdivisions) {
    dens.X <- dens.X[-1]
    dens.Y <- dens.Y[-length(dens.Y)]
    dens.sum <- dens.sum + sum(dens.X*dens.Y)
  }

  dens.sum/dens.total
}

calcCumulativeProbXGreaterThanY <- function(Xs, mu_Ys, sd_Ys,
                                            alternative=c("greater","less","two-sided"),
                                            rel.tol=.Machine$double.eps^0.25, subdivisions=100L)
{
  # numerically calculates P(X_cum>=Y_cum),
  # where X_cum variable has the PDF of joint X distribution induced to the diagonal (x_1=x_2=..x_n),
  # and Y_cum is the joint normal distribution induced to the diagonal as well
  # it's assumed that the mode of all Xs is very close to zero

  alternative <- match.arg(alternative)

  # cumulative Y distribution is Gaussian with explicitly calculated parameters 
  inv_vars <- sd_Ys^(-2)
  var_Cum <- 1 / sum( inv_vars )
  mu_cum_Y <- sum( as.numeric( mu_Ys %*% inv_vars ) * var_Cum )
  # flip Norm around zero (the mode of X) for 'two-sided' or 'less' tests
  if ( alternative == 'less' || (alternative == 'two-sided' && mu_cum_Y < 0 ) ) {
    mu_cum_Y <- -mu_cum_Y
  }
  cum_Y <- distr::Norm( mu_cum_Y, sqrt( var_Cum ) )

  d_Xs <- lapply( Xs, distr::d )
  # multiplication of all X pdfs
  # note that pdf_cum_X is not a real PDF,
  # because it's not guaranteed to integrate to 1.0
  pdf_cum_X <- function( t ) {
    # apply each d_X to arguments
    pdf_Xs <- lapply( d_Xs, function(d_X) d_X(t) )
    # multiply all PDFs for each element in t
    sapply( seq_along(t), function(ix) prod(sapply(pdf_Xs, function(pdf_X) pdf_X[[ix]])) )
  } 

  pdf_cum_XmY <- function( t ) pdf_cum_X(t) * distr::p( cum_Y )(t)

  return ( integrate( pdf_cum_XmY,
                      lower=-Inf, upper=Inf, subdivisions = subdivisions,
                      rel.tol = rel.tol )$value /
           # correct the scaling of cum_X
           integrate( pdf_cum_X,
                      lower=-Inf, upper=Inf, subdivisions = subdivisions,
                      rel.tol = rel.tol )$value )
}

distrprint <- function(X,round.digits=5) {
  paste0(class(X)," [",
         paste(sapply(setdiff(slotNames(param(X)),'name'),
                      function(x) paste(x,round(slot(param(X),x),round.digits),
                                        sep=" ")),collapse="; "),
         "]")
}

twodistr.plot <- function(X,Y,n.steps=1000,min.q=10^-3) {
  steps.seq <- seq(from=0,to=1,length.out=n.steps)
  steps <- sort(unique(c(q(X)(steps.seq),q(Y)(steps.seq))))
  ggplot(rbind(data.frame(x=steps,Distribution=paste0("X ~ ",distrprint(X)),density=d(X)(steps)),
               data.frame(x=steps,Distribution=paste0("Y ~ ",distrprint(Y)),density=d(Y)(steps)))) + 
    geom_line(aes_string(x="x",y="density",color="Distribution")) +
    ggtitle(paste0("P(X>=Y) = ",round(calcProbXGreaterThanY(X,Y),7)))
 
}

###  =========================================================================
### Ratio estimation and ratio distribution functions.
### -------------------------------------------------------------------------
###


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Distribution fit functions
###
# Distributions used are Norm and Cauchy from package 'distr'.
# See class?Norm and class?Cauchy

fitWeightedNorm <- function(x,weights) {
  new("Norm",
      mean=weightedMean(data=x,weights=weights),
      sd=sqrt(weightedVariance(data=x,weights=weights)[1])
  )
}

fitNorm <- function(x,portion=0.75) {
  x <- x[!is.na(x)]
  if (is.null(portion) || (portion <= 0)){
    bp <- boxplot.stats(x,coef=1)
    limit <- 0.5*(abs(bp$stats[1])+abs(bp$stats[5]))
  }
  else
    limit <- quantile(abs(x),prob=portion)
  good <- (x >= -limit) & (x <= limit)
  new("Norm",
      mean=mean(x[good]),
      sd=sd(x[good])
  )
}

fitCauchy <- function(x) {
  cauchy.fit <- function(theta,x){
    -sum(dcauchy(x,location=theta[1],scale=theta[2],log=TRUE),na.rm=T)
  }
  good <- !is.na(x) & !is.nan(x)
  theta.start <- c(median(x[good]),IQR(x[good])/2)
  res <- nlminb(theta.start,cauchy.fit,x=x[good],
                lower=c(-10,1e-20),upper=c(10,10)) 
  new("Cauchy",location=res$par[1],scale=res$par[2])
}

fitTlsd <- function(x) {
  t.fit <- function(theta,x){
    #-sum(dtls(x,df=theta[3],location=theta[1],scale=theta[2],log=TRUE),na.rm=T)
    -sum(log(dtls(x,df=theta[3],location=theta[1],scale=theta[2])),na.rm=T)
  }
  dtls <- function(x,df,location,scale,log = FALSE)
            1/scale * dt((x - location)/scale, df, log = log)

  good <- !is.na(x) & !is.nan(x)
  theta.start <- c(median(x[good]),sd(x[good]),2) # TODO: find good starting value
  res <- nlminb(theta.start,t.fit,x=x[good],lower=c(-1,10^(-6),2),upper=c(1,10,100))
  new("Tlsd",df=res$par[3],location=res$par[1],scale=res$par[2])
}

fitNormalCauchyMixture <- function(x) {
  gc.fit <- function(theta,x){
    -sum(log(theta[4]*dcauchy(x,location=theta[1],scale=theta[2])+(1-theta[4])*dnorm(x,mean=theta[1],sd=theta[3])))
  }
  good <- !is.na(x) & !is.nan(x)
  theta.start <- c(median(x[good]),IQR(x[good])/2,mad(x[good]),0.5)
  res <- nlminb(theta.start,gc.fit,x=x[good],
                lower=c(-10,1e-20,1e-20,0),upper=c(10,10,10,1)) 
  d1 <- Cauchy(location=res$par[1],scale=res$par[2])
  d2 <- Norm(mean=res$par[1],sd=res$par[3])
  dd <- UnivarMixingDistribution(d1,d2,mixCoeff=c(res$par[4],1-res$par[4]))
  list(mixture=dd,cauchy=d1,normal=d2)
}

fitGaussianMixture <- function(x,n=500) {
  x <- x[!is.na(x)]
  ## EM Algorithm
  ## Initialization: Guessing parameters
  theta <- list(
                weights = c(0.5,0.5),
                means = c(mean(x),mean(x)),
                sds = c(sd(x)-0.5*sd(x), sd(x)+0.5*sd(x)))

  em.gauss.mixd = function(x0,theta,same.mean=TRUE) {
    N <- length(x0)

    ## Expectation-Step
    ##prob for each datapoint to come from each distr
    p_iks <- do.call(cbind,lapply(seq_along(theta$weights),
                                  function(m) dnorm(x0,theta$means[m],theta$sds[m])*theta$weights[m]))
  
    ## membership weights
    Expectation <- p_iks / rowSums(p_iks)
    sum.weights <- colSums(Expectation)
  
    ## Maximization-Step
    theta$weights <- 1/N * colSums(Expectation) # new weights
    if (!same.mean)
      theta$means <- 1/sum.weights * colSums(Expectation*x0) # weighted mean
    theta$sds <- sqrt(1/sum.weights *
                      colSums(Expectation*(x - matrix(theta$means,byrow=T,nrow=N,ncol=2))^2))
    
    return(theta)
  }

  for(i in seq_len(n)){ theta=em.gauss.mixd(x,theta,same.mean=T) }

  d1 <- Norm(theta$means[1],theta$sds[1])
  d2 <- Norm(theta$means[2],theta$sds[2])
  dd <- UnivarMixingDistribution(d1,d2,mixCoeff=theta$weights)
  return(dd)
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### estimateRatio generic and functions.
###
setGeneric("estimateRatioNumeric",
           function(channel1, channel2, noise.model, ...)
             standardGeneric("estimateRatioNumeric"))

setGeneric("estimateRatio",
           function(ibspectra, noise.model=NULL,
                    channel1, channel2, protein, peptide,...)
           standardGeneric("estimateRatio") )

## method definitions
setMethod("estimateRatioNumeric",
    signature(channel1="numeric",channel2="numeric",noise.model="NULL"),
    function(channel1,channel2,noise.model=NULL,...) 
      estimateRatioNumeric(channel1,channel2, ...)
      )
 
setMethod("estimateRatioNumeric",
    signature(channel1="numeric",channel2="numeric",noise.model="missing"),
    function(channel1,channel2,summarize.f=median, ...) {
      if (length(channel1) != length(channel2))
        stop("length of channel 1 does not equal length of channel 2")
      if (length(channel1)==0)
        return(c(lratio=NA))
      sel <- !is.na(channel1) & !is.na(channel2) & channel1 > 0 & channel2 > 0
      
      return(c(lratio=summarize.f(log10(channel1[sel])-log10(channel2[sel]))))
    }
)

setMethod("estimateRatioNumeric",signature(channel1="numeric",channel2="numeric",
                                           noise.model="NoiseModel"),
    function(channel1,channel2,noise.model,ratiodistr=NULL,
             variance.function="maxi",
             sign.level=0.05,sign.level.rat=sign.level,sign.level.sample=sign.level,
             remove.outliers=TRUE,outliers.args=list(method="iqr",outliers.coef=1.5),
             method="isobar",fc.threshold=1.3,channel1.raw=NULL,channel2.raw=NULL,
             use.na=FALSE,preweights=NULL) {
      
      if (length(channel1) != length(channel2))
        stop("length of channel 1 does not equal length of channel 2")
      if (!is.null(channel1.raw) && length(channel1) != length(channel1.raw) ||
          !is.null(channel2.raw) && length(channel2) != length(channel2.raw)) 
        stop("length of orig channels [",length(channel1.raw),"] is not the same as channels [",length(channel1),"]")

      is.method <- function(m) identical(method,m)
      is.a.method <- function(m) m %in% method || is.method("compare.all")
      
      if (length(channel1)==0) {
        if (is.method("isobar"))
          return(c(lratio=NA,variance=NA,var_hat_vk=NA,
                   n.spectra=NA,n.na1=NA,n.na2=NA,
                   p.value.rat=NA,p.value.sample=NA,p.value=NA,is.significant=FALSE))
        if (is.method("isobar-combn"))
          return(c(lratio=NA,variance=NA,n.spectra=NA,n.na1=NA,n.na2=NA,
                   p.value.rat=NA,p.value.sample=NA,p.value=NA,is.significant=FALSE,is.significant.ev=NA,
                   p.value.combined=NA))
        else if (is.method("libra") || is.method("pep"))
          return(c(ch1=NA,ch2=NA))
        else if (is.method("multiq"))
          return(c(lratio=NA,variance=NA,n.spectra=0,isum=NA))
        else if (is.method("test") || is.method("compare.all") || length(method) > 1)
          return(NULL)
        else warning("no spectra, unknown method")
      }
      
      sel <- !is.na(channel1) & !is.na(channel2) & channel1 > 0 & channel2 > 0
      sel.notna <- !is.na(channel1) & !is.na(channel2) & channel1 > 0 & channel2 > 0
      ## Implementation of multiq and libra methods
      if (is.method("multiq")) {
        lratio <- log10(sum(channel1[sel])/sum(channel2[sel]))
        return(c(lratio, variance=NA,n.spectra=length(lratio),
                 isum=sum(channel1[sel]+channel2[sel])))
      }
      if (is.method("libra") || is.method("pep")) {
        return(c(ch1=sum(channel1[sel]),ch2=sum(channel2[sel])))
      }

      i1 <- log10(channel1)
      i2 <- log10(channel2)

      ## Compute all the spectrum ratios and their individual
      ## variance based on the noise model
      log.ratios = i2 - i1
      if (is.null(channel1.raw))
        var.i <- variance(noise.model,i1,i2)
      else 
        var.i <- variance(noise.model,
                          log10(channel1.raw),log10(channel2.raw))
 
      if (remove.outliers) {
        if (identical(outliers.args$method,"wtd.iqr")) 
          outliers.args$weights <- 1/var.i
        outliers.args$log.ratio <- log.ratios
        sel.outliers <- do.call(.sel.outliers,outliers.args)
        sel <- sel & !sel.outliers
      }
      
      # Select non-NA outlier removed spectra
      channel1 <- channel1[sel]
      channel2 <- channel2[sel]
      i1 <- i1[sel]
      i2 <- i2[sel]
      log.ratios <- log.ratios[sel]
      var.i <- var.i[sel]

      center.val <-  ifelse(is.null(ratiodistr), 0 , distr::q(ratiodistr)(0.5))
    
      ## linear regression estimation
      if (is.a.method("lm")) {
        res.lm <- .calc.lm(channel1,channel2,sign.level.sample,sign.level.rat,ratiodistr) 
        if (is.method("lm")) return (res.lm)
      }

      ## weighted linear regression estimation
      if (is.a.method("weighted lm")) {
        res.wlm <- .calc.weighted.lm(channel1,channel2,var.i,sign.level.sample,sign.level.rat,ratiodistr) 
        if (is.method("weighted lm")) return (res.wlm)
      }      
       
      # First, compute ratios on spectra with intensities for both reporter ions
      lratio.n.var <-
        .calc.weighted.ratio(log.ratios,var.i,variance.function,preweights[sel])
      
      if (is.a.method("ttest")) {
        if (length(log.ratios) < 2) 
          p.value <- 1  
        else 
          p.value <- t.test(log.ratios)$p.value

        res.ttest <- c(lratio=lratio.n.var['lratio'],
                       variance=NA,n.spectra=length(log.ratios),
                       p.value=p.value,is.significant=p.value<sign.level)
        if (is.method("ttest")) return(res.ttest)
      }
      if (is.a.method("ttest2")) {

        p.value <- .ttest.pval(mu=center.var,
                               xbar=lratio.n.var['lratio'],
                               s=sqrt(lratio.n.var['sample.variance']),
                               n=length(log.ratios))

        res.ttest2 <- c(lratio=lratio.n.var['lratio'],
                        variance=lratio.n.var['sample.variance'],
                        p.value = p.value,
                        is.significant=p.value<sign.level)
        if (is.method("ttest2"))
          return(res.ttest2)
      }
        
#if (method == "wttest" || method == "compare.all") {
#        p.value <- (length(log.ratios) < 2)? 1 : weighted.t.test(log.ratios,w=weights,weighted.ratio)$p.value
#        res.wttest <- c(
#            lratio=weighted.ratio,variance=NA,n.spectra=length(log.ratios),
#            p.value=p.value,is.significant=p.value<sign.level)
#        if (method != "compare.all") return(res.wttest)
#      }
      if (is.a.method("fc")) {
        res.fc <- c(lratio=as.numeric(lratio.n.var['lratio']),variance=NA,
                    n.spectra=length(log.ratios),
                    is.significant=abs(as.numeric(lratio.n.var['lratio']))>log10(fc.threshold))
        ratio.nw <- mean(log.ratios,na.rm=TRUE)
        res.fc.nw <- c(lratio=ratio.nw,variance=NA,n.spectra=length(log.ratios),
            is.significant=abs(ratio.nw)>log10(fc.threshold))
        if (is.method("fc")) return(res.fc)
      }
 
      weighted.ratio <- as.numeric(lratio.n.var['lratio'])
      calc.variance <- as.numeric(lratio.n.var['calc.variance'])
      if (is.a.method("isobar") || is.a.method("isobar-combn")) {
        # Check for channels where one is NA
        sel.ch1na <- is.na(channel1) & !is.na(channel2)
        sel.ch2na <- is.na(channel2) & !is.na(channel1)

        if (use.na) {
          stop("use.na currently not functional")
        } # use.na

        res.isobar <- 
          c(lratio=weighted.ratio, variance=calc.variance,
            var_hat_vk=as.numeric(lratio.n.var['var_hat_vk']),
            n.spectra=length(log.ratios),
            n.na1=sum(sel.ch1na),n.na2=sum(sel.ch2na),
            p.value.rat=calculate.ratio.pvalue(weighted.ratio, calc.variance, ratiodistr),
            p.value.sample=calculate.sample.pvalue(weighted.ratio, ratiodistr),
            p.value=if (!is.null(ratiodistr)) calcProbXDiffNormals(ratiodistr, weighted.ratio, sqrt(calc.variance), alternative="two-sided") else 1,
            is.significant=NA)
        
        if (!is.method("isobar"))
          res.isobar['is.significant.ev'] <- 
            (res.isobar['p.value.sample'] <= sign.level.sample) &&
            calculate.ratio.pvalue(weighted.ratio,lratio.n.var['estimator.variance'],
                                   ratiodistr) <= sign.level.rat

        res.isobar['is.significant'] <- 
          (res.isobar['p.value'] <= sign.level)

        if (is.method("isobar-combn")) {
          p.value.combined <- NA
          if (all(!is.na(c(res.isobar['lratio'],res.isobar['variance'])))) {
            p.value.combined <- calcProbXGreaterThanY(ratiodistr,Norm(res.isobar['lratio'],sqrt(res.isobar['variance'])))
            p.value.combined <- 2*ifelse(p.value.combined<.5,p.value.combined,1-p.value.combined)
          }
          return(c(res.isobar,p.value.combined=p.value.combined))
        }

        if (is.method("isobar")) return(res.isobar)
      }
      if (length(method) == 1 && !is.method("compare.all")) stop(paste("method",method,"not available"))

      add.res <- function(x,name) {
        if (!exists(x)) return(NULL)
        o <- get(x)
        as.numeric(o[name])
      }

      res <- c(lratio=weighted.ratio,
               lratio.lm=add.res("res.lm",'lratio'),
               lratio.wlm=add.res('res.wlm','lratio'),
               unweighted.ratio  = mean(log.ratios,na.rm=TRUE),
               n.spectra=length(log.ratios),
               variance=calc.variance,
               var.ev=as.numeric(lratio.n.var['estimator.variance']),
               var.sv=as.numeric(lratio.n.var['sample.variance']),
               var.lm=add.res('res.lm','stderr')**2,
               var.lm.w=add.res('res.wlm','stderr')**2,
               is.sign.isobar    = add.res('res.isobar','is.significant'),
               is.sign.isobar.ev = add.res('res.isobar','is.significant.ev'),
               is.sign.lm        = add.res('res.lm','is.significant'),
               is.sign.wlm       = add.res('res.wlm','is.significant'),
               is.sign.rat       = add.res('res.isobar','p.value.rat')<sign.level,
               is.sign.sample    = add.res('res.isobar','p.value.sample')<sign.level,
               is.sign.ttest     = add.res('res.ttest','is.significant'),
               is.sign.fc        = add.res('res.fc','is.significant'),
               is.sign.fc.nw     = add.res('res.fc.nw','is.significant'))
                
    }
)

calculate.ratio.pvalue <- function(lratio, variance, ratiodistr = NULL) {
  center.val <-  ifelse(is.null(ratiodistr), 0, distr::q(ratiodistr)(0.5))
  lt_mask <- !is.na(lratio) & lratio < center.val
  ge_mask <- !is.na(lratio) & !lt_mask
  res <- as.numeric( rep.int( NA, length(lratio) ) )
  res[lt_mask] <- pnorm(lratio[lt_mask],
                        mean=center.val, sd=sqrt(variance[lt_mask]),
                        lower.tail=TRUE)
  res[ge_mask] <- pnorm(lratio[ge_mask],
                        mean=center.val, sd=sqrt(variance[ge_mask]),
                        lower.tail=FALSE)
  return(res)
}

calculate.sample.pvalue <- function(lratio,ratiodistr) {
  res <- as.numeric( rep.int( NA, length(lratio) ) )
  if (!is.null(ratiodistr)) {
    center.val <- distr::q(ratiodistr)(0.5)
    lt_mask <- !is.na(lratio) & lratio < center.val
    ge_mask <- !is.na(lratio) & !lt_mask
    res[lt_mask] <- distr::p(ratiodistr)(lratio[lt_mask], lower.tail=TRUE)
    res[ge_mask] <- distr::p(ratiodistr)(lratio[ge_mask], lower.tail=FALSE)
   } else {
    #warning('Sample P-value not calculated: missing ratio distribution')
  }
  return(res)
}


calculate.mult.sample.pvalue <- function(lratio,ratiodistr,strict.pval,lower.tail,
                                         n.possible.val, n.observed.val) {
  if (is.null(ratiodistr)) 
    return(NA)

  product.p.vals <- prod(distr::p(ratiodistr)(lratio,lower.tail=lower.tail))
  if (strict.pval)
    pval <- getMultUnifPValues(product.p.vals*0.5^(n.possible.val-n.observed.val),n=n.possible.val)
  else
    pval <- getMultUnifPValues(product.p.vals,n=n.observed.val)

  return(pval)
}

adjust.ratio.pvalue <- function(quant.tbl,p.adjust,sign.level,globally=FALSE) {
  if (globally) {
    quant.tbl[,'p.value.rat.adjusted'] <- p.adjust(quant.tbl[,'p.value.rat'], p.adjust)
    quant.tbl[,'p.value.adjusted'] <- p.adjust(quant.tbl[,'p.value'], p.adjust)
    quant.tbl[,'is.significant'] <- quant.tbl[,'is.significant'] & quant.tbl[,'p.value.adjusted'] < sign.level
  } else {
    comp.cols <- c("r1","r2","class1","class2")
    comp.cols <- comp.cols[comp.cols %in% colnames(quant.tbl)]
    quant.tbl <- ddply(quant.tbl,comp.cols,function(x) {
      x[,'p.value.rat.adjusted'] <- p.adjust(x[,'p.value.rat'], p.adjust)
      x[,'p.value.adjusted'] <- p.adjust(x[,'p.value'], p.adjust)
      x[,'is.significant'] <- x[,'is.significant'] & x[,'p.value.adjusted'] < sign.level
      x
    })
  }
  quant.tbl
}

correct.peptide.ratios <- function(ibspectra, peptide.quant.tbl, protein.quant.tbl, protein.group.combined,
                                   adjust.variance=TRUE, correlation=0, recalculate.pvalue = TRUE) {

  attrs <- attributes(peptide.quant.tbl)
  if (isTRUE(attr(peptide.quant.tbl,'summarize')) || isTRUE(attr(protein.quant.tbl,'summarize')))
    stop("Ratio correction should be done before summarization! Use proteinRatio with argument before.summarize.f!")

  # map from peptides to protein group identifier
  all.q.prots <- unique(protein.quant.tbl[,'ac'])
  n.quant <- table(protein.quant.tbl[!is.na(protein.quant.tbl[,'lratio']),'ac'])
  q.protein.acs <- strsplit(all.q.prots,",")

  pi <- protein.group.combined@peptideInfo
  pi <- merge(pi,as.data.frame(isobar:::.as.matrix(protein.group.combined@indistinguishableProteins,
                                                   colnames=c("protein","protein.g")),
                               stringsAsFactors=FALSE))
  pi <- pi[pi[["protein.g"]] %in% reporterProteins(protein.group.combined),]
  pep.to.ac <- isobar:::.as.vect(unique(pi[,c('peptide','protein.g')]))

  peptide.quant.tbl[,'ac'] <- pep.to.ac[peptide.quant.tbl[,'peptide']]

  # merged peptide and protein quant table
  tbl <- merge(peptide.quant.tbl,protein.quant.tbl[,c("ac","r1","r2","lratio","variance","is.significant")],
               by = c("ac","r1","r2"), all.x = TRUE, suffixes=c(".modpep",".prot"))

  tbl[,'lratio'] <- .cn(tbl,'lratio.modpep') - .cn(tbl,'lratio.prot')

  attrs$adjust.variance <- adjust.variance
  if (adjust.variance) {
    attrs$adjust.variance.corralation <- correlation
    cov <- correlation * sqrt(.cn(tbl,'variance.modpep')) * sqrt(.cn(tbl,'variance.prot'))
    tbl[,'variance'] <- .cn(tbl,'variance.modpep') + .cn(tbl,'variance.prot') * 2 * cov
  }

  attrs$recalculate.pvalue <- recalculate.pvalue
  if (recalculate.pvalue) {
    ratiodistr <- attr(peptide.quant.tbl,'ratiodistr')
    tbl[,'p.value.rat'] <- calculate.ratio.pvalue(tbl[,'lratio'], tbl[,'variance'], ratiodistr)
    tbl[,'p.value.sample'] <- calculate.sample.pvalue(tbl[,'lratio'], ratiodistr)
    tbl[,'p.value'] <- calcProbXDiffNormals( ratiodistr, tbl[,'lratio'], sqrt(tbl[,'variance']), alternative="two-sided" )
    tbl[,'is.significant'] <- tbl[,'p.value'] <= attr(peptide.quant.tbl,'sign.level')
  }
  attrs$names <- colnames(tbl)
  attrs$row.names <- rownames(tbl)
  attributes(tbl) <- attrs

  return(tbl)
}

.calc.lm <- function(channel1,channel2,
                     sign.level.rat,sign.level.sample,ratiodistr) {
  res.lm <- NA
  if (any(!is.na(channel1) & !is.na(channel2))){
    fm <- lm(channel2~channel1+0)
    summary.fm <- summary.lm(fm)$coefficients
    ci <- confint(fm,level=1-sign.level.rat)
    res.lm <- c(lratio=log10(summary.fm[1]),
                ratio=summary.fm[1],
                stderr=summary.fm[2],
                ratio.ci.l=ci[1],
                ratio.ci.u=ci[2])

    if (!is.null(ratiodistr)) {
      rat.neg <- log10(res.lm['ratio']) < distr::q(ratiodistr)(0.5)
      ci[ci < 0] <- 0
      # TODO: fix confidence interval which goes to minus
      #message(paste(ci,collapse=":"))
      lrat.p <- log10(ifelse(rat.neg,ci[2],ci[1]))
      res.lm['p.value'] <- distr::p(ratiodistr)(lrat.p,lower.tail=rat.neg)
      res.lm['is.significant'] <- res.lm['p.value'] < sign.level.sample
    }
  }
  res.lm
}

.calc.weighted.lm <- function(channel1,channel2,var.i,
                              sign.level.rat,sign.level.sample,ratiodistr) {
  res.lm <- NA
  if (any(!is.na(channel1) & !is.na(channel2))){
    fm <- lm(channel2~channel1+0,
             weights=1/var.i)
    summary.fm <- summary.lm(fm)$coefficients
    ci <- confint(fm,level=1-sign.level.rat)
    res.lm <- c(lratio=log10(summary.fm[1]),
                ratio=summary.fm[1],
                stderr=summary.fm[2],
                ratio.ci.l=ci[1],
                ratio.ci.u=ci[2])

    if (!is.null(ratiodistr)) {
      rat.neg <- log10(res.lm['ratio']) < distr::q(ratiodistr)(0.5)
      ci[ci < 0] <- 0
      # TODO: fix confidence interval which goes to minus
      #message(paste(ci,collapse=":"))
      lrat.p <- log10(ifelse(rat.neg,ci[2],ci[1]))
      res.lm['p.value'] <- distr::p(ratiodistr)(lrat.p,lower.tail=rat.neg)
      res.lm['is.significant'] <- res.lm['p.value'] < sign.level.sample
    }
  }
  res.lm
}


.get.ri <- function(ri,ch) {
  if (is.null(ri))
    return(NULL)
  if (ch == "ALL")
    rowSums(ri,na.rm=TRUE)
  else if (ch == "AVG")
    apply(ri,1,mean,na.rm=TRUE)
  else
    ri[,ch]
}

.sel.outliers <- function(log.ratio,method="boxplot",
                          outliers.coef=1.5,outliers.trim=0.1,
                          weights=NULL) {
  # discussion: http://stats.stackexchange.com/questions/7155/rigorous-definition-of-an-outlier
  sel <- is.na(log.ratio)
  if (method == "boxplot") {
    # Tukey's (1977) method; see ?boxplot.stats and below at 'iqr'.
    # outliers.coef=1.5, typically
    bp <- boxplot.stats(log.ratio,coef=outliers.coef)
    sel | log.ratio<bp$stats[1] | log.ratio>bp$stats[5]
  } else if (method == "iqr") {
    # Tukey's (1977) method
    # applicable to skewed data (makes no distributional assumptions)
    # may not be appropriate for small sample sizes
    #   selects points which are more than outliers.coef times the interquartile range 
    #   above the third quartile or below the first quartile.  
    #  possible outliers: outlier.coef=1.5 (99.3% of the data within range in normal distribution)
    #  probable outliers: outlier.coef=3

    qs <- quantile(log.ratio,c(.25,.75),na.rm=TRUE)
    iqr <- qs[2] - qs[1]
    sel | log.ratio < qs[1]-outliers.coef*iqr | log.ratio > qs[2]+outliers.coef*iqr
  } else if (method == "wtd.iqr") {
    # weighted implementation of iqr method with weighted quantiles
    requireNamespace("Hmisc")
    qs <- Hmisc::wtd.quantile(log.ratio,weights,c(.25,.75),na.rm=TRUE)
    iqr <- qs[2] - qs[1]
    sel | log.ratio < qs[1]-outliers.coef*iqr | log.ratio > qs[3]+outliers.coef*iqr
  } else if (method == "robust.zscore") {
    # modified z-score proposed by Iglewicz and Hoaglin (1993)
    #  68% have zscores between +/- 1, 95% have zscores between +/- 2
    #  99.7% have zscores between +/- 3
    # outliers.coef=3.5, typically
    sel | 0.6745*abs(log.ratio-mean(log.ratio,na.rm=TRUE))/mad(log.ratio,constant=1,na.rm=TRUE) > outliers.coef
  } else if (method == "zscore") {
    # 99.7% of the data lie within 3 standard deviations of the mean
    # outliers.coef=3, typically
    sel | abs(log.ratio-mean(log.ratio,na.rm=TRUE))/sd(log.ratio,na.rm=TRUE) > outliers.coef

  } else if (method == "trim") {
    sel | log.ratio < quantile(log.ratio,outliers.trim/2,na.rm=TRUE) |
          log.ratio > quantile(log.ratio,1-outliers.trim/2,na.rm=TRUE)
  } else {
    stop("method ",method," not known, use one of [boxplot,iqr,wtd.iqr,zscore,trim]")
  }
}


                           
.calc.weighted.ratio <- function(log.ratio,variance,
                                 variance.function,preweights=NULL) {
  
  variance <- variance
  preweights <- preweights
  
  weights <- 1/variance
  if (!is.null(preweights))
    weights <- weights*preweights
  sum.weights <- sum(weights)
  weighted.ratio <- weightedMean(log.ratio,weights)

  # variance calculation
  estimator.variance <- 1/sum.weights
  var_hat_vk = Inf
  if (length(log.ratio) == 1)
    sample.variance <- estimator.variance^0.75
  else {
    wvar <- weightedVariance(log.ratio,weights,weighted.ratio)
    var_hat_vk = wvar[2]
    if (length(log.ratio) == 2)
      sample.variance <- max(c(wvar[1],estimator.variance^0.75))
    else
      sample.variance <- wvar[1]
  }

  calc.variance <- switch(variance.function,
      maxi = max(estimator.variance,sample.variance,na.rm=T),
      ev = estimator.variance,
      wsv = sample.variance,
      stop("unknown variance function - choose one of [maxi,ev,wsv]")
  )

  return(c(lratio=weighted.ratio,
           estimator.variance=estimator.variance,
           sample.variance=sample.variance,
           calc.variance=calc.variance,
           var_hat_vk = var_hat_vk
        ))
}

estimateRatioForProtein <- function(protein,ibspectra,noise.model,channel1,channel2,
        combine=TRUE,method="isobar",specificity=REPORTERSPECIFIC,quant.w.grouppeptides=NULL,...) {
      #message("parallel processing: ",isTRUE(getOption('isobar.parallel')))
      if (combine) {
        if (method == "multiq" || method == "libra" || method=="pep") {
          ## first compute peptide ratios, summarize then
          peptide.ratios <-
            estimateRatioForPeptide(peptides(proteinGroup(ibspectra),protein=protein),
                                    ibspectra,noise.model=noise.model,
                                    channel1=channel1,channel2=channel2,
                                    combine=FALSE,method=method)

          if (method == "libra" | method=="pep") {
            # normalize to sum
            sum.i <- peptide.ratios$ch1+peptide.ratios$ch2
            p.ch1 <- peptide.ratios$ch1/sum.i
            p.ch2 <- peptide.ratios$ch2/sum.i
            if (method=="libra") {
              ##remove outliers +2sd
              sd1 <- sd(p.ch1)
              sd2 <- sd(p.ch2)
              
              p.ch1 <- p.ch1[p.ch1 > mean(p.ch1)-2*sd1 & p.ch1 < mean(p.ch1)+2*sd1]
              p.ch2 <- p.ch2[p.ch2 > mean(p.ch2)-2*sd2 & p.ch2 < mean(p.ch2)+2*sd2]
            }
            return(lratio=log10(p.ch2/p.ch1),variance=var(log10(p.ch2/p.ch1)),
                   ch1=p.ch1,ch2=p.ch2)
             
          }
          if (method == "multiq") {
            ## TODO: filtration of peptides with low confidence score
            ## TODO: Dynamic range comp.

            ## weighted average of peptide ratios
            return(c(
              lratio=weightedMean(peptide.ratios[,'lratio'],weights=peptide.ratios$isum),
              variance=var(peptide.ratios$lratio)))
          }

        } else if (method %in% c("isobar","lm","ttest","ttest2","fc","compare.all")) {
          .call.estimateRatio(protein,"protein",ibspectra,
                             noise.model,channel1,channel2,method=method,
                             specificity=specificity,...)
        } else {
          stop("method ",method," not known")
        }
      } else {
        res <- ldply(protein,function(individual.protein) {
             if (individual.protein %in% quant.w.grouppeptides) 
               specificity <- c(GROUPSPECIFIC,specificity)

             .call.estimateRatio(individual.protein,"protein",ibspectra,
                                noise.model,channel1,channel2,method=method,...,
                                specificity=specificity)
            },.parallel=isTRUE(getOption('isobar.parallel'))
        )
        rownames(res) <- protein
        res
        #res[apply(res,2,!function(r) all(is.na(r))),]
      }
    }

estimateRatioForPeptide <- function(peptide,ibspectra,noise.model,channel1,channel2,combine=TRUE,...) {
      if (combine) {
        r <- .call.estimateRatio(peptide,"peptide",ibspectra,noise.model,
                                 channel1,channel2,...)
      } else {
        if (is.data.frame(peptide))
          peptide <- as.matrix(peptide)
        if (is.matrix(peptide)) {
          r <- ldply(seq_len(nrow(peptide)),function(p_i) 
                  .call.estimateRatio(peptide[p_i,,drop=FALSE],"peptide",ibspectra,noise.model,
                                      channel1,channel2,...),.parallel=isTRUE(getOption('isobar.parallel')))
        
        } else {
          r <- ldply(peptide,function(individual.peptide) 
                        .call.estimateRatio(individual.peptide,"peptide",ibspectra,noise.model,
                                            channel1,channel2,...),.parallel=isTRUE(getOption('isobar.parallel')))
        }
      }
      attr(r,"input") <- peptide
      attr(r,"combine") <- combine
      return(r)
}

# Calculate a T pvalue
.ttest.pval <- function(mu,xbar,s,n=2,conf.level=0.95) {
  t <- (xbar-mu)/ (s/sqrt(n))
  df <- ifelse(n==1,0.5,n-1)  ## using 0.5 df when n=1 (has no theoretical justification)
  p.value <- pt(t,df)
  p.value <- ifelse(p.value<0.5,p.value,1-p.value) # two-sided
  p.value <- p.value*2 ## correct for two-sided test
  tc <- pt(conf.level,df)/2

  if (length(xbar)>1) {cf <- cbind} else {cf <- c}
  return(cf(t.stat=t,p.value,
             ci.lower=xbar-tc*s/sqrt(n),
             ci.upper=xbar+tc*s/sqrt(n)))
}

.ttest.pval.se <- function(mu,xbar,se,n=2,conf.level=0.95) {
  t <- (xbar-mu)/ se
  df <- ifelse(n==1,0.5,n-1)  ## using 0.5 df when n=1
  p.value <- pt(t,df)
  p.value <- ifelse(p.value<0.5,p.value,1-p.value) # two-sided
  p.value <- p.value*2 ## correct for two-sided test
  if (length(xbar)>1) {cf <- cbind} else {cf <- c}
  tc <- pt(conf.level,df)/2
  return(cf(t.stat=t,p.value,
             ci.lower=xbar-tc*se,
             ci.upper=xbar+tc*se))
}


### Handling NULL protein or peptide argument
setMethod("estimateRatio",
          signature(ibspectra="IBSpectra",noise.model="ANY",
                    channel1="missing",channel2="missing",
                    protein="character",peptide="missing"),
          function(ibspectra,noise.model=NULL,protein,
                   val="lratio",summarize=FALSE,combine=TRUE,...) {
            channels <- reporterTagNames(ibspectra)
            if (combine) {
              res <- matrix(NA,nrow=length(channels),ncol=length(channels),
                            dimnames=list(channels,channels))
              for(channel1 in channels)
                for (channel2 in channels) {
                  rat <- estimateRatio(ibspectra,noise.model=noise.model,
                                       channel1=channel1,channel2=channel2,
                                       protein=protein,...)
                  res[channel1,channel2] <- rat[val]
                }
              return(res)
            } else {
              cmbn <- t(combn(channels,2))
              res <- estimateRatio(ibspectra,noise.model=noise.model,
                                   channel1=cmbn[i,1],channel2=cmbn[i,2],
                                   protein=protein,combine=FALSE,...)

              apply(cmbn,1,function(i) 
                    cbind(r1=cmbn[i,1],r2=cmbn[i,2],ac=rownames(res),res))
            }
            }
)



setMethod("estimateRatio",
          signature(ibspectra="IBSpectra",noise.model="ANY",
                    channel1="missing",channel2="missing",
                    protein="missing",peptide="character"),
          function(ibspectra,noise.model=NULL,peptide,
                   val="lratio",summarize=FALSE,combine=TRUE,...) {
            channels <- reporterTagNames(ibspectra)
            if (combine) {
              res <- matrix(NA,nrow=length(channels),ncol=length(channels),
                            dimnames=list(channels,channels))
              for(channel1 in channels)
                for (channel2 in channels) {
                  rat <- estimateRatio(ibspectra,noise.model=noise.model,
                                       channel1=channel1,channel2=channel2,
                                       peptide=peptide,...)
                  res[channel1,channel2] <- rat[val]
                }
              return(res)
            } else {
              cmbn <- t(combn(channels,2))
              res <- estimateRatio(ibspectra,noise.model=noise.model,
                                   channel1=cmbn[i,1],channel2=cmbn[i,2],
                                   peptide=peptide,combine=FALSE,...)

              apply(cmbn,1,function(i) 
                    cbind(r1=cmbn[i,1],r2=cmbn[i,2],ac=rownames(res),res))
            }
            }
)


setMethod("estimateRatio",
    signature(ibspectra="IBSpectra",noise.model="ANY",
        channel1="character",channel2="character",
        protein="character",peptide="NULL"),
    function(ibspectra,noise.model,channel1,channel2,protein,peptide=NULL,...) {
        estimateRatio(ibspectra,noise.model,channel1,channel2,protein=protein,...)
    }
)

setMethod("estimateRatio",
    signature(ibspectra="IBSpectra",noise.model="ANY",
              channel1="character",channel2="character",
              protein="NULL",peptide="character"),
    function(ibspectra,noise.model,channel1,channel2,protein=NULL,peptide,...) {
      estimateRatio(ibspectra,noise.model,channel1,channel2,peptide=peptide,...)
    }
)

setMethod("estimateRatio",
    signature(ibspectra="IBSpectra",noise.model="ANY",
              channel1="character",channel2="character",
              protein="NULL",peptide="matrix"),
    function(ibspectra,noise.model,channel1,channel2,protein=NULL,peptide,...) {
      estimateRatio(ibspectra,noise.model,channel1,channel2,peptide=peptide,...)
    }
)

### Calling estimateRatioForPeptide and calcRatioForProtein, resp.
setMethod("estimateRatio",
    signature(ibspectra="IBSpectra",noise.model="ANY",
        channel1="character",channel2="character",
        protein="character",peptide="missing"),
    function(ibspectra,noise.model,channel1,channel2,protein,...) {
        estimateRatioForProtein(protein,ibspectra,noise.model,channel1,channel2,...)
    }
)

setMethod("estimateRatio",
    signature(ibspectra="IBSpectra",noise.model="ANY",
              channel1="character",channel2="character",
              protein="missing",peptide="character"),
    function(ibspectra,noise.model,channel1,channel2,peptide,...) {
      estimateRatioForPeptide(peptide,ibspectra,noise.model,channel1,channel2,...)
    }
)

setMethod("estimateRatio",
    signature(ibspectra="IBSpectra",noise.model="ANY",
              channel1="character",channel2="character",
              protein="missing",peptide="data.frame"),
    function(ibspectra,noise.model,channel1,channel2,peptide,...) {
      estimateRatioForPeptide(peptide,ibspectra,noise.model,channel1,channel2,...)
    }
)

setMethod("estimateRatio",
    signature(ibspectra="IBSpectra",noise.model="ANY",
              channel1="character",channel2="character",
              protein="NULL",peptide="data.frame"),
    function(ibspectra,noise.model,channel1,channel2,protein=NULL,peptide,...) {
      estimateRatioForPeptide(peptide,ibspectra,noise.model,channel1,channel2,...)
    }
)



setMethod("estimateRatio",
    signature(ibspectra="IBSpectra",noise.model="ANY",
              channel1="character",channel2="character",
              protein="missing",peptide="matrix"),
    function(ibspectra,noise.model,channel1,channel2,peptide,...) {
      estimateRatioForPeptide(peptide,ibspectra,noise.model,channel1,channel2,...)
    }
)

.NULLstring <- function(x) {
  if(is.null(x))
    return("NULL")
  return(x)
}

### Helper function to estimateRatioNumeric
.call.estimateRatio <- function(x,level,ibspectra,noise.model,
                                channel1,channel2,
                                specificity=REPORTERSPECIFIC,modif=NULL,
                                n.sample=NULL,groupspecific.if.same.ac=FALSE,
                                use.precursor.purity=FALSE,do.warn=TRUE,
                                use.for.quant.only=TRUE,
                                use.vsn.transform=FALSE,...) {

  allowed.channels <- c(reporterTagNames(ibspectra),'AVG','ALL')
  allowed.channels.p <- paste(allowed.channels, collapse=", ")

  if (is.null(channel1) || !all(channel1 %in% allowed.channels))
    stop("channel 1 is ",.NULLstring(channel1),",",
         " but it should be one of ",allowed.channels.p,".")

  if (is.null(channel2) || !all(channel2 %in% allowed.channels))
    stop("channel 2 is ",.NULLstring(channel2),",",
         " but it should be one of ",allowed.channels.p,".")
  
  ## when more than one channel value is given, calculate the combination matrix
  if (length(channel1) > 1 || length(channel2) > 1) {

    ## using match.call to get the call with all arguments
    ecall <- match.call(expand.dots = TRUE)

    res  <- c()
    for (c1 in channel1) {
      for (c2 in channel2) {
        my.ecall <- ecall
        my.ecall[["channel1"]] <- c1
        my.ecall[["channel2"]] <- c2

        ## using eval() on my.ecall for recursive calling of this function
        res <- rbind(res,
                data.frame(channel1=c1,channel2=c2,
                           t(eval(my.ecall,parent.frame())),
                           stringsAsFactors=FALSE))
      }
    }
    rownames(res) <- NULL
    return(res)
  }

  ## select spectra based on quantification level
  sel <- switch(level,
                "protein"=spectrumSel(ibspectra,protein=x,specificity=specificity,
                                      do.warn=do.warn,modif=modif,
                                      groupspecific.if.same.ac=groupspecific.if.same.ac,
                                      use.for.quant.only=use.for.quant.only),
                "peptide"=spectrumSel(ibspectra,peptide=x,modif=modif,do.warn=do.warn,
                                      use.for.quant.only=use.for.quant.only),
                "peptide-probs"=stop("not implemented yet"),
                stop("level ",level," unknown"))

  ## get reporter intensities from ibspectra
  ri <- reporterIntensities(ibspectra)[sel,,drop=FALSE]

  ## get raw intensities (prior to normalization) for estimation of noise level
  ri.raw <- reporterData(ibspectra,element="ions_not_normalized")[sel,,drop=FALSE]

  ## use vsn transformed data (see apply.vsn() in IBSpectra-class.R
  if (use.vsn.transform) {
    if (!"ions_vsn_normalized" %in% assayDataElementNames(ibspectra))
      stop("Use apply.vsn() to calculate variance-stabilizing transformation, first")

    ri <- reporterData(ibspectra,element="ions_vsn_normalized")[sel,,drop=FALSE]
  }

  ## get reporter intensities
  i1 <- .get.ri(ri,channel1)
  i2 <- .get.ri(ri,channel2)
  i1.raw <- .get.ri(ri.raw,channel1)
  i2.raw <- .get.ri(ri.raw,channel2)

  # sample data for testing puposes (TP/FP estimation)
  if (!is.null(n.sample)) {
    if (n.sample <= length(i1)) {
      indices <- sample(seq_along(i1),n.sample)
      i1 <- i1[indices]
      i2 <- i2[indices]
      if (!is.null(ri.raw)) {
        i1.raw <- i1.raw[indices]
        i2.raw <- i2.raw[indices]
      }
    }
    else
      i1 <- i2 <- i1.raw <- i2.raw <- numeric()
   }

  ## use precursor purity for spectra weighting
  if (isTRUE(use.precursor.purity)) {
   if (!.SPECTRUM.COLS['PRECURSOR.PURITY'] %in% colnames(fData(ibspectra)))
    stop("Argument use.precusor.purity is TRUE, but no column '",
         .SPECTRUM.COLS['PRECURSOR.PURITY'],"' in data.")
    precursor.purity <- fData(ibspectra)[sel,"precursor.purity"]
  } else {
    precursor.purity <- NULL
  }

   estimateRatioNumeric(channel1=i1,channel2=i2,noise.model=noise.model,...,
                        preweights=precursor.purity,
                        channel1.raw=i1.raw,channel2.raw=i2.raw)
}


getMultUnifPValues <- function(product,pvals=NULL,n=NULL){
  
  if (is.null(n))
    return(NULL)
  
  if (!is.null(pvals))
    if (n == length(pvals))
      product <- prod(pvals)
    else
      return(NULL)
  
  if (n == 1)
    return(product)
  
  s <- 1
  for (i in 1:(n-1))
    s <- s + (-log(product))^i/factorial(i)
  product*s
  
} # getMultUnifPValues


getMultUnifDensity <- function(x,n=NULL){
  
  if (is.null(n))
    return(NULL)
  if (n == 1)
    return(1)
  
  i <- n-1
  (-log(x))^i/factorial(i)
  
} # getMultUnifDensity


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### proteinRatios generic and function.
###

## combn.matrix generates pairwise combinations 
##  of the values in 'x':
##  - all against all (using combn, methd="global")
##  - all combinations across classes (method="interclass")
##     used when computing ratios class A against class B
##  - all combinations within classes (method="intraclass")
##     used when computing a intra-class ratio distribution 
##     for estimation of biological variability
##  - method "versus.class": all combinations against class vs
##  - method "versus.channel": all combinations against channel vs
combn.matrix <- function(x,method="global",cl=NULL,vs=NULL) {
  if (method != "global" & is.null(cl))
    stop("No class labels cl provided.")
  if (method != "global" & !is.character(cl)) {
    warning("Class labels should be of type character - using as.character to convert.")
    cl <- as.character(cl)
  }
  if (method != "global" & length(cl) != length(x))
    stop(sprintf("cl argument does not have the same length as x (cl: %s, x: %s).",
                 length(cl),length(x)))

  # Filter NA channels
  if (!is.null(cl)) {
    x <- x[!is.na(cl)]
    cl <- cl[!is.na(cl)]
  }

  if (!is.null(vs) && !is.character(vs))
    vs <- as.character(vs)

  # Create a combn matrix with all combinations of channels to consider 
  if (method == "versus.class" || method == "versus.channel") {
    if (method == "versus.channel") {
      if (is.null(vs)) {
        vs = x[1]
        warning("vs argument is null, but method is versus.channel. Using channel '",vs,"'")
      }
      if (!vs %in% x) stop("vs argument must be one of [",paste(x,collapse=", "),"]")
      pos <- which(x==vs)
      cmbn <- rbind(vs,x[-pos])
      if (!is.null(cl)) {
        vs.class <- cl[pos]
        cmbn <- rbind(cmbn,vs.class,cl[-pos])
      }
    }
    if (method == "versus.class") {
      if (is.null(vs)) {
        vs = cl[1]
        warning("vs argument is null, but method is versus.channel. Using channel '",vs,"'")
      }

      if (!all(vs %in% cl)) stop("vs argument must be one of [",paste(cl,collapse=", "),"]")
      if (is.null(cl)) stop("class labels must be given with method versus.class")
      pos <- which(cl==vs)
      cmbn <- rbind(x[pos],rep(x[-pos],each=length(pos)))
      cmbn <- rbind(cmbn,vs,rep(cl[-pos],each=length(pos)))

    }
  } else if (method == "global") {
    cmbn <- combn(x,2) # take all combinations
    if (!is.null(cl))
      cmbn <- rbind(cmbn,combn(cl,2))
  } else {
    t <- table(cl)[unique(cl)]

    if (method == "intraclass") {
      if (length(t[t==1]) > 0)
        warning("Some class labels are not repeated - ",
                "those are ignored in intraclass ratios.")
          
      cmbn <- do.call(cbind,lapply(names(t)[t>1],
                                    function(xx) rbind(combn(x[which(cl==xx)],2),class1=xx,class2=xx)))
      
    } else if (method == "interclass") {
      if (length(t) == 1) {
        warning("Cannot compute interclass ratios when there is only one class - taking ratios vs ALL")
        cmbn <- matrix(c(x,rep("ALL",length(x))),nrow=2,byrow=TRUE)
      } else {
        cmbn <- matrix(nrow=4,ncol=0)
        for (name in names(t)) {
          pos=which(cl==name);posn=which(cl!=name);
          for (i in pos) 
            for (j in posn) {
              cc <- c(x[i],x[j],cl[i],cl[j])
              if (ncol(cmbn) > 0 &
                  any(apply(cmbn,2,function(xx) identical(rev(cc[1:2]),xx[1:2]))))
                next;
              cmbn <- cbind(cmbn,cc)
            }
        }
      }
    } else {
      stop(paste("method",method,"not implemented."))
    }
  }
  if (nrow(cmbn) == 2) rownames(cmbn) <- c("r1","r2")
  else if (nrow(cmbn) == 4) rownames(cmbn) <- c("r1","r2","class1","class2")
  return(cmbn)
}

## create a table with all protein ratios
combn.protein.tbl <- function(cmbn, reverse=FALSE, ...) {
  
  ratios <- do.call(rbind,apply(cmbn,2,function(x) {
    if (reverse)
      if (length(x) == 4)
        x <- x[c(2,1,4,3)]
      else
        x <- rev(x)

    message("ratios ",x[2]," vs ",x[1])
    r <- estimateRatio(channel1=x[1],channel2=x[2],...)
    if (class(r)=="numeric") {
      r <- t(r)
      rownames(r) <- "prot1"
    }

    if (is.matrix(attr(r,"input")))
      df <- data.frame(attr(r,"input"),r,stringsAsFactors=FALSE)
    else
     df <- data.frame(r,stringsAsFactors=FALSE)

    if (!is.null(rownames(r)) && any(rownames(r) != as.character(seq_len(nrow(r))))) 
      df[,'ac']<- rownames(r)
    rownames(df) <- NULL

    df$r1 <- x[1]; df$r2 <- x[2]
    if (length(x) == 4) {
      df$class1 <- x[3]; df$class2 <- x[4]
    }
    return(df)
  }))

  attr(ratios,"arguments") <- list(...)

  attr(ratios,"cmbn") <- cmbn
  attr(ratios,"reverse") <- reverse

  if (all(c("peptide","modif") %in% colnames(ratios))) 
    ratios <- ratios[order(ratios[,'peptide'],ratios[,'modif'],ratios$r1,ratios$r2),]
  else
    ratios <- ratios[order(ratios[,'ac'],ratios$r1,ratios$r2),]
  
  return(ratios)  
}


peptideRatios <- function(ibspectra,...,peptide=peptides(proteinGroup(ibspectra),columns=c("peptide","modif"))) {
  proteinRatios(ibspectra,...,proteins=NULL,peptide=peptide)
}

peptideRatiosNotQuant <- function(ibspectra,...,
                                  peptide=unique(fData(ibspectra)[!fData(ibspectra)[["use.for.quant"]],c("peptide","modif","site.probs")])) {
  proteinRatios(ibspectra,...,proteins=NULL,peptide=peptide,use.for.quant.only=FALSE)
}


ratiosReshapeWide <- function(quant.tbl,vs.class=NULL,sep=".",cmbn=NULL,short.names=FALSE) {
  id.cols <- c("group","ac","peptide","modif","gene_names","pep.siteprobs",
               "ID","Description","Gene","gene_names.x","gene_names.y","swissprot.ac","modification")
  id.cols <- id.cols[id.cols %in% colnames(quant.tbl)]

  # remove unneccasry column [should not be present anyway in quant.tbl created with a  current version]
  quant.tbl$ratios.subset.1..by.column. <- NULL

  if (!is.null(cmbn)) {
    sel <- paste(quant.tbl$r1,quant.tbl$r2) %in% paste(cmbn[1,],cmbn[2,])
    quant.tbl <- quant.tbl[sel,]
  }
  classes.unique <- "class1" %in% colnames(quant.tbl) &&
                    !any(is.na(quant.tbl$class1)) && !any(is.null(quant.tbl$class1)) && 
                    all(table(unique(quant.tbl[,c("r1","class1")])$class1)==1) &&
                    all(table(unique(quant.tbl[,c("r2","class2")])$class2)==1)

  if (!is.null(vs.class)) {
    if (!any(quant.tbl[,"class1"]==vs.class)) stop("vs.class set to ",vs.class,", but it is not present in quant table")
    if (length(vs.class)==1 && short.names)
      quant.tbl[,'comp']<- paste(quant.tbl$class2)
    else 
      quant.tbl[,'comp']<- paste(quant.tbl$class2,quant.tbl$class1,sep="/")

    quant.tbl <- quant.tbl[quant.tbl[["class1"]] %in% vs.class,]
  } else {
    if (classes.unique) {
      quant.tbl[,'comp']<- paste(quant.tbl$class2,quant.tbl$class1,sep="/")
    } else {
      quant.tbl[,'comp']<- paste(quant.tbl$r2,quant.tbl$r1,sep="/")
    }
  }

  ccomp <- unique(quant.tbl$comp)
  fac <- factor(apply(quant.tbl[,id.cols,drop=FALSE],1,paste,collapse="."))

  ## check that all 'comp' entries are in the right order
  if (!all(tapply(quant.tbl$comp,fac,function(x) all(x==ccomp))))
    stop("quant.tbl not in the right format - comp column values different")
  
  quant.tbl.num  <- quant.tbl[,-(c(which(colnames(quant.tbl) %in% 
                                c(id.cols,"comp","r1","r2","class1","class2"))))]

  q.coltypes <- sapply(quant.tbl.num,class)
  logical.cols <- q.coltypes == 'logical'
  if (!all(q.coltypes %in% c("numeric","logical","integer"))) {
    bad.cols <- q.coltypes[!q.coltypes %in% c("numeric","logical")]
    stop("quantification table reshape does not work - ",
         " columns ",paste0(names(bad.cols),collapse=","),
         " have types ",paste0(bad.cols,collapse=","),".")
  }

  colnames.wide <- paste(rep(colnames(quant.tbl.num),each=length(ccomp)),
                         rep(ccomp,times=ncol(quant.tbl.num)),sep=sep)

  q1 <- do.call(rbind,tapply(seq_len(nrow(quant.tbl)),fac,function(x) {
    quant.tbl[x[1],id.cols]
  },simplify=FALSE))
  
  q2 <- do.call(rbind,tapply(seq_len(nrow(quant.tbl.num)),fac,function(x) {
    unlist(quant.tbl.num[x,])
  },simplify=FALSE))
  colnames(q2) <- colnames.wide
  
  if (!all(sapply(q2,is.numeric)))
    stop("quantification table reshape did not work - columns should be numeric")
  q2 <- as.data.frame(q2)
  if (any(logical.cols)) {
    col.n <- unlist(lapply((which(logical.cols)-1)*length(ccomp)+1,function(x) seq(x,length.out=length(ccomp))))
    q2[,col.n] <- sapply(q2[,col.n],as.logical)
  }

  qq <- cbind(q1,q2)
  rownames(qq) <- NULL

  attrs <- attributes(quant.tbl)
  for (a in names(attrs))
    if (!a %in% c("row.names","names","class")) attr(qq,a) <- attrs[[a]]

  qq
}

proteinRatios <-
  function(ibspectra, noise.model, reporterTagNames=NULL,
           proteins=reporterProteins(proteinGroup(ibspectra)),peptide=NULL,
           cl=classLabels(ibspectra), combn.method="global",combn.vs=NULL,
	   symmetry=FALSE,
           summarize=FALSE,summarize.method="mult.pval",
           min.detect=NULL,strict.sample.pval=TRUE,strict.ratio.pval=TRUE,orient.div=0,
           sign.level=0.05,sign.level.rat=sign.level,sign.level.sample=sign.level,
           ratiodistr=NULL,zscore.threshold=NULL,variance.function="maxi",
           combine=FALSE,p.adjust=NULL,reverse=FALSE,
           cmbn=NULL,before.summarize.f=NULL,...) {

  if ((!is.null(proteins) && !is.null(peptide)) ||
      (is.null(proteins) && is.null(peptide)))
    stop("supply either protein or peptides!")      
  
  if (!is.null(p.adjust) && !p.adjust %in% p.adjust.methods)
    stop("p.adjust parameter must be one of '",paste(p.adjust.methods,collapse="','"),"'")
  
  if (is.null(reporterTagNames)) reporterTagNames <- reporterTagNames(ibspectra)
  if (is.null(cl) && is.null(cmbn)) stop("please supply class labels as argument cl or a cmbn matrix")

  if (is.null(cmbn))
    cmbn <- combn.matrix(reporterTagNames,combn.method,cl,vs=combn.vs)

  if (ncol(cmbn) < 1) 
    stop("No possible combination for reporters [",paste(reporterTagNames,sep=","),"]",
         " w/ classes [",paste(cl,sep=","),"] and combn.method ",combn.method," possible.",
         " summarize=",ifelse(summarize,"TRUE","FALSE"))
  
  ratios <- combn.protein.tbl(cmbn,reverse=reverse,
                              ibspectra=ibspectra,noise.model=noise.model,
                              ratiodistr=ratiodistr,
                              protein=proteins,peptide=peptide,
                              sign.level=sign.level,sign.level.rat=sign.level.rat,
                              sign.level.sample=sign.level.sample,
                              variance.function=variance.function,combine=combine,...)

  attributes(ratios) = c(attributes(ratios),list(
                         classLabels=cl,combn.method=combn.method,symmetry=symmetry,summarize=FALSE,
                         sign.level.rat=sign.level.rat,sign.level.sample=sign.level.sample,
                         ratiodistr=ratiodistr,variance.function=variance.function,
                          combine=combine,p.adjust=p.adjust,reverse=reverse))

  if (!is.null(before.summarize.f)) {
    .check.isfunction(before.summarize.f)
    ratios <- before.summarize.f(ibspectra,ratios)
  }


  if (summarize) {
    message("summarizing ratios ...")
    if (combn.method=="global")
      stop("summarization not meaningful with combn.method='global'. ",
           "Use combn.method='intraclass' or combn.method='interclass' to use ratios in or between classes.")

    ## calculate the number of combinations for each class-combination
    if (reverse)
      n.combination <- table(cmbn["class2",],cmbn["class1",])
    else
      n.combination <- table(cmbn["class1",],cmbn["class2",])

    if (nrow(n.combination)==1 & ncol(n.combination)==1) 
      n.combination <- as.numeric(n.combination)
    if (max(n.combination) < 2)
      stop("Summarize=TRUE makes no sense with only one combination, set summarize to FALSE or class labels differently.")

    if (is.null(min.detect))
      min.detect <- n.combination

    if (!is.null(proteins)) {
      by.column <- "ac"
    } else {
      by.column <- "peptide"
      if (!is.null(dim(peptide)) && "modif" %in% colnames(peptide))
       by.column <- c("peptide","modif") 
    } 
    
    ratios <- summarize.ratios(ratios,by.column,
                           summarize.method,min.detect,n.combination,
                           strict.sample.pval,strict.ratio.pval,orient.div,
                           sign.level,sign.level.rat,sign.level.sample,
                           variance.function=variance.function,
                           ratiodistr=ratiodistr)
  }

  ratios[,'zscore'] <- .calc.zscore(ratios[,'lratio'])
  if (!is.null(zscore.threshold))
    ratios[,'is.significant'] <- ratios[,'p.value'] < sign.level & abs(ratios[,'zscore']) > zscore.threshold

  if (symmetry) {
    ratios.inv <- ratios
    ratios.inv[,'lratio'] <- -ratios.inv[,'lratio']
    ratios.inv[,c('r1','r2')] <- ratios.inv[,c('r2','r1')]
    if ('class1' %in% colnames(ratios))
      ratios.inv[,c('class1','class2')] <- ratios.inv[,c('class2','class1')]
    ratios <- rbind(ratios,ratios.inv)
  }

  if (!is.null(p.adjust)) 
    ratios <- adjust.ratio.pvalue(ratios,p.adjust,sign.level.rat)

  return(ratios)
} # end proteinRatios


summarize.ratios <-
  function(ratios,by.column="ac",summarize.method="mult.pval",min.detect=NULL,
           n.combination=NULL,
           strict.sample.pval=TRUE,strict.ratio.pval=TRUE,orient.div=0,
           sign.level=0.05,sign.level.rat=sign.level,sign.level.sample=sign.level,
           variance.function="maxi",ratiodistr=NULL) {

    if (!summarize.method  %in% c("mult.pval","mean"))
      stop("Implemented summarize.methods: mult.pval, mean. ",
           "Please choose one of them instead of ",summarize.method,".")

    if (!all(c("class1","class2") %in% colnames(ratios)))
      stop("'ratios' argument must specify classes w/ columns class1 and class2!")

    classes <- unique(ratios[,c("class1","class2")])
    
    if (is.null(n.combination)) {
      cc <- unique(ratios[,c("r1","r2","class1","class2")])
      n.combination <- table(cc[,"class1"],cc[,"class2"])
      if (nrow(n.combination)==1 & ncol(n.combination)==1) 
        n.combination <- as.numeric(n.combination)
    }
    if (is.null(min.detect))
      min.detect <- n.combination

    if (!all(by.column %in% colnames(ratios))) 
      stop("'by.column' ",paste(by.column,collapse=",")," defined, but no such columns in 'ratios'.")

    mean.r <- ifelse(is.null(ratiodistr),0,distr::q(ratiodistr)(0.5))
    if (summarize.method == "mult.pval") {
      
      result <- ddply(ratios,by.column,function(ratios.subset) {

        ldply(seq_len(nrow(classes)),function(class_i) {

          class1 <- classes[class_i,1]
          class2 <- classes[class_i,2]

          n.combination.c <- ifelse(is.matrix(n.combination),
                                    n.combination[class1,class2],
                                    n.combination)

          ac.sel <- !is.na(ratios.subset$lratio) & 
                      ratios.subset$class1 == class1 & 
                      ratios.subset$class2 == class2

          if (!any(ac.sel)) { ## no data for AC and classes
            return(data.frame(lratio=NA,variance=NA,n.spectra=0,n.pos=0,n.neg=0,
                              p.value.rat=1,p.value.sample=1,p.value=1,is.significant=FALSE,r1=class1,r2=class2,
                              class1=class1,class2=class2,stringsAsFactors=FALSE))
          }
          
          n.pos <- sum(ratios.subset$lratio[ac.sel]>mean.r,na.rm=T)
          n.neg <- sum(ratios.subset$lratio[ac.sel]<mean.r,na.rm=T)
          is.pos <- (n.pos > n.neg && n.neg <= orient.div)
          is.neg <- (n.neg > n.pos && n.pos <= orient.div)

          ## ratio summarization
          good.sel <- ac.sel
          if (!strict.ratio.pval) {
            ## take only positive/negative spectra for ratio summarization
            if (is.pos) 
              good.sel <- good.sel & ratios.subset$lratio>mean.r
            if (is.neg)
              good.sel <- good.sel & ratios.subset$lratio<mean.r
          }

          ac.ratios <- ratios.subset$lratio[good.sel]
          ac.vars <- ratios.subset$variance[good.sel]
          
          sample.var <- weightedVariance(ac.ratios,weights=1/ac.vars)
          estim.var <- 1/sum(1/ac.vars)
          
          variance <- switch(variance.function,
                             maxi = max(estim.var,sample.var,na.rm=T),
                             ev = estim.var,
                             wsv = sample.var
                             )  
          lratio <- weightedMean(ac.ratios,weights=1/ac.vars)
          
          ## p-value computation
          p.value.rat <- pnorm(lratio,mean=mean.r,sd=sqrt(variance),lower.tail=lratio<mean.r)

          p.value.sample <- calculate.mult.sample.pvalue(
            ac.ratios, ratiodistr, strict.pval=strict.sample.pval, lower.tail=n.neg>n.pos,
            n.possible.val = n.combination.c, n.observed.val = sum(ac.sel))

          p.value <- calcProbXDiffNormals(ratiodistr,mean.r,sqrt(variance),alternative="two-sided")

          min.detect.c <- ifelse(is.matrix(min.detect),min.detect[class1,class2],min.detect)
          ## significance
          is.significant <- (p.value <= sign.level) &&
            (sum(ac.sel) >= min.detect.c) &&
            (is.pos | is.neg)

          return(data.frame(lratio=lratio,variance=variance,
                            n.spectra=min(ratios.subset$n.spectra[good.sel],na.rm=TRUE),n.pos=n.pos,n.neg=n.neg,
                            p.value.rat=p.value.rat,p.value.sample=p.value.sample,p.value=p.value,
                            is.significant=is.significant,r1=class1,r2=class2,
                            class1=class1,class2=class2,stringsAsFactors=FALSE))
        })
      },.parallel=isTRUE(getOption('isobar.parallel')))
      
      if (is.null(result)) stop("Error summarizing.")

      attributes(result) = c(attributes(result),
                             list(summarize=TRUE,
                                  by.column=by.column,
                                  min.detect=min.detect,
                                  orient.div=orient.div))

      attributes(result) = c(attributes(result),
                             attributes(ratios)[!names(attributes(ratios)) %in% names(attributes(result))])
      return(result)
      
    } else if (summarize.method=="mean") {
      stop("summarize method mean not anymore available.")
#      do.call(rbind,lapply(unique(ratios$ac),function(ac) {
#        ac.sel <- (ratios$ac==ac) & !is.na(ratios$lratio)
#        if (!any(ac.sel)) {
#            return(data.frame(ac=ac,lratio=NA,stringsAsFactors=FALSE))
#        }
#        return(data.frame(ac=ac,lratio=mean(ratios$lratio[ac.sel]),stringsAsFactors=FALSE))
#      }))

    } else {
      stop (paste("summarize method",summarize.method,"not implemented."))
    }
  }


.calc.zscore <- function(lratio) {
  s.median <- median(lratio,na.rm=TRUE)
  s.mad <- mad(lratio,na.rm=TRUE)
  (lratio-s.median)/s.mad
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### weightedVariance and weightedMean generics and functions
### replace by Hmisc functions?

setGeneric("weightedMean", function(data, weights,trim=0) standardGeneric("weightedMean"))
setGeneric("weightedVariance", function(data, weights, mean.estimate,trim=0) standardGeneric("weightedVariance"))

setMethod("weightedVariance",
    signature(data = "numeric", weights = "numeric", mean.estimate = "missing"),
    function(data, weights, trim=0) {
      mean.estimate <- weightedMean(data,weights,trim=trim)
      weightedVariance(data,weights=weights,mean.estimate=mean.estimate,trim=trim)
    }
)

setMethod("weightedVariance",
    signature(data = "numeric", weights = "numeric", mean.estimate = "numeric"),
    function(data, weights, mean.estimate, trim=0) {
      # Validation? if (length(data)==1) { return(0); }
      # Should we rescale the weights? weights=1/sum(weights)
      
      # weighted sample variance - for not normally distributed data
      # see http://www.gnu.org/software/gsl/manual/html_node/Weighted-Samples.html and
      # http://en.wikipedia.org/wiki/Weighted_mean#Weighted_sample_variance
      sel <- !is.na(data) & !is.na(weights)
      weights <- weights[sel]
      data <- data[sel]

      if (trim < 0 || trim > 0.5)
        stop("trim has to be between 0 and 0,5")

      if (trim > 0) {
        sel <- data > quantile(data,trim) & data < quantile(data,1-trim)
        weights <- weights[sel]
        data <- data[sel]
      }
      
      V1 <- sum(weights)
      V2 <- sum(weights**2)
      variance <- ( V1 / (V1**2 - V2) ) * sum(weights*(data-mean.estimate)**2)

      n = length(data)
      wik = (data-mean.estimate)**2
      w_bar_k = mean((data-mean.estimate)**2)
      var_hat_vk = median((wik-w_bar_k)**2)  ## for shrinkage..
      
      return(c(variance,var_hat_vk))
    }
)

setMethod("weightedMean",
    signature(data = "numeric", weights = "numeric"),
    function(data, weights, trim = 0) {
      sel <- !is.na(data) & !is.na(weights)
      weights <- weights[sel]
      data <- data[sel]

      if (trim < 0 | trim > 0.5)
        stop("trim has to be between 0 and 0,5")

      if (trim > 0) {
        sel <- data > quantile(data,trim) & data < quantile(data,1-trim)
        weights <- weights[sel]
        data <- data[sel]
      }
 
      return(
          sum(data * weights) / sum(weights)
      )
    })

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### plotting methods.
###

setGeneric("reporterMassPrecision", function(x,plot,...)
           standardGeneric("reporterMassPrecision"))
setGeneric("maplot",function(x,channel1,channel2,...) standardGeneric("maplot"))
setGeneric("plotRatio",function(x,channel1,channel2,protein,...)
           standardGeneric("plotRatio"))
setGeneric("protGgdata",function(x,relative.to,protein,...)
           standardGeneric("protGgdata"))
setGeneric("raplot",function(x,...) standardGeneric("raplot"))

setMethod("reporterMassPrecision",
          signature=c(x="IBSpectra",plot="missing"),
          function(x) reporterMassPrecision(x,TRUE))


setGeneric("reporterIntensityPlot",function(x) standardGeneric("reporterIntensityPlot"))

# Calculates and displays the deviation from the 'true' tag mass 
# - as specified in the IBSpectra object - of each channel.
setMethod("reporterMassPrecision", signature=c(x="IBSpectra",plot="logical"),
    function(x,plot=TRUE,nrow=NULL) {
      masses <- reporterMasses(x)
      melt.masses <- data.frame(reporter=factor(rep(colnames(masses),each=nrow(masses))),
                                reporter.tag.mass=rep(reporterTagMasses(x),each=nrow(masses)),
                                observed.moz=as.numeric(masses),stringsAsFactors=FALSE)
      melt.masses$mass.difference <-
          melt.masses$observed.moz - rep(reporterTagMasses(x),each=nrow(masses))
      if (plot) {
        melt.masses$reporter <-
          factor(melt.masses$reporter,
                 levels=reporterTagNames(x),
                 labels=sprintf("%s: m/z %.2f",
                   reporterTagNames(x),reporterTagMasses(x)))
       
       if (is.null(nrow))
         nrow <- ceiling(ncol(masses)/5)

       ggplot(melt.masses,aes_string(x="mass.difference")) + geom_vline(xintercept=0,alpha=0.8) +
          geom_histogram(fill="white",aes_string(colour="reporter"),alpha=0.8,
                         binwidth=1/20*(max(melt.masses$mass.difference,na.rm=TRUE)-min(melt.masses$mass.difference,na.rm=TRUE))) + 
            facet_wrap(~reporter,scales="fixed",nrow=nrow) + 
            theme_bw(base_size=10) + xlab("mass difference theoretical vs observed reporter tag mass") +
              .gg_theme(legend.position="none",
                   axis.text.x = .gg_element_text(angle=330,hjust=0,vjust=1,colour="grey50",size=7))
      } else {
        res <- ddply(melt.masses,'reporter',function(x) {
          c('true reporter mass'=x$reporter.tag.mass[1],
            'number of spectra'=sum(!is.na(x$observed.moz)),
            'mean of observed m/z'=mean(x$observed.moz,na.rm=TRUE),
            'sd of obseved m/z'=sd(x$observed.moz,na.rm=TRUE),
            'mean of mass difference * 10^3'=mean(x$mass.difference*10^3,na.rm=TRUE),
            'sd of mass difference * 10^3'=sd(x$mass.difference,na.rm=TRUE))
        })
        rownames(res) <- res$reporter
        res$reporter <- NULL
        res
      }
    }
)

.reshapeLong <- function(x,key="key",value="value") {
  res <- data.frame(key=rep(colnames(x),each=nrow(x)),
                    value=as.numeric(x),
                    stringsAsFactors=FALSE)
  colnames(res) <- c(key,value)
  res
}

setMethod("reporterIntensityPlot",
          signature=c(x="IBSpectra"),function(x) {
            intensities <- reporterIntensities(x)
            intensities.nn <- reporterData(x,element="ions_not_normalized") # null if not normalized
            
            melt.intensities <- data.frame(tag=rep(colnames(intensities),each=nrow(intensities)),
                       normalized=factor(ifelse(is.null(intensities.nn),"","2. after normalization")),
                       intensity=as.numeric(intensities),
                       stringsAsFactors=FALSE)

            if (!is.null(intensities.nn)) {
            	melt.intensities <- rbind(melt.intensities,
              	  data.frame(tag=rep(colnames(intensities.nn),each=nrow(intensities.nn)),
                             normalized="1. before normalization",
                             intensity=as.numeric(intensities.nn),
                             stringsAsFactors=FALSE))
            }
            
            ggplot(melt.intensities,aes_string(x="tag",y="intensity")) +
              geom_boxplot(aes_string(color="normalized"),size=0.5,alpha=0.6,
                           outlier.size=0.5,position=position_dodge(width=0.25)) + 
              xlab("isobaric reporter tag") +
              scale_y_log10() + theme_bw(base_size=10) + scale_color_hue("") +
              .gg_theme(axis.text.x=.gg_element_text(angle=90, hjust=1))
          }
)

# Ratio-Absolute intensity plot - will be deprecated by maplot
setMethod("raplot",signature(x="IBSpectra"),
    function(x,...) {
      ions <- reporterIntensities(x,na.rm=T)
      par(mfrow=c(ncol(ions)/2,ncol(ions)/2),mar=c(3,2,1,1))
      sums = rowSums(ions)
      for (i in colnames(ions)) {
        plot(sums,ions[,i]/sums*4,log="xy",...)
      }
    }
)

# plots ratio of one protein
setMethod("plotRatio",
    signature(x="IBSpectra",channel1="character",channel2="character",protein="character"),
    function(x,channel1,channel2,protein,noise.model=NULL,...) {
      ions <- reporterIntensities(x,na.rm=TRUE)
      R=log10(ions[,channel1])
      G=log10(ions[,channel2])
      M <- R - G
      A <- 0.5*(R + G)
      
      M <- M[order(A)]
      A <- sort(A)
      plot(A,M,pch=pch,...)
      #abline(h=0,col="red")
      if (!missing(noise.model) & !is.null(noise.model)) {
        if (is(noise.model,"NoiseModel"))
          noise.model <- c(noise.model)
        for (i in seq_along(noise.model)) {          
          lines(A,1.96*sqrt(variance(noise.model[[i]],A)),col=noise.model.col[i])
          lines(A,-1.96*sqrt(variance(noise.model[[i]],A)),col=noise.model.col[i])
        }
      }
      
      unspecific.spectra.sel    <- names(A) %in% spectrumSel(x,protein=protein,specificity=UNSPECIFIC)
      groupspecific.spectra.sel <- names(A) %in% spectrumSel(x,protein=protein,specificity=GROUPSPECIFIC)
      specific.spectra.sel      <- names(A) %in% spectrumSel(x,protein=protein,specificity=REPORTERSPECIFIC)
        
      points(A[unspecific.spectra.sel],M[unspecific.spectra.sel],pch=pch.p,col="yellow",...)
      points(A[groupspecific.spectra.sel],M[groupspecific.spectra.sel],pch=pch.p,col="orange",...)
      points(A[specific.spectra.sel],M[specific.spectra.sel],pch=pch.p,col="red",...)
        
    }
)

maplot.protein <- function(x,relative.to,protein,noise.model=NULL,
        channels=NULL,xlim=NULL,ylim=NULL,identify=FALSE,add=FALSE,pchs=NULL,log="xy",
        legend.pos="topright",names=NULL,legend.cex=0.8,cols=pchs,ltys=1,
        main=protein,xlab=NULL,ylab=NULL,type="ma",show.lm=FALSE,...) {
 	  if (is(x,"IBSpectra"))
      x <- c(x)
   
    if (!relative.to %in% reporterTagNames(x[[1]]))
      stop("relative.to must be one of the reporter tag names (",paste(reporterTagNames(x[[1]]),collapse=","),")")

    i.df <- data.frame()
    legend <- list(text=c(),lty=c(),pch=c())
    i <- 1
    if (is.null(channels)) channels <- setdiff(reporterTagNames(x[[1]]),relative.to)
    if (is.null(pchs)) pchs <- seq_along(channels)

    if (is.null(xlab) && type=="ma") xlab <- "average intensity"
    if (is.null(ylab) && type=="ma") ylab <- "ratio"

    if (is.null(xlab) && type!="ma") xlab <- paste("intensity",relative.to)
    if (is.null(ylab) && type!="ma") ylab <- paste("intensity",paste(channels,collapse=", "))
    

    for (ib in x) {
      ions <- reporterIntensities(ib,na.rm=FALSE,protein=protein)
      if (length(ions) == 0 || all(is.na(ions)))
        next;

      if (is.null(xlim))
        xlim <- range(ions,na.rm=TRUE)
      if (is.null(ylim) && type!="ma") ylim <- xlim
      if (any(!is.finite(xlim))) xlim  <- NULL
      if (type == "ma")
        ions[which(is.na(ions))] <- 0

      for (channel_i in seq_along(channels)) {
        channel <- channels[channel_i]
        channel.rt <- ifelse(length(relative.to) == 1,relative.to,relative.to[channel_i])
        if (length(ions[,channel])>0) {
          if (is.null(names))
            legend$text <- c(legend$text,sprintf("%s/%s",channel,channel.rt))
          else
            legend$text <- c(legend$text,names[i])
          legend$pch <- c(legend$pch,pchs[i])
          legend$lty <- c(legend$lty,ltys[i])
        }

        if (type=="ma") {
          div <- ions[,channel]/ions[,channel.rt]
          avg <- (ions[,channel]+ions[,channel.rt])/2
          div[div == Inf] = 100
          div[div == -Inf] = 0.01
          div[div > 100] = 90
          div[div < 0.01] = 1/90
          x <- avg
          y <- div
        } else {
          x <- ions[,channel.rt]
          y <- ions[,channel]
        }

        if (channel_i == 1 && add == FALSE) 
          plot(x,y,xlim=xlim,ylim=ylim,pch=pchs[i],log=log,main=main,col=cols[i],
               xlab=xlab,ylab=ylab,...)
        else
          points(x,y,pch=pchs[i],col=cols[i],...)

        add = TRUE
          if (identify)
            i.df <- rbind(i.df,data.frame(x=x,y=y,spectrum=rownames(ions),
                                          peptide=peptides(x=ib,spectrum=rownames(ions)),stringsAsFactors=T))

        if (!is.null(noise.model)) {
          ratio <- estimateRatio(ib,noise.model,channel.rt,channel,protein=protein)
          ratio.lm <- estimateRatio(ib,noise.model,channel.rt,channel,protein=protein,method="lm")
          #abline(h=10^ratio[1],lty=ltys[i],col=cols[i],...)

          if (type=="ma") {
            abline(h=10^ratio[1],lty=1,col=cols[i],...)
            if (show.lm)
              abline(h=ratio.lm['ratio'],lty=2,col=cols[i],...)
          } else {
            #abline(0,10^ratio[1],lty=1,col=cols[i],...)
            abline(0,10^ratio[1],lty=1,col=cols[i],untf=TRUE,...)
            if (show.lm)
              abline(0,ratio.lm['ratio'],lty=2,col=cols[i],untf=TRUE,...)
          }

          if (!is.na(ratio[1]))
            legend$text[length(legend$text)] <- 
              sprintf("%s = %.2f +/- %.2f",legend$text[length(legend$text)],10^ratio[1],10^sqrt(ratio[2]))
          #abline(h=10^(ratio[1]+1.96*sqrt(ratio[2])),lty=channel_i,lwd=0.5)
          #abline(h=10^(ratio[1]-1.96*sqrt(ratio[2])),lty=channel_i,lwd=0.5)
        }
        i <- i + 1
      }
    }
    if (length(legend$text) > 0 ) {
    legend(legend.pos,legend=legend$text,pch=legend$pch,lty=legend$lty,cex=legend.cex,col=cols)
    if (identify) {
      identify(x=i.df$x,y=i.df$y,labels=i.df$peptide)
    }}
}

.get.ri <- function(ri,ch) {
  if (ch == "ALL")
    rowSums(ri,na.rm=TRUE)
  else if (ch == "AVG")
    apply(ri,1,mean,na.rm=TRUE)
  else
    ri[,ch]
}

# MA plots
setMethod("maplot",
          signature(x="IBSpectra",channel1="character",channel2="character"),
          function(x,channel1,channel2,noise.model=NULL,colorize.protein=NULL,
              h=NULL,v=NULL,col.h="green",col.v="green",
              ylab=paste("ratio channel",channel1,"vs",channel2),
              pch=".",noise.model.col=c("#E41A1C","#377EB8","#4DAF4A","#984EA3","#FF7F00","#FFFF33","#A65628","#F781BF","#999999"),
              pch.p=1,protein=NULL,peptide=NULL,smooth=FALSE,
              set.na.to=NULL,set.na.to.lim=NULL,na.rm=is.null(set.na.to),
              identify=FALSE,identify.column="spectrum",
              x.axis.labels=TRUE,y.axis.labels=TRUE,
              col="black",...) {
            if (is.null(protein)) 
              ions <- reporterIntensities(x,na.rm=FALSE)
            else
              ions <- reporterIntensities(x,na.rm=FALSE,protein=protein)

            fd <- fData(x)
            
            # remove points which are NA in both channels
            if (channel2=="ALL" | channel1=="ALL")
              sel <- apply(is.na(ions),1,all)
            else 
              sel <- is.na(ions[,channel1]) & is.na(ions[,channel2])
            ions <- ions[!sel,,drop=FALSE]
            fd <- fd[!sel,,drop=FALSE]
            
            if (na.rm) {
              if (channel2=="ALL" | channel1=="ALL")
                sel <- apply(is.na(ions),1,any)
              else 
                sel <- is.na(ions[,channel1])| is.na(ions[,channel2])
              if (any(sel))
                warning(sprintf("removing %s NA points",sum(sel)))
              ions <- ions[!sel,,drop=FALSE]
              fd <- fd[!sel,,drop=FALSE]
            }
            
            if (channel1=="ALL" && channel2=="ALL")
              stop("Cannot set both channels to ALL")

            if (channel1=="ALL") {
              R=as.numeric(ions[,setdiff(colnames(ions),channel2)])
              G=rep(ions[,channel2],ncol(ions)-1)
            } else if (channel2=="ALL") {
              R=rep(ions[,channel1],ncol(ions)-1)
              G=as.numeric(ions[,setdiff(colnames(ions),channel1)])
            } else {
              R=ions[,channel1]
              G=ions[,channel2]
            }
            
            M <- R/G
            A <- log10(sqrt(R*G))
            

            if (isTRUE(set.na.to)) {
              set.na.to <- ceiling(10^(max(abs(log10(M)), na.rm = TRUE)+0.2))
              set.na.to.lim <- ceiling(10^(max(abs(log10(M)), na.rm = TRUE)+0.1))
            }
            if (!is.null(set.na.to)) {
              sel <- is.na(R);
              M[sel] <- set.na.to + 1;
              A[sel] <- log10(G[sel])
              sel <- is.na(G); 
              M[sel] <- 1/set.na.to
              A[sel] <- log10(R[sel])
            }

            if (length(M)==0) {
              plot(1,1,cex=0,xaxt="n",yaxt="n",bty="n",xlab="",ylab="")
              text(1,1,"no datapoints")
              return()
            }
            
            if (smooth) {
#              smoothScatter(A,M,log="y",...)
              stop("option smooth deprecated!")
            } else {
              plot(A,M,pch=pch,log="y",
                  xlab=expression(paste(log[10]," average intensity")),
                  ylab=ylab,col=col,axes=FALSE,...)
              axis(side=1,labels=x.axis.labels)

              if (!is.null(set.na.to)) {
                y.axis <- axTicks(2,log=TRUE)
                y.axis <- y.axis[y.axis >= 1/set.na.to.lim & y.axis <= set.na.to.lim]
                y.axis <- y.axis[seq(from=2,to=length(y.axis)-1)]
                if (isTRUE(y.axis.labels))
                  y.axis.labels <- c(-Inf,y.axis,Inf)
                axis(side=2,at=c(1/set.na.to.lim,y.axis,set.na.to.lim),
                     labels= y.axis.labels,las=2)
                abline(h = y.axis, col = "#000000F0",lwd=0.15)
                abline(h = c(1/set.na.to.lim,set.na.to.lim), col = "blue", lty = 3)
              } else {
                axis(side=2,labels=y.axis.labels)
              }

            }
            if (!is.null(h)) {
              sel <- M > h | M < 1/h
              points(A[sel],M[sel],col=col.h,pch=pch,...)
            }
            if (!is.null(v)) {
              sel <- A < v
              points(A[sel],M[sel],col=col.v,pch=pch,...)
            }
            
            #abline(h=0,col="red")
            if (!missing(noise.model) & !is.null(noise.model)) {
              if (is(noise.model,"NoiseModel"))
                noise.model <- c(noise.model)

              ss <- seq(min(A,na.rm=TRUE),max(A,na.rm=TRUE),length.out=100)
              for (i in seq_along(noise.model)) {          
                lines(ss,10^(1.96*sqrt(variance(noise.model[[i]],ss))),col=noise.model.col[i],lwd=2)
                lines(ss,10^(-1.96*sqrt(variance(noise.model[[i]],ss))),col=noise.model.col[i],lwd=2)
              }
            }
            
            if (!is.null(colorize.protein)) {
              if (is.list(colorize.protein)) {
                cp.legend <- c()
                cp.legend.col <- c()
                for (i in names(colorize.protein)) {
                   data <- colorize.protein[[i]]
                   protein <- data$protein
                   p.col <- data$col
                   p.us.col <- data$us.col
                   cp.legend <- c(cp.legend,i)
                   cp.legend.col <- c(cp.legend.col,p.col)

                   protein.sel <- names(A) %in% spectrumSel(x,protein=protein,
                                                            specificity=c(GROUPSPECIFIC,REPORTERSPECIFIC),
                                                            spectrum.titles=TRUE)

                   protein.us.sel <- names(A) %in% spectrumSel(x,protein=protein,specificity=UNSPECIFIC,
                                                                spectrum.titles=TRUE)


                   points(A[protein.sel],M[protein.sel],pch=pch.p,col=p.col,...)
                   points(A[protein.us.sel],M[protein.us.sel],pch=pch.p,col=p.us.col,...)
                }
                legend("topright",legend=cp.legend,col=cp.legend.col,pch=pch.p)
              } else {
                unspecific.spectra.sel    <- names(A) %in% spectrumSel(x,protein=colorize.protein,specificity=UNSPECIFIC,spectrum.titles=T)
                groupspecific.spectra.sel <- names(A) %in% spectrumSel(x,protein=colorize.protein,specificity=GROUPSPECIFIC,spectrum.titles=T)
                specific.spectra.sel      <- names(A) %in% spectrumSel(x,protein=colorize.protein,specificity=REPORTERSPECIFIC,spectrum.titles=T)
                other.sel <- !unspecific.spectra.sel & !groupspecific.spectra.sel & !specific.spectra.sel
              
                points(A[unspecific.spectra.sel],M[unspecific.spectra.sel],pch=pch.p,col="yellow",...)
                points(A[groupspecific.spectra.sel],M[groupspecific.spectra.sel],pch=pch.p,col="orange",...)
                points(A[specific.spectra.sel],M[specific.spectra.sel],pch=pch.p,col="green",...)
  #points(A[other.sel],M[other.sel],pch=".",col="blue",...)
              }
              
            }
            if (identify)
              identify(x=A,y=M,labels=fd[,identify.column])
          }
)






setMethod("maplot",
    signature=c(x="missing",channel1="numeric",channel2="numeric"),
    function(channel1,channel2,noise.model=NULL,pch=".",noise.model.col=1:10,...) {
      sel <- !is.na(channel1) & !is.na(channel2)
      channel1 <- channel1[sel]
      channel2 <- channel2[sel]

      if (length(channel1)==0) {
        plot(1,1,cex=0,xaxt="n",yaxt="n",bty="n",xlab="",ylab="")
        text(1,1,"no datapoints")
        return()
      }
      
      M <- log10(channel1) - log10(channel2)
      A <- 0.5*(log10(channel1) + log10(channel2))
      M <- M[order(A)]
      ch1 <- log10(channel1)[order(A)]
      ch2 <- log10(channel2)[order(A)]
      A <- sort(A)
      plot(A,M,pch=pch,xlab=expression(paste(log[10]," average intensity")),
          ylab="ratio",...)
      #abline(h=0,col="red")
      if (!missing(noise.model) & !is.null(noise.model)) {
        if (is(noise.model,"NoiseModel"))
          noise.model <- c(noise.model)
        ss <- seq(min(A),max(A),by=0.05)
        for (i in seq_along(noise.model)) {          
          lines(ss,1.96*sqrt(variance(noise.model[[i]],ss)),col=noise.model.col[i],lwd=1.5)
          lines(ss,-1.96*sqrt(variance(noise.model[[i]],ss)),col=noise.model.col[i],lwd=1.5)
          #lines(A,1.96*sqrt(variance(noise.model[[i]],channel1,channel2)),col=noise.model.col[i])
          #lines(A,-1.96*sqrt(variance(noise.model[[i]],channel1,channel2)),col=noise.model.col[i])
        }
      }
    }
)

setMethod("maplot",
          signature(x="IBSpectra",channel1="missing",channel2="missing"),
          function(x,noise.model=NULL,pairs=TRUE,xlim="fixed",ylim="fixed",...){
            ions <- reporterIntensities(x)
              set.na.to <- NULL
              set.na.to.lim <- NULL
              if (identical(xlim,"fixed")) 
                xlim <- log10(range(ions,na.rm=T)) 
              if (identical(ylim,"fixed")) {
                y.max <- max(apply(ions,1,function(x) {
                                   y <- x[!is.na(x)]
                                   if (length(y) < 2) NA
                                   else max(y)/min(y)
                                   }),na.rm=T)
                set.na.to <- ceiling(10^(log10(y.max)+0.2))
                set.na.to.lim <- ceiling(10^(log10(y.max)+0.1))
                y.max <- ceiling(10^(log10(y.max)+0.3))
                ylim <- c(1/y.max,y.max)

              }

            if (pairs) {
              histlimits=log10(range(ions,na.rm=TRUE))
              par(mfrow=c(ncol(ions),ncol(ions)),mar=c(2.5,2.25,0,0),
                  bty="n",fg="darkgray")
              for (i in colnames(ions)) {
                for (j in colnames(ions)) {
                  if (i == j) { 
                    if (all(is.na(ions[,i])))
                      plot(histlimits,c(1,1),type="n",bty="n", xaxt="n", yaxt="n", main="",ylim=c(0,1))
                    else
                      hist(log10(ions[,i]), col="#EEEEEE", freq=FALSE, 
                           xlim=histlimits,cex=0.5,breaks=20,
                           bty="n", xaxt="n", yaxt="n", main="",ylim=c(0,1))
                    text(sum(histlimits)/2,0.5,i,col="black",cex=1.2,font=2); 
                    #text(sum(histlimits)/2,0.4,sprintf("n = %s",sum(!is.na(ions[,i]))),col="black",cex=1,font=2); 
                    #legend("center",
                    #       legend=c(sprintf("%3s",i),
                    #                sprintf("n = %s",sum(!is.na(ions[,i])))),
                    #       bty="n",text.col="black",
                    #       cex=1,text.font=c(2,1))

                  } else if (i > j) {
                    plot.labels.x = is.null(xlim) || (i == tail(colnames(ions),1) && j == head(colnames(ions),1))
                    plot.labels.y = is.null(ylim) || (i == tail(colnames(ions),1) && j == head(colnames(ions),1))
                    maplot(x=x,channel1=as.character(i),channel2=as.character(j),
                           noise.model=noise.model,xlim=xlim,ylim=ylim,
                           set.na.to=set.na.to,set.na.to.lim=set.na.to.lim,
                           x.axis.labels = plot.labels.x,
                           y.axis.labels = plot.labels.y,
                           ...)
                    
                  }
                  else if (i < j) {
                    plot(c(0,1),c(0,1),type="n",ylab="",xlab="",xaxt="n",yaxt="n",bty="n")
                  }
                        
                }
              }
            } else {
              par(mfrow=c(1,ncol(ions)),mar=c(2.5,2.25,2.25,0),
                  bty="n",fg="darkgray")
              for (i in colnames(ions)) {
                plot.labels.y = is.null(ylim) || (i == colnames(ions)[1])
                maplot(x=x,channel1=as.character(i),channel2="ALL",
                       main=substitute(paste(i," ",italic(vs)," ALL",sep="")),
                       noise.model=noise.model,xlim=xlim,ylim=ylim,
                       set.na.to=set.na.to,set.na.to.lim=set.na.to.lim,
                       x.axis.labels = TRUE,
                       y.axis.labels = plot.labels.y,
                       ...)

              }
            }
          }
)



# Should protGgdata be deprecated?
setMethod("protGgdata",
    signature(x="ANY",relative.to="character",protein="character"),
    function(x,relative.to,protein,noise.model=NULL,
        channels=setdiff(reporterTagNames(x),relative.to),
        names=NULL,legend.cex=0.8,...) {
      
      dfs <- data.frame()
      i <- 1
      for (ib in x) {
            ions <- reporterIntensities(ib,na.rm=FALSE,protein=protein)
            ions[which(is.na(ions))] <- 0
            
            for (channel_i in seq_along(channels)) {
              channel <- channels[channel_i]
              channel.rt <- ifelse(length(relative.to) == 1,
                                   relative.to,relative.to[channel_i])
              
              div <- ions[,channel]/ions[,channel.rt]
              avg <- (ions[,channel]+ions[,channel.rt])/2
              
              if (length(div)>0 & sum(avg,na.rm=T) > 0) {
                if (!is.null(noise.model)) {
                  ratio <- estimateRatio(ib,noise.model,channel.rt,channel,protein=protein)
                  
                  dfs <-
                    rbind(dfs,data.frame(channel=channel,channel.rt=channel.rt,
                                         average=avg,difference=div,ratio=10^ratio[1],
                                         name=ifelse(is.null(names),
                                          sprintf("%s/%s",channel,channel.rt),names[i]),
                                         lower=10^(ratio[1]-sqrt(ratio[2])),
                                         upper=10^(ratio[1]+sqrt(ratio[2]))))
                  
                }
              }
              i <- i + 1
            }
          }
          return(dfs)
          
    }
)

### R code from vignette source 'isobar.Rnw'

###################################################
### code chunk number 1: init
###################################################
  require(distr)
  require(ggplot2)
  dir.create(file.path("graphics"), showWarnings = FALSE)


###################################################
### code chunk number 2: isobar.Rnw:55-58 (eval = FALSE)
###################################################
## if (!requireNamespace("BiocManager", quietly=TRUE))
##     install.packages("BiocManager")
## BiocManager::install("isobar")


###################################################
### code chunk number 3: isobar.Rnw:67-69 (eval = FALSE)
###################################################
## library(devtools)
## install_github("fbreitwieser/isobar")


###################################################
### code chunk number 4: load-isobar
###################################################
  library(isobar) ## load the isobar package


###################################################
### code chunk number 5: isobar.Rnw:79-80 (eval = FALSE)
###################################################
## packageVersion("isobar")


###################################################
### code chunk number 6: isobar.Rnw:83-84 (eval = FALSE)
###################################################
## packageDescription("isobar")$Version


###################################################
### code chunk number 7: isobar.Rnw:116-126 (eval = FALSE)
###################################################
##   ## generating IBSpectra object from ID.CSV and MGF
##   ib <- readIBSpectra("iTRAQ4plexSpectra",list.files(pattern=".id.csv"),
##           list.files(pattern=".mgf"))
## 
##   ## write in tabular IBSPECTRA.CSV format to file
##   write.table(as.data.frame(ib),sep="\t",row.names=F,
##           file="myexperiment.ibspectra.csv")
## 
##   ## generate from saved IBSPECTRA.CSV - MGF does not have to be supplied
##   ib.2 <- readIBSpectra("iTRAQ4plexSpectra","myexperiment.ibspectra.csv")


###################################################
### code chunk number 8: load-ibspiked
###################################################
  data(ibspiked_set1)
  ceru.human <- protein.g(proteinGroup(ibspiked_set1),"CERU_HUMAN")
  ceru.rat <- protein.g(proteinGroup(ibspiked_set1),"CERU_RAT")
  ceru.mouse <- protein.g(proteinGroup(ibspiked_set1),"CERU_MOUSE")
  ceru.proteins <- c(ceru.human,ceru.rat,ceru.mouse)


###################################################
### code chunk number 9: isobar.Rnw:233-238 (eval = FALSE)
###################################################
## readIBSpectra(...,
##               mapping.file="mapping2.csv",
##               mapping=c(identification.spectrum="id-spectrum-ms2",
##                         quantification.spectrum="quant-spectrum-ms3")
##               )


###################################################
### code chunk number 10: isobar.Rnw:251-258 (eval = FALSE)
###################################################
## readIBSpectra("TMT6plexSpectra",
##               id.file="cid.identifications.csv",
##               peaklist.file="hcd.peaklist.mgf",
##               id.file.domap="hcd.identifications.csv",
##               mapping.file="mapping.csv",
##               ...
##               )


###################################################
### code chunk number 11: isobar.Rnw:275-281 (eval = FALSE)
###################################################
## precursors <- read.delim("file1_precursors.txt",header=TRUE,sep="\t",stringsAsFactors=FALSE)
## idfile <- read.delim("file1.id.csv",header=TRUE,sep="\t",stringsAsFactors=FALSE)
## 
## idfile <- merge(idfile,precursors,
##                 by.x=c("fixed_spectrum","scans.from"),
##                 by.y=c("fixed_spectrum","SCAN"),all.x=TRUE) # FIXME see /work/analysis/gwinter/raw-spectra/analysis.R


###################################################
### code chunk number 12: isobar.Rnw:300-302 (eval = FALSE)
###################################################
##   as(ibspectra,"MSnSet")
##   as(msnset,"IBSpectra")


###################################################
### code chunk number 13: isobar.Rnw:336-338 (eval = FALSE)
###################################################
##   as(ibspectra,"MSnSet")
##   as(msnset,"IBSpectra")


###################################################
### code chunk number 14: show-reporterMasses
###################################################
sprintf("%.4f",reporterTagMasses(ibspiked_set1))  ## expected masses
mass <- assayData(ibspiked_set1)[["mass"]] ## observerd masses
apply(mass,2,function(x) sprintf("%.4f",quantile(x,na.rm=TRUE,probs=c(0.025,0.975))))


###################################################
### code chunk number 15: print-reporterMasses
###################################################
  print(reporterMassPrecision(ibspiked_set1))


###################################################
### code chunk number 16: normalize-ibspiked
###################################################
ib.old <- ibspiked_set1
ibspiked_set1 <- correctIsotopeImpurities(ibspiked_set1)
ibspiked_set1 <- normalize(ibspiked_set1)


###################################################
### code chunk number 17: fig-maplot-normalize1
###################################################
png("graphics/fig_maplot_normalize.png",width=6.6,height=3,units="in",res=300,pointsize=8)


###################################################
### code chunk number 18: fig-maplot-normalize2
###################################################
par(mfrow=c(1,2))
maplot(ib.old,channel1="114",channel2="117",ylim=c(0.5,2),
       main="before normalization")
abline(h=1,col="red",lwd=2)
maplot(ibspiked_set1,channel1="114",channel2="117",ylim=c(0.5,2),
       main="after normalization")
abline(h=1,col="red",lwd=2)


###################################################
### code chunk number 19: fig-maplot-normalize3
###################################################
dev.off()


###################################################
### code chunk number 20: calc-noisemodel
###################################################
ib.background <- subsetIBSpectra(ibspiked_set1,protein=ceru.proteins,direction="exclude")
noise.model <- NoiseModel(ib.background)


###################################################
### code chunk number 21: ceru-noisemodel
###################################################
ib.ceru <- subsetIBSpectra(ibspiked_set1,protein=ceru.proteins,
                            direction="include",
                            specificity="reporter-specific")
nm.ceru <- NoiseModel(ib.ceru,one.to.one=FALSE,pool=TRUE)


###################################################
### code chunk number 22: maplot-noisemodel1
###################################################
png("graphics/fig_maplot_noisemodel.png",width=3.3,height=3,units="in",res=300,pointsize=8)


###################################################
### code chunk number 23: maplot-noisemodel2
###################################################
maplot(ib.background,noise.model=c(noise.model,nm.ceru),
       channel1="114",channel2="115",ylim=c(0.2,5),
       main="95% CI noise model")


###################################################
### code chunk number 24: maplot-noisemodel3
###################################################
dev.off()


###################################################
### code chunk number 25: estimateRatio
###################################################
## Calculate ratio based on all spectra of peptides specific 
##  to CERU_HUMAN, CERU_RAT or CERU_MOUSE. Returns a named
##  numeric vector.
10^estimateRatio(ibspiked_set1,noise.model,
                 channel1="114",channel2="115",
                 protein=ceru.proteins)['lratio']

## If argument 'combine=FALSE', estimateRatio returns a data.frame 
##  with one row per protein
10^estimateRatio(ibspiked_set1,noise.model,
                 channel1="114",channel2="115",
                 protein=ceru.proteins,combine=FALSE)[,'lratio']
## spiked material channel 115 vs 114: 
##                 CERU_HUMAN (P00450): 1:1
##                 CERU_RAT   (P13635): 2:1  = 2
##                 CERU_MOUSE (Q61147): 5:10 = 0.5

## Peptides shared between rat and mouse
pep.shared <- peptides(proteinGroup(ibspiked_set1),
                       c(ceru.rat,ceru.mouse),set="intersect",
                       columns=c('peptide','n.shared.groups'))
## remove those which are shared with other proteins
pep.shared <- pep.shared$peptide[pep.shared$n.shared.groups==2]

## calculate ratio: it is between the rat and mouse ratios
10^estimateRatio(ibspiked_set1,noise.model,
                 channel1="114",channel2="115",
                 peptide=pep.shared)['lratio']


###################################################
### code chunk number 26: protein-ratios
###################################################
protein.ratios <- proteinRatios(ibspiked_set1,noise.model,cl=c("1","0","0","0"))

## defined class 114 and 115 as class 'T', 116 and 117 as class 'C'
classLabels(ibspiked_set1) <- c("T","T","C","C")

proteinRatios(ibspiked_set1,noise.model,protein=ceru.proteins,
              cl=classLabels(ibspiked_set1),combn.method="interclass",
              summarize=T)[,c("ac","lratio","variance")]


###################################################
### code chunk number 27: ratiodistribution
###################################################
#protein.ratios <- proteinRatios(ibspiked_set1,noise.model)
protein.ratiodistr.wn <- fitWeightedNorm(protein.ratios[,'lratio'],
                                         weights=1/protein.ratios[,'variance'])
protein.ratiodistr.cauchy <- fitCauchy(protein.ratios[,"lratio"])


###################################################
### code chunk number 28: plot-ratiodistr
###################################################
library(distr) # required library
limits=seq(from=-0.5,to=0.5,by=0.001)
curve.wn <- data.frame(x=limits,y=d(protein.ratiodistr.wn)(limits))
curve.cauchy<-data.frame(x=limits,y=d(protein.ratiodistr.cauchy)(limits))

g <- ggplot(data.frame(protein.ratios),aes(x=lratio)) +
  geom_histogram(colour = "darkgreen", fill = "white",aes(y=..density..),
                 binwidth=0.02) + geom_rug() +
  geom_line(data=curve.wn,aes(x=x,y=y),colour="blue") +
  geom_line(data=curve.cauchy,aes(x=x,y=y),colour="red")
print(g)


###################################################
### code chunk number 29: significant-ratios
###################################################
rat.list <-
  estimateRatio(ibspiked_set1,noise.model=noise.model,channel1="114",channel2="115",
                protein=reporterProteins(proteinGroup(ibspiked_set1)),combine=F,
                ratiodistr=protein.ratiodistr.cauchy)
rat.list[rat.list[,"is.significant"]==1,]


###################################################
### code chunk number 30: shared-pep
###################################################
## peptides shared between CERU_RAT and CERU_MOUSE have been computed before
pep.shared
## peptides specific to CERU_RAT
pep.rat <- peptides(proteinGroup(ibspiked_set1),protein=ceru.rat,
                    specificity="reporter-specific")

## create an IBSpectra object with only CERU_RAT and shared peptides
ib.subset <- subsetIBSpectra(ibspiked_set1,
                             peptide=c(pep.rat,pep.shared),direction="include")

## calculate shared ratios
sr <- shared.ratios(ib.subset,noise.model,
                    channel1="114",channel2="117",
                    ratiodistr=protein.ratiodistr.cauchy)
sr



###################################################
### code chunk number 31: isobar.Rnw:631-634
###################################################
  ## plot significantly different protein groups where 90% CI does not overlap
  ## CERU_MOUSE and CERU_RAT is detected, as expected.
  shared.ratios.sign(sr,z.shared=1.282)


###################################################
### code chunk number 32: isobar.Rnw:658-660 (eval = FALSE)
###################################################
## create.reports(type="iTRAQ4plexSpectra",
##                identifications="my.id.csv",peaklist="my.mgf")


###################################################
### code chunk number 33: isobar.Rnw:684-687 (eval = FALSE)
###################################################
## ## execute to find the path and file location in your installation.
## system.file("report",package="isobar") ## path
## list.files(system.file("report",package="isobar")) ## files


###################################################
### code chunk number 34: isobar.Rnw:761-772
###################################################
#p <- readLines(system.file("report","properties.R",package="isobar"))
#p <- sub("(#.*)","\\\\textcolor{comment}{\\1}",p)
#for (i in c("TRUE","FALSE","NULL"))
#  p <- sub(i,paste("\\\\textcolor{special}{",i,"}",sep=""),p)
#p <- sub('(".*?")',"\\\\textcolor{string}{\\1}",p)
#p <- gsub("#","\\#",p,fixed=TRUE)
#p <- gsub("_","\\_",p,fixed=TRUE)
#cat("\\begin{lstlisting}")
#cat(p,sep="\n")
#cat("\\end{lstlisting}")
cat(sprintf("\\lstinputlisting{%s}\n",system.file("report","properties.R",package="isobar")))


###################################################
### code chunk number 35: isobar.Rnw:793-796 (eval = FALSE)
###################################################
## ## execute to find the path and file location in your installation.
## system.file("pl",package="isobar") ## path
## list.files(system.file("pl",package="isobar")) ## files


###################################################
### code chunk number 36: isobar.Rnw:804-810 (eval = FALSE)
###################################################
## ## execute on your system
## system(paste("perl",system.file("pl","mascotParser2.pl",package="isobar"),
##             "--help"))
## 
## print(paste("perl",system.file("pl","pidresParser2.pl",package="isobar"),
##             "--help"))


###################################################
### code chunk number 37: isobar.Rnw:818-819
###################################################
  toLatex(sessionInfo())

testPdflatex <- function(pdflatex.cmd) {
  if (system(pdflatex.cmd) != 0) 
    stop("pdflatex does not seem to be installed. ",
         "Install LaTeX using the TeXLive or MiKTex distribution to generate ",
         "PDF reports with isobar. Set 'write.qc.report=FALSE' and 'write.report=FALSE', otherwise.")
}
    

create.reports <- function(properties.file="properties.R",
                           global.properties.file=system.file("report","properties.R",package="isobar"),
                           args=NULL,...,recreate.properties.env=TRUE,recreate.report.env=TRUE) {
  ow <- options("warn")
  if (!exists("properties.env") || recreate.properties.env) {
    properties.env <- load.properties(properties.file,
                                      global.properties.file,
                                      args=args,...)
    assign("properties.env",properties.env,envir=.GlobalEnv)
  }
  options(warn=properties.env[["warning.level"]])

  if (!exists("report.env") || recreate.report.env) {
    report.env <- .GlobalEnv
    initialize.env(report.env,properties.env)
  }

  zip.files <- c(properties.file)

  ## generate XLS report
  if(property('write.xls.report',properties.env)) {
    message("Writing isobar-analysis.xls")
    write.xls.report(properties.env,report.env)
    zip.files <- c(zip.files,"isobar-analysis.xls")
  }
 
  ## generate Latex/Sweave report
  if(property('write.qc.report',properties.env)) {
    message("Weaving isobar-qc report")
    Sweave(system.file("report","isobar-qc.Rnw",package="isobar"))
    if (property('use.name.for.report',properties.env)) {
      qc.name <- .sanitize.sh(sprintf("%s.qc",property('name',properties.env)))
    	file.rename("isobar-qc.tex",sprintf("%s.tex",qc.name))
    } else {
        qc.name <- "isobar-qc"
    }

    zip.files <- c(zip.files,sprintf("%s.tex",qc.name))
    if (properties.env[["compile"]]) 
      zip.files <- .compile.tex(qc.name,zip.files)
  }

  if(property('write.report',properties.env) && properties.env[["report.level"]] != 'peptide') {
    message("Weaving isobar-analysis report")
    name <- switch(properties.env[["report.level"]],
                   protein="isobar-analysis",
                   peptide="isobar-peptide-analysis",
                   stop(properties.env[["report.level"]]," report.level not known",
                        " - choose protein or peptide"))
    Sweave(system.file("report",paste(name,".Rnw",sep=""),package="isobar"))

    if (property('use.name.for.report',properties.env)) {
    	tex.name <- sprintf("%s.tex",name)
      name <- .sanitize.sh(sprintf("%s.quant",property('name',properties.env)))
    	file.rename(tex.name,sprintf("%s.tex",name))
    } else {
    }


    zip.files <- c(zip.files,sprintf("%s.tex",name))
    if (properties.env[["compile"]])
      zip.files <- .compile.tex(name,zip.files)
  }

  if (properties.env[["zip"]]) {
    zip.f <- sprintf("%s.zip",property('name',properties.env))
    zip(zip.f,zip.files)
    message("Created zip archive ",zip.f)
  }

  options(ow) 
  message("\nSUCCESSFULLY CREATED REPORTS\n")
}

  .compile.tex <- function(name,zip.files) {
    r.cmd = shQuote(file.path(R.home("bin"),"R"))
    testPdflatex(paste(r.cmd,"CMD","pdflatex -version"))
    
    dir <- tempdir()
    cat("compiling ",name,".tex ...  1",sep="")
    .call.cmd(sprintf("%s CMD pdflatex -halt-on-error -output-directory=%s %s.tex",r.cmd,dir,name),
              paste(dir,"/",basename(name),".stdout",sep=""))
    cat(" 2")
    .call.cmd(sprintf("%s CMD pdflatex -halt-on-error -output-directory=%s %s.tex",r.cmd,dir,name),
              paste(dir,"/",basename(name),".stdout",sep=""))
    cat(" done!\n\n")
    file.copy(file.path(dir,paste0(name,".pdf")),file.path(getwd(),paste0(name,".pdf")),overwrite=TRUE)
    file.remove(file.path(dir,paste0(name,".pdf")))
    c(zip.files,sprintf("%s.pdf",name))
  }
 
#- load properties

load.properties <- function(properties.file="properties.R",
                            global.properties.file=system.file("report","properties.R",package="isobar"),
                            args=NULL,...) {

  properties.env <- new.env()
  tmp.properties.env <- new.env()

  message("Loading global properties file ",global.properties.file," ...")
  tryCatch(sys.source(global.properties.file,properties.env),
           error=function(e) stop("\nCould not read properties file:\n", e) )

  message("Looking for local properties file ",properties.file," ...")
  if (!is.null(properties.file) && file.exists(properties.file)) {
    message("  Loading local properties file ...")
    ## load properties file
    tryCatch(sys.source(properties.file,tmp.properties.env),
             error=function(e) stop("\nCould not read properties file:\n", e) )
    .env.copy(properties.env,tmp.properties.env)
    properties.env[["properties.file"]] <- properties.file
    properties.env[["properties.file.content"]] <- readLines(properties.file)
  } else {
    message("  No local properties file.")
  }
  
  ## command argument parsing
  tmp.properties.env <- new.env()
  if (length(args) > 0)
    message("parsing command line arguments ...")
  for (arg in args) {
    if (grepl("^--",arg)) {

      arg.n.val <- strsplit(substring(arg,3),"=")[[1]]
      if (length(arg.n.val) == 1)
        tmp.properties.env[[arg.n.val]] <- TRUE
      else if (length(arg.n.val) == 2) 
        tmp.properties.env[[arg.n.val[1]]] <- switch(arg.n.val[2],'TRUE'=TRUE,'FALSE'=FALSE,arg.n.val[2])
      else
        stop("Could not parse command line argument ",arg)
    } else {
        stop("Could not parse command line argument ",arg)
    }
    .env.copy(properties.env,tmp.properties.env)
  }

  ## ... parsing
  dotargs <- list(...)
  if (length(dotargs) > 0) {
    message("parsing function arguments")
    .env.copy(properties.env,list2env(dotargs))
  }

  return(properties.env)
}

#- initialize environment
initialize.env <- function(env,properties.env) {
  ## get property and exists property convenience functions
  get.property <- function(name) .get.property(name,properties.env)

  if (file.exists(get.property('cachedir'))) {
    if (!file.info(get.property('cachedir'))$isdir)
      stop("Cannot write into cachedir [",get.property('cachedir'),"]",
           " - it is a file and no directory")
  } else {
    ret <- dir.create(get.property('cachedir'))
    if (!ret) stop("Error creating cachedir [",get.property('cachedir'),"]")
  }

  ib.name <- file.path(.get.property('cachedir',properties.env),"ibspectra.rda")
  if (file.exists(ib.name)) {
    load(ib.name)
    env[["ibspectra"]] <- get("ibspectra")
  } else {
    env[["ibspectra"]] <- .create.or.load.ibspectra(properties.env)
    save(list='ibspectra',envir=env,file=ib.name)
  }
  if ("site.probs" %in% colnames(fData(env[["ibspectra"]])) 
      && ! "pep.siteprobs" %in% colnames(fData(env[["ibspectra"]])))
    fData(env[["ibspectra"]])$pep.siteprobs <- 
      .convertPhosphoRSPepProb(fData(env[["ibspectra"]])[,'peptide'],fData(env[["ibspectra"]])[,'site.probs'],
                               round.to.frac=20)


  env[["noise.model"]] <- .create.or.load.noise.model(env,properties.env)
  env[["ratiodistr"]] <- .create.or.load.ratiodistr(env,properties.env)
  if (identical(properties.env[["report.level"]],"peptide") ) {
    env[["ptm.info"]]  <- .create.or.load.ptm.info(properties.env,proteinGroup(env[["ibspectra"]]))
    if (all(c('use.for.quant','pep.siteprobs') %in% colnames(fData(env[['ibspectra']]))) && !all(fData(env[['ibspectra']])[['use.for.quant']]))
      env[["quant.tbl.notlocalized"]] <- 
        .create.or.load.quant.table(env,properties.env,name="quant.tbl.notlocalized",type="other-sites")
  }
  env[["quant.tbl"]] <- .create.or.load.quant.table(env,properties.env)
  if (!"ac" %in% colnames(env[["quant.tbl"]]) && "protein" %in% colnames(env[["quant.tbl"]]))
    env[["quant.tbl"]][,'ac'] <- env[["quant.tbl"]]$protein

  ## required for TeX
  if (property('write.report',properties.env) && identical(properties.env[["report.level"]],"protein"))
    env[["my.protein.infos"]] <- .create.or.load.my.protein.infos(env,properties.env)

  if (property('write.xls.report',properties.env)) {
    xls.table.name <- paste0("xls.quant.tbl.",properties.env[["xls.report.format"]])
    env[["xls.quant.tbl"]] <- 
      .create.or.load.xls.quant.tbl(env,properties.env,xls.table.name,"quant.tbl")
    if (exists("quant.tbl.notlocalized",envir=env)) {
      xls.ntable.name <- paste0("xls.quant.tbl.notlocalized.",properties.env[["xls.report.format"]])
      env[["xls.quant.tbl.notlocalized"]] <- 
        .create.or.load.xls.quant.tbl(env,properties.env,xls.ntable.name,"quant.tbl.notlocalized")
    }
  }
}

#- property loading helper functions
.DOES.NOT.EXIST = "NOT AVAILABLE"

.get.or.load <- function(name,envir,msg.f=name,class=NULL,null.ok=FALSE,do.load=FALSE) {
  if (.exists.property(name,envir,null.ok=null.ok)) {
    o <- .get.property(name,envir)
    if (is.null(o) && null.ok) return(NULL)
    if (!is.null(class) && is(o,class)) {
      return(o)
    } else if (is(o,"character")) {
      file.name <- o
    } else {
      stop("property ",name," is neither 'character' nor of a specified class")
    }
  } else {
    file.name <- sprintf("%s/%s.rda",.get.property('cachedir',envir),name)
  }
  if (file.exists(file.name) && (!envir[["regen"]] || do.load)) {
    message(sprintf("  loading %s from %s ... ",msg.f, file.name),appendLF=FALSE)
    x <- .load.property(name,file.name)
    message("done")
    return(x)
  } else {
    stop("Cannot get or load property [",name,"] - would expect it in [",file.name,"], but file does not exist.")
  }
}

.create.or.load <- function(name,envir,f,msg.f=name,regenerate=FALSE,do.load=FALSE,class=NULL,error=stop,default.value=NULL,...) {
  x <- if ( !regenerate ) tryCatch(.get.or.load(name,envir,msg.f,class,do.load=do.load),error=function(e) .DOES.NOT.EXIST) else .DOES.NOT.EXIST
  if (identical(x,.DOES.NOT.EXIST)) {
    message(paste("  creating",msg.f,"... "),appendLF=FALSE)
    tryCatch({
      x <- f(...)
      assign(name,x)
      file.name <- sprintf("%s/%s.rda",.get.property('cachedir',envir),name)
      message("done ",appendLF=FALSE)
      save(list=c(name),file=file.name)
      message("& saved to ",file.name)

    },error=error)

    if (!exists(name,inherits=FALSE)) {
      x <- default.value
      assign(name,x)
    }
  }
  if (!is.null(class) && !is(x,class))
    stop("property [",name,"] should be of class [",class,"] but is of class [",class(x),"]")

  return(x)
}
 
.get.property <- function(x,envir) get(x,envir=envir,inherits=FALSE)

property <- function(x, envir, null.ok=TRUE,class=NULL) {
  if (!.exists.property(x,envir)) 
    stop("property ",x," does not exist!")

  o <- get(x,envir=envir,inherits=FALSE)

  if (!null.ok && is.null(o))
    stop("property ",x," may not be NULL, but it is!")

  if (!is.null(class) && !is(o,class))
    stop("property ",x," does not have class ",class,"!")

  return(o)
}

.exists.property <- function(x,envir,null.ok=TRUE) 
  exists(x,envir=envir,inherits=FALSE) &&
    (null.ok || !is.null(.get.property(x,envir)))

.check.property <- function(x,envir,inherits=FALSE,msg="",print=TRUE,valid=NULL,check.file=FALSE,...) {
  def <- grep(x,envir[["properties.file.content"]],value=TRUE)
  if (length(def) > 0)
    def <- paste("\n  Corresponding line in ",envir[["properties.file"]],":\n\t",
                 paste(def,collapse="\n\t"),"\n\n",sep="")

  p <- .get.property(x,envir=envir)
  does.not.exist <- !exists(x,envir=envir,inherits=inherits,...)
  is.null.p <- does.not.exist || is.null(.get.property(x,envir=envir))
  length.0 <- does.not.exist || length(.get.property(x,envir=envir))==0
  if (does.not.exist || is.null.p || length.0) 
    stop ("  property '",x,"' not defined or did not evaluate in the properties file, but it is required",
          ifelse(is.null(valid),"",paste(" to be one of \n\t",paste(valid,collapse="\n\t"),
          sep="")),def)
  
  if (check.file && !file.exists(.get.property(x,envir=envir)))
    stop("  property '",x,"' should be assigned to a file, but it is not:",def)

  if (!is.null(valid)) {
    isnt.valid <- !p %in% valid
    if (isnt.valid) stop(" property '",x,"' must be one of \n\t",paste(valid,collapse="\n\t"),",\n\n",
                         "  but it is ",p )
  }
  if (print) {
    message("  property '",x,"' defined",ifelse(is.character(p),paste0(": ",paste(p,collapse="; ")),"."))
  }
}

.load.property <- function(name,file) {
  tmp.env <- new.env()
  load(file,envir=tmp.env)
  if (exists(name,envir=tmp.env))
    return(get(name,envir=tmp.env))
  else if (length(ls(envir=tmp.env)) == 1)
    return(get(ls(envir=tmp.env)[1],envir=tmp.env))
  else
    stop("Could not load property ",name," from file ",file)
}

.create.or.load.ibspectra <- function(properties.env) {
  get.property <- function(name) .get.property(name,properties.env)
  check.property <- function(name,...) .check.property(name,properties.env,...)
  ## check that required properties are defined
  check.property('ibspectra')
  check.property('type',valid=IBSpectraTypes())

  readIBSpectra.args <- get.property('readIBSpectra.args')
  if (is.null(readIBSpectra.args)) readIBSpectra.args <- list()
  for (name in names(readIBSpectra.args)) {
      arg <- readIBSpectra.args[[name]]
      if (!is.null(arg)) {
          if (!is.null(names(arg)))
            arg <- paste(names(arg),arg,collapse="=")
          if (is.function(arg)) { message("    ",name,": function")
          } else {
          message("    ",name,": ",
                  paste(arg,collapse=ifelse(length(arg)>2,"\n\t",", ")))
          }
      }
  }
  readIBSpectra.args[["type"]]=get.property('type')
  readIBSpectra.args[["id.file"]]=get.property('ibspectra')
  readIBSpectra.args[["fragment.precision"]]=get.property('fragment.precision')
  readIBSpectra.args[["fragment.outlier.prob"]]=get.property('fragment.outlier.prob')
  readIBSpectra.args[["proteinGroupTemplate"]]=.get.or.load('protein.group.template',properties.env,"ProteinGroup",null.ok=TRUE)

  if (is.data.frame(get.property('ibspectra')) || all(file.exists(get.property('ibspectra')))) {
    if (is.data.frame(get.property('ibspectra'))) {
      readIBSpectra.args[["id.file"]] <- get.property('ibspectra')
      ibspectra <- do.call(readIBSpectra,readIBSpectra.args)
    } else if (grepl(".csv",get.property('ibspectra'))) {
        message("ibspectra ends on .csv:\n",
                sprintf('ibspectra <- readIBSpectra("%s",%s) ...',
                        get.property('type'),
                        ifelse(length(get.property('ibspectra'))==1,
                               paste0('"',get.property('ibspectra'),'"'),
                               paste0('c("',get.property('ibspectra'),'")',collapse='","')
                               )))
        ibspectra <- do.call(readIBSpectra,readIBSpectra.args)
    } else if (grepl(".rda",get.property("ibspectra"))) {
        message("ibspectra ends on .rda: ","loading")
        ibspectra <- .load.property("ibspectra",get.property("ibspectra"))
    } else {
        stop("weird naming of ibspectra: ",get.property("ibspectra"),". use XXX.csv or XXX.rda")
    }

  } else {
    # create ibspectra file
    message(sprintf('\n  file %s does not exist, creating ibspectra',get.property('ibspectra')))
    check.property('peaklist',print=FALSE,check.file=TRUE)
    check.property('identifications',print=FALSE,check.file=TRUE)
    message(sprintf('    id.files = %s%s\n    peaklist.files = %s%s\n',
          ifelse(length(get.property('identifications'))>1,"\n      ",""),
          paste(get.property('identifications'),collapse="\n      "),
          ifelse(length(get.property('peaklist'))>1,"\n      ",""),
          paste(get.property('peaklist'),collapse="\n      ")))

    readIBSpectra.args[["id.file"]]=get.property('identifications')
    readIBSpectra.args[["peaklist.file"]]=get.property('peaklist')

    ibspectra <- do.call(readIBSpectra,readIBSpectra.args)
    if (grepl(".csv",get.property('ibspectra'))) {
        write.table(as.data.frame(ibspectra),sep="\t",row.names=F,file=get.property('ibspectra'),quote=FALSE)
    } else if (grepl(".rda",get.property("ibspectra"))) {
        save(ibspectra,file=get.property("ibspectra"),compress=TRUE)
    }  else {
        stop("weird naming of ibspectra: ",get.property("ibspectra"),". use XXX.csv or XXX.rda")
    }
  }

  if (!is.null(get.property("isotope.impurities"))) {
    isotopeImpurities(ibspectra) <- get.property("isotope.impurities")
  }

  if (property('correct.isotope.impurities',properties.env))
    ibspectra <- correctIsotopeImpurities(ibspectra)
  if (property('normalize',properties.env)) {
    ibspectra <-
      normalize(ibspectra,
                use.protein=property('normalize.use.protein',properties.env),
                exclude.protein=property('normalize.exclude.protein',properties.env),
                peptide.specificity=property('peptide.specificity',properties.env),
                f=property('normalize.function',properties.env),
                channels=property('normalize.channels',properties.env),
                na.rm=property('normalize.na.rm',properties.env),
                normalize.factors=property('normalize.factors',properties.env))
  }

  class.labels <- LETTERS[1:length(reporterTagNames(ibspectra))]
  if (.exists.property('class.labels',properties.env,null.ok=FALSE))
    class.labels <- get.property('class.labels')
  if (!any(table(class.labels)>1) && property('summarize',properties.env)) {
    stop("When summarize=TRUE, the must be more then one channel per class")
  }
  if (!is.character(class.labels)) {
    stop("Please provide class.labels of class character!")
  }
  classLabels(ibspectra) <- class.labels

  if (!is.null(property('protein.info.f',properties.env)))
    tryCatch({
    proteinInfo(proteinGroup(ibspectra)) <- 
      .create.or.load("protein.info",envir=properties.env,
                      f=property('protein.info.f',properties.env),
                      x=proteinGroup(ibspectra),
                      do.load=TRUE, msg.f="protein.info",
                      error=warning,default.value=proteinInfo(proteinGroup(ibspectra)))
    },error=function(e) stop("Error creating proteinInfo using function defined in [protein.info.f]: ",e,"Set protein.info.f to NULL if you want to create a report anyway."))

  if ("site.probs" %in% colnames(fData(ibspectra)) 
      && ! "pep.siteprobs" %in% colnames(fData(ibspectra)))
     fData(ibspectra)$pep.siteprobs <- 
        .convertPhosphoRSPepProb(fData(ibspectra)[,'peptide'],fData(ibspectra)[,'site.probs'],
                                 round.to.frac=20)
 
  return(ibspectra)
}

.create.or.load.noise.model <- function(env,properties.env) {
  noise.model.channels <- .get.property("noise.model.channels",properties.env)
  if (is.null(noise.model.channels)) 
    noise.model.channels <- reporterTagNames(env[["ibspectra"]])[!is.na(classLabels(env[["ibspectra"]]))]
  is.one.to.one <- isTRUE(.get.property("noise.model.is.technicalreplicates",properties.env))

  noise.model <- .get.property("noise.model",properties.env)
  if (!is(noise.model,"NoiseModel")) {
    noise.model.f <- .get.property("noise.model",properties.env)
    if (!file.exists(noise.model.f)) {
      message("estimating noise model as ",ifelse(is.one.to.one,"","non"),
              " one-to-one from channels ",paste0(noise.model.channels,collapse=", ")," ...")
      noise.model <- new("ExponentialNoiseModel",env[["ibspectra"]],one.to.one=is.one.to.one,
                         reporterTagNames=noise.model.channels,
                         min.spectra=property('noise.model.minspectra',properties.env))
      save(noise.model,file=noise.model.f,compress=TRUE)
    } else {
      message(sprintf("  loading noise model from %s ...",noise.model.f))
      noise.model <- .load.property("noise.model",noise.model.f)
    }
  }
  return(noise.model)
}

.create.or.load.ptm.info <- function(properties.env,protein.group) {
  if (is.null(property('ptm.info.f',properties.env))) 
    properties.env[["ptm.info.f"]] <- getPtmInfoFromNextprot

  return(.create.or.load("ptm.info",envir=properties.env,
                         f=property('ptm.info.f',properties.env),
                         protein.group=protein.group))
}

.create.or.load.ratiodistr <- function(env,properties.env) {
  
  return(.create.or.load("ratiodistr",envir=properties.env,class="Distribution",
                         msg.f="biological variability ratio distribution",f=function(){

    cl <- classLabels(env[["ibspectra"]])
    if (!is.null(property('ratiodistr.class.labels',properties.env)))
      cl <- property('ratiodistr.class.labels',properties.env)

    ratios.for.distr.fitting <- .create.or.load.ratiodistr.ratios(env,properties.env,cl)

    if (all(is.na(ratios.for.distr.fitting[,'lratio']))) 
      stop("All ratios for distr fitting are NA - are the correct class labels used?")

    if (!is.function(property('ratiodistr.fitting.f',properties.env)))
      stop("ratiodistr.fitting.f must be set to a function [e.g. fitCauchy or fitTd]")

    ratiodistr <- property('ratiodistr.fitting.f',properties.env)(ratios.for.distr.fitting[,'lratio'])
    ratiodistr <- .round.distr(ratiodistr,digits=5)
    attr(ratiodistr,"combn.method") <- attr(ratios.for.distr.fitting,"combn.method")
    attr(ratiodistr,"cl") <- attr(ratios.for.distr.fitting,"cl")
    attr(ratiodistr,"tagNames") <- reporterTagNames(env[["ibspectra"]])
    ratiodistr
  }))
}

.round.distr <- function(distr,digits) {
  for (s in slotNames(param(distr))) 
    if (s != "name" && is.numeric(slot(param(distr),s))) 
      slot(distr@param,s) <- round(slot(param(distr),s),digits)
  distr
}

.create.or.load.ratiodistr.ratios <- function (env,properties.env,cl) {
  .create.or.load("ratios.for.distr.fitting",envir=properties.env,class="data.frame",
                  msg.f="ratios for biological variability distribution fitting",f=function() {

    message("  Using class labels ",paste(cl,collapse=", "))
    if (property('summarize',properties.env) || any(table(cl)>=2)) {
      method <- "intraclass"
    } else {
      message(" WARNING: ratiodistr will be computed based on global ratios")
      method <- "global"
    }
    if (any(table(properties.env[["class.labels"]])>2))
      do.summarize <- properties.env[["summarize"]]
    else
      do.summarize <- FALSE

    if (identical(properties.env[["report.level"]],"peptide"))
      ratios.for.distr.fitting <- peptideRatios(env[["ibspectra"]],noise.model=env[["noise.model"]],do.warn=FALSE,
                                  cl=cl,combn.method=method,symmetry=isTRUE(properties.env[["ratiodistr.symmetry"]]),summarize=do.summarize)
    else
      ratios.for.distr.fitting <- proteinRatios(env[["ibspectra"]],noise.model=env[["noise.model"]],do.warn=FALSE,
                                      cl=cl,combn.method=method,symmetry=isTRUE(properties.env[["ratiodistr.symmetry"]]),summarize=do.summarize)

    if (all(is.nan(ratios.for.distr.fitting[["lratio"]])))
      stop("Cannot compute protein ratio distribution - no ratios available.\n",
           "Probably due to missing reporter intensities.")

    attr(ratios.for.distr.fitting,"combn.method") <- method
    attr(ratios.for.distr.fitting,"cl") <- cl

    ratios.for.distr.fitting
  })
}


.set <- function(x,name,list) {
  if (name %in% names(list)) {
    stop(name," already assigned in list!")
  }
  x
}

.create.or.load.quant.table <- function(env,properties.env,name="quant.tbl",type='confident-sites') {
  protein.group <- proteinGroup(env[["ibspectra"]])
  protein.info <- proteinInfo(protein.group)
  isoforms <- protein.group@isoformToGeneProduct
  
  .create.or.load(name,envir=properties.env,class="data.frame",
                  msg.f=paste("table of ratios of",properties.env[["report.level"]],"as",name),f=function() {
    if (!is.null(property('ratios.opts',properties.env)$summarize)) {
        message("WARNING: ratio.opts$summarize will be overwritten,",
                " define it outside of ratio.opts!")
        warning("ratio.opts$summarize will be overwritten,",
                " define it outside of ratio.opts!")
    }
    ratios.opts <- property('ratios.opts',properties.env)
    set.ratioopts <- function(x,name=names(x)) {
      if (is.null(intersect(name,names(ratios.opts))))
          stop("property '",intersect(name,names(ratios.opts)),
               "' already assigned in ratios.opts list - check your properties file!")

      if (is.list(x))
        ratios.opts <<- c(ratios.opts,x)
      else
        ratios.opts[[name]] <<- x
    }

    set.ratioopts(name="ibspectra",env[["ibspectra"]])
    set.ratioopts(name="noise.model",env[["noise.model"]])
    set.ratioopts(name="ratiodistr",env[["ratiodistr"]])
    
    if(identical(properties.env[["report.level"]],"peptide")) {
      if (!"use.for.quant" %in% colnames(fData(env[["ibspectra"]])))
        fData(env[["ibspectra"]])[["use.for.quant"]] <- TRUE
      if (type=='confident-sites')
        pep.n.modif <- unique(apply(fData(env[["ibspectra"]])[fData(env[["ibspectra"]])[["use.for.quant"]],
                                    c("peptide","modif")],2,cbind))
      else {
        pep.n.modif <- unique(apply(fData(env[["ibspectra"]])[!fData(env[["ibspectra"]])[["use.for.quant"]],
                                    c("peptide","modif","pep.siteprobs")],2,cbind))
        set.ratioopts(list(use.for.quant.only=FALSE))
      }


      set.ratioopts(list(peptide=pep.n.modif,
                         proteins=NULL))

    } else if (identical(properties.env[["report.level"]],"protein")) {
      set.ratioopts(list(peptide=NULL,
                         proteins=reporterProteins(proteinGroup(env[["ibspectra"]])),
                         quant.w.grouppeptides=property('quant.w.grouppeptides',properties.env)))
    } else {
      stop("don't known level ",properties.env[["report.level"]])
    }

    if (is.null(property('cmbn',properties.env)) & 
	!is.null(property('vs.class',properties.env)))
      properties.env[["cmbn"]] <- combn.matrix(reporterTagNames(env[["ibspectra"]]),
					   "versus.class",
					   property('class.labels',properties.env),
					   vs=property('vs.class',properties.env))

    set.ratioopts(list(combn.method=property('combn.method',properties.env),
                       cl=classLabels(env[["ibspectra"]]),
                       summarize=property('summarize',properties.env),
                       cmbn=property('cmbn',properties.env),
                       use.na=property('use.na',properties.env),
                       zscore.threshold=properties.env[["zscore.threshold"]],
                       do.warn=FALSE))

    if (!is.null(property('correct.peptide.ratios.with',properties.env))) {
      protein.quant.tbl <- .get.or.load("correct.peptide.ratios.with",properties.env)
      correct.protein.group <- if ( !is.null(property('correct.peptide.ratios.with_protein.group',properties.env)) ) {
        .get.or.load("correct.peptide.ratios.with_protein.group",properties.env)
      } else {
        warning( '"correct.peptide.ratios.with_protein.group" property not found, ',
                 'using the existing protein group' )
        .get.or.load("protein.group.template",properties.env)
      }
      ratios.opts[["before.summarize.f"]] <- function(...)
        correct.peptide.ratios(..., protein.quant.tbl=protein.quant.tbl,
                               protein.group.combined = correct.protein.group,
                               correlation = property('peptide.protein.correlation',properties.env))
    }

    quant.tbl <- do.call("proteinRatios",ratios.opts)

    quant.tbl[,"sd"] <- sqrt(quant.tbl[,"variance"])
    
#    quant.tbl$sign.string <- "not significant"
#    quant.tbl$sign.string[quant.tbl$is.significant] <- "is significant"
    
#    if (length(property('preselected',properties.env)) > 0) {
#      preselected <- unique(ip[sub("-.*","",names(ip)) %in% property('preselected',properties.env)])
#      quant.tbl$is.preselected <- quant.tbl$protein %in% preselected

#      quant.tbl$sign.string <- "not significant"
#      quant.tbl$sign.string[is.sign & !quant.tbl$is.preselected] <- "is significant [ours]"
#      quant.tbl$sign.string[is.sign & quant.tbl$is.preselected] <- "is significant [both]"
#      quant.tbl$sign.string[!is.sign & quant.tbl$is.preselected] <- "is significant [theirs]"
#    }

    if (identical(properties.env[["report.level"]],"protein")) {
      quant.tbl[,"gene_names"] <- sapply(quant.tbl[,"ac"], function(x) {
        if (length(protein.info) == 0) return("")
        allreporter <- indistinguishableProteins(protein.group,protein.g=x)
        acs <- unique(isoforms[allreporter,"proteinac.wo.splicevariant"])
        paste(sort(unique(protein.info[protein.info[["accession"]] %in% acs,"gene_name"])),
              collapse=", ")
      })
      sort.genenames <- quant.tbl[,"gene_names"]
      sort.genenames[sort.genenames==""] <- quant.tbl[sort.genenames=="","ac"]
      quant.tbl <- quant.tbl[order(sort.genenames,quant.tbl[,"r1"],quant.tbl[,"r2"]),]
      quant.tbl[,"group"] <- as.numeric(factor(quant.tbl[,"ac"],levels=unique(quant.tbl[,"ac"])))
    }
  
    if (all(is.na(quant.tbl[["lratio"]])))
      stop("All ratios are NA")

    return(quant.tbl)
  })
}

.create.or.load.my.protein.infos <- function(env,properties.env) {
  .create.or.load("my.protein.infos",envir=properties.env,
                  f=function() {

    protein.groupnames <-unique(env[["quant.tbl"]][,"ac"])
    ## if (is.null(protein.info)) { stop("protein info is null!")}                
    my.protein.infos <- llply(protein.groupnames, .do.create.protein.info, protein.group=proteinGroup(env[["ibspectra"]]), 
                              .parallel=isTRUE(getOption('isobar.parallel')))
    #my.protein.infos <- lapply(protein.groupnames, .do.create.protein.info, protein.group=protein.group)
    names(my.protein.infos) <- protein.groupnames
    return(my.protein.infos)
  })
}
  
.do.create.protein.info <- function(x,protein.group) {
    protein.group.table <- proteinGroupTable(protein.group)
      allgroupmember <- indistinguishableProteins(protein.group, protein.g =
                                                  protein.group.table[protein.group.table[,'reporter.protein'] %in% x,"protein.g"])
     
      reporter.protein.info <- my.protein.info(protein.group,x)
      collapsed.gene_name <- human.protein.names(reporter.protein.info)
      peptides <- peptides(protein.group,protein=x,do.warn=FALSE)
      peptides.gs <- peptides(protein.group,protein=x,
                              specificity=GROUPSPECIFIC,do.warn=FALSE)
      peptides.rs <- peptides(protein.group,protein=x,
                              specificity=REPORTERSPECIFIC,do.warn=FALSE)
      n.spectra <- length(names(spectrumToPeptide(protein.group))[spectrumToPeptide(protein.group)%in%peptides])
      
      tbl.protein.name <- sort(collapsed.gene_name[,'protein_name'])[1];
      if (length(unique(collapsed.gene_name[,'protein_name'])) > 1)
        tbl.protein.name <- paste(tbl.protein.name,", ...",sep="")
      
      list(n.reporter = nrow(reporter.protein.info),
           n.groupmember = length(allgroupmember),
           reporter.protein.info = reporter.protein.info,
           n.peptides=length(peptides),
           n.spectra=n.spectra,
           collapsed.gene_name = collapsed.gene_name,
           table.name = ifelse(
             length(collapsed.gene_name[,'ac_link'])>3,
             paste(paste(collapsed.gene_name[1:3,'ac_link'],collapse=", "),
                   ", \\dots",sep=""),
             paste(collapsed.gene_name[,'ac_link'],collapse=", ")),
           section.name = sanitize(paste(unique(collapsed.gene_name[,'name_nolink']),
             collapse=", ")),
           table.protein.name = tbl.protein.name,
           gene.name = paste(sort(unique(reporter.protein.info[,'gene_name'])),collapse=", ")
           )
    }

## copys objects from env into parentenv.
## Those objects MUST exist in parentenv before.
.env.copy <- function (parentenv,env) {
  printval <- function(x) if(is.character(x)) capture.output(dput(x)) else toString(x)

  for (object in ls(envir=env)) {
    if (exists(object,envir=parentenv)) {
      ## assign object to parent env
      val <- get(object,envir=env)
      if (is.list(val)) {
        val.s <- paste0("list(",paste(names(val),sapply(val,printval),sep=": ",collapse=";"),")")
      } else {
        val.s <- tryCatch(printval(val),error=function(e) capture.output(print(val)))
      }
      message(sprintf("%30s = %s",object,paste(val.s,collapse=paste0("\n",paste(rep(" ",35),collapse="")))))
      assign(object, value=get(object,envir=env),envir=parentenv)
      ## remove(object,envir=env)
    } else {
      stop("Illegal property ",object,"!\n",
           "Call script with --help for usage details.\n\n",
           "Available properties:\n",
           "\t",paste(ls(envir=parentenv),collapse="\n\t"),"\n")
                      
      ## more verbose list:
      ##cat(paste(unlist(lapply(objects(),function(o) {sprintf("%s [%s]",o,class(get(o)))})),collapse="\n\t"))
    }
  }
}

%\VignetteIndexEntry{isobar for developers}
%\VignetteDepends{}
%\VignetteKeywords{Documentation}
%\VignettePackage{isobar}
\documentclass[11pt]{article}

\usepackage{times}
\usepackage{hyperref}

\textwidth=6.2in
\textheight=8.5in
%\parskip=.3cm
\oddsidemargin=.1in
\evensidemargin=.1in
\headheight=-.3in

\newcommand{\scscst}{\scriptscriptstyle}
\newcommand{\scst}{\scriptstyle}
\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{{\textit{#1}}}

\author{Florian P.\ Breitwieser, Jacques Colinge}

\bibliographystyle{plainnat}

\begin{document}

\title{\emph{isobar} for developers}
\maketitle
\tableofcontents

\section{Introduction}
This documents highlights the structure of the S4 classes and methods in the
\texttt{isobar} pacakge.

<<>>=
  library(isobar)
@

\section{Classes}
\subsection{IBSpectra}
<<>>=
  getClass("IBSpectra")
@ MSMS isobarically tagged proteomics data consists of spectrum
identifications and quantitative values. Spectrums are identified as
stemming from distinct peptides, and quantitative information of each
spectrum are extracted from a certain \texttt{m/z} region.

\Robject{IBSpectra} class holds this qualitative and quantitative
information. It is a virtual class. It extends \Robject{eSet} from
Biobase to store meta-information of spectrum identifications and
quantitative information (\texttt{m/z} and \texttt{intensity}) of
reporter tags. \Robject{eSet} is extended by slots for protein
grouping, tag names, tag masses and isotope impurity correction
matrix.

\Robject{ProteinGroup} objects store the mapping and grouping of
peptide level identifications to protein identifications.

\Robject{IBSpectra} is a virtual class. Currently used isobaric
tagging kits iTRAQ 4plex and 8plex, and TMT 2plex and 6plex are
implemented in the \Robject{iTRAQ4plexSpectra},
\Robject{iTRAQ8plexSpectra}, \Robject{TMT2plexSpectra},
\Robject{TMT6plexSpectr} and \Robject{TMT10plexSpectr}, respectively. These are subclasses of
\Robject{iTRAQSpectra} and \Robject{TMTSpectra}, resp.\, which in turn
are virtual subclasses of \Robject{IBSpectra}.

\subsection{ProteinGroup}
<<>>=
  getClass("ProteinGroup")
@ Bottom-up MSMS proteomics experiments identify peptides which are
mapped back to proteins. This mapping leads to protein groups, which
explain the observed peptides according to the parsimony law.

A \Robject{ProteinGroup} object is generated when a
\Robject{IBSpectra} object is created by
\Rfunction{readIBSpectra}. Protein to peptide to spectrum mapping is
extracted from a suitable identication format\footnote{IBSpectra CSV,
  and MzIdentML format. Mascot DAT and Phenyx pidres.xml format
  converters to IBSpectra format are provided.}

\subsection{NoiseModel}
<<>>=
  getClass("NoiseModel")
@ \Robject{NoiseModel} objects model the technical variance observed
in the spectrum-level ratios of a certain experimental setup.


\section{Session Information}

The version number of R and packages loaded for generating the
vignette were:

<<echo=FALSE,results=tex>>=
  toLatex(sessionInfo())
@ 

\end{document}

% vim: set expandtab ts=2 sw=2:
\SweaveOpts{eps=FALSE}
\SweaveOpts{strip.white=false}
<<initialization,echo=FALSE,results=hide>>=
# this Rnw file should be called by create_reports.R. create_reports 
# initializes the correct data objects to be used in the report.
options(SweaveHooks=list(fig=function()
         par(mar=c(5.1, 4.1, 1.1, 2.1))))
suppressPackageStartupMessages(library(isobar))
options(encoding="UTF-8")  

load.tex.properties(.GlobalEnv) # loads environments 

@
\documentclass[10pt,oneside]{article}
\usepackage[margin=2cm]{geometry}
\usepackage[utf8]{inputenc}
\usepackage{amsmath}
\usepackage{graphicx}
\usepackage{verbatim}
\usepackage{array}
\usepackage{rotating}
\usepackage{tikz}
\usetikzlibrary{arrows,backgrounds,positioning,matrix,shapes.geometric}
\tikzstyle{every picture}+=[remember picture]

% hyper reference
\usepackage{hyperref}
\definecolor{darkblue}{rgb}{0,0,.35}
\definecolor{darkred}{rgb}{0.25,0,0}
\definecolor{darkgreen}{rgb}{0,0.25,0}
\hypersetup{pdftex=true, colorlinks=true, breaklinks=true, linkcolor=darkblue, menucolor=darkblue, pagecolor=darkblue, urlcolor=darkblue}

% tables
\usepackage{ltxtable}
\usepackage{bigstrut}
\usepackage{multirow}
\usepackage{ctable} % for midrule, etc
%\setlength{\tabcolsep}{5pt} %default value: 6pt
%\usepackage{varwidth} %for the varwidth minipage environment
\newcolumntype{x}[1]{>{\raggedright}p{#1}}
\newcolumntype{f}{>{\footnotesize}r} 

<<echo=FALSE,results=tex,strip.white=FALSE>>=
write.tex.commands() # prints analysisname, analysisauthor, and isobarthanks commands
my.file <- system.file("report","report-utils.tex",package="isobar")
if (my.file == "")
 stop("Could not find file ISOBARDIR/report/report-utils.tex")

cat(readLines(my.file),sep="\n")
@

<<definitions,echo=FALSE,results=tex>>=
protein.group <- proteinGroup(ibspectra)
protein.info <- proteinInfo(protein.group)
isoforms <- protein.group@isoformToGeneProduct

bndt <- 5
bnd <- round(log10(bndt),1)
cat(sprintf("\\newcommand{\\bnd}{%s}",bnd),"\n")
cat(sprintf("\\newcommand{\\bndt}{%s}\n",bndt),"\n")

if (!"class1" %in% colnames(quant.tbl)) quant.tbl[["class1"]] <- quant.tbl[["r1"]]
if (!"class2" %in% colnames(quant.tbl)) quant.tbl[["class2"]] <- quant.tbl[["r2"]]
cmbn <- t(unique(quant.tbl[,c("r1","r2")]))
cmbn.class <- sanitize(t(unique(quant.tbl[,c("class1","class2")])))
is.single.comparision <- ncol(cmbn) ==1

cat(sprintf("\\newcommand{\\mr}[1]{\\multirow{%s}{*}{#1}}\n",ncol(cmbn)))
cat(sprintf("\\newcommand{\\mrp}[1]{\\multirow{%s}{*}{\\parbox{\\linewidth}{#1}}}\n",ncol(cmbn)))
@

\makeatletter
\renewcommand\paragraph{%
   \@startsection{paragraph}{4}{0mm}%
      {-\baselineskip}%
      {.5\baselineskip}%
      {\normalfont\normalsize\bfseries}}
\makeatother

\usepackage{fancyhdr}
\pagestyle{fancy}
\lfoot{\today}
\cfoot{\thepage}
\rfoot{\analysisname}

\setlongtables
\setlength\LTleft{0pt}
\setlength\LTright{0pt}
\renewcommand*{\arraystretch}{1.25}

\title{Isobaric Tag Quantification Report\\ \analysisname\isobarthanks}
\author{\analysisauthor}
\date{\today}
\raggedbottom
\setcounter{tocdepth}{1}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% BEGIN DOCUMENT

\begin{document}
\maketitle

%% Print class labels table, and column descriptions table
<<echo=FALSE,results=tex>>=
print_classlabels_tbl(classLabels(ibspectra),reporterTagNames(ibspectra))
@
\quad
\grpdescriptionstbl{}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% SIGNIFICANT PROTEINS TABLE
<<significantly.dfferent.proteins,echo=FALSE,results=tex>>=
if (is.single.comparision) {
  title.comp <- paste(cmbn.class[2,1],"/",cmbn.class[1,1])
} else {
  title.comp <- NULL
}

if (properties.env[["show.significant.proteins"]]) { 
 
  cat("\\section{Significantly Regulated Proteins",title.comp,"}\n\n")
  cat("\\secsubtitle{Number of significant proteins: ",
      get_n_proteins(quant.tbl,sign=TRUE),"}\\vspace{.5cm}\n" )
  
  if (any(quant.tbl[["is.significant"]] == 1,na.rm=TRUE)) { 
    print_sign_proteins_tbl(file="sign-proteins.tex",
                        cmbn,protein.group,
                        quant.tbl,
                        my.protein.infos,bnd)
  
    cat("\\small\n")
    cat("\\noindent\\LTXtable{\\textwidth}{sign-proteins.tex}\n")
    cat("\\normalsize\n")
    
  } else {
    cat ("No significantly regulated proteins present in sample.")
  }
}
@

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% PROTEIN QUANTIFICATION TABLE
<<echo=FALSE,results=tex>>=
cat("\\section{Protein Relative Quantitation",title.comp,"}")
cat("\\secsubtitle{Number of quantified proteins: ",
    get_n_proteins(quant.tbl),"}\\vspace{.5cm}\n")
@

\subsection{Quantified Proteins}
{
 \setlength{\extrarowheight}{1.5pt}

<<proteintbl,echo=FALSE,results=tex>>=

print_protein_quant_tbl(file="protein-quant.tex",
                    cmbn,protein.group,
                    quant.tbl,
                    my.protein.infos,bnd)

cat("\\small\n")
cat("\\noindent\\LTXtable{\\textwidth}{protein-quant.tex}\n")
cat("\\normalsize\n")

@
}

<<not.quantified,echo=FALSE,results=tex>>=
tt <- table(quant.tbl[is.na(quant.tbl[["lratio"]]),"ac"])
proteins.notquantified <- names(tt)[tt==ncol(cmbn)]

if (length(proteins.notquantified) > 0) {
cat("\\subsection{Not Quantified Proteins}\n")
n.quant <- ddply(quant.tbl,c("class1","class2"),function(x) sum(is.na(x[["lratio"]])))
cat("\\secsubtitle{Number of identified but not quantified proteins: ",
    get_n_proteins(quant.tbl,is.na=TRUE),"}\\vspace{.5cm}")

print_protein_notquant_tbl(file="protein-notquant.tex",
                    cmbn,protein.group,
                    quant.tbl,
                    my.protein.infos)

cat("\\small\n")
cat("\\LTXtable{\\textwidth}{protein-notquant.tex}\n")
cat("\\normalsize\n")
}
@

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% PROTEIN GROUP TABLE
<<proteindetails,echo=FALSE,results=tex>>=
if (properties.env[["show.protein.details"]]) {

message("Writing protein group details ")
print_protein_grp_info()

proteins.n.group <- unique(quant.tbl[,c('ac','group')])
proteins <- proteins.n.group[order(proteins.n.group[["group"]]),"ac"]
print_protein_grp_tbl("protein-groups.tex",proteins,protein.group)

cat("\\small\n")
cat("\\LTXtable{\\textwidth}{protein-groups.tex}\n")
cat("\\normalsize\n")

}
@

\end{document}

%\VignetteIndexEntry{isobar package for iTRAQ and TMT protein quantification}
%\VignetteDepends{}
%\VignetteKeywords{Documentation}
%\VignettePackage{isobar}
\documentclass[11pt]{article}

\SweaveOpts{keep.source=TRUE}
\SweaveOpts{prefix.string=graphics/plot}
<<init,echo=FALSE>>=
  require(distr)
  require(ggplot2)
  dir.create(file.path("graphics"), showWarnings = FALSE)
@ 

\usepackage{times}
\usepackage{hyperref}
\usepackage[utf8]{inputenc}
\usepackage{xcolor}
\usepackage{listings}

\textwidth=6.2in
\textheight=8.5in
%\parskip=.3cm
\oddsidemargin=.1in
\evensidemargin=.1in
\headheight=-.3in

\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{{\texttt{#1}}}
\newcommand{\isobar}{\Rpackage{isobar}}

\title{\emph{isobar} package for iTRAQ and TMT protein quantification}
\author{Florian P.\ Breitwieser, Jacques Colinge}

\begin{document}

\maketitle
\tableofcontents

\section{Introduction}
The \Rpackage{isobar} package is an extensible and interactive
environment for the data analysis and exploration of iTRAQ and TMT
data.  \Rpackage{isobar} implements the theory presented in
Breitwieser et al., Journal of Proteome Research 2011. Further
extensions for the analysis of \small{PTM} data are presented in
Breitwieser and Colinge, Journal of Proteomics 2013.

\subsection{Installation}
\Rpackage{isobar} is part of
\href{http://www.bioconductor.org/packages/release/bioc/html/isobar.html}{Bioconductor}. To
install the latest stable version available for your version of R, start R and
enter:

<<eval=FALSE>>=
if (!requireNamespace("BiocManager", quietly=TRUE))
    install.packages("BiocManager")
BiocManager::install("isobar")
@ 

Bioconductor has semi-annual releases, and the releases are tied to
specific to R versions. The latest development version of
\Rpackage{isobar} are always available at GitHub
(\url{https://github.com/fbreitwieser/isobar}). Use the
\Rpackage{devtools} package to install it from the source:

<<eval=FALSE>>=
library(devtools)
install_github("fbreitwieser/isobar")
@ 

\subsection{Loading package}
The first thing you need to do is load the package.
<<load-isobar,results=hide>>=
  library(isobar) ## load the isobar package
@ 

Check the installed version of isobar:
<<eval=FALSE>>=
packageVersion("isobar")
@ 
or equivalently for older versions of R:
<<eval=FALSE>>=
packageDescription("isobar")$Version
@ 

\section{Loading data}\label{sec:loading-data}
\Rpackage{isobar} can read identifications and quantifications from
tab-separated and MGF files. Perl scripts are supplied to generate a
tab-separated version from the vendor formats of Mascot and Phenyx,
see appendix \ref{sec:dependencies}.  The ``ID.CSV'' format is
described in appendix \ref{app:fileformats}.  Rockerbox and MSGF+ CSV
file formats can be read, too, and are recognized by their file
extension (\texttt{peptides.csv} and \texttt{msgfp.csv}).
Experimental support for the mzIdentML format is also available -
please contact the maintainer in case of problems.

\begin{description}
\item[\small ID.CSV] tab-separated file containing peptide-spectra matches and
  spectrum meta-information such as retention time, m/z and charge. 
  Generated by parser scripts.
\item[\small MGF] contains peak lists from which quantitative
  information on reporter tags are extracted. Must be centroided.
\item[\small IBSPECTRA.CSV] tab-separated file containing the same columns as
  {\small ID.CSV} plus \emph{quantitative information} extracted from {\small MGF}
  file - that means the reporter tag masses and intensities as
  additional columns.
\end{description}

\Robject{readIBSpectra} is the primary function to generate a
\Robject{IBSpectra} object. The first argument is one of {\small
  iTRAQ4plexSpectra}, {\small iTRAQ8plexSpectra}, {\small
  TMT2plexSpectra}, {\small TMT6plexSpectra} and {\small TMT10plexSpectra} denotes the tag
type and therefore class. 

<<echo=TRUE,eval=FALSE>>=
  ## generating IBSpectra object from ID.CSV and MGF
  ib <- readIBSpectra("iTRAQ4plexSpectra",list.files(pattern=".id.csv"),
          list.files(pattern=".mgf"))

  ## write in tabular IBSPECTRA.CSV format to file
  write.table(as.data.frame(ib),sep="\t",row.names=F,
          file="myexperiment.ibspectra.csv")

  ## generate from saved IBSPECTRA.CSV - MGF does not have to be supplied
  ib.2 <- readIBSpectra("iTRAQ4plexSpectra","myexperiment.ibspectra.csv")
@

In case the MGF file is very big, it can be advanteguous to generate a
smaller version containing only meta- and quantitative information
before import in \texttt{R}. On Linux, the tool \texttt{grep} is
readily available.

\begin{verbatim}
egrep '^[A-Z]|^1[12][0-9]\.' BIG.mgf > SMALL.mgf
\end{verbatim}

\subsection{\Robject{ibspiked} test samples}
The examples presented are based on the dataset
\Robject{ibspiked\_set1} which has been designed to test
\Rpackage{isobar}'s functionality and searched against the Swissprot
human database with Mascot and Phenyx. \Robject{ibspiked\_set1} is an
iTRAQ 4-plex data set comprised of a complex background (albumin- and
IgG-depleted human plasma) and spiked proteins. MS analysis was
performed in ThermoFisher Scientific LTQ Orbitrap HCD instrument with
2D shotgun peptide separation (see original paper for more
details). The samples used for each iTRAQ channel are as follows:
\begin{itemize}
\item Depleted human plasma background (>150 protein detected);
\item Spiked-in proteins with the following ratios
  \begin{itemize}
    \item \texttt{CERU\_HUMAN (P00450)} at concentrations $1:1:1:1$;
    \item \texttt{CERU\_RAT (P13635)} at concentrations $1:2:5:10$;
    \item \texttt{CERU\_MOUSE (Q61147)} at concentrations $10:5:2:1$.
  \end{itemize}
\end{itemize}

A second data set with ratios 1:10:50:100 is available as
\Robject{ibspiked\_set2} from
\url{http://bininformatics.cemm.oeaw.ac.at/isobar}.

The Ceruplasmins have been selected as the share peptides. Hereafter,
we load the data pacakage and the ceru protein IDs are identified via
the \Rfunction{protein.g} function, which provides a mean to retrieve
data from \Robject{ProteinGroup} objects. \Robject{ProteinGroup} is a 
slot of \Robject{IBSpectra} objects and contains informations on proteins
and their grouping. See \ref{ss:proteingrouping}.

<<load-ibspiked>>=
  data(ibspiked_set1)
  ceru.human <- protein.g(proteinGroup(ibspiked_set1),"CERU_HUMAN")
  ceru.rat <- protein.g(proteinGroup(ibspiked_set1),"CERU_RAT")
  ceru.mouse <- protein.g(proteinGroup(ibspiked_set1),"CERU_MOUSE")
  ceru.proteins <- c(ceru.human,ceru.rat,ceru.mouse)
@ 

\subsection{Protein information and grouping in \Robject{ProteinGroup}}
\label{ss:proteingrouping}
When an \texttt{ibspectra.csv} is read, protein are grouped to
identify proteins which have unique peptides. By default, only
peptides with unique peptides are grouped.

The algorithm to infer protein groups works as follows:
\begin{enumerate}
\item Group proteins together which have been seen with exactly the
  same peptides (\Robject{indistinguishableProteins}) - these are 
  the \Robject{protein.g} identifiers.
\item Create protein groups (\Robject{proteinGroupTable}):
  \begin{enumerate}
  \item Define proteins with specific peptides as reporters
    (\Robject{reporterProteins})
  \item Get proteins which are contained \footnote{That means these
      proteins have a subset of the peptides of the reporter} by
    reporterProteins and group them below.
  \end{enumerate}
\item Create protein groups for proteins without specific peptides as
  above.
\end{enumerate}


\subsection{Loading data from CID/HCD (or CID/MS3, etc) experiments}

A combined CID/HCD approach, in which for each precursor two fragmentation
spectra are acquired, has proven useful to increase the number of identified
and quantified peptide-spectrum matches. Usually, the reporter intensity 
information is taken from the HCD spectrum, and the peptide is identified 
based on the fragment ions in the CID spectrum. 

To import these experiments, a comma-separated 
\emph{mapping file} is
needed, which contains the association from the \texttt{identification} to the
\texttt{quantification} spectrum title.

\paragraph{Example mapping file (\texttt{mapping.csv}):}
\begin{verbatim}
"hcd","cid"
"spectrum 1","spectrum 2"
"spectrum 3","spectrum 4"
\end{verbatim}

By calling \Rfunction{readIBSpectra} with \texttt{mapping.file="mapping.csv"}, the spectra titles
in the \texttt{id.file} are matched to those in the \texttt{peaklist.file}. If the column names 
for the quantification and identification spectrum are not \texttt{hcd} and \texttt{cid}, resp., 
they can be set with the argument \texttt{mapping}=\texttt{c(identification.spectrum="column name 1",quantification.spectrum="column name 2")}.

\textbf{Example mapping file \texttt{mapping2.csv}:}
\begin{verbatim}
"quant-spectrum-ms3","id-spectrum-ms2"
"spectrum 1","spectrum 2"
"spectrum 3","spectrum 4"
\end{verbatim}

<<eval=FALSE>>=
readIBSpectra(...,
              mapping.file="mapping2.csv",
              mapping=c(identification.spectrum="id-spectrum-ms2",
                        quantification.spectrum="quant-spectrum-ms3")
              )
@

The argument \texttt{mapping.file} can take multiple files as argument (which are read and concatenated), or a \texttt{data.frame}.

\subsection{Loading data from CID/HCD experiments with full HCD
  spectrum}

If a full HCD spectrum was acquired, and both the HCD and CID spectrum
are searched against a protein database, the argument
\texttt{id.file.domap} to \texttt{readIBSpectra} can be used to merge
both CID and HCD identifications:

<<eval=FALSE>>=
readIBSpectra("TMT6plexSpectra",
              id.file="cid.identifications.csv",
              peaklist.file="hcd.peaklist.mgf",
              id.file.domap="hcd.identifications.csv",
              mapping.file="mapping.csv",
              ...
              )
@

Here, the CID (\texttt{cid.identifications.csv}) and HCD
identifications (\texttt{hcd.identifications.csv}) are combined and
mapped according to \texttt{mapping.csv}. Diverging identifications
are discarded.

\subsection{Integrating and thresholding precursor purity measures}
Savitzki et al., 2011 describe a measure for the precursor purity of
fragment spectra. An implementation of the algorithm for the
\texttt{multiplierz} platform is provided at
\url{http://sourceforge.net/apps/trac/multiplierz/wiki/Precursors}. The
script creates \texttt{\_precursors.txt} files for each \texttt{RAW}
file. To integrate and threshold based on the \texttt{s2i} measure,
use the following code:

<<eval=FALSE>>=
precursors <- read.delim("file1_precursors.txt",header=TRUE,sep="\t",stringsAsFactors=FALSE)
idfile <- read.delim("file1.id.csv",header=TRUE,sep="\t",stringsAsFactors=FALSE)

idfile <- merge(idfile,precursors,
                by.x=c("fixed_spectrum","scans.from"),
                by.y=c("fixed_spectrum","SCAN"),all.x=TRUE) # FIXME see /work/analysis/gwinter/raw-spectra/analysis.R
@ 

\subsection{\Rpackage{MSnbase} integration}
NOTE: It has been reported that the conversion functions are not
functional for current versions of the package. We will try to resolve
this.

\Rpackage{MSnbase} by Laurent Gatto provides data manipulation and
processing methods for MS-based proteomics data. It provides import,
representation and analysis of raw MS data stored in \texttt{mzXML},
\texttt{mzML} and \texttt{mzData} using the \Rpackage{mzR} package and
centroided and un-centroided {\small \texttt{MGF}} peak lists. It
allows to use and preprocess raw data whereas \Rpackage{isobar}
requires centroided peak lists. In the future, the \Rpackage{isobar}
class \Robject{IBSpectra} might be based on or replaced by
\Robject{MSnbase}'s class \Robject{MSnSet}.  For now, methods for
coercion are implemented:

<<eval=FALSE>>=
  as(ibspectra,"MSnSet")
  as(msnset,"IBSpectra")
@ 



\section{Data Analysis}
\subsection{Reporter mass precision}
The distribution of observed masses from the reporter tags can be used
to visualize the precision of the MS setup on the fragment level and
used to set the correct window for isolation.

The expected masses of the reporter tags are in the slot
\Rfunction{reporterTagMasses} of the implementations of the
\Robject{IBSpectra} class. The experimental masses are in the matrix
\Robject{mass} of \Robject{AssayData}; they can also be accessed by
the method \Rfunction{reporterMasses(x)}.


\subsection{\Rpackage{MSnbase} integration}
NOTE: It has been reported that the conversion functions are not
functional for current versions of the package. We will try to resolve
this.

\Rpackage{MSnbase} by Laurent Gatto provides data manipulation and
processing methods for MS-based proteomics data. It provides import,
representation and analysis of raw MS data stored in \texttt{mzXML},
\texttt{mzML} and \texttt{mzData} using the \Rpackage{mzR} package and
centroided and un-centroided {\small \texttt{MGF}} peak lists. It
allows to use and preprocess raw data whereas \Rpackage{isobar}
requires centroided peak lists. In the future, the \Rpackage{isobar}
class \Robject{IBSpectra} might be based on or replaced by
\Robject{MSnbase}'s class \Robject{MSnSet}.  For now, methods for
coercion are implemented:

<<eval=FALSE>>=
  as(ibspectra,"MSnSet")
  as(msnset,"IBSpectra")
@ 



\section{Data Analysis}
\subsection{Reporter mass precision}
The distribution of observed masses from the reporter tags can be used
to visualize the precision of the MS setup on the fragment level and
used to set the correct window for isolation.

The expected masses of the reporter tags are in the slot
\Rfunction{reporterTagMasses} of the implementations of the
\Robject{IBSpectra} class. The experimental masses are in the matrix
\Robject{mass} of \Robject{AssayData}; they can also be accessed by
the method \Rfunction{reporterMasses(x)}.

<<show-reporterMasses>>=
sprintf("%.4f",reporterTagMasses(ibspiked_set1))  ## expected masses
mass <- assayData(ibspiked_set1)[["mass"]] ## observerd masses
apply(mass,2,function(x) sprintf("%.4f",quantile(x,na.rm=TRUE,probs=c(0.025,0.975))))
@

\Rfunction{reporterMassPrecision} provides a plot of the distribution.

<<print-reporterMasses,fig=TRUE,width=6.6,height=3,include=FALSE>>=
  print(reporterMassPrecision(ibspiked_set1))
@ 

\begin{center}
\begin{figure}[htp]
\includegraphics{graphics/plot-print-reporterMasses}
\caption{Reporter mass precision plot.}
\label{fig:reporterMassPrecision}
\end{figure}
\end{center}

\subsection{Normalization and isotope impurity correction} \label{sec:norm-isot-impur}
Isotope impurity correction factors are supplied by labelling reagent
manufacturers. Default values that can be modified by the user are
available in \Rpackage{isobar} and corrections are obtained by simple
linear algebra.

Due to differences between samples it is advisable to normalize data
before further processing. By default, \Rfunction{normalize} corrects
by a factor such that the median intensities in all reporter channels
are equal.

See figure ~\ref{fig:normalization}.

<<normalize-ibspiked>>=
ib.old <- ibspiked_set1
ibspiked_set1 <- correctIsotopeImpurities(ibspiked_set1)
ibspiked_set1 <- normalize(ibspiked_set1)
@ 

<<fig-maplot-normalize1,echo=FALSE>>=
png("graphics/fig_maplot_normalize.png",width=6.6,height=3,units="in",res=300,pointsize=8)
@
<<fig-maplot-normalize2>>=
par(mfrow=c(1,2))
maplot(ib.old,channel1="114",channel2="117",ylim=c(0.5,2),
       main="before normalization")
abline(h=1,col="red",lwd=2)
maplot(ibspiked_set1,channel1="114",channel2="117",ylim=c(0.5,2),
       main="after normalization")
abline(h=1,col="red",lwd=2)
@
<<fig-maplot-normalize3,echo=FALSE,results=hide>>= 
dev.off()
@

\begin{center}
\begin{figure}[htp]
\includegraphics{graphics/fig_maplot_normalize}
\caption{Ratio versus intensity plots ('MA plots') before and after applying normalization.}
\label{fig:normalization}
\end{figure}
\end{center}

\subsection{Fitting a noise model}
A noise model is a approximation of the expected technical variation
based on signal intensity. It is stable for a certain experimental
setup and thus can be learned once. Noise is observed directly when
comparing identical samples in multiple channels (1:1 iTRAQ/TMT
sample) and we can use \Robject{ibspiked\_set1} background proteins as
a 1:1 sample. Therefore we exclude the ceruplasmins before fitting a
noise model using \Rfunction{NoiseModel}. See figure
~\ref{fig:maplots}.

<<calc-noisemodel>>=
ib.background <- subsetIBSpectra(ibspiked_set1,protein=ceru.proteins,direction="exclude")
noise.model <- NoiseModel(ib.background)
@ 

Though only recommended when sufficient data are available, a method
exist for the estimation of a noise model without a 1:1 dataset. It
takes longer time as it first computes all the protein ratios to shift
spectrum ratios to 1:1. To examplify this procedure, we only take rat
and mouse CERU proteins from \Robject{ibspiked\_set1}, see figure
~\ref{fig:maplots}. The resultant noise model is a rough approximation
only because of the very limitted data, see Breitwieser et
al. Supporting Information, submitted, for a real example.

<<ceru-noisemodel>>=
ib.ceru <- subsetIBSpectra(ibspiked_set1,protein=ceru.proteins,
                            direction="include",
                            specificity="reporter-specific")
nm.ceru <- NoiseModel(ib.ceru,one.to.one=FALSE,pool=TRUE)
@ 

<<maplot-noisemodel1,echo=FALSE>>=
png("graphics/fig_maplot_noisemodel.png",width=3.3,height=3,units="in",res=300,pointsize=8)
@
<<maplot-noisemodel2,echo=TRUE>>=
maplot(ib.background,noise.model=c(noise.model,nm.ceru),
       channel1="114",channel2="115",ylim=c(0.2,5),
       main="95% CI noise model")
@ 
<<maplot-noisemodel3,echo=FALSE,results=hide>>=
dev.off()
@ 

\begin{center}
\begin{figure}[htp]
\includegraphics{graphics/fig_maplot_noisemodel}

\caption{Red lines denote the 95 \% confidence interval as estimated
  by the noise model on background proteins. The blue line is
  estimated as non 1:1 noise model based on only spectra of CERU
  proteins.}
\label{fig:maplots}
\end{figure}
\end{center}

\subsection{Protein and peptide ratio calculation}
\Rfunction{estimateRatio} calculates the relative abundance of a
peptide or protein in one tag compared to another. It calculates a
weighted average (after outlier removal) of the spectrum ratios. The
weights are the inverse of the spectrum ratio variances. It requires a
\Robject{IBSpectra} and \Robject{NoiseModel} object and definitions of
channel1, channel2, and the protein or peptide. The result is
channel2/channel1.

<<estimateRatio>>=
## Calculate ratio based on all spectra of peptides specific 
##  to CERU_HUMAN, CERU_RAT or CERU_MOUSE. Returns a named
##  numeric vector.
10^estimateRatio(ibspiked_set1,noise.model,
                 channel1="114",channel2="115",
                 protein=ceru.proteins)['lratio']

## If argument 'combine=FALSE', estimateRatio returns a data.frame 
##  with one row per protein
10^estimateRatio(ibspiked_set1,noise.model,
                 channel1="114",channel2="115",
                 protein=ceru.proteins,combine=FALSE)[,'lratio']
## spiked material channel 115 vs 114: 
##                 CERU_HUMAN (P00450): 1:1
##                 CERU_RAT   (P13635): 2:1  = 2
##                 CERU_MOUSE (Q61147): 5:10 = 0.5

## Peptides shared between rat and mouse
pep.shared <- peptides(proteinGroup(ibspiked_set1),
                       c(ceru.rat,ceru.mouse),set="intersect",
                       columns=c('peptide','n.shared.groups'))
## remove those which are shared with other proteins
pep.shared <- pep.shared$peptide[pep.shared$n.shared.groups==2]

## calculate ratio: it is between the rat and mouse ratios
10^estimateRatio(ibspiked_set1,noise.model,
                 channel1="114",channel2="115",
                 peptide=pep.shared)['lratio']
@ 

When examining the global differences and differences in between
classes, \Rfunction{proteinRatios} can be used. It is also suitable to
inspect sample variability. The argument \Robject{cl} can be used to
define class labels. If \Robject{combn.method='interclass'} or
\Robject{intraclass} and \Robject{summarize=TRUE},
\Rfunction{proteinRatios} return a single summarized ratio across and
within classes, resp..

<<protein-ratios>>=
protein.ratios <- proteinRatios(ibspiked_set1,noise.model,cl=c("1","0","0","0"))

## defined class 114 and 115 as class 'T', 116 and 117 as class 'C'
classLabels(ibspiked_set1) <- c("T","T","C","C")

proteinRatios(ibspiked_set1,noise.model,protein=ceru.proteins,
              cl=classLabels(ibspiked_set1),combn.method="interclass",
              summarize=T)[,c("ac","lratio","variance")]
@ 

\subsection{Protein ratio distribution and selection}
Protein ratio distributions can be calculated ideally on biological
replicated. To examine differentially expressed proteins, both sample
variability information (random protein ratios) as a
\emph{fold-change} constraint, and ratio \emph{precision} can be used.
For a experimental setup with biolgical replicates in the same
experiment (but different channels), the distribution of biological
variability can be learned by computing the ratios between the
replicates. With no replicates available, one has the choice to (a)
model the actual protein ratios and just select the most extreme
ratios; (b) learn the distribution from a previous experiment; or (c)
assume a standard Cauchy distribution with location $0$ and scale
$0.1$, $0.05$, and $0.025$, which correspond with $\alpha=0.05$
roughly to fold changes of $4$, $2$, and $1.5$. 

A Cauchy distribution fits accurately this type of random protein
ratio distribution: Cauchy is displayed in red, Gaussian in blue. In
the case of \Robject{ibspiked\_set1}, the many 1:1 proteins provide us
with adequate data to learn the random protein ratio distribution,
however only of the \textit{technical} variation.


<<ratiodistribution>>=
#protein.ratios <- proteinRatios(ibspiked_set1,noise.model)
protein.ratiodistr.wn <- fitWeightedNorm(protein.ratios[,'lratio'],
                                         weights=1/protein.ratios[,'variance'])
protein.ratiodistr.cauchy <- fitCauchy(protein.ratios[,"lratio"])
@ 

\begin{center}
\begin{figure}[htp]
<<plot-ratiodistr,fig=TRUE,width=4,height=2.5>>=
library(distr) # required library
limits=seq(from=-0.5,to=0.5,by=0.001)
curve.wn <- data.frame(x=limits,y=d(protein.ratiodistr.wn)(limits))
curve.cauchy<-data.frame(x=limits,y=d(protein.ratiodistr.cauchy)(limits))

g <- ggplot(data.frame(protein.ratios),aes(x=lratio)) +
  geom_histogram(colour = "darkgreen", fill = "white",aes(y=..density..),
                 binwidth=0.02) + geom_rug() +
  geom_line(data=curve.wn,aes(x=x,y=y),colour="blue") +
  geom_line(data=curve.cauchy,aes(x=x,y=y),colour="red")
print(g)
@
\caption{Histogram of all protein ratios in
  \Robject{ibspiked\_set1}. A fit with a Gaussian and Cauchy
  probability density function is shown in blue and red,
  respectively.}
\label{fig:ratiodistr}
\end{figure}
\end{center}

Now, when supplying a \Robject{ratiodistr} parameter to
\Rfunction{estimateRatio} and \Rfunction{proteinRatios}, sample and
signal p-values are calculated, what we illustrate in the code below

<<significant-ratios>>=
rat.list <-
  estimateRatio(ibspiked_set1,noise.model=noise.model,channel1="114",channel2="115",
                protein=reporterProteins(proteinGroup(ibspiked_set1)),combine=F,
                ratiodistr=protein.ratiodistr.cauchy)
rat.list[rat.list[,"is.significant"]==1,]
@

\subsection{Detection of proteins with no specific peptides}
It is well known that MS analysis only reveals the presence of
so-called protein groups, defined as sets of proteins identified by
the same set of peptides. The protein that contains all the peptides
is the group reporter (there are possibly several group reporters) and
if it has one specific peptide at least then its presence in the
sample is certain. The status of the other proteins in the group is in
general impossible to determine.  When quantitative information is
available, there is a potential to elucidate the structure of part of
the protein groups.

In the example below, a subset \Rfunction{IBSpectra} object is created, 
   containing only peptides shared between CERU\_RAT and CERU\_MOUSE, 
   and those specific to CERU\_RAT.

<<shared-pep>>=
## peptides shared between CERU_RAT and CERU_MOUSE have been computed before
pep.shared
## peptides specific to CERU_RAT
pep.rat <- peptides(proteinGroup(ibspiked_set1),protein=ceru.rat,
                    specificity="reporter-specific")

## create an IBSpectra object with only CERU_RAT and shared peptides
ib.subset <- subsetIBSpectra(ibspiked_set1,
                             peptide=c(pep.rat,pep.shared),direction="include")

## calculate shared ratios
sr <- shared.ratios(ib.subset,noise.model,
                    channel1="114",channel2="117",
                    ratiodistr=protein.ratiodistr.cauchy)
sr

@ 
\begin{center}
\begin{figure}[htp]
<<fig=TRUE,width=6.6,height=1.8>>=
  ## plot significantly different protein groups where 90% CI does not overlap
  ## CERU_MOUSE and CERU_RAT is detected, as expected.
  shared.ratios.sign(sr,z.shared=1.282)
@ 
\caption{Peptides of spiked ceruplasmins have significantly different ratios 
    between groups. Group \textit{reporter} consists of peptides specific 
        to \texttt{CERU\_RAT (P13635)}, group \textit{member} are peptides 
        shared between \texttt{CERU\_RAT} and \texttt{CERU\_MOUSE (Q61147)}.}
\label{fig:sharedratios}
\end{figure}
\end{center}



\section{Report generation}
\Rpackage{isobar} provides a rich interface for creating Excel and PDF
reports for further analysis and quality control. The main entry
function is \texttt{create.reports}. Alternatively the Rscript
\texttt{create\_reports.R} can be used. It is located in the
\texttt{report} folder of the \isobar{} installation, and reads the
properties from a file in the working directory.

The posible values are defined in the \texttt{report/properties.R} file in the
isobar installation. To generate a report with standard properties the 
following code should do the trick:

<<eval=FALSE>>=
create.reports(type="iTRAQ4plexSpectra",
               identifications="my.id.csv",peaklist="my.mgf")
@ 

The properties can also be defined in a \texttt{properties.R} file
which is located in the working directory. The properties are set in
the following order:

\begin{itemize}
  \item 'global' properties in \texttt{ISOBAR-DIRECTORY/report/properties.R}\footnote{located in \texttt{system.file('report','properties.R',package='isobar')}}
  \item 'local' properies in \texttt{WORKING-DIRECTORY/properties.R}
  \item command line arguments to \texttt{create\_reports.R} or 
        \texttt{create.reports} function
\end{itemize}

Appendix \ref{sec:properties} provides a syntax-highlighted version of
the properties file supplied with isobar, which sets the default
parameters and provides some help in the comments. The number of
paramters which can be set may seem a lot at first, however most
times only a few are needed. 

For successful completion, \LaTeX - for the PDF reports - and Perl -
for the Excel reports - need to be installed.

\subsection{Files used for report generation}
<<eval=FALSE>>=
## execute to find the path and file location in your installation.
system.file("report",package="isobar") ## path
list.files(system.file("report",package="isobar")) ## files
@ 

\begin{description}
  \item[create\_reports.R] R script which can be used to create QC and PDF reports
    It initializes the environment, reads properties and calls \Rfunction{Sweave}
    on QC and DA Sweave files. Additionally it generates a Excel data analysis report
    by calling \texttt{tab2xls.pl}.
  \item[isobar-qc.Rnw] Sweave file with quality control plots.
  \item[isobar-analysis.Rnw] Sweave file for generating a data analysis report
    with the list of all protein ratios and list of significantly different proteins.
  \item[properties.R] Default configuration for \texttt{create\_reports.R}.
    It is parsed as R code.
  \item[report-utils.R] Helper R functions used in Sweave documents.
  \item[report-utils.tex] Helper \LaTeX{} functions used in Sweave documents.
\end{description}

\begin{appendix}

\section{File formats}
\label{app:fileformats}
\subsection{ID CSV file format}
The Perl parsers create ID CSV files - identification information for all
matched spectra without quantitative information. You can create your own parser, 
the resulting file should be tab-delimited and contain the following columns.
Only bold columns are obligatory. The information is redundant - that means if
a peptide may stem from two different proteins the information of the identification
is repeated.

\begin{tabular}{r|l}
\textbf{accession} & Protein AC \\
\textbf{peptide} & Peptide sequence\\
modif & Peptide modification string\\
charge & Charge state\\
theo.mass & Theoretical peptide mass\\
exp.mass & Experimentally observed mass\\
parent.intens & Parent intensity\\
retention.time & Retention time\\
\textbf{spectrum} & Spectrum identifier\\
search.engine & Protein search engine and score\\
\end{tabular}

\subsection{IBSpectra CSV file format}
IBSpectra file format has the same columns as the ID CSV format and additionally 
columns containing the quantitation information, namely
\texttt{X\textit{tagname}\_mass} and \texttt{X\textit{tagname}\_ions},
for mass and intensity of each tag \textit{tagname}. Below an example of the 
further columns for an \texttt{iTRAQ 4plex} IBSpectra.


\begin{tabular}{r|l}
\textbf{X114\_mass} & reporter ion mass\\
\textbf{X115\_mass} & reporter ion mass\\
\textbf{X116\_mass} & reporter ion mass\\
\textbf{X117\_mass} & reporter ion mass\\
\textbf{X114\_ions} & reporter ion intensity\\
\textbf{X115\_ions} & reporter ion intensity\\
\textbf{X116\_ions} & reporter ion intensity\\
\textbf{X117\_ions} & reporter ion intensity\\
\end{tabular}

\section{\texttt{properties.R} for report generation}
\label{sec:properties}
\definecolor{comment}{HTML}{000065}
\definecolor{string}{HTML}{DD0000}
\definecolor{special}{HTML}{0A0AF0}
{
\lstset{ %
language=R,                % the language of the code
basicstyle={\footnotesize \tt},       % the size of the fonts that are used for the code
commentstyle=\color{comment},
stringstyle=\color{string},
linewidth=\textwidth
}
<<echo=FALSE,results=tex>>=
#p <- readLines(system.file("report","properties.R",package="isobar"))
#p <- sub("(#.*)","\\\\textcolor{comment}{\\1}",p)
#for (i in c("TRUE","FALSE","NULL"))
#  p <- sub(i,paste("\\\\textcolor{special}{",i,"}",sep=""),p)
#p <- sub('(".*?")',"\\\\textcolor{string}{\\1}",p)
#p <- gsub("#","\\#",p,fixed=TRUE)
#p <- gsub("_","\\_",p,fixed=TRUE)
#cat("\\begin{lstlisting}")
#cat(p,sep="\n")
#cat("\\end{lstlisting}")
cat(sprintf("\\lstinputlisting{%s}\n",system.file("report","properties.R",package="isobar")))
@ 

}

\section{Dependencies}\label{sec:dependencies}
\subsection{\LaTeX{} and PGF/TikZ}
\LaTeX{} is a high-quality typesetting system; it includes features designed for the production of technical and scientific documentation. It is available as free software\footnote{\url{http://www.latex-project.org}}. PGF is a \TeX{} macro package for generating graphics It comes with a user-friedly syntax layer called TikZ\footnote{\url{http://sourceforge.net/projects/pgf}}.

\LaTeX{} is used for creating PDF analysis reports, with the PGF package creating the graphics. Go to \url{http://www.latex-project.org} to get information on how to download and install a \LaTeX{} system and packages.


\subsection{Perl}\label{app:installing-perl}
Perl is a high-level, general-purpose, interpreted, dynamic programming language. Perl is required for two tasks:
\begin{itemize}
  \item Conversion of Pidres XML and Mascot DAT files to ID CSV format;
  \item Creation of Microsoft Excel format data analysis report.
\end{itemize}
Go to \url{http://www.perl.org} to download and get help on the installation of Perl on your Operating System.
For file format conversion, perl module \texttt{Statistics::Lite} is required. For Excel export \texttt{Spreadsheet::WriteExcel}. All Perl scripts are in the subdirectory \texttt{pl} of the isobar package installation.

<<eval=FALSE>>=
## execute to find the path and file location in your installation.
system.file("pl",package="isobar") ## path
list.files(system.file("pl",package="isobar")) ## files
@ 

\texttt{mascotParser2.pl} and \texttt{pidresParser2.pl} convert from respective protein search outputfiles to a XML file format, which can be converted into a CSV file readable by \textit{isobar} by using \texttt{psx2tab2.pl}.

\texttt{mascotParser2.pl} coverts from Mascot format, and requires the file \texttt{modifconv.csv} as a definition of modification names. \texttt{pidresParser2.pl} converts from Phenyx output and requires the file \texttt{parsersConfig.xml}. \texttt{tab2xls.pl} converts csv file to different sheets of an Excel spreadsheet.

\label{app:itraqmascotparser-help}
<<eval=FALSE>>=
## execute on your system
system(paste("perl",system.file("pl","mascotParser2.pl",package="isobar"),
            "--help"))

print(paste("perl",system.file("pl","pidresParser2.pl",package="isobar"),
            "--help"))
@


\section{Session Information}

The version number of R and packages loaded for generating the vignette were:

<<<echo=FALSE,results=tex>>=
  toLatex(sessionInfo())
@ 
\end{appendix}

\end{document}

\bibliographystyle{plainnat}

% vim: set expandtab ts=3 sw=3:
\SweaveOpts{eps=FALSE}
\SweaveOpts{strip.white=false}
<<initialization,echo=FALSE,results=hide>>=
suppressPackageStartupMessages(library(distr))
suppressPackageStartupMessages(require(ggplot2))
suppressPackageStartupMessages(library(isobar))
suppressPackageStartupMessages(library(gridExtra))
suppressPackageStartupMessages(library(plyr))
options(encoding="UTF-8")  
pdf.options(pointsize=6)

load.tex.properties(.GlobalEnv) # loads environments 
n.tags <- length(reporterTagNames(ibspectra))
@

\documentclass[11pt,oneside,nogin]{article}
\usepackage[margin=2cm]{geometry}
\usepackage[utf8]{inputenc}
\usepackage{amsmath}
\usepackage{longtable}
\usepackage[pdftex]{graphicx}
\usepackage{float}
\usepackage{verbatim}
\usepackage{fancyhdr}
\renewcommand{\headrulewidth}{0pt}
\pagestyle{fancy}
\lfoot{\today}
\cfoot{\thepage}

<<echo=FALSE,results=tex>>=
write.tex.commands() # prints analysisname, analysisauthor, and isobarthanks commands
@

\rfoot{Isobar QC Report \analysisname}

\usepackage{hyperref}
\hypersetup{
  colorlinks=true,
  pdfauthor=\analysisauthor
}
\usepackage{titling}
\setlength{\droptitle}{-5em}   % less space for title

\title{Isobar QC Report \analysisname\isobarthanks}
\author{\analysisauthor}
\date{\today}

\raggedbottom

\begin{document}

\maketitle

\subsection*{Reporter Mass Precision}
Histogram representing the distribution of the delta mass (in
\emph{m/z}) for each reporter tag between theoretical and observed
mass. Reporter tag ions were extracted from MS2 fragment spectra at
$\pm$ \Sexpr{properties.env$fragment.precision/2} $m/z$ around the
theoretical masses.

<<fig-reporterMassPrecisionv,fig=T,echo=F,width=6.5,height=2.5,results=hide>>=
 print(reporterMassPrecision(ibspectra))
@

\subsection*{Reporter Intensities Normalization}

<<echo=FALSE,results=tex>>=
reporter.tags <- reporterTagNames(ibspectra)
n.factors <- setNames(rep("n.n.", length(reporter.tags)),reporter.tags)
nn.factors <- ibspectra@log[grep("normalization.multiplicative.factor",
                                 rownames(ibspectra@log)),"Message"]
names(nn.factors) <- sub(".* channel ","",names(nn.factors))
n.factors[names(nn.factors)] <- nn.factors

ri <- reporterIntensities(ibspectra)
ri.nn <- reporterData(ibspectra,element="ions_not_normalized") # null if not normalized
if (is.null(ri.nn)) {
    cat("Box-plot of the reporter intensities per labeled channel. No normalization was performed.
        
        ")
} else {
  cat("Box-plot of the reporter intensities before and after the
       normalization process. Isobar applies a normalization factor to impose
       equal median or summed intensity in each channel. Some channels may be
       excluded from normalization, as defined in the \\texttt{properties.R}.
       
       ")
}
@ 

\begin{minipage}[c]{0.4\textwidth}
<<fig-reporterIntensityPlot,fig=T,echo=F,width=2.5,height=2.3,results=hide>>=
 print(reporterIntensityPlot(ibspectra) + isobar:::.gg_theme(legend.position="bottom"))
@
\end{minipage}
\hfill
\begin{minipage}[c]{0.6\textwidth}
\vspace{0pt} \small
<<normalization factors,echo=FALSE,results=tex,strip.white=false>>=
if (is.null(ri.nn)) { 
   cat("\\begin{tabular}{rrrr}\n",
       paste("","",
             "\\multicolumn{2}{c}{intensites}",sep=" & "),"\\\\",
       paste("tag",
             "NA\\footnote{Percentage of spectra with no quantitative information}",
             "mean",
             "median",
             sep=" & "),
       " \\\\ \\hline",
       paste(names(n.factors),
             sprintf("%.2f",100*colSums(is.na(ri))/nrow(ri)),
             sprintf("%.2f",apply(ri,2,median,na.rm=TRUE)),
             sprintf("%.2f",colMeans(ri,na.rm=TRUE)),             
             sep=" & ",collapse="\\\\ \n"),
       "\\end{tabular}",sep="\n")

} else {
    cat("\\begin{tabular}{r@{\\hspace{1em}}rr@{\\hspace{1em}}rr@{\\hspace{1em}}rr}\n",
       paste("","",
             "\\multicolumn{2}{c}{before norm.}",
             "\\multicolumn{2}{c}{after norm.}",
#"\\multicolumn{2}{c}{after n.\\footnote{Intensities before and after normalization}}",
             "",sep=" & "),"\\\\",
       paste("tag",
             "NA\\footnote{Percentage of spectra with no quantitative information}",
             "mean",
             "median",
             "mean",
             "median",
             "factor\\footnote{n.n. \\ldots~excluded from normalization}",sep=" & "),
       " \\\\ \\hline",
       paste(names(n.factors),
             sprintf("%.2f",100*colSums(is.na(ri.nn))/nrow(ri.nn)),
             sprintf("%.1f",colMeans(ri.nn,na.rm=TRUE)),
             sprintf("%.1f",apply(ri.nn,2,median,na.rm=TRUE)),
             sprintf("%.1f",colMeans(ri,na.rm=TRUE)),             
             sprintf("%.1f",apply(ri,2,median,na.rm=TRUE)),
             n.factors,sep=" & ",collapse="\\\\ \n"),
       "\\end{tabular}",sep="\n")
   
}
@
\normalsize
\end{minipage}

\subsection*{Ratio Distribution}
Distribution of the computed protein ratios (bars) and the probability density function of 
a \Sexpr{class(ratiodistr)}  distribution (solid line) fitted as described below.

\begin{minipage}{0.45\textwidth}
<<fig-ratiodistribution,fig=TRUE,echo=FALSE,width=3,height=2>>=

#quant.tbl$ratio <- 10^ quant.tbl$lratio
limits <- seq(from=min(quant.tbl$lratio,na.rm=TRUE),to=max(quant.tbl$lratio,na.rm=TRUE),length.out=1000)
#breaks=c(0.1,0.2,1,5,10)
#breaks <- breaks[log10(breaks) %inrange% range(quant.tbl$lratio)]

g <- ggplot(quant.tbl,aes(x=lratio)) +
  geom_histogram(colour = "darkgreen", fill = "white",
                 aes(y=..density..),alpha=0.8,binwidth=1/20*(max(limits)-min(limits))) +
  geom_rug(alpha=0.1) +
#  scale_x_continuous("Ratio",breaks=log10(breaks),labels=breaks) +
  scale_x_continuous(expression(paste(log[10]," ratio"))) +
  theme_bw(base_size=10) +
  isobar:::.gg_theme(legend.position="none",
       axis.text.x = isobar:::.gg_element_text(colour="grey50"))

if (!is.null(ratiodistr)) {
  curve.distr <- data.frame(x=limits,y=d(ratiodistr)(limits))
  g <- g + geom_line(data=curve.distr,aes(x=x,y=y),colour="black")
}

print(g)

@
\end{minipage}
\hfill
<<show-ratiodistribution,echo=FALSE,results=tex,strip.white=false>>=
cat("\\begin{minipage}{0.5\\textwidth}\n")
if (!is.null(ratiodistr)) {
  cat("\\begin{verbatim}\n")
  print(ratiodistr)
  cat("\\end{verbatim}\n\n")
  if (!is.null(attr(ratiodistr,"method"))) {
     m <- attr(ratiodistr,"method")
     classLabels <- attr(ratiodistr,"cl")
     t <- table(classLabels)
     tagNames <- attr(ratiodistr,"tagNames")

     cat("Distribution calculated on \\emph{",m,"} ratios",sep="")
     classTxt <- sapply(names(t),function(class) 
                        sprintf("\\emph{%s} (tags %s)",class,
                                paste(tagNames[!is.na(classLabels) & classLabels==class],collapse=", ")))
     if (m == "intraclass") {
        cat(" within classes ",paste(classTxt[names(t)[t>1]],collapse=" and "),".")
     } else {
        cat(" with classes ",paste(classTxt[names(t)],collapse=", "),".")
     }
  }
} else {
   cat("No ratio distribution available.")
}
cat("\\end{minipage}")
if (!is.null(ratiodistr)) {
  probs <- c(0.005,0.01,0.025,0.05,0.95,0.975,0.99,0.995)
  qs <- q(ratiodistr)(probs)
  cat("\n\n \\small",
      "\\begin{tabular}{r",rep("r",length(probs)),"}","\n",
      "probabilities:&",paste(probs*100,"\\%",sep="",collapse="&"),"\\\\ \n",
      "distribution quantiles ($log_{10}$):&",paste(round(qs,2),collapse="&"),"\\\\ \n",
      "(normal scale):&",paste(round(10**qs,3),collapse="&"),"\\\\ \n",
      "\\end{tabular} \\normalsize",
      sep="")
}
@

\subsection*{Individual Ratio Distributions}

<<reporter-intensities-2,fig=TRUE,width=6.5,height=4.5,echo=F,results=hide>>=
if (all(c("class1","class2") %in% colnames(quant.tbl)) && 
    (!all(quant.tbl[,'class1']==quant.tbl[,'r1']))) {
  quant.tbl$comp <- sprintf("%s [%s] /\n %s [%s]",
                            quant.tbl$r2,quant.tbl$class2,
                            quant.tbl$r1,quant.tbl$class1)
  quant.tbl$channels <- sprintf("%s / %s",quant.tbl$r2,quant.tbl$r1)
} else {
  quant.tbl$comp <- sprintf("%s / %s",quant.tbl$r2,quant.tbl$r1)
}

g <- ggplot(quant.tbl,aes(x=lratio)) +
  geom_histogram(
                 aes(y=..density..,fill=comp),alpha=0.8,binwidth=1/20*(max(limits)-min(limits))) + facet_wrap(~comp,nrow=1) +
#  scale_x_continuous("Ratio",breaks=log10(breaks),labels=breaks) +
  scale_x_continuous(expression(paste(log[10]," ratio"))) +
  theme_bw(base_size=10) +
  isobar:::.gg_theme(legend.position="none",axis.text.x = isobar:::.gg_element_text(colour="grey50"))  + scale_fill_brewer(palette="Set1")

if (!is.null(ratiodistr)) {
  curve.distr <- data.frame(x=limits,y=d(ratiodistr)(limits))
  g <- g + geom_line(data=curve.distr,aes(x=x,y=y),colour="black",alpha=.5)
}

#gg <- grid.arrange(g,
#tableGrob(ddply(quant.tbl,c('channels','class2','class1')[c('channels','class2','class1') %in% colnames(quant.tbl)],
#                 function(x) c('number of ratios'=sum(!is.na(x$lratio)),
#                               'significantly\ndownregulated'=sum(x$is.significant[x$lratio<0],na.rm=TRUE),
#                               'significantly\nupregulated'=sum(x$is.significant[x$lratio>0],na.rm=TRUE))),
#           gpar.coretext = gpar(fontsize=9), gpar.coltext=gpar(fontsize=10), gpar.rowtext=gpar(fontsize=10)),
#ncol=1)

print(g)
#width <- max(6.5,nrow(unique(quant.tbl[,c('r1','r2')])) * 2)
#width<- 6.5
#pdf("reporter-intensities-2.pdf",width=width,height=4.5)
#print(gg)
#dev.off()
print(ddply(quant.tbl,c('channels','class2','class1')[c('channels','class2','class1') %in% colnames(quant.tbl)],
                 function(x) c('number of ratios'=sum(!is.na(x$lratio)),
                               'significantly\ndownregulated'=sum(x$is.significant[x$lratio<0],na.rm=TRUE),
                               'significantly\nupregulated'=sum(x$is.significant[x$lratio>0],na.rm=TRUE))))
@
%\includegraphics{reporter-intensities-2}

\subsection*{Ratio-intensity Plot}
Display of the ratio (\emph{y-axis}) versus the log10 average signal intensity (\emph{x-axis}) for all the reporter ratio combinations. The noise model used by Isobar is indicated as a solid red line.

<<fig-maplot,echo=FALSE,results=hide>>=
ppi <- 300
png("fig_maplot.png",width=6.5*ppi,height=4.5*ppi,res=ppi,pointsize=12)
maplot(ibspectra,noise.model=noise.model,bty="l",pairs=properties.env$qc.maplot.pairs)
dev.off()
@
\includegraphics{fig_maplot}

<<fig-ratio-variance,fig=TRUE,echo=FALSE,width=6,height=4,eval=FALSE>>=
#\section{Ratio Variance Plot}
#if (ep("preselected.proteins"))

quant.tbl <- merge(quant.tbl,ddply(quant.tbl,"sign.string",nrow),all.x=TRUE,by="sign.string")
quant.tbl$sign <- paste(quant.tbl$sign.string,": ",quant.tbl$V1,sep="")
#message(paste(unique(quant.tbl$sign),collapse="\n"))

g <- ggplot(quant.tbl,aes(x=lratio,y=sqrt(variance))) +
  geom_point(aes(colour=factor(is.significant),shape=factor(is.significant)),alpha=0.8) +
  scale_shape(solid=FALSE) + facet_grid(class1~class2) + theme_bw()
print(g)
@

\end{document}

%\VignetteIndexEntry{Usecases for isobar package}
%\VignetteDepends{}
%\VignetteKeywords{Documentation}
%\VignettePackage{isobar}
\documentclass[11pt]{article}

\SweaveOpts{keep.source=TRUE}
<<init,echo=FALSE>>=
  require(distr)
  require(ggplot2)
@ 

\usepackage{times}
\usepackage{hyperref}
\usepackage[utf8]{inputenc}
\usepackage{xcolor}
\usepackage{listings}

\textwidth=6.2in
\textheight=8.5in
\oddsidemargin=.1in
\evensidemargin=.1in
\headheight=-.3in

\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{{\texttt{#1}}}
\newcommand{\isobar}{\Rpackage{isobar}}

\title{Usecases for the \emph{isobar} package}
\author{Florian P.\ Breitwieser, Jacques Colinge}

\begin{document}

\maketitle
\tableofcontents

\section{}


\end{document}

\bibliographystyle{plainnat}

%\VignetteIndexEntry{isobar for quantification of PTM datasets}
%\VignetteDepends{}
%\VignetteKeywords{Documentation}
%\VignettePackage{isobar}
\documentclass[11pt]{article}

\usepackage{hyperref}
\usepackage{tikz}

\SweaveOpts{keep.source=TRUE}
\SweaveOpts{prefix.string=graphics/plot}
<<init,echo=FALSE>>=
  require(ggplot2)
  dir.create(file.path("graphics"), showWarnings = FALSE)
@ 

\usepackage[utf8]{inputenc}
\textwidth=6.2in
\textheight=8.5in
%\parskip=.3cm
\oddsidemargin=.05in
\evensidemargin=.05in
\headheight=-.3in
\newcommand{\isobar}{\texttt{isobar}}
\newcommand{\isobarPtm}{\texttt{isobar}$^{\sc ptm}$}

\title{\emph{isobar} for quantification of PTM datasets}
\author{Florian P.\ Breitwieser, Jacques Colinge}

\begin{document}

\maketitle
\tableofcontents

\section{Introduction}
\isobar{} \cite{Breitwieser_Colinge2011-JPR} version $2$ includes
modules to facilitate PTM quantification. This vignette describes its
parts, and how to use it to generate quantification reports.

<<load-isobar,results=hide>>=
  library(isobar) ## load the isobar package
@

Using \isobar{}, automatic report generation is straight-forward
given proper input files using the script \texttt{report/create\_reports.R}.
When called, it parses the globabl properties file \texttt{report/properties.R}
and then the \texttt{properties.R} in the current directory. Below is a small 
example \texttt{properties.R} for creating a PDF Quality Control and XLSX 
analysis report:

\begin{verbatim}
type="iTRAQ4plexSpectra"
## peaklist files for quantitation, by default all mgf file in directory
peaklist=list.files(pattern="*\\.mgf$")

## id files, by default all id.csv files in directory
identifications=list.files(pattern="*\\.id.csv$")

modif="PHOS" # modification to track (eg PHOS, ACET, MET)
ptm.info.f <- getPtmInfoFromNextprot
spreadsheet.format="xlsx"
\end{verbatim}

Reports will be generated calling
\texttt{path\_to\_isobar/report/create\_reports.R --peptide} from the
directory containing the peaklists, identifications and
\texttt{properties.R}.

\section{Modification Site Localization}
\isobar{} supports PhosphoRS \cite{Taus_Mechtler2011-JPR} and Delta
Score \cite{Savitski_Kuster2011-MCP} for modification site localization.

\paragraph{PhosphoRS integration}
The standalone Java version of PhosphoRS can be downloaded from
\url{http://cores.imp.ac.at/uploads/media/PhosphoRS.zip}. It features
a command line interface to a script which rescores localizations of
the modification for each peptide-spectrum match. It uses \textsc{xml}
files for input and output, which can be generated and parsed by
\isobar{}. 

<<eval=FALSE>>=
# Generate PhosphoRS XML input file based on MGF and identification file
#  massTolerance: fragment ion mass tolerance (in Da)
#  activationType: CID, HCD, or ETD
writePhosphoRSInput("phosphors.in.xml",
                      "identifications.id.csv","peaklist.mgf",
                       massTolerance=0.5,activationType="CID")
@

After calling PhosphoRS (\texttt{java -jar phosphoRS.jar
  phosphors.in.xml phosphors.out.xml}), the resulting \textsc{xml}
file can be read:

<<eval=FALSE>>=
# Read PhosphoRS XML output file
#   simplify: if TRUE, a data.frame is returned, else a list
#   besthit.only: if TRUE, only the best localization per spectrum is returned
readPhosphoRSOutput("phosphors.out.xml",simplify=TRUE,besthit.only=TRUE)
@

\texttt{getPhosphoRSProbabilities} is a convenience function calling 
the writer, the script, and the reader in succession. 

<<eval=FALSE>>=
getPhosphoRSProbabilities("identifications.id.csv","peaklist.mgf",
                            massTolerance=0.5,activationType="CID",
                            phosphors.cmd="java -jar phosphoRS.jar")
@ 

\paragraph{Delta Score calculation}
The Mascot Delta Score can be calculated directly by the parser
\texttt{mascotParser2.pl} and thresholded (\emph{e. g.}
\texttt{--minDeltaScore=10}). For CSV identification files which
contain all hits for each spectrum (not just the best one), the
function \texttt{calc.delta.score} within the R package is provided.

\paragraph{Using PhosphoRS and Delta Score in Report Generation.}
When generating an IBSpectra object from peaklist and identifications,
via \texttt{readIBSpectra}'s argument \texttt{annotate.spectra.f} a
function can be plugged in to extend or modify the identification
information. This can be used to calculate scores and filter
localization scores with \texttt{filterSpectraDeltaScore}) or
\texttt{annotateSpectraPhosphoRS}.

<<eval=FALSE>>=
# filterSpectraDeltaScore calls calc.delta.score 
#   if no column named delta.score is present in the data frame
# identifications below a min.delta.score are REMOVED
ib <- readIBSpectra("identifications.id.csv","peaklist.mgf",
                      annotate.spectra.f=function(...) 
                        filterSpectraDeltaScore(...,min.delta.score=10))


# filterSpectraPhosphoRS calls PhosphoRS to caluclate PhosphoRS probabilities
# identifications below a min.prob (PhosphoRS peptide isoform probability) 
# are marked to be NOT QUANTIFIED (use.for.quant=FALSE), but not removed
ib <- readIBSpectra("identifications.id.csv","peaklist.mgf",
                      annotate.spectra.f=
                        function(...) filterSpectraPhosphoRS(...,min.prob=0.9,
                           phosphors.cmd="java -jar PhosphoRS.jar"))
@

This can be used in report generation, too, where the
\texttt{readIBSpectra.args} can be set accordingly in the report
properties file \texttt{properties.R}:

\begin{verbatim}
readIBSpectra.args = list(annotate.spectra.f=filterSpectraDeltaScore)
\end{verbatim}
or
\begin{verbatim}
readIBSpectra.args = list(annotate.spectra.f=filterSpectraPhosphoRS)
\end{verbatim}

\section{Peptide Ratio Calculation}

All functions which are available to calculate ratios on protein level
can also be used for peptides. The same noise model is appropriate for
both.

<<>>=
data(ib_phospho)
data(noise.model.hcd)
head(proteinGroup(ib_phospho)@peptideInfo)
10^estimateRatio(ib_phospho,noise.model.hcd,peptide="SPLSPTETFSWPDVR")
@ 

By giving a matrix to \texttt{estimateRatio}, we can calculate ratios for peptides with specific modifications:

<<>>=
pep.n.modif <- unique(apply(fData(ib_phospho)[,c("peptide","modif")],2,cbind))
print(head(pep.n.modif))
estimateRatio(ib_phospho,noise.model.hcd,channel1="114",channel2="115",
                peptide=head(pep.n.modif),combine=FALSE)[,c("lratio","variance",
                                                            "n.spectra","p.value.rat")]

@ 

A ratio distribution can be calculated based on peptide ratios:

<<ratiodistr,fig=TRUE>>=
suppressPackageStartupMessages(library(distr))
suppressPackageStartupMessages(library(ggplot2))
peptide.ratios <- peptideRatios(ib_phospho,noise.model=noise.model.hcd,
                                  cmbn=matrix(c("114","116"),ncol=1))

lim <- max(abs(peptide.ratios$lratio),na.rm=TRUE)
peptide.distr.cauchy <- fitCauchy(peptide.ratios$lratio)

pseq <- seq(from=-lim,to=lim,length.out=1000)
ggplot() + 
  geom_histogram(aes(x=lratio,y=..density..),data=peptide.ratios,binwidth=0.05,
                 color="darkgreen",fill="white") +
  geom_line(aes(x=x,y=y),color="black",
            data=data.frame(x=pseq,y=d(peptide.distr.cauchy)(pseq)))
@ 


\paragraph{Correction with protein ratios.}
The observed change in concentration of modified peptides in one
condition versus another is integrating two separate modes of
regulation \cite{Wu_Gygi2011-MCP}:

\begin{enumerate}
  \item Protein expression change
  \item Modification state change
\end{enumerate}

In many cases, it thus can be advisable to conduct separate MS
quantification runs of the peptides enriched for the modification of
interest, AND the global proteome quantification. 

In the report
generation, data from other experiments can be integrated using the
property \texttt{compare.to.quant} in \texttt{properties.R}:

%\fcolorbox{gray}{gray}%
% {\color{white} 
\begin{verbatim}
load("../proteome/quant.tbl.rda")         # load proteome quantification table
compare.to.quant=list(proteome=quant.tbl) # set property
rm(quant.tbl)
\end{verbatim}
% }

Peptide ratios can also be corrected with proteome ratios of a
separate experiment, when giving as \texttt{peptide} argument a
\texttt{matrix} or \texttt{data.frame} with columns for 'peptide',
'modif', and 'correct.ratio'. 'correct.ratio' is a $log_{10}$ ratio
which will be used to adjust the one calculated for the specific
modified peptide.

<<>>=
peptides <- pep.n.modif[1:5,]

orig.ratio <- estimateRatio(ib_phospho,noise.model.hcd,channel1="114",channel2="115",
                              peptide=peptides,combine=FALSE)[,c("lratio","variance")]

peptides.c <- cbind(peptides,correct.ratio=c(0,-1,1,2,-2))
corr.ratio <- estimateRatio(ib_phospho,noise.model.hcd,channel1="114",channel2="115",
                              peptide=peptides.c,combine=FALSE)[,c("lratio","variance")]

data.frame(peptides.c,orig.ratio,corr.ratio)
@

As appearent, the variance stays the same also for corrected
ratios. If a fourth column \texttt{variance} of the \texttt{peptide}
argument reports the variance of the correction ratio, it is added to
the calculated ratio's variance (assuming independence).

% Todo: add part for report generation

\section{Harvesting public PTM databases}
neXtProt \cite{Lane_Bairoch2012-NAR} and PhosphoSitePlus
\cite{Hornbeck_Sullivan2012-NAR} provide information on experimentally
determined post-translational modifications. neXtProt focuses on man,
and PhosphoSitePlus on man and mouse. Both are manually curated and
annotate thousands of residues of post-translationally modified
proteins. 

\isobar{} provides functions to gather their information on identified
proteins.

<<eval=FALSE>>=
ptm.info <- getPtmInfoFromPhosphoSitePlus(proteinGroup(ib_phospho),modif="PHOS")
ptm.info <- getPtmInfoFromNextprot(proteinGroup(ib_phospho))
@ 
<<>>=
head(ptm.info)
@ 

For reports, the function can be selected via the property \texttt{ptm.info.f} in \texttt{properties.R}:

\begin{verbatim}
protein.info.f = getPtmInfoFromNextprot
\end{verbatim}

For PhosphoSitePlus, define the modification to get the correct dataset:
\begin{verbatim}
ptm.info.f <- function(...) getPtmInfoFromPhosphoSitePlus(...,modification="PHOS")
\end{verbatim}

PhosphoSitePlus datasets will be downloaded from their website to 'Phosphorylation\_site\_dataset.gz' or
'Acetylation\_site\_dataset.gz', etc (see mapping property of getPtmInfoFromPhosphoSitePlus) unless a file
with that name exists.

\bibliographystyle{abbrv}
\bibliography{isobar_lib}

\end{document}

\name{spectra.count2}
\alias{spectra.count2}
\title{
  Spectral count for peptides and proteins in ProteinGroup objects.
}
\description{
  Spectral count for peptides and proteins in ProteinGroup objects. 
  It can - other than \code{\link{spectra.count}} - quantify the spectra count on
  the level of peptides, potenitally modifed, too,
}
\usage{
spectra.count2(ibspectra, value = reporterProteins(protein.group), 
               type = "protein.g", specificity = c("reporter-specific", "group-specific", "unspecific"), 
               modif = NULL, combine = FALSE, subset = NULL, require.quant = NULL, ...)
}
\arguments{
  \item{ibspectra}{IBSpectra object.}
  \item{value}{List of protein group identifiers or peptides.}
  \item{type}{Either 'protein.g' or 'peptide'.}
  \item{specificity}{Specificity of peptides.}
  \item{modif}{Only count peptides having a certain modification.}
  \item{combine}{If TRUE, only one combined result is returned.}
  \item{subset}{Allows to specify an \code{\link{expression}} to subset \code{link{featureData}} of the ibspectra.}
  \item{require.quant}{If not NULL, it may be 'any' or 'all' to only consider
  spectra with quantitative information in at least one or all channels.}
  \item{\dots}{Further arguments to \code{\link{peptides}}}
}
\author{Florian P Breitwieser}

\seealso{
  \code{\link{spectra.count}},
  \code{\link{ProteinGroup}}
}
\examples{
  data(ibspiked_set1)
  pg <- proteinGroup(ibspiked_set1)
  protein.gs <- protein.g(pg,"CERU")
  sc <- spectra.count2(ibspiked_set1,protein.gs)
  sc.ik <- spectra.count2(ibspiked_set1,protein.gs,modif="iTRAQ4plex_K")
  rbind(spectra.counts=sc,spectra.counts_iTRAQk=sc.ik)
}

\name{ProteinGroup-class}
\docType{class}

% Class
\alias{class:ProteinGroup}
\alias{ProteinGroup-class}
\alias{ProteinGroup}

% Constructor-like function:
\alias{ProteinGroup}
\alias{ProteinGroup,data.frame,missing-method}
\alias{ProteinGroup,data.frame,NULL-method}
\alias{ProteinGroup,data.frame,ProteinGroup-method}
\alias{readProteinGroup}
\alias{readProteinGroup2}

% Coercion methods:
\alias{coerce,data.frame,ProteinGroup-method}
\alias{as.data.frame,ProteinGroup-method}
\alias{as.data.frame.ProteinGroup}
\alias{proteinGroup.as.concise.data.frame}

% Accessor-like methods:
\alias{peptideSpecificity}
\alias{peptideSpecificity,ProteinGroup-method}
\alias{peptideNProtein}
\alias{peptideNProtein,ProteinGroup-method}
\alias{proteinGroupTable}
\alias{proteinGroupTable,ProteinGroup-method}
\alias{peptideSpecificity}
\alias{peptideSpecificity,ProteinGroup-method}
\alias{peptideInfo}
\alias{peptideInfo-methods}
\alias{peptideInfo,ProteinGroup-method}

\alias{indistinguishableProteins}
\alias{indistinguishableProteins,ProteinGroup-method}
\alias{indistinguishableProteins,ProteinGroup,ANY,ANY-method}
\alias{indistinguishableProteins,ProteinGroup,character,missing-method}
\alias{indistinguishableProteins,ProteinGroup,missing,character-method}
\alias{indistinguishableProteins,ProteinGroup,missing,missing-method}
\alias{spectrumToPeptide}
\alias{spectrumToPeptide,ProteinGroup-method}
\alias{get.pep.group}
\alias{peptides}
\alias{peptides,ProteinGroup,missing-method}
\alias{peptides,ProteinGroup,character-method}
\alias{reporterProteins}
\alias{reporterProteins,ProteinGroup-method}
\alias{protein.ac}
\alias{protein.g}
\alias{protein.ac,ProteinGroup,character-method}
\alias{protein.ac,ProteinGroup,missing-method}
\alias{protein.g,ProteinGroup,character-method}
\alias{protein.g,ProteinGroup,character,character-method}

% show and method:
\alias{show,ProteinGroup-method}
\alias{summary.ProteinGroup}


\title{ProteinGroup objects}

\description{
  The ProteinGroup class is a container for identified peptides and 
	  proteins, and groups them to distinguish proteins with specific 
	  peptides.
}

\usage{
ProteinGroup(from,template=NULL,proteinInfo=data.frame())

protein.ac(x, protein.g)
protein.g(x, pattern, variables=c("AC","name"), ...)
}

\arguments{
  \item{from}{\code{data.frame} object to create a ProteinGroup from.
  See Details from column specifications}
  \item{template}{'template' ProteinGroup object for grouping.}
  \item{x}{ProteinGroup object}
  \item{protein}{character string}
  \item{proteinInfo}{data.frame for proteinInfo slot}
  \item{protein.g}{character string, denoting a 'protein group'.}
  \item{pattern}{character string, see \code{\link{grep}} for details.}
  \item{variables}{\code{AC} maps a protein accession code to a protein group.
  \code{name} maps using protein information from \code{proteinInfo}.}
  \item{\dots}{Passed on to \code{\link{grep}}.}
}

\details{
  The ProteinGroup class stores spectrum to peptide to protein mapping.
  
  The proteins are grouped by their evidence, i. e. peptides:
  \itemize{
    \item Peptides with changes only from Leucin to Isoleucin are
      considered the same, as they cannot be distinguished by MS.
    \item Proteins which are detected with the same peptides are
      grouped together to a 'indistinguishable protein'- normally
      these are splice variants.
    \item Proteins with specific peptides are 'reporters'.
    \item Proteins with no specific peptides are grouped under these 'reporters.
  }

  This information is stored in six slots:
  \describe{
    \item{\code{spectra.n.peptides}}{a named 'character' vector, names
  being spectrum identifier and values are peptides.}
    \item{\code{peptide.n.proteins}}{a 'data.frame' containing the
  number of proteins the peptides could derive from.}
    \item{\code{peptide.n.protein}}{a character 'matrix' linking
  peptides to proteins.}
    \item{\code{indistinguishable.proteins}}{a 'matrix' contain.}
  }
}

\section{Constructor}{
  \describe{
    \item{}{
      \code{ProteinGroup(tbl.prot.pep,template=NULL)}:
      Creates a ProteinGroup object.
      \describe{
        \item{\code{tbl.prot.pep}}{A 'data.frame' with three columns:
	  1. Protein, 2. Peptide, 3. Spectrum.}
        \item{\code{template}}{Optional ProteinGroup object the grouping
  is based upon.}
      }
    }
  }
}

\section{Coercion}{
  In the code snippets below, \code{x} is a ProteinGroup object.

  \describe{
    \item{\code{as(from, "ProteinGroup")}:}{
      Creates a ProteinGroup object from a data.frame.
    }
    \item{\code{as.data.frame(x, row.names = NULL, optional = FALSE)}:}{
      Creates a data.frame with columns \code{protein} (character),
      \code{peptide} (character), \code{spectrum}.}
    \item{\code{as.concise.data.frame(from)}:}{
      Creates a 'concise' data.frame with one spectrum per row, and protein ACs combined
    }
  }
}

\section{Accessors}{
  In the following code snippets, \code{x} is a ProteinGroup object.

  \describe{
    \item{\code{spectrumToPeptide(x)}:}{ Gets spectrum to peptide assignment.}
    \item{\code{peptideInfo(x)}:}{ Peptide information such as protein start position.}
    \item{\code{peptideSpecificity(x)}:}{ Gets a 'data.frame'
    containing the peptide specificity: they can be reporter-specific,
    group-specific, or non-specific.}
    \item{\code{peptideNProtein(x)}:}{ Gets peptide to protein assignment.}
    \item{\code{indistinguishableProteins(x)}:}{
      Gets the proteins which cannot be distinguished based on peptide evidence.
    }
    \item{\code{proteinGroupTable}:}{ Gets the protein grouping, listing
    reporters and group members. }
    \item{\code{peptides(x,protein=NULL,specificity=c("reporter-specific",
	"group-specific","unspecific"),columns="peptide",set=union)}:}{
      Gets all peptides detected, or just those for a protein with the
      defined specificity. \code{columns} might define multiple columns of
      \code{peptideSpecificity(x)}. set=union returns the union of
      peptides of all proteins defined, set=intersect returns the intersection.
      
    }
  }
}

\author{Florian P. Breitwieser}

\seealso{
  \link[isobar]{IBSpectra}
}

\examples{

tbl <- data.frame(spectrum=1:14,peptide=c(rep(letters[1:3],4),"a","x"),
                  modif=":",start.pos=1,
                  protein=c(rep(c("A","B"),each=6),"C","D"))
pg <- ProteinGroup(tbl)
pg
proteinGroupTable(pg)

data(ibspiked_set1)
pg <- proteinGroup(ibspiked_set1)
ceru.proteins <- protein.g(pg,"CERU")

## all ceru peptides
peptides(pg,ceru.proteins)

## peptides shared by all ceru proteins
peptides(pg,ceru.proteins, set=intersect) 

}

\name{human.protein.names}
\alias{human.protein.names}
\alias{my.protein.info}
\title{
    Info on proteins
}
\description{
    Gather human readable information from protein group codes.
}
\usage{

my.protein.info(x, protein.g)


human.protein.names(my.protein.info)
}
\arguments{
  \item{x}{ProteinGroup object}
  \item{protein.g}{protein}
  \item{my.protein.info}{Return value of function my.protein.info}
}
\author{Florian P Breitwieser}

\name{writeIBSpectra}
\alias{writeIBSpectra}
\title{
  Write IBSpectra file as CSV in a format readable by readIBSpectra.
}
\description{
  Write IBSpectra file using write.table with defaults in a format readable by readIBSpectra.
}
\usage{
writeIBSpectra(ibspectra, file, sep = "\t", row.names = FALSE, ...)
}
\arguments{
  \item{ibspectra}{IBSpectra object}
  \item{file}{file name.}
  \item{sep}{field separator string.}
  \item{row.names}{indicates whether row.names should be written.}
  \item{\dots}{further arguments to \code{\link{write.table}}}
}
\author{
  Florian P Breitwieser
}

\name{distr-methods}
\alias{distrprint}
\alias{twodistr.plot}
\alias{calcProbXGreaterThanY}
\alias{calcProbXDiffNormals}
\alias{calcCumulativeProbXGreaterThanY}

\alias{UnivariateDistribution-class}
\alias{AbscontDistribution-class}
\alias{Parameter-class}
\alias{Distribution-class}

\title{Functions for distribution calculations
}
\description{
  calcProbXGreaterThanY calculates the probability that X >= Y.
  calcProbXDiffNormals calculates the probabilities of a set of normals, defined 
  by the vectors mu_Y and sd_Y are greater or less than the reference distribution
  Y.                                                              
}
\usage{
  calcProbXGreaterThanY(X, Y, rel.tol = .Machine$double.eps^0.25, subdivisions = 100L)
  calcProbXDiffNormals(X, mu_Y, sd_Y, ..., alternative = c("greater", "less", "two-sided"), progress = FALSE)
  #calcCumulativeProbXGreaterThanY(Xs, mu_Ys, sd_Ys, alternative = c("greater", "less", "two-sided"), rel.tol = .Machine$double.eps^0.25, subdivisions = 100L)
  distrprint(X, round.digits = 5)
  twodistr.plot(X, Y, n.steps = 1000, min.q = 10^-3)
}
\arguments{
  \item{X}{ Object of the class Distribution. }
  \item{Y}{ Object of the class Distribution. }
  \item{min.q}{ minimum quantile }
  \item{n.steps}{ Number of steps. }
  \item{mu_Y}{ Numeric vector of parameter mu of a Normal. }
  \item{sd_Y}{ Numeric vector of parameter sd of a Normal. }
  \item{subdivisions}{the maximum number of subintervals}
  \item{rel.tol}{relative accuracy requested}
  \item{\dots}{ Additional arguments to calcProbXGreaterThanY. }
  \item{alternative}{ "less", "greater", or "two-sided". }
  \item{progress}{Show text progress bar?}
  \item{round.digits}{ Round digits for printing. }
}
\author{ Florian P. Breitwieser }

\examples{
  library(distr)
  calcProbXGreaterThanY(Norm(0,.25),Norm(1,.25))
}

\name{isobar-preprocessing}

\alias{isobar-preprocessing}
\alias{correctIsotopeImpurities}
\alias{correctIsotopeImpurities,IBSpectra-method}
\alias{normalize}
%\alias{normalize,IBSpectra-method}
\alias{subtractAdditiveNoise}
\alias{subtractAdditiveNoise,IBSpectra-method}
\alias{exclude}
\alias{exclude,IBSpectra,character-method}

\title{IBSpectra preprocessing}

\description{
  Preprocessing is a necessary step prior to analysis of data.
  In a sequential order, it is often neccassary to correct isotope
  impurities, to normalize, and subtract additive noise.
}

\section{Isotope impurity correction}{
  \describe{
    \item{\code{correctIsotopeImpurities(x)}:}{
      Returns impurity corrected IBSpectra object by solving a linear
      system of equations.
      See also \link{isotopeImpurities}.
      \describe{
        \item{\code{}}{}
    }
  }}
}

\section{Normalization}{
  \describe{
    \item{\code{normalize(x,f=median,target="intensity",exclude.protein=NULL,
	use.protein=NULL,f.doapply=TRUE,log=TRUE,channels=NULL,na.rm=FALSE)}:}{
      Normalizes the intensities for multiplicative errors. Those changes
      are most likely produced by pipetting errors, and different
      hybridization efficencies, but can also be due to biological
      reasons. 
      By default, tag intensities are multiplied by a factor so that the
      median intensity is equal across tags.
      \describe{
        \item{\code{f}:}{\code{f} is applied to each column, unless f.doapply is FALSE. 
          Then \code{f} is supposed to compute column-wise statistics of the matrix of 
          intensities. E.g. \code{colSums} and \code{colMeans}.}
	    \item{\code{target}:}{One of "intensity" and "ratio".}
      	\item{\code{exclude.proteins}}{
          Spectra of peptides which might come from these proteins are excluded. Use for example 
          for contaminants and proteins depleted in the experiment.}
        \item{\code{use.protein}:}{
          If specified, only spectra coming from this protein are used. Use when a protein 
          is spiked-in as normalization control.}
        \item{\code{f.isglobal}:}{If true, \code{f} is applied on each column. 
            If false, \code{f} is supposed to compute column-wise statistics of the matrix 
            of intensities. E.g. \code{colSums} and \code{colMeans}.} 
        \item{\code{log}:}{Used when target=ratio.}
      }
    }
  }
}

\section{Substract additive noise}{
  \describe{
    \item{\code{subtractAdditiveNoise(x,method="quantile",shared=TRUE,prob=0.01)}:}{
      
      \describe{
        \item{\code{method}}{'quantile' method is supported for now. It
	  take's the prob (0.01) quantile to estimate the noise level.
	  This value is subtracted from all intensities, and all
	  remaining intensities have to be at least that value.
	}
	\item{\code{prob}}{See 'method'.}
	\item{\code{shared}}{
	  If channels are assumed similar in intensity and hence a
	  shared noise level is reasonable.
	  If not, then one level per channel is necessary.
	}

      }
    }
  }
}


\section{Exclusion of proteins}{
  \describe{
    \item{\code{exclude(x,proteins.to.exclude)}:}{
      Removes spectra which are assigned to proteins in
      protein.to.exclude from the object. This can be useful to remove
      contaminants. It create a new grouping based on the data which
      is left.
      \describe{
        \item{\code{proteins.to.exclude}}{Proteins to exclude.}
    }
  }}
}


\author{Florian P. Breitwieser, Jacques Colinge}

\seealso{
  \link{ProteinGroup},
  \link{IBSpectra},
  \link{isobar-analysis},
  \link{isobar-plots}
}

\examples{
data(ibspiked_set1)
maplot(ibspiked_set1,main="IBSpiked, not normalized")
maplot(normalize(ibspiked_set1),main="IBSpiked, normalized")

}

\name{sanitize}

\alias{sanitize}

\title{
        Helper function for LaTeX export
}
\description{
        Sanitizes strings for LaTeX
}

\usage{
    sanitize(str, dash = TRUE)
}
\arguments{
  \item{str}{character string to be escaped}
  \item{dash}{shoud a dash ('-') should be escaped to a '\\nobreakdash-'?}
}
\value{
    escaped character
}
\author{iQuantitator,Florian P Breitwieser
}
\examples{
    sanitize("\\textbf{123-123}")
}

\name{calcPeptidePosition}
\alias{calcPeptidePosition}
\title{Recalculate peptide start positions based on protein sequence
}
\description{Function to recalculate start position of peptide in protein when it is missing or wrong.
}
\usage{
calcPeptidePosition(peptide.info, protein.info, calc.il.peptide)
}
\arguments{
  \item{peptide.info}{Peptide info object of ProteinGroup. }
  \item{protein.info}{Protein info object of ProteinGroup. }
  \item{calc.il.peptide}{Should the 'real' peptide (I/L difference) be calculated?}
}

\name{TlsParameter-class}
\Rdversion{1.1}
\docType{class}
\alias{TlsParameter-class}

\title{Class \code{"TlsParameter"}}
\description{The parameter of a location scale t distribution, used by Tlsd-class
}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("TlsParameter", ...)}.
Usually an object of this class is not needed on its own, it is generated automatically when an object of the class
\code{Tlsd} is instantiated. 
}
\section{Slots}{
  \describe{
    \item{\code{df}:}{Object of class \code{"numeric"} ~~ }
    \item{\code{location}:}{Object of class \code{"numeric"} ~~ }
    \item{\code{scale}:}{Object of class \code{"numeric"} ~~ }
    \item{\code{name}:}{Object of class \code{"character"} ~~ }
  }
}
\section{Extends}{
Class \code{"\linkS4class{Parameter}"}, directly.
Class \code{"\linkS4class{OptionalParameter}"}, by class "Parameter", distance 2.
}
\section{Methods}{
No methods defined with class "TlsParameter" in the signature.
}
\author{
  Florian P. Breitwieser, based on original TParameter class.
}
\seealso{
\code{\linkS4class{Tlsd}}
}
\examples{
showClass("TlsParameter")
}
\keyword{classes}

\name{Protein and peptide ratio calculation and summarization}
\alias{proteinRatios}
\alias{peptideRatios}
\alias{peptideRatiosNotQuant}
\alias{combn.matrix}
\alias{combn.protein.tbl}
\alias{summarize.ratios}
\alias{weightedMean}
\alias{weightedVariance}
\alias{weightedVariance,numeric,numeric,missing-method}
\alias{weightedVariance,numeric,numeric,numeric-method}
\alias{weightedMean,numeric,numeric-method}

\title{Calculating and Summarizing Protein and Peptide Ratios}
\description{
  A set of functions to create ratios within groups and summarize them. 
  \code{proteinRatios} serves as hub and calls \code{combn.matrix}, 
  \code{combn.protein.tbl} and \code{summarize.ratios} successively. It can be 
  used to calculate intra-class and inter-class ratios, to assess ratios and
  variability within and over cases.
}
\usage{
proteinRatios(ibspectra, noise.model, reporterTagNames = NULL,
                 proteins = reporterProteins(proteinGroup(ibspectra)),
                 peptide = NULL, cl = classLabels(ibspectra),
                 combn.method = "global", combn.vs = NULL, 
                 symmetry = FALSE, summarize =
                 FALSE, summarize.method = "mult.pval", min.detect =
                 NULL, strict.sample.pval = TRUE, strict.ratio.pval =
                 TRUE, orient.div = 0, sign.level = 0.05,
                 sign.level.rat = sign.level, sign.level.sample =
                 sign.level, ratiodistr = NULL, zscore.threshold =
                 NULL, variance.function = "maxi", combine = FALSE,
                 p.adjust = NULL, reverse = FALSE, cmbn = NULL,
                 before.summarize.f = NULL, ...)

peptideRatiosNotQuant(ibspectra, ..., peptide = unique(fData(ibspectra)[!fData(ibspectra)[["use.for.quant"]], c("peptide", "modif", "site.probs")]))
peptideRatios(ibspectra, ..., peptide = peptides(proteinGroup(ibspectra), columns = c("peptide", "modif")))
              
combn.matrix(x, method = "global", cl = NULL, vs = NULL)

combn.protein.tbl(cmbn, reverse = FALSE, ...)

summarize.ratios(ratios, by.column = "ac", summarize.method = "mult.pval",
                 min.detect = NULL, n.combination = NULL, 
                 strict.sample.pval = TRUE, strict.ratio.pval = TRUE, 
                 orient.div = 0, sign.level = 0.05, sign.level.rat = 
                 sign.level, sign.level.sample = sign.level, 
                 variance.function = "maxi", ratiodistr = NULL)

}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{ibspectra}{IBSpectra object}
  \item{x}{for combn.matrix: reporter names. See reporterTagNames. argument of proteinRatios.}
  \item{ratios}{result of combn.protein.tbl}
  \item{by.column}{Column(s) which are the identifiers. Usually 'ac', 'peptide' or c('peptide','modif') }
  \item{cmbn}{result of combn.matrix}
  \item{before.summarize.f}{Function which is called after calculating ratios before summarizing them.}
  \item{noise.model}{NoiseModel for spectra variances}
  \item{reporterTagNames}{Reporter tags to use. By default all reporterTagNames of ibspectra object.}
  \item{proteins}{proteins for which ratios are calculated - defaults to all 
     proteins with peptides specific to them.}
  \item{peptide}{peptides for which ratios are calculated.}
  \item{cl}{Class labels. See also ?classLabels. }
  \item{vs}{Class label or reporter tag name. When \code{combn.method} is "versus.class", 
            all combinations against class \code{vs} are computed, 
            when \code{combn.method} is "verus.channel", all combinations 
            against channel \code{vs}.}
  \item{combn.method}{"global", "interclass", "intra-class", "versus.class" or "versus.channel". Defines which ratios
     are computed, based on class labels cl}
  \item{method}{See combn.method}
  \item{combn.vs}{vs argument for combn, if combn.method is "versus.class" or "versus.channel".}
  \item{symmetry}{If true, reports also the inverse ratio}
  \item{summarize}{If true, ratios for each protein are summarized.}
  \item{summarize.method}{"isobar", for now.}
  \item{min.detect}{How many times must a ratio for a protein be present when
     summarizing? When NULL, defaults to the maximum number of combinations.}
  \item{strict.sample.pval}{If true, missing ratios are penalized by giving 
     them a sample.pval of 0.5.}
  \item{strict.ratio.pval}{If true, take all ratios into account. If false, only
    take ratios into account which are in the same direction as the majority of 
    ratios}
  \item{orient.div}{Number of ratios which might go in the wrong direction.}
  \item{sign.level}{Significance level}
  \item{sign.level.rat}{Significance level on ratio p-value}
  \item{sign.level.sample}{Significance level on sample p-value}
  \item{ratiodistr}{Protein ratio distribution}
  \item{variance.function}{Variance function}
  \item{zscore.threshold}{z-score threshold to apply}
  \item{\dots}{Passed to estimateRatio()}
  \item{combine}{If true, a single ratio for all proteins and peptides, resp., is calculated. See \code{\link{estimateRatio}}.}
  \item{p.adjust}{Set to one of p.adjust.methods to adjust ratio p-values for multiple comparisions. See \code{\link{p.adjust}}.}
  \item{reverse}{reverse}
  \item{n.combination}{number of combinations possible}
}
\value{
'data.frame':	11 variables:
\item{lratio }{log ratio}
\item{variance }{variance}
\item{n.spectra }{Number of spectra used for quantification}
\item{p.value.rat }{Signal p-value (NA if ratiodistr is missing)}
\item{p.value.sample }{Sample p-value (NA if ratiodistr is missing)}
\item{is.significant }{Is the ratio significant? (NA if ratiodistr is
  missing)}
\item{protein }{Protein quantified}
\item{r1 }{r1}
\item{r2 }{r2}
}
\author{
  Florian P Breitwieser, Jacques Colinge
}
\seealso{
  \link{IBSpectra},
  \link{isobar-preprocessing}
  \link{isobar-analysis}
}
\examples{

combn.matrix(114:117,method="interclass",cl=as.character(c(1,1,2,2)))
combn.matrix(114:117,method="interclass",cl=as.character(c(1,1,2,2)))
combn.matrix(114:117,method="global")

data(ibspiked_set1)
data(noise.model.hcd)

ceru.proteins <- c("P13635","Q61147")
proteinRatios(ibspiked_set1,noise.model=noise.model.hcd,proteins=ceru.proteins,cl=c("T","T","C","C"),combn.method="interclass",summarize=TRUE)

}

\name{getPeptideModifContext}
\alias{getPeptideModifContext}
\title{Get context of modification}
\description{Gets neighboring amino acids around modification which can be used to find enriched motifs.
}
\usage{
getPeptideModifContext(protein.group, modif, n.aa.up = 7, n.aa.down = 7)
}
\arguments{
  \item{protein.group}{ProteinGroup object.}
  \item{modif}{Modification of interest.}
  \item{n.aa.up}{Number of AA downstream to report.}
  \item{n.aa.down}{Number of AA upstream to report.}
}

\name{getPtmInfo}
\alias{getPtmInfoFromPhosphoSitePlus}
\alias{getPtmInfoFromNextprot}

\title{
  Get PTM site information for idenfied proteins from public databases.
}
\description{
  Get PTM site information for idenfied proteins from public databases.
}
\usage{
  getPtmInfoFromPhosphoSitePlus(protein.group, file.name = NULL, modif = "PHOS", 
                                psp.url = "http://www.phosphosite.org/downloads/", 
                                mapping = c(PHOS = "Phosphorylation_site_dataset.gz", 
                                            ACET = "Acetylation_site_dataset.gz", 
                                            METH = "Methylation_site_dataset.gz", 
                                            SUMO = "Sumoylation_site_dataset.gz", 
                                            UBI = "Ubiquitination_site_dataset.gz"))

  getPtmInfoFromNextprot(protein.group, nextprot.url = "http://www.nextprot.org/rest/entry/NX_XXX/ptm?format=json",
                         url.wildcard = "XXX")
}
\arguments{
  \item{protein.group}{ProteinGroup object.}
  \item{file.name}{File name to save downloaded data, defaults to the original file name (see mapping).}
  \item{modif}{Selects dataset to download (see mapping).}
  \item{psp.url}{PhosphoSitePlus main URL for datasets.}
  \item{mapping}{Names of PhosphoSitePlus modification datasets, mapped by modif name.}
  \item{nextprot.url}{URL for fetching Nextprot results. \code{url.wildcard} will be replaced by the Uniprot Protein AC.}
  \item{url.wildcard}{wildcard to replace with Uniprot protein AC in \code{nextprot.url}.}
}
\details{PhosphoSitePlus datasets are downloaded and written to the working directory with its original name (see mapping) 
  unless a file with that name exists, which is then parsed into a data.frame of suitable format.
}
\value{
  data.frame with (at least) the columns:
    isoform_ac, description, evidence, position
}
\references{
  PhosphoSitePlus: a comprehensive resource for investigating the structure and function of experimentally determined post-translational modifications in man and mouse.
  Hornbeck PV, Kornhauser JM, Tkachev S, Zhang B, Skrzypek E, Murray B, Latham V, Sullivan M.
  Nucleic Acids Res. 2012 Jan;40(Database issue):D261-70. Epub 2011 Dec 1.

  neXtProt: a knowledge platform for human proteins.
  Lane L, Argoud-Puy G, Britan A, Cusin I, Duek PD, Evalet O, Gateau A, Gaudet P, Gleizes A, Masselot A, Zwahlen C, Bairoch A.
  Nucleic Acids Res. 2012 Jan;40(Database issue):D76-83. Epub 2011 Dec 1.
}
\author{
  Florian P. Breitwieser
}
\note{
  PhosphoSitePlus is licensed under Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License and is freely available for non-commercial purpose,
  see http://www.phosphosite.org/staticDownloads.do.

  neXtProt is licensed under the Creative Commons Attribution-NoDerivs License, see: http://creativecommons.org/licenses/by-nd/3.0.

  Please read the conditions and use the data only if you agree.
}

\examples{

\dontrun{
  data(ib_phospho)
  ptm.info.np <- getPtmInfoFromNextprot(proteinGroup(ib_phospho))
  ptm.info.np <- ptm.info.np[grep("Phospho",ptm.info.np$modification.name),]
  ptm.info.psp <- getPtmInfoFromPhosphoSitePlus(proteinGroup(ib_phospho),modif="PHOS")

  str(ptm.info.np)
  str(ptm.info.psp)

  }
}

\name{shared.ratios}
\alias{shared.ratios}
\title{
  Shared ratio calculation
}
\description{
  Calculate ratios of reporter proteins and subset proteins with shared peptides.
}
\usage{
shared.ratios(ibspectra, noise.model, channel1 , channel2 , protein = reporterProteins(proteinGroup(ibspectra)), ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{ibspectra}{IBspectra object.}
  \item{noise.model}{NoiseModel object.}
  \item{channel1}{channel1 to compare.}
  \item{channel2}{channel2 to compare.}
  \item{protein}{proteins for which the calculation should be made.}
  \item{...}{Additional arguments passed to estimteRatio.}
}
\value{
  data.frame
}
\author{
  Florian P.\ Breitwieser
}

\seealso{
  \code{\link{shared.ratios.sign}}
}

\name{ratiosReshapeWide}
\alias{ratiosReshapeWide}

\title{Reshape output of proteinRatios into wide format}
\description{
  Reshape output of proteinRatios into wide format
}

\usage{
ratiosReshapeWide(quant.tbl, vs.class = NULL, sep = ".", cmbn = NULL,
                  short.names = FALSE)
}
\arguments{
  \item{quant.tbl}{Output of proteinRatios or peptideRatios.}
  \item{vs.class}{Only return ratios where class1 is vs.class}
  \item{sep}{Separator for column names in the reshape.}
  \item{cmbn}{Not functional.}
  \item{short.names}{If vs.class is set and short.names=TRUE, then the comparision 
    name will be i.e. 'class2' instead of 'class2/class1'.}
}
\author{Florian P. Breitwieser}

\name{groupMemberPeptides}
\alias{groupMemberPeptides}
\title{
    Peptide info for protein group members
}
\description{
    For a given reporter protein group identifier, information on
    its peptides is returned. It contains information on how the
    peptides are shared and in which member they occur.
}
\usage{
groupMemberPeptides(x, reporter.protein.g, ordered.by.pos = TRUE, only.first.pos = TRUE)
}
\arguments{
  \item{x}{ProteinGroup object}
  \item{reporter.protein.g}{group reporter protein}
  \item{ordered.by.pos}{if TRUE, start position of peptides in 
      proteins is exported and peptides are ordered by position
}
  \item{only.first.pos}{
      if TRUE, only first occurence of peptide in protein is reported
}
}
\value{
    list of two:
        [1] peptide.info: data.frame
             peptide
             specificity
             n.shared.groups
             n.shared.proteins
             start.pos
        [2] group.member.peptides: data.frame
             each column corresponds to a group member, and
             each row to a peptide
}
\author{Florian P Breitwieser}

\examples{
    data(ibspiked_set1)
    protein.group <- proteinGroup(ibspiked_set1)
    ceru.rat <- protein.g(protein.group,"CERU_RAT")
    groupMemberPeptides(protein.group,ceru.rat)

## find protein groups with members
    t <- table(proteinGroupTable(protein.group)$reporter.protein)
    t[t>2]
    protein.g <- names(t)[t>2][1]
    groupMemberPeptides(protein.group,protein.g)
}

\name{specificities}
\alias{specificities}
\alias{group-specific}
\alias{reporter-specific}
\alias{unspecific}
\alias{UNSPECIFIC}
\alias{GROUPSPECIFIC}
\alias{REPORTERSPECIFIC}
\alias{SPECIFICITIES}
\docType{data}
\title{
    Peptide specificities
}
\description{
    Peptides can appear in multiple proteins and therefore have different specificities.
}
\details{
    reporter specific: peptides specific to reporter.
    group specific: peptides specific to the group.
    unspecific: peptides shared with other proteins.
}
\keyword{datasets}

\name{isobar.data}
\alias{ibspiked_set1}
\alias{ibspiked_set2}
\alias{noise.model.hcd}
\alias{ib_phospho}
\docType{data}
\title{
  Isobar Data packages
}
\description{
  ibspiked_set1 and ibspiked_set2 are objects of class iTRAQ4plexSpectra.
  It contains over 160 protein groups, over 1600 peptides from about 15,000 spectra each,
  mainly from background proteins and three spiked-in Ceruplasmins 
  (CERU_HUMAN, CERU_MOUSE, CERU_RAT).
}
\usage{
  data(ibspiked_set1)
  data(ibspiked_set2)
  data(ib_phospho)
}
\format{
  iTRAQ4plexSpectra objects.
}
\source{
  isobar publication. Acquired on Orbitrap instrument w/ 20 offline-fractions 
  and HCD fragmentation.
}
\examples{
data(ibspiked_set1)
print(ibspiked_set1)
}
\keyword{datasets}

\name{shared.ratios.sign}
\alias{shared.ratios.sign}
\title{
  Plot and get significantly shared ratios.
}
\description{
  Plot and get significantly shared ratios.
}
\usage{
shared.ratios.sign(ress, z.shared, min.spectra = 1, plot = TRUE)
}
\arguments{
  \item{ress}{Result of shared.ratios.}
  \item{z.shared}{z. }
  \item{min.spectra}{Minimal number of spectra needed.}
  \item{plot}{plot. }
}
\author{
  Florian P.\ Breitwieser
}

\seealso{
  \code{\link{shared.ratios}}.
}

\name{number.ranges}
\alias{number.ranges}
\title{
    Helper function to transform number lists to ranges
}
\description{
    1,2,3,4,5,8,9,10 -> 1-5,8-10
}

\usage{
number.ranges(numbers)
}
\arguments{
  \item{numbers}{numeric}
}
\value{character}
\examples{
    number.ranges(c(1,2,3,9,3,10,8,11))
}
\author{Florian P Breitwieser}

\name{writeHscoreData}
\alias{writeHscoreData}
\title{
  Write identifications into a format suitable for Hscore.
}
\description{
  Write identifications into a format suitable for Hscore.
}
\usage{
writeHscoreData(outfile, ids, massfile = "defs.txt")
}
\arguments{
  \item{outfile}{Output file. }
  \item{ids}{ IBSpectra identifications data.frame (ie fData). } 
  \item{massfile}{ Definition file for Hscore. }
}
\author{Florian P. Breitwieser }

\name{isobar-import}

\alias{isobar-import}

% readIBSpectra
\alias{readIBSpectra}
\alias{readIBSpectra,character,character,character-method}
\alias{readIBSpectra,character,character,missing-method}
\alias{readIBSpectra,character,data.frame,missing-method}
\alias{readIBSpectra,character,data.frame,character-method}
\alias{readIBSpectra,character,character-method}
\alias{read.mzid}

\title{Loading data into IBSpectra objects using readIBSpectra}

\description{
        Read ibspectra-csv files and peaklist files as an
	IBSpectra object of type 'type' (see \code{\link{IBSpectra}},
	e.g. iTRAQ4plexSpectra or TMT6plexSpectra). If
	peaklist.file is missing, it is assumed that id.file contains
	intensity and m/z columns for the reporter tags.
}
      
\usage{
\S4method{readIBSpectra}{character,character}(type,id.file)

# reads id file
\S4method{readIBSpectra}{character,character,character}(
                 type, id.file, peaklist.file, sep = "\t", mapping.file
                 = NULL, mapping = c(quantification.spectrum = "hcd",
                 identification.spectrum = "cid"), id.file.domap =
                 NULL, identifications.format = NULL, decode.titles =
                 FALSE, ...)

# reads peaklist file
\S4method{readIBSpectra}{character,data.frame,character}(
                 type, id.file, peaklist.file, annotate.spectra.f =
                 NULL, peaklist.format = NULL, scan.lines = 0,
                 fragment.precision = NULL, fragment.outlier.prob =
                 NULL, ...)
}

\arguments{
  \item{type}{
    Name of class of new IBSpectra object: 
    \code{\link{iTRAQ4plexSpectra}}, \code{\link{iTRAQ8plexSpectra}},
    \code{\link{TMT2plexSpectra}}, \code{\link{TMT6plexSpectra}}, or \code{\link{TMT10plexSpectra}}
  }
  \item{id.file}{
    Database search results file in \code{ibspectra.csv} or
   \code{mzIdentML} format. See \code{identifications.format}. See the vignette for
   information on converting Mascot dat and Phenyx pidres files into
   ibspectra format.
  }
  \item{peaklist.file}{
    Peaklist file, typically in MGF format, see \code{peaklist.format}.
    MGF must be centroid!
  }
%  \item{proteinGroupTemplate}{
%    When having technical or biological repeats: First a template
%    protein group is created which uses information from all runs,
%    then this template is applied.
%    It should increase comparability across runs.
%  }
  \item{mapping.file}{
    If defined, spectum titles from the peaklist file are linked to
    the identifications via this file. This can be used when running HCD
    runs for quantification and CID runs for identification. See
    Koecher et al., 2009 for details.
  }
  \item{mapping}{
    Named character vector defining the names of columns
    in mapping.file. The names must be 'peaklist' and 'id', and
    the values must correspond to colnames of the mapping files.
  }
%  \item{mapping.file.readopts}{ 
%    Read options for \code{\link{read.table}} when reading files specified
%    in \code{mapping.file}.
%  }
  \item{id.file.domap}{
    When using HCD-CID or a method akin and every spectrum is used for 
    identification, the ID result files  of the HCD run can be
    specfied in \code{id.file.domap}. Then, the results are merged after
    mapping the identification results.
  }
  \item{annotate.spectra.f}{
    Function which changes or annotates the spectra feature data
    before it is written to IBpectra object. This can be used to calculate 
    and threshold additional scores, for example localization scores of post-
    translational modifications such as Delta Score (filterSpectraDeltaScore) or
    PhosphoRS site localization probabilities (annotateSpectraPhosphoRS).
  }
  \item{peaklist.format}{
    "mgf" (Mascot Generic format) or "mcn" (iTracker Machine Readable
    output). When NULL, it detects the format on file name extension.
  }
  \item{identifications.format}{
    "ibspectra.csv" or "mzid"  (PSI MzIdentML format). When NULL, file
    format is guessed based on extension.
  }
  \item{fragment.precision}{
    Fragment precision for extraction of reporter tags: for each tag
    and spectrum the m/z-intensity pair with it's mass closest to the
    known reporter tag mass is extracted within the window
    true_mass +/- fragment.precision/2.
  }
  \item{fragment.outlier.prob}{
    Fragment outlier probability filter: After all m/z-intensity pairs
    have been extracted, those pairs with the fragment.outlier.prob/2
    most unprecise m/z values are filtered out.
  }
  \item{decode.titles}{
    Boolean. Decode spectrum titles in identification file using \code{\link{URLdecode}}. 
    When extracting the  DAT file from Mascot web interface, the spectrum titles are
    encoded - \%20 instead of space, etc. Set decode.titles to TRUE to
    map these titles to the unescaped MGF titles.
  }
  \item{scan.lines}{
    Read files sequentially scan.lines lines at a time. Can help in
    case of memory issues, set to 10000 or higher, for example.
  }
  \item{sep}{sep argument of read.table}
  \item{\dots}{
    Further arguments handed down to \code{initialize}.
  }
}


\author{Florian P. Breitwieser, Jacques Colinge}

\seealso{
  \link{ProteinGroup},
  \link{IBSpectra},
  \link{isobar-preprocessing},
  \link{isobar-analysis},
  \link{isobar-plots}
}

\examples{
data(ibspiked_set1)

# get identifier for Ceruplasmin proteins
ceru.acs <- protein.g(proteinGroup(ibspiked_set1),"CERU")
# create a smaller ibspectra w/ only Ceruplasmins
ib.ceru <- subsetIBSpectra(ibspiked_set1,protein=ceru.acs,direction="include")

# write it to a file
tf <- tempfile("isobar")
write.table(as.data.frame(ib.ceru),sep="\t",file=tf,quote=FALSE)

# read it again into an IBSpectra object
ib.ceru2 <- readIBSpectra("iTRAQ4plexSpectra",tf,identifications.format="ibspectra")
ib.ceru2

unlink(tf)
}

\name{peptide.count}
\alias{peptide.count}
\alias{spectra.count}
\alias{sequence.coverage}
\title{
  Peptide counts, spectral counts and sequence coverage for ProteinGroup objects.
}
\description{
  Report the peptide count, spectral count and sequence coverage for supplied proteins.
}
\usage{
peptide.count(protein.group, protein.g = reporterProteins(protein.group), 
              specificity = c("reporter-specific", "group-specific", "unspecific"), ...)

spectra.count(protein.group, protein.g = reporterProteins(protein.group), 
              specificity = c("reporter-specific", "group-specific", "unspecific"), 
              modif = NULL, ...)

sequence.coverage(protein.group, protein.g = reporterProteins(protein.group), 
                  specificity = c("reporter-specific", "group-specific", "unspecific"), 
                  simplify = TRUE, ...)
}
\arguments{
  \item{protein.group}{ProteinGroup object.}
  \item{protein.g}{Protein group identifier.}
  \item{specificity}{Specificity of peptides.}
  \item{modif}{Only count peptides having a certain modification.}
  \item{simplify}{If simplify=TRUE, a named numeric vector is returned, 
                  with the mean sequence coverage of the ACs of each protein.g supplied.
                  Else, a list with the length of protein.g is returned having the sequence 
                  coverage for each protein AC.}
  \item{\dots}{Further arguments to \code{\link{peptides}}}
}
\author{
  Florian P Breitwieser
}

\seealso{
  \code{\link{calculate.emPAI}},
  \code{\link{calculate.dNSAF}},
  \code{\link{ProteinGroup}}
}
\examples{
  data(ibspiked_set1)
  sc <- spectra.count(proteinGroup(ibspiked_set1))
  pc <- peptide.count(proteinGroup(ibspiked_set1)) 
  plot(jitter(sc),jitter(pc),log="xy")
}

\name{calculate.emPAI}
\alias{calculate.emPAI}
\alias{n.observable.peptides}
\alias{observable.peptides}
\title{
  emPAI approximate abundance calculations.
}
\description{
  The Exponentially Modified Protein Abundance Index (emPAI) is a
  label free quantitative measure of protein abundance based on protein
  coverage by peptide matches. The original publication is 
  Ishihama Y, et al., Proteomics (2005).
}
\usage{
calculate.emPAI(protein.group, protein.g = reporterProteins(protein.group), normalize = FALSE,
                observed.pep = c("pep", "mod.charge.pep"), use.mw = FALSE, combine.f = mean, 
                ..., nmc = 0, report.all = FALSE)
n.observable.peptides(...)
observable.peptides(seq, nmc = 1, min.length = 6, min.mass = 600, max.mass = 4000, 
                    custom = list(code = c("B", "Z", "J", "U"),
                                  mass = c(164.554862, 278.61037, 213.12392, 150.953636)), ...)
}
\arguments{
  \item{protein.group}{ProteinGroup object. Its \code{@proteinInfo}
    slot \code{data.frame} must contain a \code{sequence} column to
    calculate the number of observable peptides per protein.
  }
  \item{protein.g}{Protein group identifiers.}
  \item{normalize}{Normalize to sum = 1?.}
  \item{observed.pep}{What counts as observed peptide?}
  \item{report.all}{TOADD}
  \item{use.mw}{Use MW to normalize for protein size}
  \item{combine.f}{How to handle proteins seen only with shared peptides?}
  \item{seq}{Protein sequence.}
  \item{nmc}{Number of missed cleavages.}
  \item{min.length}{Minimum length of peptide.}
  \item{min.mass}{Minimum mass of peptide.}
  \item{max.mass}{Maximum mass of peptide.}
  \item{custom}{User defined residue for \code{\link[OrgMassSpecR]{Digest}}.}
  \item{\dots}{Further arguments to \code{\link{observable.peptides}}/\code{\link[OrgMassSpecR]{Digest}}.}
}
\details{
  The formula is \deqn{emPAI = 10^{\frac{N <- {observed}}{N <- {observable}}} -1}{emPAI = 10^(N_observed/N_observable) -1}
  N_observed is the number of observed peptides - we use the count
  of unique peptide without consideration of charge state.
  N_observable is the number of observable peptides. 
  Sequence cleavage is done using \code{\link[OrgMassSpecR]{Digest}}.
}
\value{
  Named numeric vector of emPAI values.
}
\references{
  Ishihama Y, et al., Proteomics (2005)
}
\author{
  Florian P Breitwieser
}

\seealso{
  \code{\link[OrgMassSpecR]{Digest}},
  \code{\link{proteinInfo}},
  \code{\link{getProteinInfoFromUniprot}},
  \code{\link{calculate.dNSAF}},
  \code{\link{ProteinGroup}}
}
\examples{
data(ibspiked_set1)
protein.group <- proteinGroup(ibspiked_set1)
calculate.emPAI(protein.group,protein.g=protein.g(protein.group,"CERU"))
}
\keyword{ ~emPAI }

\name{Isobar util functions}

\alias{paste0}
\alias{\%inrange\%}

\title{
  Isobar util functions
}

\description{
  Utility functions.
  paste0 as a shorthand to paste(...,sep="") in versions of R pre 2.14.
}

\usage{
paste0(..., sep = "")
a \%inrange\% b
}

\arguments{
  \item{\dots}{Arguments to paste.}
  \item{sep}{Separator.}
  \item{a}{values.}
  \item{b}{range.}
}

\author{
  Florian P Breitwieser
}

\examples{
  1:10 %inrange% c(2,5)
}

\name{getPhosphoRSProbabilities}
\alias{getPhosphoRSProbabilities}
\alias{readPhosphoRSOutput}
\alias{writePhosphoRSInput}
\alias{filterSpectraPhosphoRS}

\title{
  Generate input files for PhosphoRS, call it, and get modification site probabilities
}
\description{
  Get phosphorylation site localization probabilities by calling PhosphoRS
  and parsing its output. \code{getPhosphoRSProbabilities} generates a XML
  input file for PhosphoRS calling \code{writePhosphoRSInput}, then executes
  phosphoRS.jar with java, and parses the XML result file with 
  \code{readPhosphoRSOutput}.
}
\usage{
  getPhosphoRSProbabilities(id.file, mgf.file, massTolerance, activationType,
                 simplify = FALSE, mapping.file = NULL, mapping =
                 c(peaklist = "even", id = "odd"), pepmodif.sep =
                 "##.##", besthit.only = TRUE, phosphors.cmd =
                 paste("java -jar", system.file("phosphors",
                 "phosphoRS.jar", package = "isobar")), file.basename =
                 tempfile("phosphors."))


  writePhosphoRSInput(phosphoRS.infile, id.file, mgf.file, massTolerance,
                 activationType, mapping.file = NULL, mapping =
                 c(peaklist = "even", id = "odd"), pepmodif.sep =
                 "##.##", modif.masses = rbind(c("PHOS", "1",
                 "1:Phospho:Phospho:79.966331:PhosphoLoss:97.976896:STY"),
                 c("Oxidation_M", "2",
                 "2:Oxidation:Oxidation:15.994919:null:0:M"),
                 c("Cys_CAM", "3",
                 "3:Carbamidomethylation:Carbamidomethylation:57.021464:null:0:C"),
                 c("iTRAQ4plex", "4",
                 "4:iTRAQ4:iTRAQ4:144.1544:null:0:KX"), c("iTRAQ8plex",
                 "5", "5:iTRAQ8:iTRAQ8:304.308:null:0:KX"),
                 c("TMT6plex", "7",
                 "7:TMT6:TMT6:229.162932:null:0:KX"), c("TMTsixplex",
                 "6", "6:TMT6:TMT6:229.162932:null:0:KX")))


  readPhosphoRSOutput(phosphoRS.outfile, simplify = FALSE, pepmodif.sep = "##.##", besthit.only = TRUE)

  filterSpectraPhosphoRS(id.file, mgf.file, ..., min.prob = NULL, do.remove=FALSE)
}
\arguments{
  \item{id.file}{Database search results file in \code{ibspectra.csv} or
    \code{mzIdentML} format. See \code{\link{IBSpectra}} and isobar vignette for
    information on converting Mascot dat and Phenyx pidres files
    into ibspectra format.}
  \item{mgf.file}{Peaklist file}
  \item{massTolerance}{Fragment ion mass tolerance (in Da)}
  \item{activationType}{Activation types of spectra. CID, HCD, or ETD.}
  \item{simplify}{If \code{TRUE}, returns a \code{data.frame} instead of a \code{list}.}
  \item{mapping.file}{Mapping file. See also \code{\link{readIBSpectra}}.}
  \item{mapping}{Mapping columns. }
  \item{besthit.only}{Only show best hit, simplifies result to data.frame instead of list.}
  \item{phosphors.cmd}{PhosphoRS script.}
  \item{file.basename}{Base name for creating phosphoRS input and output files.}
  \item{phosphoRS.infile}{PhosphoRS input XML file name.}
  \item{phosphoRS.outfile}{PhosphoRS output XML file name.}
  \item{pepmodif.sep}{separator of peptide and modification in XML id}
  \item{modif.masses}{masses and ID used for PhosphoRS}
  \item{min.prob}{Threshold for PhosphoRS peptide probability to consider it for quantification}
  \item{...}{Further arguments to getPhosphoRSProbabilities}
  \item{do.remove}{If TRUE, spectra below the min.prob threshold are not just set as 'use.for.quant=FALSE' but removed.}
}

\details{
  PhosphoRS is described in Taus et al., 2011.
  It can be downloaded from http://cores.imp.ac.at/protein-chemistry/download/ 
  and used as Freeware. Java is required at runtime.
}
\value{
  If \code{simplify=TRUE}, a \code{data.frame} with the following columns:
    \code{spectrum, peptide, modif, PepScore, PepProb, seqpos}

  If \code{simplify=FALSE}, a list (of spectra) of lists (of peptide 
    identifications) of lists (with information about identification 
    and localization).
    spectrum -> {peptide 1, peptides 2, ...} -> {peptide}.
    First level:
      - spectrum
    Second level:
      - peptide identifications for spectrum (might be more than one)
    Third level:
      - peptide: vector with peptide sequence and modification stirng
      - site.probs: matrix with site probabilities for each phospho site
      - isoforms: peptide score and probabilities for each isoform  
}
\references{
  Taus et al., 2011
}
\author{
  Florian P Breitwieser
}

\keyword{ ~phospho }

\name{calculate-pvalues}
\alias{calculate.sample.pvalue}
\alias{calculate.ratio.pvalue}
\alias{calculate.mult.sample.pvalue}
\alias{adjust.ratio.pvalue}
\title{
  Calculate and Adjust Ratio and Sample p-values.
}

\description{
  Functions for calculating and adjusting ratios and sample p-values.
  Usually, these are called by proteinRatios or peptideRatios.
}

\usage{
  calculate.ratio.pvalue(lratio, variance, ratiodistr = NULL)
  calculate.sample.pvalue(lratio, ratiodistr)

  calculate.mult.sample.pvalue(lratio, ratiodistr, strict.pval, 
                               lower.tail, n.possible.val, n.observed.val)

  adjust.ratio.pvalue(quant.tbl, p.adjust, sign.level, globally = FALSE)
}

\arguments{
  \item{lratio}{log 10 protein or peptide ratios.}
  \item{ratiodistr}{Fitted ratio distribution/}
  \item{variance}{Variance of lratios.}

  \item{strict.pval}{If FALSE, missing ratios are ignored. If TRUE, missing ratios are penalized by giving them a sample.pval of 0.5.}
  \item{lower.tail}{lower.tail of distribution?}
  \item{n.possible.val}{Number of possible ratios.}
  \item{n.observed.val}{Number of observed ratios.}

  \item{quant.tbl}{Quantification table (from proteinRatios or peptideRatios).}
  \item{p.adjust}{p-value adjustment method (see ?p.adjust).}
  \item{sign.level}{Ratio significance level.}
  \item{globally}{Whether the p-values should be adjusted over all conditions,
                  or individually in each condition.}

}
\author{Florian P. Breitwieser}

\seealso{
 \code{\link{proteinRatios}},\code{\link{peptideRatios}}
}

\examples{
  lratio <- c(-1,-1,seq(from=-1,to=1,by=.25),1,1)
  variance <- c(0,1,rep(0.1,9),0,1)
  ratiodistr.precise <- new("Norm",mean=0,sd=.25)
  ratiodistr.wide <- new("Norm",mean=0,sd=.5)

  # ratio p-value is impacted only by the variance
  # sample p-value captures whether the ratio distribution is narrow ('precise')
  #  or wide
  data.frame(lratio, variance,
             ratio.pvalue=calculate.ratio.pvalue(lratio, variance),
             sample.pvalue.precise=calculate.sample.pvalue(lratio,ratiodistr.precise),
             sample.pvalue.wide=calculate.sample.pvalue(lratio,ratiodistr.wide))

}

\name{proteinInfo-methods}
\alias{getProteinInfoFromBiomart}
\alias{getProteinInfoFromUniprot}
\alias{getProteinInfoFromTheInternet}
\alias{getProteinInfoFromNextProt}
\alias{getProteinInfoFromEntrez}
\alias{getProteinInfoFromBioDb}
\alias{proteinInfoIsOnSpliceVariants}
\alias{proteinInfo}
\alias{proteinInfo-methods}
\alias{proteinInfo,ProteinGroup-method}
\alias{proteinInfo,ProteinGroup,missing,missing-method}
\alias{proteinInfo,ProteinGroup,character,missing-method}
\alias{proteinInfo,ProteinGroup,missing,character-method}
\alias{proteinInfo<-}
\alias{proteinInfo<-,ProteinGroup-method}

\title{Methods for Function proteinInfo}
\description{
  proteinInfo slot in Proteingroup objects contains information about proteins. 
  \code{proteinInfo} method allows to get and set it.

  \code{getProteinInfoFromUniprot} downloads information of contained proteins
  from Uniprot, \code{getProteinInfoFromBiomart} from Biomart.

}
\usage{

  \S4method{proteinInfo}{ProteinGroup}(x)

  \S4method{proteinInfo}{ProteinGroup,character,missing}(x, protein.g, select="name", collapse=", ",
                                                 simplify = TRUE, do.warn = TRUE)

  \S4method{proteinInfo}{ProteinGroup,missing,character}(x, protein.ac, select="name", collapse=", ",
                                                 simplify = TRUE, do.warn = TRUE)


  proteinInfoIsOnSpliceVariants(protein.info)

#  getProteinInfoFromUniprot(x, splice.by = 200, fields = c(accession = "id", name
#                 = "entry%20name", protein_name = "protein%20names",
#                 gene_name = "genes", organism = "organism", length =
#                 "length", sequence = "sequence"))


  getProteinInfoFromTheInternet(x)

  getProteinInfoFromNextProt(x)

  getProteinInfoFromBiomart(x, database = "Uniprot")

  getProteinInfoFromBioDb(x, ..., con = NULL)

  getProteinInfoFromEntrez(x, splice.by = 200)
}
\arguments{
  \item{x}{ProteinGroup object}
  \item{protein.g}{Protein group identifier. If supplied, only information for 
    these proteins is returned.}
  \item{protein.ac}{Protein ACs. If supplied, only information for 
    these proteins is returned.}
  \item{select}{indicating columns to select. See Details.}
  \item{collapse}{passed to \code{\link{paste}} to concatenate information of
    multiple protein in one protein group.}
  \item{simplify}{If true, a vector or matrix is returned, with the pasted protein information.
                  If false, a list is returned.}
  \item{do.warn}{If true, report diagnostic warning messages.}
  \item{splice.by}{Chunk size for query of Uniprot database.}
  \item{database}{database from which the ACs stem from. Only Uniprot is supported for now.}
  \item{con}{database connection}
  \item{fields}{mapping of CSV field names to proteinInfo field names}
  \item{...}{arguments to build database connection.}
  \item{protein.info}{protein info data.frame}
}

\details{proteinInfo contains columns \code{accession}, \code{name}, 
  \code{gene_name}, \code{protein_name}, and possibly \code{length}
  and \code{sequence}. \code{accession} is mapped with
  the entry AC is mapped to the entry AC in the database.
  \code{getProteinInfoFromUniprot} is the preferred methods to get the information.
  \code{getProteinInfoFromBioDb} is an example how to implement the query on a
  local database. Depending on the database, protein information might be
  available on protein ACs or also on the specific splice variants. This can be
  queried with the \code{proteinInfoIsOnSpliceVariants} function.
}
\seealso{
  \link{protein.g}
}
\examples{
data(ibspiked_set1)
pg <- proteinGroup(ibspiked_set1)

\dontrun{
  proteinInfo(pg) <- getProteinInfoFromUniprot(pg)
  proteinInfo(pg) <- getProteinInfoFromBiomart(pg)
}

proteinInfo(pg,protein.g="P13635")
protein.g(pg,"CERU")
}
\keyword{methods}

\name{isobar-analysis}

\alias{isobar-analysis}

% estimateRatio methods
\alias{estimateRatio}
\alias{estimateRatio,IBSpectra,ANY,character,character,character,missing-method}
\alias{estimateRatio,IBSpectra,ANY,missing,missing,missing,character-method}
\alias{estimateRatio,IBSpectra,ANY,missing,missing,character,missing-method}
\alias{estimateRatio,IBSpectra,ANY,character,character,missing,matrix-method}
\alias{estimateRatio,IBSpectra,ANY,character,character,NULL,matrix-method}
\alias{estimateRatio,IBSpectra,ANY,character,character,character,NULL-method}
\alias{estimateRatio,IBSpectra,ANY,character,character,missing,character-method}
\alias{estimateRatio,IBSpectra,ANY,character,character,NULL,character-method}
\alias{estimateRatio,IBSpectra,ANY,character,character,NULL,data.frame-method}
\alias{estimateRatio,IBSpectra,ANY,character,character,missing,data.frame-method}
\alias{estimateRatioNumeric}
\alias{estimateRatioNumeric,numeric,numeric,missing-method}
\alias{estimateRatioNumeric,numeric,numeric,NULL-method}
\alias{estimateRatioNumeric,numeric,numeric,NoiseModel-method}

\alias{estimateRatioForProtein}
\alias{estimateRatioForPeptide}

% multUnifPValues and multUnifDensity
\alias{getMultUnifPValues}
\alias{getMultUnifDensity}

\title{IBSpectra analysis: Protein and peptide ratio calculation}
\description{Calculates the relative abundance of a peptide or protein in one tag compared to another.}
\usage{
estimateRatio(ibspectra, noise.model = NULL, channel1, channel2, protein, peptide, ...)
estimateRatioForPeptide(peptide, ibspectra, noise.model, channel1, channel2, combine = TRUE, ...)
estimateRatioForProtein(protein, ibspectra, noise.model, channel1, channel2, combine = TRUE, method = "isobar", specificity = REPORTERSPECIFIC, quant.w.grouppeptides = NULL, ...)

\S4method{estimateRatioNumeric}{numeric,numeric,missing}(channel1,channel2,summarize.f=median, ...)

\S4method{estimateRatioNumeric}{numeric,numeric,NoiseModel}(channel1,channel2,noise.model,ratiodistr=NULL,variance.function="maxi",
                                                            sign.level=0.05,sign.level.rat=sign.level,sign.level.sample=sign.level,
                                                            remove.outliers=TRUE,outliers.args=list(method = "iqr", outliers.coef = 1.5),
                                                            method="isobar",fc.threshold=1.3,
                                                            channel1.raw=NULL,channel2.raw=NULL,use.na=FALSE,preweights=NULL)

\S4method{estimateRatio}{IBSpectra,ANY,character,character,character,missing}(ibspectra,noise.model,channel1,channel2,
                                                                              protein,peptide,...)

\S4method{estimateRatio}{IBSpectra,ANY,character,character,character,NULL}(ibspectra,noise.model,channel1,channel2,
                                                                           protein,peptide=NULL,...)

\S4method{estimateRatio}{IBSpectra,ANY,character,character,missing,character}(ibspectra,noise.model,channel1,channel2,protein,peptide,...)
\S4method{estimateRatio}{IBSpectra,ANY,character,character,NULL,character}(ibspectra,noise.model,channel1,channel2,protein=NULL,peptide,...)

}
\arguments{
  \item{ibspectra}{ IBSpectra object. } 
  \item{noise.model}{ NoiseModel object.}
  \item{channel1}{ Tag channel 1. 
    Can either be a character denoting a 'reporter name' or a numeric vector
    whose value should be summarized.Ratio is calculated as channel2/channel1. }
  \item{channel2}{ Tag channel 2.
    Can either be a character denoting a 'reporter name' or a numeric vector
    whose value should be summarized. Ratio is calculated as channel2/channel1. }
  \item{protein}{ Protein(s) of interest. If present, channel1 and channel2 
    must be reporter names. Provide either proteins or peptides.  }
  \item{peptide}{ Peptide(s) of interest. If present, channel1 and channel2 
    must be reporter names. Provide either proteins or peptides. }
  \item{combine}{ If true, a single ratio is returned even for multiple
    peptides/spectra. If false, a data.frame with a row for each peptide/protein
    is returned.}
  \item{specificity}{ See \code{\link{specificities}}. }
  \item{quant.w.grouppeptides}{Proteins which should be quantified with group specific peptides.
        Normally, only reporter specific peptides are used.}
  \item{ratiodistr}{\code{distr} object of ratio distribution.}
  \item{variance.function}{Defines how the variance for ratio is calculated. 
    'ev' is the estimator variance and thus 1/sum(1/variances).
    'wsv' is the weighted sample variance.
    'maxi' method takes the maximum of the former two variances.
  }
  \item{sign.level}{Significiance level. }
  \item{sign.level.rat}{Signal p-value significiance level. }
  \item{sign.level.sample}{Sample p-value significiance level. }
  \item{remove.outliers}{Should outliers be removed? }
  \item{outliers.args}{Arguments for outlier removal, see OUTLIERS function (TODO).}
  \item{method}{ method taken for ratio computation and selection: one of
    'isobar','libra','multiq','pep','ttest' and 'compare.all'. }
  \item{fc.threshold}{When method equals fc, takes this as fold change threshold. }
  \item{summarize.f}{A method for summarizing spectrum ratios when no other 
    information is available. For example \code{median} or \code{mean}. }
  \item{channel1.raw}{When given, noise estimation is based on channel1.raw and channel2.raw. 
                      These are the intensities of the channels before normalization. }
  \item{channel2.raw}{See channel1.raw. }
  \item{use.na}{Use NA values to calculate ratio. Experimental feature - use with caution. }
  \item{preweights}{Specifies weigths for each spectrum. Experimental feature - use with caution. }
%  \item{correct.ratio}{Can be used to scale ratio - should be in log 10. }
  \item{\dots}{Passed down to \code{estimateRatioNumeric} methods.}
}
\value{
  In general, a named character vector with the following elements:
    - lratio: log ratio
    - variance
    - n.spectra: number of spectra available in the ratio calculation
    - p.value.rat: Signal p-value. NA if called w/o ratiodistr
    - p.value.sample: Sample p-value. NA if called w/o ratiodistr
    - is.significant: NA if called w/o ratiodistr

   If combine=FALSE, estimateRatio returns a data.frame, with columns as 
   described above.
}

\author{Florian P. Breitwieser, Jacques Colinge}

\seealso{
  \link{ProteinGroup},
  \link{IBSpectra},
  \link{isobar-preprocessing},
  \link{isobar-plots}
  \link{proteinRatios}
}

\examples{

  data(ibspiked_set1)
  data(noise.model.hcd)
  ceru.human <- protein.g(proteinGroup(ibspiked_set1),"CERU_HUMAN")
  ceru.rat <- protein.g(proteinGroup(ibspiked_set1),"CERU_RAT")
  ceru.mouse <- protein.g(proteinGroup(ibspiked_set1),"CERU_MOUSE")
  ceru.proteins <- c(ceru.human,ceru.rat,ceru.mouse)

## Calculate ratio based on all spectra of peptides specific 
##  to CERU_HUMAN, CERU_RAT or CERU_MOUSE. Returns a named
##  numeric vector.
10^estimateRatio(ibspiked_set1,noise.model.hcd,
                 channel1="114",channel2="115",
                 protein=ceru.proteins)['lratio']

## If argument 'combine=FALSE', estimateRatio returns a data.frame 
##  with one row per protein
10^estimateRatio(ibspiked_set1,noise.model.hcd,
                 channel1="114",channel2="115",
                 protein=ceru.proteins,combine=FALSE)[,'lratio']
## spiked material channel 115 vs 114: 
##                 CERU_HUMAN (P00450): 1
##                 CERU_RAT   (P13635): 2
##                 CERU_MOUSE (Q61147): 0.5

}

\name{proteinNameAndDescription}
\alias{proteinNameAndDescription}
\alias{proteinDescription}
\alias{proteinGeneName}
\alias{proteinID}

\title{Get protein gene names and description from protein info of protein group.}
\description{
  Convenience functions to retrieve protein gene names and description 
  for a list of protein group identifiers.
}
\usage{
proteinNameAndDescription(protein.group, protein.g = reporterProteins(protein.group), collapse = FALSE)
proteinGeneName(protein.group, protein.g = reporterProteins(protein.group))
proteinDescription(protein.group, protein.g = reporterProteins(protein.group))
proteinID(protein.group, protein.g = reporterProteins(protein.group))
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{protein.group}{ProteinGroup object.}
  \item{protein.g}{protein group identifier.}
  \item{collapse}{If TRUE, the information for all protein.gs is combined.}
}
\author{Florian P Breitwieser}

\examples{
data(ibspiked_set1)
pg <- proteinGroup(ibspiked_set1)
protein.gs <- protein.g(pg,"CERU")
protein.gs
proteinNameAndDescription(pg,protein.gs)
proteinNameAndDescription(pg,protein.gs,collapse=TRUE)
proteinGeneName(pg,protein.gs)
proteinDescription(pg,protein.gs)  
proteinID(pg,protein.gs)  
}

\name{calc.delta.score}
\alias{calc.delta.score}
\alias{calc.pep.delta.score}
\alias{filterSpectraDeltaScore}

\title{Calculate Delta Score from Ion Score}
\usage{
calc.delta.score(my.data)
filterSpectraDeltaScore(my.data, min.delta.score=10, do.remove=FALSE)
}

\description{
  Calculates delta score from raw search engine score by substracting the best matching hit with the second best matching.
  data needs to have not only the best hit per spectrum, but multiple, to be able to calculate the delta score.
  filterSpectraDeltaScore calls \code{calc.delta.score} and filters spectra below a minum delta score.}

\arguments{
  \item{my.data}{IBSpectra data frame.}
  \item{min.delta.score}{Minimum delta score.}
  \item{do.remove}{If TRUE, spectra below the min.prob threshold are not just set as 'use.for.quant=FALSE' but removed.}
}
\value{
  Returns data with additional column 'delta.score'.
}
\author{
  Florian P. Breitwieser
}

\name{calculate.dNSAF}
\alias{calculate.dNSAF}

\title{
  dNSAF approximate abundance calculations.
}
\description{
  Distributed normalized spectral abundance factor (dNSAF) is a 
  label free quantitative measure of protein abundance based on 
  spectral counts which are corrected for peptides shared by multiple
  proteins. Original publication: Zhang Y et al., Analytical Chemistry
  (2010).
}
\usage{
calculate.dNSAF(protein.group, use.mw = FALSE, normalize = TRUE,
                combine.f = mean)
}
\arguments{
  \item{protein.group}{ProteinGroup object. Its \code{@proteinInfo}
    slot \code{data.frame} must contain a \code{length} column.}
  \item{use.mw}{Use MW to account for protein size}
  \item{normalize}{Normalize dSAF to dNSAF?}
  \item{combine.f}{How to handle proteins seen only with shared peptides?}
}
\value{
  Named numeric vector of dNSAF values.
}
\references{
  Zhang Y et al., Analytical Chemistry (2010)
}
\author{
  Florian P Breitwieser
}

\seealso{
  \code{\link{proteinInfo}},
  \code{\link{getProteinInfoFromUniprot}},
  \code{\link{calculate.emPAI}},
  \code{\link{ProteinGroup}}
}

\examples{
data(ibspiked_set1)
protein.group <- proteinGroup(ibspiked_set1)
calculate.dNSAF(protein.group)
}

\keyword{ ~dNSAF }

\name{reporter.protein-methods}
\docType{methods}
\alias{reporter.protein-methods}
\alias{reporter.protein}
\alias{reporter.protein,ProteinGroup,character-method}
\title{Get reporter protein group identifier for protein group identifier}
\description{
 Methods for function \code{reporter.protein} in package \pkg{isobar}
}
\section{Methods}{
\describe{

\item{\code{signature(x = "ProteinGroup", protein.g = "character")}}{
 Get reporter protein for protein group identifier.
}
}}
\keyword{methods}
\keyword{ ~~ other possible keyword(s) ~~ }

\name{fit distributions}
\alias{fitCauchy}
\alias{fitTlsd}
\alias{fitNorm}
\alias{fitWeightedNorm}
\alias{fitNormalCauchyMixture}
\alias{fitGaussianMixture}
\title{Fit weighted and unweighted Cauchy and Normal distributions}
\description{
  Functions to fit the probability density functions
    on ratio distribution.
}
\usage{
fitCauchy(x)
fitNorm(x, portion = 0.75)
fitWeightedNorm(x, weights)
fitNormalCauchyMixture(x)
fitGaussianMixture(x, n = 500)
fitTlsd(x)
}
\arguments{
  \item{x}{Ratios}
  \item{weights}{Weights}
  \item{portion}{Central portion of data to take for computation}
  \item{n}{number of sampling steps}
}
\value{
  \code{\link[distr]{Cauchy}},\code{\link[distr]{Norm}}
}
\author{
  Florian P Breitwieser, Jacques Colinge.
}

\seealso{
  \link{proteinRatios}
}
\examples{
library(distr)
data(ibspiked_set1)
data(noise.model.hcd)
# calculate protein ratios of Trypsin and CERU_HUMAN. Note: this is only
#  for illustration purposes. For estimation of sample variability, data
#  from all protein should be used
pr <- proteinRatios(ibspiked_set1,noise.model=noise.model.hcd,
                    cl=as.character(c(1,1,2,2)),combn.method="intraclass",protein=c("136429","P00450"))

# fit a Cauchy distribution
ratiodistr <- fitCauchy(pr$lratio)
plot(ratiodistr)
}

\name{isobar-reports}

\alias{isobar-reports}
\alias{connect.nodes}
\alias{create.reports}
\alias{create.meta.reports}
\alias{draw.boxplot}
\alias{draw.protein.group}
\alias{draw.proteingroup.row}
\alias{initialize.env}
\alias{load.properties}
\alias{modifs}
\alias{print_longtablehdr}
\alias{tikz.proteingroup}
\alias{write.xls.report}
\alias{print_longtablehdr_peptide}
\alias{transform_pepmodif}
\alias{property}
\alias{write.tex.commands}
\alias{load.tex.properties}

\alias{get_n_proteins}
\alias{print_classlabels_tbl}
\alias{print_groupsize}
\alias{print_protein_grp_info}
\alias{print_protein_grp_tbl}
\alias{print_protein_notquant_tbl}
\alias{print_protein_quant_tbl}
\alias{print_sign_proteins_tbl}

\alias{testPdflatex}
\alias{testPerl}

\title{Isobar reports}

\description{
  Generation of LaTeX and XLS reports is helped with
  functions which facilitate the gathering of relevant information and
  creation of tikz plots. \code{create.reports} parses properties (by
  calling \code{load.properties}) and initialize environments and
  computations (by calling \code{initialize.env}) required by the
  reports, calls Sweave and pdflatex.
}

\usage{
create.reports(properties.file = "properties.R", 
               global.properties.file = system.file("report","properties.R", package = "isobar"),
               args = NULL, ...,
               recreate.properties.env = TRUE, recreate.report.env = TRUE)

load.properties(properties.file = "properties.R",
                global.properties.file = system.file("report","properties.R",package="isobar"),
                args = NULL, ...)

initialize.env(env, properties.env)

}

\arguments{
  \item{properties.file}{File which holds the parameters for data
    analysis and report generation. It is parsed as R code after the
    global report configuration file \code{global.properties.file}  
    and defines peaklists, identification files, significance levels,
    etc. See the global properties file for the available options and
    values.}
  \item{global.properties.file}{
    \code{system.file("report","properties.R",package="isobar")}}
  \item{args}{Additional (command line) arguments which overrids those
    in \code{properties.file}. }
  \item{...}{Additional properties.}
  \item{recreate.properties.env}{Whether a properties.env existing in the global
  environment should be used, or it should be recreated.}
  \item{recreate.report.env}{Whether a report.env existing in the global
  environment should be used, or it should be recreated.}
  \item{env}{Item to be initialized.}
  \item{properties.env}{Environment into which properties are read.}
}

\details{
  The directory \code{inst} in the isobar installation directory 
  \code{system.file("inst",package="isobar")} contains R, Sweave, and LaTeX
  files as examples of how to create XLS and PDF reports using \code{isobar}.
  \describe{
    \item{create_reports.R}{Call with Rscript. It is the main file which
      \enumerate{
	\item{parses command line options. \code{--compile} and
	  \code{--zip} are parsed directly and given as arguments to
	  \code{create.reports}. Other arguments are given
	  \code{\link{load.properties}}.}	
	\item{calls a perl script to generate a XLS report}
	\item{generates a LaTeX quality control and analysis report}
	}
	for the XLS report the script pl/tab2xls.pl is used, which
	concetenates CSV files to a XLS. See Perl requirements.
	Sweave is called on report/isobar-qc.Rnw and
	report/isobar-analysis.Rnw. All files are written the
	working directory.}
    \item{isobar-qc.Rnw}{Quality control Sweave file.}
    \item{isobar-analysis.Rnw}{Data analysis Sweave file.}
    \item{properties.R}{Default configuration for data analysis.}
    \item{report-utils.tex}{LaTeX functions for plotting tikz graphics, etc.}
  }
}

\author{Florian P Breitwieser}

\seealso{
  \link{IBSpectra},
  \link{isobar-preprocessing}
  \link{isobar-analysis}
}

\name{isobar-plots}

\alias{isobar-plots}
\alias{reporterMassPrecision}
\alias{reporterMassPrecision,IBSpectra,missing-method}
\alias{reporterMassPrecision,IBSpectra,logical-method}
\alias{reporterIntensityPlot}
\alias{reporterIntensityPlot-methods}
\alias{reporterIntensityPlot,IBSpectra-method}
\alias{raplot}
\alias{raplot,IBSpectra-method}
\alias{plotRatio}
\alias{plotRatio,IBSpectra,character,character,character-method}
\alias{maplot}
\alias{maplot,IBSpectra,missing,missing-method}
\alias{maplot,IBSpectra,character,character-method}
\alias{maplot,missing,numeric,numeric-method}
\alias{maplot2}
\alias{maplot2,ANY,character,character-method}
\alias{maplot2,list,character,character-method}
\alias{protGgdata}
\alias{protGgdata,ANY,character,character-method}

\title{IBSpectra plots}

\description{
  Various plots are implement to assure data quality, and accompany
  preprocessing  and analysis.
}

\section{reporterMassPrecision}{
  \describe{
    \item{\code{reporterMassPrecision(x)}:}{
      Calculates and displays the deviation from the 'true' tag mass - 
      as specified in the IBSpectra object - of each channel.
  }}
}

\section{reporterIntensityPlot}{
  \describe{
    \item{\code{reporterIntensityPlot(x)}:}{
      Displays boxplots of intensity of channels before and after
      normalization - useful to check the result of normalization.
  }}
}


\section{raplot}{
  \describe{
    \item{\code{raplot(x,...)}:}{
      Ratio-Absolute intensity plot - will be deprecated by maplot
      \describe{
        \item{\code{x}}{IBSpectra object}
      	\item{\code{...}}{Parameters to plot function.}
    }
  }}
}

\section{plotRatio}{
  \describe{
    \item{\code{plotRatio(x,channel1,channel2,protein,...)}:}{
      Plots abundances of one protein
      
      \describe{
        \item{\code{x}}{IBSpectra object}
        \item{\code{channel1}}{}
        \item{\code{channel2}}{}
        \item{\code{protein}}{}
      	\item{\code{...}}{Parameters to plot function.}
      }
    }
  }
}


\section{maplot}{
  \describe{
    \item{\code{maplot(x,channel1,channel2,...)}:}{
      Creates a ratio-versus-intensity plot.
      \describe{
        \item{\code{x}}{IBSpectra object.}
    }
  }}
}

\section{maplot2}{
  \describe{
    \item{\code{maplot2()}:}{
  }}
}


\author{Florian P. Breitwieser, Jacques Colinge}

\seealso{
  \link{IBSpectra},
  \link{isobar-preprocessing}
  \link{isobar-analysis}
}

\examples{
data(ibspiked_set1)
maplot(ibspiked_set1,main="IBSpiked, not normalized")
maplot(normalize(ibspiked_set1),main="IBSpiked, normalized")

}

\name{isobar-package}
\alias{isobar-package}
\alias{isobar}
\docType{package}
\title{
Analysis and quantitation of isobarically tagged MSMS proteomics data
}
\description{
\code{isobar} provides methods for preprocessing, normalization, and
report generation for the analysis of quantitative mass spectrometry
proteomics data labeled withOA isobaric tags, such as iTRAQ and TMT.
}
\details{
\tabular{ll}{
Package: \tab isobar\cr
Version: \tab 1.1.2\cr
biocViews: \tab Proteomics, MassSpectrometray, Bioinformatics, MultipleComparisons, QualityControl\cr
Depends: \tab R (>= 2.9.0), Biobase, stats, methods, ggplot2\cr
Imports: \tab distr, biomaRt\cr
Suggests: \tab MSnbase,XML\cr
LazyLoad: \tab yes\cr
License: \tab LGPL-2\cr
URL: \tab http://bioinformatics.cemm.oeaw.ac.at\cr
Collate: \tab utils.R
ProteinGroup-class.R
IBSpectra-class.R
NoiseModel-class.R
ratio-methods.R
sharedpep-methods.R
MSnSet-methods.R
zzz.R\cr
}

Index:
\preformatted{
IBSpectra-class         IBSpectra objects
NoiseModel-class        NoiseModel objects
ProteinGroup-class      ProteinGroup objects
do.log                  Log functions for IBSpectra objects
fitCauchy               Fit weighted and unweighted Cauchy and Normal
                        distributions
groupMemberPeptides     Peptide info for protein group members
human.protein.names     Info on proteins
ibspiked_set1           Isobar Data packages
isobar-analysis         IBSpectra analysis: Protein and peptide ratio
                        calculation
isobar-import           Loading data into IBSpectra objects using
                        readIBSpectra
isobar-package          Analysis and quantitation of isobaric tag
                        Proteomics data
isobar-plots            IBSpectra plots
isobar-preprocessing    IBSpectra preprocessing
isobar-reports          Isobar reports
maplot.protein          MAplot for individual proteins
number.ranges           Helper function to transform number lists to
                        ranges
proteinInfo-methods     Methods for Function proteinInfo
proteinRatios           protein and peptide ratios
sanitize                Helper function for LaTeX export
shared.ratios           Shared ratio calculation
shared.ratios.sign      Plot and get significantly shared ratios.
}

Further information is available in the following vignettes:
\tabular{ll}{
\code{isobar} \tab Isobar Overview (source, pdf)\cr
\code{isobar-devel} \tab Isobar for developers (source, pdf)\cr
}
}
\author{
Florian P Breitwieser <fbreitwieser@cemm.oeaw.ac.at> and
Jacques Colinge <jcolinge@cemm.oeaw.ac.at>, with contributions from
Xavier Robin <xavier.robin@unige.ch>

Maintainer: Florian P Breitwieser <fbreitwieser@cemm.oeaw.ac.at>
}
\keyword{ package }

\name{Tlsd-class}
\Rdversion{1.1}
\docType{class}
\alias{Tlsd-class}
\alias{Tlsd}

\title{Class \code{"Tlsd"}}
\description{
  Location scale family T distribution, based on the original T function.
}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("Tlsd", df, location, scale)}.
%%  ~~ describe objects here ~~ 
}
\section{Slots}{
  \describe{
    \item{\code{gaps}:}{Object of class \code{"OptionalMatrix"} ~~ }
    \item{\code{img}:}{Object of class \code{"rSpace"} ~~ }
    \item{\code{param}:}{Object of class \code{"OptionalParameter"} ~~ }
    \item{\code{r}:}{Object of class \code{"function"} ~~ }
    \item{\code{d}:}{Object of class \code{"OptionalFunction"} ~~ }
    \item{\code{p}:}{Object of class \code{"OptionalFunction"} ~~ }
    \item{\code{q}:}{Object of class \code{"OptionalFunction"} ~~ }
    \item{\code{.withSim}:}{Object of class \code{"logical"} ~~ }
    \item{\code{.withArith}:}{Object of class \code{"logical"} ~~ }
    \item{\code{.logExact}:}{Object of class \code{"logical"} ~~ }
    \item{\code{.lowerExact}:}{Object of class \code{"logical"} ~~ }
    \item{\code{Symmetry}:}{Object of class \code{"DistributionSymmetry"} ~~ }
  }
}
\section{Extends}{
Class \code{"\linkS4class{AbscontDistribution}"}, directly.
Class \code{"\linkS4class{UnivariateDistribution}"}, by class "AbscontDistribution", distance 2.
Class \code{"\linkS4class{AcDcLcDistribution}"}, by class "AbscontDistribution", distance 2.
Class \code{"\linkS4class{Distribution}"}, by class "AbscontDistribution", distance 3.
Class \code{"\linkS4class{UnivDistrListOrDistribution}"}, by class "AbscontDistribution", distance 3.
}
\section{Methods}{
No methods defined with class "Tlsd" in the signature.
}

\author{
  Florian P. Breitwieser, based on original T distribution class.
}

\examples{
showClass("Tlsd")
}
\keyword{classes}

\name{observedKnownSites}
\alias{observedKnownSites}
\alias{modif.site.count}
\alias{modif.sites}

\title{
  Observed modification sites.
}
\description{
  Functions to display the modification sites observed for each protein isoform
  and count the number of modified residues per protein.
}
\usage{
observedKnownSites(protein.group, protein.g, ptm.info, modif, modification.name = NULL)

modif.site.count(protein.group, protein.g = reporterProteins(protein.group), modif, take = max)

modif.sites(protein.group, protein.g = reporterProteins(protein.group), modif)

}

\arguments{
  \item{protein.group}{ProteinGroupb object.}
  \item{protein.g}{protein group identifier.}
  \item{ptm.info}{ptm information data.frame, see ?getPtmInfo.}
  \item{modif}{Modification to track, e.g. 'PHOS'.}
  \item{modification.name}{Value to filter 'modification.name' column in ptm.info.}
  \item{take}{should be either max or min: When multiple isoforms are present, which value should be taken for the count?}
}
\author{
  Florian P. Breitwieser
}
\examples{
data(ib_phospho)
data(ptm.info)

# Modification sites of reporter proteins:
#  a list of protein groups, 
#     containing sub-lists of identified sites for each isoform
protein.modif.sites <- sort(modif.site.count(proteinGroup(ib_phospho),modif="PHOS"))

# Details on modification sites of proteins
#   detected with most modifications
modif.sites(proteinGroup(ib_phospho),modif="PHOS",protein.g=names(tail(protein.modif.sites)))

# How many sites are known, and how many known sites have been observed?
observedKnownSites(proteinGroup(ib_phospho),modif="PHOS",protein.g=names(tail(protein.modif.sites)),ptm.info=ptm.info,modification.name="Phospho")
}

\name{NoiseModel-class}
\docType{class}

% Class:
\alias{class:NoiseModel}
\alias{NoiseModel-class}
\alias{ExponentialNoiseModel-class}
\alias{ExponentialNoANoiseModel-class}
\alias{InverseNoiseModel-class}
\alias{InverseNoANoiseModel-class}
\alias{GeneralNoiseModel-class}

% Constructor:
\alias{initialize,NoiseModel-method}
\alias{NoiseModel}
\alias{NoiseModel,IBSpectra-method}

% Accessor methods:
\alias{variance}
\alias{variance,NoiseModel,numeric,numeric-method}
\alias{variance,NoiseModel,numeric,missing-method}

\alias{stddev}
\alias{stddev,NoiseModel-method}

\alias{noiseFunction}
\alias{noiseFunction,NoiseModel-method}

\alias{parameter}
\alias{parameter<-}
\alias{parameter,NoiseModel-method}
\alias{parameter<-,NoiseModel-method}

\alias{lowIntensity}
\alias{lowIntensity<-}
\alias{lowIntensity,NoiseModel-method}
\alias{lowIntensity<-,NoiseModel-method}

\alias{naRegion}
\alias{naRegion<-}
\alias{naRegion,NoiseModel-method}
\alias{naRegion<-,NoiseModel-method}

% show method:
\alias{show,NoiseModel-method}

% plot method:
\alias{plot.NoiseModel}

\title{NoiseModel objects}

\description{
  A NoiseModel represent the technical variation which is dependent on
  signal intensity.
}


\section{Constructor}{
  \describe{
    \item{\code{new(type,ibspectra,reporterTagNames=NULL,one.to.one=TRUE,min.spectra=10,plot=FALSE,
                    pool=FALSE)}:}{
    Creates a new NoiseModel object based on ibspectra object.
      \describe{
        \item{\code{type}:}{A non-virtual class deriving from NoiseModel:
          \code{ExponentialNoiseModel}, \code{ExponentialNoANoiseModel},
          \code{InverseNoiseModel}, \code{InverseNoANoiseModel}}
        \item{\code{reporterTagNames}:}{When NULL, all channels from ibspectra are taken
          (i.e. \code{sampleNames(ibspectra)}). Otherwise, specify
	  subset of names, or a matrix which defines the desireed combination of channels (nrow=2).}
	\item{\code{one.to.one}:}{Set to false to learn noise model one a non
	  one-to-one dataset}
	\item{\code{min.spectra}:}{When one.to.one=FALSE, only take proteins
	  with min.spectra to learn noise model.}
	\item{\code{plot}:}{Set to true to plot data the noise model is learnt on.}
    \item{\code{pool}:}{If false, a NoiseModel is estimated on each combination
         of channels indivdually, and then the parameters are averaged. If true,
         the ratios of all channels are pooled and then a NoiseModel is estimated. }
      }
    }
  }
}

\section{Accessor methods}{
  \describe{
    \item{\code{noiseFunction}:}{Gets the noise function.}
    \item{\code{parameter}:}{Gets and sets the parameters for the noise function.}
    \item{\code{variance}:}{Gets the variance for data points based on
      the noise function and parameters.}
    \item{\code{stddev}:}{Convenience function, \code{sqrt(variance(...))}.}
    \item{\code{lowIntensity}:}{Gets and sets the low intensity slot, denoting the noise region.}
    \item{\code{naRegion}:}{Gets and sets the na.region slot.}
  }
}

\examples{

data(ibspiked_set1)

ceru.proteins <- protein.g(proteinGroup(ibspiked_set1),"CERU")

# normalize
ibspiked_set1 <- normalize(correctIsotopeImpurities(ibspiked_set1))

# remove spiked proteins
ibspiked_set1.noceru <- exclude(ibspiked_set1,ceru.proteins)
ibspiked_set1.justceru <- subsetIBSpectra(ibspiked_set1,protein=ceru.proteins,direction="include")

# learn noise models
nm.i <- new("InverseNoiseModel",ibspiked_set1.noceru)
nm.e <- new("ExponentialNoiseModel",ibspiked_set1.noceru)

#learn on non-one.to.one data: not normalized, with spiked proteins
nm.n <- new("ExponentialNoiseModel",ibspiked_set1.justceru,one.to.one=FALSE)

maplot(ibspiked_set1,noise.model=c(nm.e,nm.i,nm.n),ylim=c(0.1,10))

}

\name{maplot.protein}
\alias{maplot.protein}
\title{
    Ratio intensity plot for individual proteins
}
\description{
    Plots ratio-versus-intensity for a selected protein against a reference channel.
}
\usage{
maplot.protein(x, relative.to, protein, noise.model = NULL, channels = NULL, 
               xlim = NULL, ylim = NULL, identify = FALSE, add = FALSE, 
               pchs = NULL, log="xy", legend.pos = "topright", names = NULL, 
               legend.cex = 0.8, cols = pchs, ltys = 1, main = protein, 
               xlab = NULL, ylab = NULL, type="ma", show.lm = FALSE, ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{IBSpectra object}
  \item{relative.to}{
      a character vector specifying reporter tag names. Either of 
      length 1 or same length as channels.
  }
  \item{protein}{
      Protein group identifier.
  }
  \item{noise.model}{
      NoiseModel object.
  }
  \item{channels}{
       Reporter tag names.
  }
  \item{xlim}{
      See par.
  }
  \item{ylim}{
      See par.
  }
  \item{identify}{
      boolean. If \code{true}, \code{\link{identify}} is called with peptide labels.
  }
  \item{add}{
%%     ~~Describe \code{add} here~~
  }
  \item{pchs}{
    a vector of the same length as \code{channels}. See \code{pch} in
    \code{\link{plot.default}}.
  }
  \item{log}{
    a character string which contains \code{x} if the x axis is to
    be logarithmic, \code{y} if the y axis is to be logarithmic and
    \code{xy} or \code{yx} if both axes are to be logarithmic.
  }
  \item{legend.pos}{
    see \code{pos} in \code{\link{legend}}.
  }
  \item{names}{
    a character string of the same length as \code{channels}, legend text.
  }
  \item{legend.cex}{
    see \code{cex} in \code{\link{legend}}.
  }
  \item{cols}{
    a vector of the same length as \code{channels}. See \code{col} in
    \code{\link{plot.default}}.
  }
  \item{ltys}{
    a vector of the same length as \code{channels}. See \code{lty} in
    \code{\link{plot.default}}.
  }
  \item{main}{
    a main title for the plot  
  }
  \item{xlab}{
    a label for the x axis, defaults to a description of \code{x}.
  }
  \item{ylab}{
    a label for the y axis, defaults to a description of \code{y}.
  }
  \item{type}{type of plot}
  \item{\dots}{
    passed to \code{\link{plot}}.
  }
  \item{show.lm}{show LM}
}
\author{Florian P. Breitwieser}

\name{subsetIBSpectra}
\alias{subsetIBSpectra}
\title{
  Subset IBSpectra objects
}
\description{
  Returns an IBSpectra object which is a subset of the input, excluding
  or exclusively containing the peptides or proteins supplied.
}
\usage{
subsetIBSpectra(x, protein = NULL, peptide = NULL,
                direction = "exclude",
                specificity = c(REPORTERSPECIFIC, GROUPSPECIFIC, UNSPECIFIC), ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{IBSpectra object.}
  \item{protein}{Protein group identifiers. Use \code{\link{protein.g}}
    to get protein group identifiers from protein database ACs.}
  \item{peptide}{Peptide sequences.}
  \item{direction}{either 'include' or 'exclude'.}
  \item{specificity}{When 'protein' is supplied: Which peptides should
    be selected? See \code{\link{specificities}}.}
  \item{\dots}{Further arguments passed to \code{\link{spectrumSel}}}
}
\author{Florian P Breitwieser}

\seealso{
  \code{\link{protein.g}}, \code{\link{spectrumSel}}, \code{\link{specificities}}
}
\examples{
data(ibspiked_set1)

# get Keratin proteins
keratin.proteins <- protein.g(proteinGroup(ibspiked_set1),"Keratin")

# exclude Keratin proteins
subsetIBSpectra(ibspiked_set1,protein=keratin.proteins,direction="exclude")
}

\name{IBSpectra.log}
\alias{do.log}
\alias{do.log,IBSpectra,character-method}
\alias{get.log}
\alias{get.log,IBSpectra,character-method}
\alias{is.logged}
\alias{is.logged,IBSpectra,character-method}
\title{
    Log functions for IBSpectra objects
}
\description{
    The slot \code{log} of \code{IBSpectra} objects contains a matrix with
    two columns which contain a timestamp and message. Rownames
    relate to the item logged. 

    Used by \code{\link{correctIsotopeImpurities}} and \code{\link{normalize}}.
}
\usage{
do.log(x, name, msg)

get.log(x, name)

is.logged(x, name)

}
\arguments{
  \item{x}{IBSpectra object}
  \item{name}{Name of property to be logged (translates to row name).}
  \item{msg}{Message to be logged for name.}
}
\details{
    A warning message will be displayed if a already logged property
    is logged again.
}
\value{
    do.log:  IBSpectra object with updated log.
    get.log: 
}
\author{Florian P Breitwieser}

\seealso{IBSpectra-class}
\examples{
    data(ibspiked_set1)
    ib <- normalize(correctIsotopeImpurities(ibspiked_set1))
    ib@log
}

\name{IBSpectra-class}
\docType{class}

% Class
\alias{class:IBSpectra}
\alias{IBSpectra-class}
\alias{iTRAQSpectra-class}
\alias{iTRAQSpectra}
\alias{iTRAQ4plexSpectra-class}
\alias{iTRAQ4plexSpectra}
\alias{iTRAQ8plexSpectra-class}
\alias{iTRAQ8plexSpectra}
\alias{TMTSpectra-class}
\alias{TMTSpectra}
\alias{TMT2plexSpectra-class}
\alias{TMT2plexSpectra}
\alias{TMT6plexSpectra-class}
\alias{TMT6plexSpectra2-class}
\alias{TMT6plexSpectra}
\alias{TMT6plexSpectra2}
\alias{TMT10plexSpectra-class}
\alias{TMT10plexSpectra}
\alias{IBSpectra}

% Constructor and readIBSpectra function:
\alias{initialize,IBSpectra-method}
\alias{VARMETADATA}

% Coercion methods:
\alias{coerce,IBSpectra,data.frame-method}
\alias{as.data.frame,IBSpectra-method}
\alias{as.data.frame.IBSpectra}

% Accessor-like methods:
\alias{proteinGroup}
\alias{proteinGroup,IBSpectra-method}
\alias{proteinGroup<-}
\alias{proteinGroup<-,IBSpectra-method}
\alias{isotopeImpurities}
\alias{isotopeImpurities,IBSpectra-method}
\alias{isotopeImpurities<-}
\alias{isotopeImpurities<-,IBSpectra-method}
\alias{reporterIntensities}
\alias{reporterIntensities,IBSpectra-method}
\alias{reporterIntensities<-}
\alias{reporterIntensities<-,IBSpectra-method}
\alias{reporterMasses}
\alias{reporterMasses,IBSpectra-method}
\alias{reporterMasses<-}
\alias{reporterMasses<-,IBSpectra-method}
\alias{reporterTagNames}
\alias{reporterTagNames,IBSpectra-method}
\alias{reporterTagMasses}
\alias{reporterTagMasses,IBSpectra-method}
\alias{reporterData}
\alias{reporterData,IBSpectra-method}
\alias{reporterData<-}
\alias{reporterData<-,IBSpectra-method}
\alias{classLabels}
\alias{classLabels,IBSpectra-method}
\alias{classLabels<-}
\alias{classLabels<-,IBSpectra-method}

% Util methods:

\alias{spectrumTitles}
\alias{spectrumTitles,IBSpectra-method}
\alias{spectrumSel}
\alias{spectrumSel,IBSpectra,missing,missing-method}
\alias{spectrumSel,IBSpectra,matrix,missing-method}
\alias{spectrumSel,IBSpectra,character,missing-method}
\alias{spectrumSel,IBSpectra,missing,character-method}
\alias{spectrumSel,IBSpectra,data.frame,missing-method}
\alias{IBSpectraTypes}
\alias{writeData}
\alias{writeData,IBSpectra-method}
\alias{ibSpectra.as.concise.data.frame}

% show method:
\alias{show,IBSpectra-method}

\title{IBSpectra Class for Isobarically Tagged Quantitative MS
  Proteomics Data}

\description{ This class represents a quantitative MS proteomics
  experiment labeled using Isobaric tags (iTRAQ, TMT). \code{IBSpectra}
  is a abstract class which is implemented in the
  \code{\link{IBSpectraTypes}} classes \code{iTRAQ4plexSpectra},
  \code{iTRAQ8plexSpectra}, \code{TMT2plexSpectra}, \code{TMT6plexSpectra} and \code{TMT10plexSpectra}.
    
  It contains per-spectrum meassurements of the reporter tag intensity and 
  m/z in assayData, and protein grouping in proteinGroup.
}

\section{Objects from the Class}{
  \code{IBSpectra} objects are typically created using the
  \code{\link{readIBSpectra}} method or by calls of the form
  \code{new("iTRAQ4plexSpectra",data=NULL,data.ions=NULL,...).}
}

\section{Slots}{
  \code{IBSpectra} extends \code{\link[Biobase:class.eSet]{eSet}} which
  is a container for high-throughput assays and experimental metadata.
  Slots introduced in \code{eSet} (for more details on slots and methods
  refer to \code{\link{eSet}} help):
   \describe{
     \item{\code{assayData}:}{Contains matrices 'ions' and 'mass
       storing reporter tag intensities and m/z values for each tag and
       spectrum. Can be accessed by \code{\link{reporterIntensities}}
       and \code{\link{reporterMasses}}.  Class: \code{\link{AssayData}}}
      \item{\code{phenoData}:}{Contains experimenter-supplied variables
	describing phenotypes behind reporter tags.
	Class: \code{\link{AnnotatedDataFrame-class}}}
      \item{\code{featureData}:}{Describes the spectra's retention time,
	charge, peptide sequence, etc and can be accessed by
	\code{\link{fData}}. Class: \code{\link{AnnotatedDataFrame}}} 
      \item{\code{experimentData}:}{Contains details of experimental
	methods. Class: \code{\link{MIAME}}}
      \item{\code{annotation}:}{UNUSED. Label associated with the annotation
	package used in the experiment. Class: \code{character}}
      \item{\code{protocolData}:}{UNUSED. Contains equipment-generated
      variables describing reporter tags. Class: \code{\link{AnnotatedDataFrame}}}
      \item{\code{log}:}{\code{character matrix} logging isotope impurity
        correction, normalization, etc.}
   }
  
  Slots introduced in \code{IBSpectra}:
  \describe{
    \item{\code{proteinGroup}:}{A \code{\link{ProteinGroup}}
      object describing peptide and protein identifications
      grouped by shared peptides.}
    \item{\code{reporterTagNames}:}{A character vector denoting the
      reporter tag labels.}
    \item{\code{reporterMasses}:}{The 'true' m/z of the reporter tags
      in the MS/MS spectrum, used to isolate m/z-intensity pairs from
      peaklist.}
    \item{\code{isotopeImpurities}:}{Manufacturer supplied isotope
      impurities, need to be set per batch and used for correction
      by \code{\link{correctIsotopeImpurities}}.
    }
  }
}


\section{Constructor}{
  See \code{\link{readIBSpectra}} for creation based on peaklist
  (e.g. MGF format) and identification files (Mascot and Phenyx output).
  \describe{
    \item{\code{new(type,data)}:}{
      Creates a IBSpectra object.
      \describe{
        \item{\code{type}}{Denotes the type of IBSpectra, either
	  'iTRAQ4plexSpectra','iTRAQ8plexSpectra','TMT2plexSpectra', 'TMT6plexSpectra'
	  or 'TMT10plexSpectra'. Call IBSpectraTypes() to see a list of
	  the implemented types.
	}
        \item{\code{data}}{A 'data.frame' in a ibspectra-csv format.}
      }
    }
  }
}

\section{Coercion}{
  In the code snippets below, \code{x} is a IBSpectra object.
  \code{IBSpectra} object can be coerced to 
  \describe{
    \item{\code{as(x, "data.frame")}:}{ Creates a data.frame containing
      all identification and quantitation information. Peptide matching
      to multiple proteins produce multiple lines. }
    \item{\code{ibSpectra.as.concise.data.frame(x)}:}{ Creates a data.frame
      containing all identification and
      quantitation information. Proteins are concatenated - so the
      resulting \code{data.frame} has one line per spectrum.}
    \item{\code{as(x, "MSnSet")}:}{
      Coerces to a \code{\link[MSnbase]{MSnSet}} object (package
      \code{\link[MSnbase]{MSnbase}}). }
    \item{\code{as(msnset,"IBSpectra")}:}{
      Coerces a \code{\link[MSnbase]{MSnSet}} to \code{IBSpectra} object.
    }
  }
}

\section{Accessors}{
  In the following code snippets, \code{x} is a IBSpectra object.

  \describe{
    \item{\code{proteinGroup(x)}:}{Gets and sets the ProteinGroup.}
    \item{\code{isotopeImpurities(x)}:}{Gets and sets the isotope
    impurities of the isobaric tags as defined by the manufacturers
    per batch.}  
    \item{\code{reporterData(x,element="ions",na.rm=FALSE,na.rm.f='any',...)}:}{
      Gets and sets the element ('ions' or 'mass') for each tag and spectrum.
      '...' is handed down to spectrumSel, so it is possible to
      select for peptides or proteins.
      If na.rm is TRUE, than spectra missing quantitative information in 'any' or 'all'
      channels (parameter na.rm.f) are removed.
    }
    
    \item{\code{reporterIntensities(x,...)}:}{
      Convenience function, calls reporterData(...,element="ions")
    }
    \item{\code{reporterMasses(x,...)}:}{
      Convenience function, calls reporterData(...,element="mass")
    }

    \item{\code{spectrumTitles(x,...)}:}{
      Gets the spectrum titles. '...' is passed down to spectrumSel.
    }
    \item{\code{classLabels(x)}:}{
      Gets and sets the class labels in phenoData. Used for
      summarization, see also \code{\link{estimateRatio}} and
      \code{\link[Biobase]{phenoData}}.
    }
  }
}

\section{Methods}{
  In the following code snippets, \code{x} is a IBSpectra object.

  \describe{
    \item{\code{subsetIBSpectra(x, protein=NULL, peptide=NULL, direction="exclude",specificity)}:}{
      Get a 'subset' of IBSpectra: include or exclude proteins
      or peptides. When selection  is based on proteins, it can be
      defined to exclude only peptides which are specific to the protein
      ('reporter-specific'), specific to the group ('group-specific') or
      which are shared with other proteins ('unspecific'). See \code{\link{subsetIBSpectra}}.}
    \item{\code{spectrumSel(x,peptide,protein,specificity="reporter-specific")}:}{
      Gets a boolean vector selecting the corresponding spectra:
      If peptide is given, all spectra assigned to this peptide.
      If protein is given, all spectra assigned to peptides of this
      protein with specificity 'specificity'.
      See also \code{\link[isobar]{ProteinGroup}}.
    }
}
}
\author{Florian P. Breitwieser}

\seealso{
  \link{ProteinGroup},
  \link{isobar-preprocessing},
  \link{isobar-analysis},
  \link{isobar-plots}
}

\examples{


data(ibspiked_set1)
ibspiked_set1
head(reporterIntensities(ibspiked_set1))
head(reporterMasses(ibspiked_set1))
proteinGroup(ibspiked_set1)
isotopeImpurities(ibspiked_set1)

# create new object
set.seed(123)
data <- data.frame(spectrum=letters,
                   peptide=sample(c("pepA","pepB","pepC"),26,TRUE),
                   start.pos=1,
                   modif=sample(c("::X:::",":Y::::","::Z:::"),26,TRUE),
                   accession=c("protein1","protein2"))
data.ions <- matrix(rnorm(26*2,1000,50),
                    ncol=2,dimnames=list(letters,NULL))
data.mass <- matrix(rep(c(126.1,127.1),26),
                    ncol=2,byrow=TRUE,dimnames=list(letters,NULL))
ib <- new("TMT2plexSpectra",data,data.ions,data.mass)
ib
reporterIntensities(ib)
isotopeImpurities(ib) <- matrix(c(0.8,0.1,0.2,0.9),nrow=2)
reporterIntensities(correctIsotopeImpurities(ib))
}

\name{correct.peptide.ratios}
\alias{correct.peptide.ratios}

\title{Correct peptide ratios with protein ratios from a separate experiment.}
\description{Correct peptide ratios with protein ratios from a separate experiment.}

\usage{
correct.peptide.ratios(ibspectra, peptide.quant.tbl, protein.quant.tbl,
         protein.group.combined, adjust.variance = TRUE,
         correlation = 0, recalculate.pvalue = TRUE)
}

\arguments{
  \item{ibspectra}{IBSpectra object.}
  \item{peptide.quant.tbl}{Calculated with peptideRatios.}
  \item{protein.quant.tbl}{Calculated with proteinRatios.}
  \item{protein.group.combined}{ProteinGroup object generated on both PTM and protein data.}
  \item{adjust.variance}{Adjust variance of ratios.}
  \item{correlation}{Assumed correlation between peptide and protein ratios for variance adjustment.}
  \item{recalculate.pvalue}{Recalculate p-value after variance adjustment.}
}
\author{Florian P. Breitwieser}
